/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.compiler;

import java.util.*;
import java.io.*;
import kmy.jint.reflect.*;
import kmy.jint.parser.*;
import kmy.jint.tree.*;

import kmy.jint.gen.*;
import kmy.jint.util.*;
import kmy.jint.typer.*;

import kmy.jint.constants.TokenConst;
import kmy.jint.constants.DefinitionConst;
import kmy.jint.constants.ErrorCodes;

import kmy.regex.tree.RNode;
import kmy.regex.compiler.RCompiler;
import kmy.regex.jvm.*;

public class CompilerContext extends Context implements TokenConst, DefinitionConst, ErrorCodes
{

  CodeGenerator gen;
  CodeGeneratorFactory factory;
  JintClassTable classTable;

  int regexCount = 1;
  int formatCount = 1;

  BitSet registerMask;

  /**
   * Type of LHS of '.' operator when evaluating RHS
   * (which is always identifier)
   */ 
  AbstractClass baseType;

  int selfRegister = -1; // register where 'self' variable is kept, if any

  AbstractClass constructorOuterClass; // outer class passed to constructor

  static Random cacheID = new Random();

  static final Integer I_ONE  = new Integer(1);
  static final Integer I_MONE = new Integer(-1);
  static final Integer I_ZERO = new Integer(0);
  static final Long    L_ONE  = new Long(1);
  static final Long    L_ZERO = new Long(0);
  static final Float   F_ONE  = new Float(1);
  static final Float   F_ZERO = new Float(0);
  static final Double  D_ONE  = new Double(1);
  static final Double  D_ZERO = new Double(0);

  static final String ANY_UTILITIES = AnyUtilities.class.getName().replace( '.', '/' );
  static final String ANY_CALL_METHOD = "callMethod";
  static final String ANY_CALL_METHOD_SIG = 
   "(Ljava/lang/Object;Ljava/lang/String;[Ljava/lang/Class;" +
   "[Ljava/lang/Object;I)Ljava/lang/Object;";
  static final String ANY_CALL_METHOD_FULL_SIG = 
   "(Ljava/lang/Object;Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/Class;" + 
   "[Ljava/lang/Object;I)Ljava/lang/Object;";
  static final String ANY_OP1_METHOD = "op";
  static final String ANY_OP1_METHOD_SIG = "(ILjava/lang/Object;)Ljava/lang/Object;";
  static final String ANY_OP2_METHOD = "op";
  static final String ANY_OP2_METHOD_SIG = 
   "(ILjava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;";
  static final String ANY_COND_METHOD = "cond";
  static final String ANY_COND_METHOD_SIG = "(ILjava/lang/Object;Ljava/lang/Object;)Z";
  static final String ANY_SHIFT_METHOD = "shift";
  static final String ANY_SHIFT_METHOD_SIG = "(ILjava/lang/Object;I)Ljava/lang/Object;";
  static final String ANY_FLDGET_METHOD = "getField";
  static final String ANY_FLDGET_METHOD_SIG = 
   "(Ljava/lang/Object;Ljava/lang/String;I)Ljava/lang/Object;";
  static final String ANY_FLDPUT_METHOD = "setField";
  static final String ANY_FLDPUT_METHOD_SIG = 
    "(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/String;I)V";
  static final String ANY_GETELEM_METHOD = "getElement";
  static final String ANY_GETELEM_METHOD_SIG = 
    "(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;";
  static final String ANY_PUTELEM_METHOD = "putElement";
  static final String ANY_PUTELEM_METHOD_SIG = 
    "(Ljava/lang/Object;Ljava/lang/Object;Ljava/lang/Object;)V";
  static final String ANY_TEXTSEARCH_METHOD = "textSearch";
  static final String ANY_TEXTSEARCH_METHOD_SIG = 
    "(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;";

  static final String DISPATCH_OBJECT = DispatchObject.class.getName().replace( '.', '/' );

  private boolean reassignRegisterFlag = true;

  private boolean trashTheValue = false;

  private Stack      nestedLabels; // used to process break, continue, return and finally
  private LabelInfo  label; // used to process labeled statements

  private int contextState = STATE_LOAD;

  final static int STATE_LOAD  = 0;
  final static int STATE_BASE  = 1;
  final static int STATE_STORE = 2;
  final static int STATE_LOAD2 = 3;

  final static Object TRASHED = new Object();

  Identifier INIT;
  Identifier THIS;
  Identifier SUPER;
  Identifier CLINIT;

  Hashtable classRefs = null;
  AbstractClass currentClass;
  AbstractClass selfClass;

  private ErrorStorage errorStorage;
  
  boolean compileRegexToGivenClass;

  static class LabelInfo
  {
    StatementNode stmt;
    boolean       loop;
    AbstractMark  againMark;
    AbstractMark  finishMark;
    int           syncLock;
  }

  static class EmbedRegexAttachment
  {
    // for embedded regexes
    int      forkRegister;
    int      cellRegister;
  }

  public CompilerContext( JintClassTable ct,
			  CodeGeneratorFactory f ) throws IOException
  {
    classTable = ct;
    INIT = ct.getNameTable().INIT;
    THIS = ct.getNameTable().THIS;
    SUPER = ct.getNameTable().SUPER;
    CLINIT = ct.getNameTable().CLINIT;
    factory = f;
  }

  public void setErrorStorage( ErrorStorage errorStorage )
  {
    this.errorStorage = errorStorage;
  }

  public void storeError( CompilerException e )
  {
    if( errorStorage != null )
      errorStorage.storeError( e );
    else
      System.err.println( "Line " + e.getLine() + ", pos " + 
			 e.getColumn() + ": " + e.getMessage() );
  }  

  public void storeError( int code, int fpos )
  {
    storeError( new CompilerException( code, fpos ) );
  }  

  public void storeError( int code, int fpos, Object add )
  {
    Object[] addInfo = { add };
    storeError( new CompilerException( code, fpos, addInfo ) );
  }  

  public void storeError( int code, BasicNode node )
  {
    storeError( new CompilerException( code, node.getFilePos() ) );
  }  

  public void storeError( int code, BasicNode node, Object add )
  {
    Object[] addInfo = { add };
    storeError( new CompilerException( code, node.getFilePos(), addInfo ) );
  }  

  public void storeError( int code, BasicNode node, Object add1, Object add2 )
  {
    Object[] addInfo = { add1, add2 };
    storeError( new CompilerException( code, node.getFilePos(), addInfo ) );
  }  

  private Object evalExpr( ExpressionNode expr )
  {
    try
      {
	if( expr != null )
	  return expr.eval( this );
      } 
    catch( CompilerException e )
      {
	storeError( e );
      }
    catch( Exception e )
      {
	internalError( e, expr );
      }
   return null;
  }

  private void evalAndPop( ExpressionNode expr )
  {
    if( expr == null )
      return;
    if( expr instanceof IdentifierNode )
      {
	IdentifierNode ind = (IdentifierNode)expr;
	if( ind.idList.length <= 1 || ((Object[])ind.attachment).length == 0 )
	  return;
      }
    trashTheValue = (expr instanceof InfixNode || expr instanceof PrefixNode
		     || expr instanceof PostfixNode );
    if( evalExpr( expr ) != TRASHED && expr.clazz != classTable.voidType )
      {
	try
	  {
	    gen.pop( expr.clazz.getDescriptor(), 1 );
	  }
	catch( Exception e )
	  {
	    System.err.println( "Failed to pop " + expr.clazz );
	    internalError( e, expr );
	  }
      }
    trashTheValue = false;
  }

  private AbstractClass evalAndWrap( ExpressionNode expr )
  {
    AbstractClass wrapper = classTable.getWrapper( expr.clazz );
    evalAndWrapOrCast( expr, wrapper );
    return wrapper;
  }

  private static String getTypeName( char c )
  {
    switch( c )
      {
	case 'Z' :
	  return "boolean";
	case 'I' :
	  return "int";
	case 'J' :
	  return "long";
	case 'S' :
	  return "short";
	case 'C' :
	  return "char";
	case 'D' :
	  return "double";
	case 'F' :
	  return "float";
	case 'B' :
	  return "byte";
      }
    throw new IllegalArgumentException( "Unexpected descriptor: " + c );
  }

  private void evalAndWrapOrCast( ExpressionNode expr, AbstractClass to )
  {
    evalAndWrapOrCast( expr, to, false );
  }

  private void evalAndWrapOrCast( ExpressionNode expr, AbstractClass to, boolean doCast )
  {
    try
      {
	AbstractClass clazz = expr.clazz;
	if( clazz == to || to == null )
	  {
	    evalExpr( expr );
	    return;
	  }
	if( clazz == null )
	  {
	    if( to.isPrimitive() )
	      storeError( ERR_C_CVTNULLTOPRIM, expr );
	    evalExpr( expr );
	    return;
	  }
	if( to.isPrimitive() )
	  {
	    evalExpr( expr );
	    if( clazz.isPrimitive() )
	      {
		gen.cast( clazz.getDescriptor(), to.getDescriptor() );
	      }
	    else
	      {
		String td = to.getDescriptor();
		char tdc = td.charAt(0);
		if( clazz == classTable.anyClass )
		  {
		    if( tdc == 'Z' )
		      {
			gen.cast( "Ljava/lang/Object;", "Ljava/lang/Boolean;" );
			gen.invokevirtual( "java/lang/Boolean", "booleanValue", "()Z" );
		      }
		    else
		      {
			gen.dup( "Ljava/lang/Object;" );
			gen.instnceof( "java/lang/Character" );
			AbstractMark m1 = gen.newMark();
			AbstractMark done = gen.newMark();
			gen.jumpIf( true, gen.TOKEN_EE, "I", m1 );
			gen.cast( "Ljava/lang/Object;", "Ljava/lang/Character;" );
			gen.invokevirtual( "java/lang/Character", "charValue", "()C" );
			if( tdc != 'C' )
			  gen.cast( "C", td );
			gen.jump( done );
			gen.mark( m1 );
			gen.cast( "Ljava/lang/Object;", "Ljava/lang/Number;" );
			gen.invokevirtual( "java/lang/Number", 
					   getTypeName( tdc ) + "Value", "()" + tdc );
			gen.mark( done );
		      }
		  }
		else
		  {
		    String wd = classTable.getWrappedType( clazz ).getDescriptor();
		    char wdc = wd.charAt(0);
		    if( wdc == 'C' || wdc == 'Z' || tdc == 'C' || tdc == 'Z' )
		      {
			gen.invokevirtual( clazz.getDescriptor(), getTypeName( wdc ) + "Value",
					   "()" + wdc );
			if( wdc != tdc )
			  gen.cast( wd, td );
		      }
		    else
		      gen.invokevirtual( clazz.getDescriptor(), getTypeName( tdc ) + "Value",
					 "()" + tdc );
		  }
	      }
	  }
	else
	  {
	    if( clazz.isPrimitive() )
	      {
		if( to == classTable.anyClass || to == classTable.objectClass )
		  to = classTable.getWrapper( clazz );
		gen.newobject( to.getJVMName() );
		gen.dup( to.getDescriptor() );
		evalExpr( expr );
		String cd = clazz.getDescriptor();
		AbstractClass wrapped = classTable.getWrappedType( to );
		if( wrapped != clazz )
		  {
		    String cd1 = wrapped.getDescriptor();
		    gen.cast( cd, cd1 );
		    cd = cd1;
		  }
		gen.invokespecial( to.getJVMName(), "<init>", "(" + cd + ")V" );
	      }
	    else if( clazz == classTable.stringClass && to == classTable.charStringClass )
	      {
		gen.newobject( to.getJVMName() );
		gen.dup( to.getDescriptor() );
		evalExpr( expr );
		gen.invokespecial( to.getJVMName(), "<init>", "(" + clazz.getDescriptor() + ")V" );
	      }
	    else
	      {
		evalExpr( expr );
		if( to != classTable.anyClass && !clazz.isSubclassOf( to ) )
		  {
		    if( to != classTable.stringClass || doCast )
		      gen.cast( clazz.getDescriptor(), to.getDescriptor() );
		    else
		      gen.invokevirtual( "java/lang/Object", "toString", "()Ljava/lang/String;" );
		  }
	      }
	  }
      }
    catch( Exception e )
      {
	internalError( e, expr );
      }
  }

  private boolean ifZero( ExpressionNode expr )
  {
    if( ! (expr instanceof ConstNode) )
      return false;
    Object v = ((ConstNode)expr).val;
    if( v instanceof Number )
      return ((Number)v).doubleValue() == 0.0;
    if( v instanceof Boolean )
      return ((Boolean)v).booleanValue() == false;
    if( v instanceof Character )
      return ((Character)v).charValue() == 0;
    return v == null;
  }

  private void evalCond( ExpressionNode expr, AbstractMark trueMark, AbstractMark falseMark )
  {
    try
      {
	if( expr instanceof InfixNode )
	  {
	    InfixNode infix = (InfixNode)expr;
	    int op = infix.op;
	    AbstractClass common = (AbstractClass)infix.attachment;
	    switch( op )
	      {
	        case TOKEN_EE :
	        case TOKEN_NE :
	        case '>' :
	        case '<' :
	        case TOKEN_GE :
	        case TOKEN_LE :
		  {
		    if( common == null )
		      common = classTable.anyClass;
		    boolean anyFlag = common == classTable.anyClass ||
		      !(common.isPrimitive() || op == TOKEN_EE || op == TOKEN_NE);
		    String cdscr = common.getDescriptor();
		    if( anyFlag )
		      gen.loadConst( new Integer( op ) );

 		    boolean zero1 = !anyFlag && ifZero( infix.arg1 );
 		    boolean zero2 = !anyFlag && ifZero( infix.arg2 );
		    if( ! zero1 )
		      evalAndWrapOrCast( infix.arg1, common );
		    if( ! zero2 || zero1 )
		      evalAndWrapOrCast( infix.arg2, common );

		    if( zero1 && common != classTable.floatClass && 
			common != classTable.doubleClass )
		      {
			zero2 = true;
			boolean swapLabels = true;
			switch( op )
			  {
			    case '>' :
			      op = TOKEN_LE;
			      break;
			    case '<' :
			      op = TOKEN_GE;
			      break;
			    case TOKEN_GE :
			      op = '<';
			      break;
			    case TOKEN_LE :
			      op = '>';
			      break;
			    default:
			      swapLabels = false;
			  }
			if( swapLabels )
			  {
			    AbstractMark tmpMark = trueMark;
			    trueMark = falseMark;
			    falseMark = tmpMark;
			  }
		      }

		    if( anyFlag )
		      {
			gen.invokestatic( ANY_UTILITIES, ANY_COND_METHOD, ANY_COND_METHOD_SIG );
			expr = null;
		      }
		    else
		      {
			if( trueMark != null )
			  {
			    gen.jumpIf( true, zero2, op, cdscr, trueMark );
			    if( falseMark != null )
			      gen.jump( falseMark );
			  }
			else if( falseMark != null )
			  gen.jumpIf( false, zero2, op, cdscr, falseMark );
			return;
		      }
		  }
		  break;
	        case TOKEN_OR :
		  {
		    AbstractMark mark = ( trueMark == null ? gen.newMark() : trueMark );
		    evalCond( infix.arg1, mark, null );
		    evalCond( infix.arg2, trueMark, falseMark );
		    if( mark != trueMark )
		      gen.mark( mark );
		    return;
		  }
	        case TOKEN_AND :
		  {
		    AbstractMark mark = ( falseMark == null ? gen.newMark() : falseMark );
		    evalCond( infix.arg1, null, mark );
		    evalCond( infix.arg2, trueMark, falseMark );
		    if( mark != falseMark )
		      gen.mark( mark );
		    return;
		  }
	      }
		
	  }
	else if( expr instanceof PrefixNode )
	  {
	    PrefixNode prefix = (PrefixNode)expr;
	    if( prefix.op == '!' )
	      {
		evalCond( prefix.arg, falseMark, trueMark );
		return;
	      }
	  }

	if( expr != null ) // if null, already on stack
	  {
	    // default behaviour
	    expr.eval( this );
	    AbstractClass clazz = expr.clazz;
	    if( clazz.isPrimitive() )
	      {
		if( clazz != classTable.booleanType )
		  throw new CompilerException( ERR_C_NOTBOOL, expr.getFilePos() );
	      }
	    else
	      {
		if( clazz != classTable.booleanClass )
		  gen.cast( "Ljava/lang/Object;", "Ljava/lang/Boolean;" );
		gen.invokevirtual( "java/lang/Boolean", "booleanValue", "()Z" );
	      }
	  }
	if( trueMark != null )
	  {
	    gen.jumpIf( true, true, gen.TOKEN_NE, "Z", trueMark );
	    if( falseMark != null )
	      gen.jump( falseMark );
	  }
	else if( falseMark != null )
	  gen.jumpIf( true, true, gen.TOKEN_EE, "Z", falseMark );
      }
    catch( CompilerException e )
      {
	storeError( e );
      }
    catch( Exception e )
      {
	internalError( e, expr );
      }
  }

  private int toJVMFlags( AbstractClass scope )
  {
    int flags = scope.getModifiers();
    return toJVMFlags( flags );
  }

  private int toJVMFlags( int modifiers )
  {
    int result = 0;
    if( (modifiers & DefinitionNode.PUBLIC_FLAG ) != 0 )
      result |= gen.ACC_PUBLIC;
    if( (modifiers & DefinitionNode.PRIVATE_FLAG) != 0 )
      result |= gen.ACC_PRIVATE;
    if( (modifiers & DefinitionNode.STATIC_FLAG) != 0 )
      result |= gen.ACC_STATIC;
    if( (modifiers & DefinitionNode.PROTECTED_FLAG) != 0 )
      result |= gen.ACC_PROTECTED;
    if( (modifiers & DefinitionNode.FINAL_FLAG) != 0 )
      result |= gen.ACC_FINAL;
    if( (modifiers & DefinitionNode.ABSTRACT_FLAG) != 0 )
      result |= gen.ACC_ABSTRACT;
    if( (modifiers & DefinitionNode.SYNCHRONIZED_FLAG) != 0 )
      result |= gen.ACC_SYNCHRONIZED;
    if( (modifiers & DefinitionNode.VOLATILE_FLAG) != 0 )
      result |= gen.ACC_VOLATILE;
    if( (modifiers & DefinitionNode.TRANSIENT_FLAG) != 0 )
      result |= gen.ACC_TRANSIENT;
    if( (modifiers & DefinitionNode.INTERFACE_FLAG) != 0 )
      result |= gen.ACC_INTERFACE;
    return result;
  }

  private String baseType( ExpressionNode expr )
  {
    if( expr instanceof InfixNode )
      {
	InfixNode inf = (InfixNode)expr;
	if( inf.op == '[' )
	  return "2";
	if( inf.op == '.' )
	  expr = inf.arg2;
      }
    if( expr instanceof IdentifierNode )
      {
	AbstractVariable[] av = (AbstractVariable[])expr.attachment;
	if( av == null || av.length == 0 )
	  return "V";
	else
	  return ((CompilerVariable)av[av.length-1]).getBase( this ).getDescriptor();
      }
    throw new CompilerException( ERR_C_CANTASSIGN, expr.getFilePos() );
  }

  private Number get1OfType( AbstractClass clazz )
  {
    switch( clazz.getDescriptor().charAt(0) )
      {
        case 'I' :
        case 'C' :
        case 'Z' :
        case 'B' :
        case 'S' :
        case 'L' :
        case 'A' :
	  return I_ONE;
        case 'J' :
	  return L_ONE;
        case 'F' :
	  return F_ONE;
        case 'D' :
	  return D_ONE;
      }
    throw new IllegalArgumentException( "Never reached" );
  }

  private Object opEq( int op, ExpressionNode arg1, ExpressionNode arg2, AbstractClass type, 
		     boolean resultAfter, boolean trash ) throws IOException
  {

    boolean shift = op == TOKEN_GGE || op == TOKEN_GGGE || op == TOKEN_LLE;

    contextState = STATE_BASE;
    arg1.eval( this );

    String baseType = baseType( arg1 );
    if( baseType != "V" )
      gen.dup( baseType );
 
    contextState = STATE_LOAD2;

    evalAndWrapOrCast( arg1, type );
    
    if( !trash && !resultAfter)
      {
	if( baseType == "V" )
	  gen.dup( type.getDescriptor() );
	else
	  gen.insert( type.getDescriptor(), baseType );
      }

    if( type == classTable.anyClass )
      {
	gen.loadConst( new Integer( op ) );
	gen.swap();
      }

    contextState = STATE_LOAD;

    if( type == classTable.stringClass )
      {
	if( op != TOKEN_PE )
	  throw new IllegalArgumentException( "type == stringClass and op is not +=" );
	gen.newobject( "java/lang/StringBuffer" );
	gen.insert( "Ljava/lang/StringBuffer;", "Ljava/lang/String;" );
	gen.swap();
	gen.invokespecial( "java/lang/StringBuffer", "<init>",
			   "(Ljava/lang/String;)V" );
      }

    if( arg2 != null )
      {
	if( type != classTable.stringClass )
	  evalAndWrapOrCast( arg2, ( shift ? classTable.intType : type ) );
      }
    else
      {
	if( type == classTable.anyClass )
	  {
	    gen.newobject( "java/lang/Integer" );
	    gen.dup( "Ljava/lang/Integer;" );
	    gen.loadConst( I_ONE );
	    gen.invokespecial( "java/lang/Integer", "<init>", "(I)V" );
	  }
	else
	  gen.loadConst( get1OfType( type ) );
      }

    if( type == classTable.anyClass )
      if( shift )
	gen.invokestatic( ANY_UTILITIES, ANY_SHIFT_METHOD, ANY_SHIFT_METHOD_SIG );
      else
	gen.invokestatic( ANY_UTILITIES, ANY_OP2_METHOD, ANY_OP2_METHOD_SIG );
    else
      if( type == classTable.stringClass )
	{
	  evalAndAppend( arg2 );
	  gen.invokevirtual( "java/lang/StringBuffer", "toString",
			   "()Ljava/lang/String;" );
	}
      else
	gen.op( op, type.getDescriptor() );

    if( arg1.clazz != type && ! arg1.clazz.isSubclassOf( type ) )
      gen.cast( type.getDescriptor(), arg1.clazz.getDescriptor() );
    
    if( !trash && resultAfter )
      {
	if( baseType == "V" )
	  gen.dup( arg1.clazz.getDescriptor() );
	else
	  gen.insert( arg1.clazz.getDescriptor(), baseType );
      }
    contextState = STATE_STORE;
    arg1.eval( this );
    contextState = STATE_LOAD;
    return ( trash ? TRASHED : null );
  }

  private LabelInfo nestedLabelsPush( StatementNode stmt )
  {
    LabelInfo myInfo = new LabelInfo();
    myInfo.stmt = stmt;
    if( nestedLabels == null )
      nestedLabels = new Stack();
    nestedLabels.push( myInfo );
    return myInfo;
  }

  private void nestedLabelsPop( StatementNode check )
  {
    if( check != null && check != ((LabelInfo)nestedLabels.pop()).stmt )
      storeError( ERR_C_NESTING, check );
  }

  private boolean needToJumpOut()
  {
    if( nestedLabels == null )
      return false;
    Enumeration list = nestedLabels.elements();
    while( list.hasMoreElements() )
      {
	StatementNode stmt = ((LabelInfo)list.nextElement()).stmt;
	if( stmt instanceof TryNode || stmt instanceof SynchronizedNode )
	  return true;
      }
    return false;
  }

  private void jumpOut( StatementNode from, StatementNode to ) throws IOException
  {
    LabelInfo myInfo = null;
    boolean returnFlag = from instanceof ReturnNode;
    boolean breakFlag = from instanceof BreakNode;
    if( nestedLabels != null )
      {
	for( int l = nestedLabels.size() - 1 ; l >= 0 ; l -- )
	  {
	    LabelInfo info = (LabelInfo)nestedLabels.elementAt(l);
	    StatementNode stmt = info.stmt;
	    if( stmt instanceof TryNode )
	      {
		if( info.finishMark != null )
		  gen.jsr( info.finishMark, 0 );
		if( stmt != to )
		  continue;
	      }
	    else if( stmt instanceof SynchronizedNode )
	      {
		gen.load( info.syncLock, "Ljava/lang/Object;" );
		gen.monitorexit();
		if( stmt != to )
		  continue;
	      }
	    if( returnFlag )
	      continue;
	    if( stmt != to )
	      {
		if( to != null )
		  continue;
		if( stmt instanceof BodyNode )
		  continue;
	      }
	    if( breakFlag || !(stmt instanceof SwitchNode) )
	      {
		myInfo = info;
		break;
	      }
	  }
      }
    if( myInfo == null )
      {
	if( !returnFlag )
	  storeError( ERR_C_NOENCLOSING, from );
      }
    else
      {
	if( breakFlag )
	  {
	    if( myInfo.finishMark == null )
	      myInfo.finishMark = gen.newMark();
	    gen.jump( myInfo.finishMark );
	  }
	else
	  {
	    if( myInfo.againMark == null )
	      myInfo.againMark = gen.newMark();
	    gen.jump( myInfo.againMark );
	  }
      }
  }

  private void genArgArray( ExpressionNode[] args, int off, int len )
    throws IOException
  {
    gen.loadConst( new Integer( len ) );
    gen.newarray( "Ljava/lang/Object;" );
    for( int i = 0 ; i < len ; i++ )
      {
	gen.dup( "[Ljava/lang/Object;" );
	gen.loadConst( new Integer(i) );
	ExpressionNode arg = args[i+off];
	if( arg.clazz != null && arg.clazz.isPrimitive() )
	  evalAndWrap( arg );
	else
	  evalExpr( arg );
	gen.putelement( "Ljava/lang/Object;" );
      }
  }

  private void genArgClassArray( ExpressionNode[] args, int off, int len )
    throws IOException
  {
    gen.loadConst( new Integer( len ) );
    gen.newarray( "Ljava/lang/Class;" );
    for( int i = 0 ; i < len ; i++ )
      {
	AbstractClass clazz = args[i+off].clazz;
	if( clazz != classTable.anyClass && clazz != null )
	  {
	    gen.dup( "[Ljava/lang/Class;" );
	    gen.loadConst( new Integer(i) );
	    genClass( clazz );
	    gen.putelement( "Ljava/lang/Class;" );
	  }
      }
  }

  void genClass( AbstractClass clazz ) throws IOException
  {
    String name = clazz.getJVMName();
    if( clazz.isPrimitive() )
      {
	AbstractClass wrapper = classTable.getWrapper( clazz );
	gen.getstatic( wrapper.getJVMName(), "TYPE", "Ljava/lang/Class;" );
      }
    else
      {
	if( classRefs == null )
	  classRefs = new Hashtable();
	String field;
	if( name.endsWith( ";" ) )
	  field = "c$" + name.substring(0,name.length()-1).replace( '/', '$' ).replace( '[', '1' );
	else
	  field = "c$" + name.replace( '/', '$' ).replace( '[', '1' );
	name = name.replace( '/', '.' );
	classRefs.put( field, field ); // no duplicates!
	String currentName = currentClass.getJVMName();
	gen.getstatic( currentName, field, "Ljava/lang/Class;" );
	gen.dup( "Ljava/lang/Class;" );
	AbstractMark ifHave = gen.newMark();
	gen.jumpIf( true, gen.TOKEN_NE, "Ljava/lang/Class;", ifHave );
	gen.pop( "Ljava/lang/Class;", 1 );
	gen.loadConst( name );
	gen.invokestatic( currentName, "c$", "(Ljava/lang/String;)Ljava/lang/Class;" );
	gen.dup( "Ljava/lang/Class;" );
	gen.putstatic( currentName, field, "Ljava/lang/Class;" );
	gen.mark( ifHave );
      }
  }

  /**
   *  Loads an instance of outer class, specified by parameter 'need'
   */
  private void loadHiddenNewArg( AbstractClass need, AbstractClass base,
	     boolean objectIsNull, boolean subclassOK ) throws IOException
  {
    if( need == null )
      return;
    if( need == classTable.objectClass && objectIsNull )
      {
	// special case
	if( base != null )
	  gen.pop( "Ljava/lang/Object;", 1 );
	gen.loadConst( null );
	return;
      }
    if( selfRegister == -1 )
      {
	// static initializer; for now just cheat
	gen.loadConst( null );
	return;
      }

    // System.out.println( "Need: " + need + "; Base: " + base + "; Self:"
    //			+ selfClass + "; Curr: " + currentClass );
    AbstractClass have;
    if( base != null )
      have = base;
    else if( constructorOuterClass != null &&
	     need != selfClass && need != currentClass && 
	    !currentClass.isSubclassOf( need ) )
      {
	// special case: self and current instances might not be initialized yet!
	if( currentClass.getOuter() != constructorOuterClass )
	  throw new CompilerException( ERR_C_NEVER, -1 ); 
	gen.load( 1, constructorOuterClass.getDescriptor() );
	have = constructorOuterClass;
      }
    else if( selfClass != currentClass )
      {
	// nonstatic method
	if( selfClass == need )
	  {
	    gen.load( selfRegister, selfClass.getDescriptor() );
	    return;
	  }
	else
	  {
	    have = currentClass; // start with 'this', not 'self'
	    gen.load( 0, have.getDescriptor() );
	    if( need == have || have.isSubclassOf( need ) )
	      return;
	  }
      }
    else
      {
	have = currentClass;
	gen.load( selfRegister, have.getDescriptor() );
	if( need == have || have.isSubclassOf( need ) )
	  return;
      }

    AbstractClass env = have;
    while( need != have && !( subclassOK && have.isSubclassOf( need ) ) )
      {
	AbstractField fld = have.getOuterField();
	if( fld == null )
	  throw new RuntimeException( "Cannot find outer class " + need + " in " + env );
	have = fld.getType();
	gen.getfield( fld.getDeclaringClass().getJVMName(), fld.getName(),
		      have.getDescriptor() );
      }
  }

  private AbstractClass loadHiddenConstructorArg( 
       AbstractClass scope, AbstractClass target ) throws IOException
  {
    AbstractClass need = target.getOuter();
    if( need == null )
      return null;
    if( need == classTable.objectClass )
      {
	// special case
	gen.loadConst( null );
	return need;
      }
    AbstractClass have = scope.getOuter();
    gen.load( 1, have.getDescriptor() );
    while( need != have && !have.isSubclassOf( need ) )
      {
	AbstractField fld = have.getOuterField();
	have = fld.getType();
	gen.getfield( fld.getDeclaringClass().getJVMName(), fld.getName(),
		      have.getDescriptor() );
      }

    return need;
  }

  void freeRegister( BitSet rmask, int reg, AbstractClass type )
  {
    rmask.clear( reg );
    if( type.getNCells() == 2 )    
      rmask.clear( reg+1 );
  }

  void freeRegister( int reg, AbstractClass type )
  {
    freeRegister( registerMask, reg, type );
  }

  int allocRegister( BitSet rmask, AbstractClass type )
  {
    int i;
    if( type == null || type.getNCells() == 1 )
      {
	for( i = 0 ; rmask.get(i) ; i++ )
	  ;
      }
    else
      {
	i = 1;
	while( true )
	  {
	    if( rmask.get(i) )
	      {
		i += 2;
		continue;
	      }
	    i--;
	    if( rmask.get(i) )
	      {
		i += 2;
		continue;
	      }
	    break;
	  }
	rmask.set(i+1);
      }
    rmask.set(i);
    return i;
  }

  int allocRegister( AbstractClass type )
  {
    return allocRegister( registerMask, type );
  }

  void assignRegister( CompilerVariable var )
  {
    if( reassignRegisterFlag )
      var.setRegister( allocRegister( var.getType() ) );
  }

  void genDefaultConstructorCall( BasicNode node ) throws IOException
  {
    AbstractClass sc = currentClass.getSuperclass();
    AbstractClass sco = sc.getOuter();
    Enumeration constr;
    gen.load( 0, currentClass.getDescriptor() );
    if( sco == null )
      constr = sc.getMethods( INIT, 0 );
    else
      {
	loadHiddenConstructorArg( currentClass, sc );
	constr = sc.getMethods( INIT, 1 );
      }
    if( constr == null || !constr.hasMoreElements() )
      {
	gen.pop( "Ljava/lang/Object;", ( sco == null ? 1 : 2 ) ); // to avoid spurious errors later
	storeError( ERR_C_NOCNSTRWOPRM, node );
      }
    else
      gen.invokespecial( sc.getJVMName(), "<init>", 
			 ((AbstractMethod)constr.nextElement()).getTypeDescriptor() );
  }

  void genInit() throws IOException
  {
    AbstractField outer = currentClass.getOuterField();
    if( outer != null )
      {
	String type = outer.getType().getDescriptor();
	gen.load( 0, currentClass.getDescriptor() );
	gen.load( 1, type );
	gen.putfield( currentClass.getJVMName(), outer.getName(), type );
      }
    if( (currentClass.getModifiers() & METHOD_SCOPE_FLAG) == 0 )
      {
	gen.load( 0, currentClass.getDescriptor() );
	gen.invokespecial( currentClass.getJVMName(), "i$init", "()V" );
      }
  }

  private void genRegexesInit( BitSet rmask, CodeGenerator g, Enumeration elems ) throws IOException
  {
    int forkRegister = -1;
    int cellRegister = -1;
    int maxCells = 0;
    while( elems.hasMoreElements() )
      {
	RegexNode tpn = (RegexNode)elems.nextElement();
	if( tpn.embed )
	  {
	    EmbedRegexAttachment att = new EmbedRegexAttachment();
	    tpn.attachment = att;
	    if( forkRegister < 0 )
	      forkRegister = allocRegister( rmask, classTable.objectClass );
	    att.forkRegister = forkRegister;
	    int needCells = tpn.regexNode.getNCells();
	    if( needCells > 0 )
	      {
		if( cellRegister < 0 )
		  cellRegister = allocRegister( rmask, classTable.objectClass );
		att.cellRegister = cellRegister;
	      }
	    else
	      att.cellRegister = -1;
	    if( needCells > maxCells )
	      maxCells = needCells;
	  }
      }
    if( forkRegister >= 0 )
      {
	g.loadConst( new Integer(4) );
	g.newarray( "I" );
	g.store( forkRegister, "[I" );
      }
    if( cellRegister >= 0 )
      {
	g.loadConst( new Integer(maxCells) );
	g.newarray( "I" );
	g.store( cellRegister, "[I" );
      }
  }

  private void genSubscopes( StatementNode[] statements, final Vector regexList )
  {
    if( statements == null || statements.length == 0 )
      return;

    TraverserContext traverser = new TraverserContext() {
      public Object evalClass( ClassNode classNode )
      {
	CompilerContext.this.evalClass( classNode );
	return classNode;
      }
      public Object evalMethod( MethodNode methodNode )
      {
	if( methodNode.name.idList[0] != CLINIT )
	  CompilerContext.this.evalMethod( methodNode );
	return methodNode;
      }
      public Object evalRegex( RegexNode regexNode )
      {
	regexList.addElement( regexNode );
	return regexNode;
      }
    };

    for( int i = 0 ; i < statements.length ; i++ )
      statements[i].eval( traverser );
  }

  void evalScope( JintClass scope, StatementNode[] statements, BasicNode node )
  {
    boolean privateInit = (scope.getModifiers() & METHOD_SCOPE_FLAG) == 0;
    CodeGenerator oldGen = gen;
    AbstractClass oldClass = currentClass;
    AbstractClass oldSelfClass = selfClass;
    Stack oldLabels = nestedLabels;
    Hashtable oldClassRefs = classRefs;
    BitSet oldRegisterMask = registerMask;
    AbstractClass oldCOC = constructorOuterClass;
    int oldSelfRegister = selfRegister;
    nestedLabels = null;
    gen = null;
    classRefs = null;
    currentClass = scope;
    constructorOuterClass = null;
    selfClass = scope;
    registerMask = null;
    selfRegister = -1;
    try
      {
	String clazzName = scope.getJVMName();
	boolean isInterface = scope.isInterface();
	gen = factory.createCodeGenerator( toJVMFlags( scope ),
			  clazzName, scope.getSuperclass().getJVMName() );
	AbstractClass[] impl = scope.getImplementedInterfaces();
	if( impl == null || impl.length == 0 )
	  gen.setInterfaces( new String[0] );
	else
	  {
	    String[] arr = new String[impl.length];
	    for( int i = 0 ; i < arr.length ; i++ )
	      arr[i] = impl[i].getJVMName();
	    gen.setInterfaces( arr );
	  }
	AbstractField[] fields = scope.getFields();
	for( int i = 0 ; i < fields.length ; i++ )
	  {
	    JintField f = (JintField)fields[i];
	    int accessFlags = toJVMFlags(f.getModifiers());

	    // cannot initialize non-static final fields properly!!!
	    boolean isStatic = (accessFlags & gen.ACC_STATIC) != 0;
	    if( !isStatic )
	       accessFlags &= ~gen.ACC_FINAL;
	    String dscr = f.getType().getDescriptor();
	    String fldname = f.getName();
	    gen.addField( accessFlags, fldname, dscr, f.getConstValue() );
	    String accessor = f.getAccessor();
	    if( accessor != null )
	      {
		gen.startMethod( (isStatic ? gen.ACC_STATIC : 0 ) | gen.ACC_FINAL,
				 accessor,
				 "()" + dscr, null );
		if( isStatic )
		  gen.getstatic( clazzName, fldname, dscr );
		else
		  {
		    gen.load( 0, "L" + clazzName + ";" );
		    gen.getfield( clazzName, fldname, dscr );
		  }
		gen.retrn( dscr );
		gen.endMethod();

		gen.startMethod( (isStatic ? gen.ACC_STATIC : 0 ) | gen.ACC_FINAL,
				 accessor,
				 "(" + dscr + ")V", null );
		if( isStatic )
		  {
		    gen.load( 0, dscr );
		    gen.putstatic( clazzName, fldname, dscr );
		  }
		else
		  {
		    gen.load( 0, "L" + clazzName + ";" );
		    gen.load( 1, dscr );
		    gen.putfield( clazzName, fldname, dscr );
		  }
		gen.retrn( "V" );
		gen.endMethod();
	      }
	  }


	// new location -- check if it's all right to have it here
	Vector regexList = new Vector();
	genSubscopes( statements, regexList );

	AbstractField outer = scope.getOuterField();
	if( privateInit )
	  {
	    boolean needClinit = false;
	    MethodNode clinitMethod = null;
	    registerMask = new BitSet();
	    registerMask.set( 0 ); // self
	    selfRegister = 0;
	    if( ! isInterface )
	      {
		gen.startMethod( gen.ACC_PRIVATE,
				 "i$init", "()V", null );
		genRegexesInit( registerMask, gen, regexList.elements() );
	      }
	    if( statements != null )
	      {
		for( int i = 0 ; i < statements.length ; i++ )
		  {
		    StatementNode s = statements[i];
		    if( s instanceof DefinitionNode )
		      {
			if( s instanceof MemberNode )
			  {
			    if( (((MemberNode)s).getFlags() & STATIC_FLAG) == 0 )
			      if( isInterface )
				storeError( ERR_C_NONSTATINIT, s );
			      else
				s.eval( this );
			    else
			      needClinit = true;
			  }
			if( s instanceof MethodNode && ((MethodNode)s).name.idList[0] == CLINIT )
			  clinitMethod = (MethodNode)s;
		      }
		    else
		      if( isInterface )
			storeError( ERR_C_NONSTATINIT, s );
		      else
			s.eval( this );
		  }
	      }
	    if( ! isInterface )
	      {
		gen.retrn( "V" );
		gen.endMethod();
	      }
	    selfRegister = -1;
	    if( needClinit || clinitMethod != null )
	      {
		gen.startMethod( gen.ACC_STATIC, "<clinit>", "()V", null );
		for( int i = 0 ; i < statements.length ; i++ )
		  {
		    StatementNode s = statements[i];
		    if( s instanceof MemberNode )
		      {
			if( (((MemberNode)s).getFlags() & STATIC_FLAG) != 0 )
			  s.eval( this );
		      }
		  }
		if( clinitMethod == null )
		  gen.retrn( "V" );
		else
		  clinitMethod.eval( this );
		gen.endMethod();
	      }
	    registerMask = null;
	  }
	else
	  genRegexesInit( oldRegisterMask, oldGen, regexList.elements() );

	Enumeration constr = scope.getMethods( INIT, outer == null ? 0 : 1 );
	if( constr != null && constr.hasMoreElements() )
	  {
	    if( isInterface )
	      storeError( ERR_C_INTERFCONSTR, node );
	    AbstractMethod def = (AbstractMethod)constr.nextElement();
	    if( scope.getMethodBody( def.getDescriptor() ) == null )
	      {
		// default constructor
		gen.startMethod( toJVMFlags(def.getModifiers()) & (~gen.ACC_ABSTRACT),
				 "<init>", def.getTypeDescriptor(), null );
		genDefaultConstructorCall( node );
		genInit();
		gen.retrn( "V" );
		gen.endMethod();
	      }
	  }

	if( classRefs != null )
	  {
	    int access = gen.ACC_STATIC|gen.ACC_PRIVATE;
	    Enumeration flist = classRefs.elements();
	    while( flist.hasMoreElements() )
	      gen.addField( access, (String)flist.nextElement(), "Ljava/lang/Class;", null );
	    gen.startMethod( access, "c$", "(Ljava/lang/String;)Ljava/lang/Class;", null );
	    gen.load( 0, "Ljava/lang/String;" );
	    AbstractMark sp = gen.newMark();
	    gen.mark( sp );
	    gen.invokestatic( "java/lang/Class", "forName", "(Ljava/lang/String;)Ljava/lang/Class;" );
	    AbstractMark ep = gen.newMark();
	    gen.mark( ep );
	    gen.retrn( "Ljava/lang/String;" );
	    AbstractMark h = gen.newMark();
	    gen.addExceptionHandler( sp, ep, h, "java/lang/ClassNotFoundException" );
	    gen.mark( h );
	    gen.invokevirtual( "java/lang/Throwable", "getMessage", "()Ljava/lang/String;" );
	    gen.store( 0, "Ljava/lang/String;" );
	    gen.newobject( "java/lang/NoClassDefFoundError" );
	    gen.dup( "Ljava/lang/NoClassDefFoundError;" );
	    gen.load( 0, "Ljava/lang/String;" );
	    gen.invokespecial( "java/lang/NoClassDefFoundError", "<init>", "(Ljava/lang/String;)V" );
	    gen.athrow();
	    gen.endMethod();
	  }
      }
    catch( Exception e )
      {
	internalError( e, node );
      }
    finally
      {
	constructorOuterClass = oldCOC;
	currentClass = oldClass;
	selfClass = oldSelfClass;
	classRefs = oldClassRefs;
	if( gen != null )
	  factory.close( gen, scope.getJVMName() );
	gen = oldGen;
	nestedLabels = oldLabels;
	registerMask = oldRegisterMask;
	selfRegister = oldSelfRegister;
      }
  }

  private void internalError( Exception e, BasicNode node )
  {
    //e.printStackTrace();
    if( e instanceof WrapperException )
      throw (WrapperException)e;
    if( e instanceof CompilerException )
      {
	CompilerException ce = (CompilerException)e;
	storeError( ce );
	if( ce.getFilePos() == -1 )
	  ce.setFilePos( node.getFilePos() );
      }
    else
      storeError( new CompilerException( ERR_C_INTERNAL, node.getFilePos(), e ) );
    throw new WrapperException( e );
  }

  private void evalAndAppend( ExpressionNode expr ) throws IOException
  {
    expr.eval( this );
    AbstractClass clazz = expr.clazz;
    String dscr = clazz.getDescriptor();
    String typeDscr;
    switch( dscr.charAt( 0 ) )
      {
        case 'B' :
        case 'S' :
        case 'I' :
	  typeDscr = "(I)Ljava/lang/StringBuffer;";
	  break;
        case 'Z' :
	  typeDscr = "(Z)Ljava/lang/StringBuffer;";
	  break;
        case 'L' :
        case 'A' :
	  if( clazz == classTable.stringClass )
	    typeDscr = "(Ljava/lang/String;)Ljava/lang/StringBuffer;";
	  else
	    typeDscr = "(Ljava/lang/Object;)Ljava/lang/StringBuffer;";
	  break;
        case 'J' :
	  typeDscr = "(J)Ljava/lang/StringBuffer;";
	  break;
        case 'D' :
	  typeDscr = "(D)Ljava/lang/StringBuffer;";
	  break;
        case 'F' :
	  typeDscr = "(F)Ljava/lang/StringBuffer;";
	  break;
        case 'C' :
	  typeDscr = "(C)Ljava/lang/StringBuffer;";
	  break;
        case '[' :
	  if( dscr.charAt(1) == 'C' )
	    typeDscr = "([C)Ljava/lang/StringBuffer;";
	  else
	    typeDscr = "(Ljava/lang/Object;)Ljava/lang/StringBuffer;";
	  break;
        default:
	  throw new IllegalArgumentException( "Unexpected operand type: " + dscr );
      }
    gen.invokevirtual( "java/lang/StringBuffer", "append", typeDscr );
  }

  void genNextCID() throws IOException
  {
    int cid;
    do
      {
	cid = cacheID.nextInt();
      }
    while( cid == 0 );
    gen.loadConst( new Integer( cid ) );
  }

  void genGetOrPutElement( boolean get, AbstractClass clazz ) throws IOException
  {
    if( clazz.getDimension() > 0 )
      {
	AbstractClass eclass = clazz.getElementClass();
	if( eclass == classTable.anyClass )
	  {
	    if( get )
	      gen.invokestatic( "java/lang/reflect/Array", "get",
			       "(Ljava/lang/Object;I)Ljava/lang/Object;" );
	    else
	      gen.invokestatic( "java/lang/reflect/Array", "set", 
			       "(Ljava/lang/Object;ILjava/lang/Object;)V" );
	  }
	else
	  {
	    String type = eclass.getDescriptor();
	    if( get )
	      gen.getelement( type );
	    else
	      gen.putelement( type );
	  }
      }
    else
      {
	if( clazz.isPrimitive() )
	  throw new CompilerException( ERR_C_PRIMELEM, -1 );
	if( clazz == classTable.vectorClass || clazz.isSubclassOf( classTable.vectorClass ) )
	  {
	    if( get )
	      gen.invokevirtual( "java/util/Vector", "elementAt", "(I)Ljava/lang/Object;" );
	    else
	      {
		gen.swap();
		gen.invokevirtual( "java/util/Vector", "setElementAt", "(Ljava/lang/Object;I)V" );
	      }
	  }
	else if( classTable.abstractListClass != null && 
		 ( clazz == classTable.abstractListClass || 
		   clazz.isSubclassOf( classTable.abstractListClass ) ) )
	  {
	    if( get )
	      gen.invokevirtual( "java/util/AbstractList", "get", "(I)Ljava/lang/Object;" );
	    else
	      {
		gen.invokevirtual( "java/util/AbstractList", "set", 
				 "(ILjava/lang/Object;)Ljava/lang/Object;" );
		gen.pop( "Ljava/lang/Object;", 1 );
	      }
	  }
	else if( clazz == classTable.dictionaryClass || clazz.isSubclassOf( classTable.dictionaryClass ) )
	  {
	    if( get )
	      gen.invokevirtual( "java/util/Dictionary", "get", "(Ljava/lang/Object;)Ljava/lang/Object;" );
	    else
	      {
		gen.invokevirtual( "java/util/Dictionary", "put", 
				 "(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;" );
		gen.pop( "Ljava/lang/Object;", 1 );
	      }
	  }
	else
	  {
	    if( get )
	      gen.invokestatic( ANY_UTILITIES, ANY_GETELEM_METHOD, ANY_GETELEM_METHOD_SIG );
	    else
	      gen.invokestatic( ANY_UTILITIES, ANY_PUTELEM_METHOD, ANY_PUTELEM_METHOD_SIG );
	  }
      }
  }

  public int intValue( ConstNode cn )
  {
    Object v = cn.val;
    if( v instanceof Character )
      return ((Character)v).charValue();
    return ((Number)v).intValue();
  }

  public void evalVarList( AbstractVariable[] iavars, ExpressionNode node, 
			   boolean autoPrefix ) throws IOException
  {
    if( iavars == null )
      {
	if( baseType != null )
	  return;

	if( contextState != STATE_LOAD )
	  storeError( ERR_C_THISASSIGN, node );

	loadHiddenNewArg( node.clazz, null, false, false );

	/*
	else if( selfClass != node.clazz )
	  internalError( new RuntimeException( 
		"Class mismatch " + selfClass + " " + node.clazz), node );
	else
	  gen.load( selfRegister, selfClass.getDescriptor() );
	*/

	return;
      }
    
    int last = iavars.length - 1;
    
    if( contextState == STATE_STORE )
      ((CompilerVariable)iavars[last]).genPut( gen, this );
    else if( contextState == STATE_LOAD2 )
      ((CompilerVariable)iavars[last]).genGet( gen, this );
    else
      {
	int l = last;
	if( baseType == null )
	  {
	    AbstractClass needed;
	    if( l < 0 )
	      needed = node.clazz;
	    else
	      needed = ((CompilerVariable)iavars[0]).getBase( this );
	    if( needed != null && needed != classTable.voidType && autoPrefix )
	      loadHiddenNewArg( needed, null, false, false /*### was true */ );
	  }
	if( contextState == STATE_BASE )
	  l--;
	for( int i = 0 ; i <= l ; i++ )
	  ((CompilerVariable)iavars[i]).genGet( gen, this );
      }
  }

  public void embedRegexMatch( RegexNode tpn, EmbedRegexAttachment att,
    boolean fullBuffer ) throws IOException
  {
    registerMask = (BitSet)registerMask.clone();
    RLocalAllocator varalloc = new RLocalAllocator() {
      public int alloc()
      {
	return allocRegister( classTable.intType );
      }
      public void free( int local )
      {
	freeRegister( local, classTable.intType );
      }
    };
    int[] commVar = new int[RJavaClassMachine.NVARS];

    AbstractMark fMark = gen.newMark();
    AbstractMark tMark = gen.newMark();
    
    RJavaClassMachine jcm = new RJavaClassMachine();
    
    // must be on stack
    commVar[jcm.I_STRING] = allocRegister( classTable.objectClass );
    gen.store( commVar[jcm.I_STRING], "[C" );			

    // start of string
    if( fullBuffer )
      commVar[jcm.I_START] = -1; // is always 0, no need for register
    else
      {
	// must be on stack
	commVar[jcm.I_START] = allocRegister( classTable.intType );
	gen.store( commVar[jcm.I_START], "I" );
      }
    
    commVar[jcm.I_END] = allocRegister( classTable.intType );
    if( fullBuffer )
      {
	gen.load( commVar[jcm.I_STRING], "[C" );
	gen.arraylength();
      }
    else
      {
	// must be on stack
      }
    gen.store( commVar[jcm.I_END], "I" );
					
    // position where to try to match first
    commVar[jcm.I_BEGIN] = allocRegister( classTable.intType );
    if( tpn.regexNode.isEndAnchored() && tpn.regexNode.maxLeft < 2048 )
      {
	gen.load( commVar[jcm.I_END], "I" );
	gen.loadConst( new Integer( tpn.regexNode.maxLeft ) );
	gen.op( '-', "I" );
	gen.dup( "I" );
	AbstractMark skip = gen.newMark();
	if( commVar[jcm.I_START] < 0 )
	  gen.jumpIf( true, gen.TOKEN_GE, "I", skip );
	else
	  {
	    gen.load( commVar[jcm.I_START], "I" );
	    gen.jumpIf( false, gen.TOKEN_GE, "I", skip );
	  }
	gen.pop( "I", 1 );
	if( commVar[jcm.I_START] < 0 )
	  gen.loadConst( I_ZERO );
	else
	  gen.load( commVar[jcm.I_START], "I" );
	gen.mark( skip );
      }
    else
      {
	if( commVar[jcm.I_START] < 0 )
	  gen.loadConst( I_ZERO );
	else
	  gen.load( commVar[jcm.I_START], "I" );
      }
    gen.store( commVar[jcm.I_BEGIN], "I" );
    
    // last position where to try to match 
    if( tpn.regexNode.isStartAnchored() )
      {
	commVar[jcm.I_MAXSTART] = commVar[jcm.I_START];
	gen.load( commVar[jcm.I_BEGIN], "I" );
	if( commVar[jcm.I_START] < 0 )
	  {
	    gen.jumpIf( true, '>', "I", fMark );
	  }
	else
	  {
	    gen.load( commVar[jcm.I_MAXSTART], "I" );
	    gen.jumpIf( false, '>', "I", fMark );
	  }
      }
    else
      {
	if( tpn.regexNode.minLeft == 0 )
	  commVar[jcm.I_MAXSTART] = commVar[jcm.I_END];
	else
	  {
	    commVar[jcm.I_MAXSTART] = allocRegister( classTable.intType );
	    gen.load( commVar[jcm.I_END], "I" );
	    gen.loadConst( new Integer( tpn.regexNode.minLeft ) );
	    gen.op( '-', "I" );
	    gen.store( commVar[jcm.I_MAXSTART], "I" );
	  }
      }
    
    commVar[jcm.I_HEAD] = allocRegister( classTable.intType );
    gen.loadConst( I_ZERO );
    gen.store( commVar[jcm.I_HEAD], "I" );
    
    commVar[jcm.I_CELLS] = att.cellRegister;
    int tmpVar = allocRegister( classTable.intType );
    int nCells = tpn.regexNode.getNCells();
    if( nCells > 0 )
      {
	gen.loadConst( new Integer( nCells-1 ) );
	gen.store( tmpVar, "I" );
	AbstractMark tmpMark = gen.newMark();
	gen.mark( tmpMark );
	gen.load( att.cellRegister, "[I" );
	gen.load( tmpVar, "I" );
	gen.loadConst( I_MONE );
	gen.putelement( "I" );
	gen.iinc( tmpVar, -1 );
	gen.load( tmpVar, "I" );
	gen.jumpIf( true, '<', "I", tmpMark );
      }
    freeRegister( tmpVar, classTable.intType );
    
    commVar[jcm.I_FORKS] = att.forkRegister;
    
    commVar[jcm.I_FORKPTR] = allocRegister( classTable.intType );
    gen.loadConst( I_ZERO );
    gen.store( commVar[jcm.I_FORKPTR], "I" );
    
    int[] extVars;
    AbstractMark initRefForward = null;
    AbstractMark initRefBack = null;
    
    int initStackDepth = gen.getStackDepth();
    
    if( tpn.refList == null )
      extVars = null;
    else
      {
	extVars = new int[tpn.refList.length];
	initRefForward = gen.newMark();
	initRefBack = gen.newMark();
	for( int i = 0 ; i < extVars.length ; i++ )
	  extVars[i] = allocRegister( classTable.intType );
	gen.jump( initRefForward );
	gen.mark( initRefBack, initStackDepth );
      }
    
    jcm.embed( true, gen, varalloc, commVar, extVars, fMark, tMark );
    
    RCompiler rcomp = new RCompiler( jcm );
    rcomp.compile( tpn.regexNode, tpn.str );
    
    if( tpn.refList != null )
      {
	int v_cells = commVar[jcm.I_CELLS];
	gen.mark( initRefForward );
	String[] refList = tpn.refList;
	for( int i = 0 ; i < refList.length ; i++ )
	  {
	    String var = refList[i];
	    if( ! Character.isLetter(var.charAt(0)) )
	      continue;
	    AbstractVariable[] av = (AbstractVariable[])tpn.vars.get(var);
	    int begin = jcm.getVariableHandle( var, true );
	    int end = jcm.getVariableHandle( var, false );
	    int extCell = jcm.getExtVariableHandle( var );
	    AbstractClass type = ((CompilerVariable)av[av.length-1]).getType();
	    if( type == classTable.charStringClass )
	      {
		String cscn = classTable.charStringClass.getJVMName();
		String cscd = classTable.charStringClass.getDescriptor();
		evalVarList( av, tpn, true );
		int tmpr = allocRegister( classTable.intType );
		gen.store( tmpr, cscd );

		gen.load( v_cells, "[I" );
		gen.loadConst( new Integer( begin ) );
		gen.load( tmpr, cscd );
		gen.getfield( cscn, "first", "I" );
		gen.putelement( "I" );

		gen.load( v_cells, "[I" );
		gen.loadConst( new Integer( end ) );
		gen.load( tmpr, cscd );
		gen.getfield( cscn, "last", "I" );
		gen.putelement( "I" );

		gen.load( tmpr, cscd );
		gen.getfield( cscn, "buf", "[C" );
		gen.store( extVars[extCell], "[C" );

		freeRegister( tmpr, classTable.intType );
	      }
	    else
	      {
		gen.load( v_cells, "[I" );
		gen.loadConst( new Integer( begin ) );
		gen.loadConst( I_ZERO );
		gen.putelement( "I" );
		gen.load( v_cells, "[I" );
		gen.loadConst( new Integer( end ) );
		contextState = STATE_LOAD;
		evalVarList( av, tpn, true );
		if( type.isPrimitive() )
		  gen.invokestatic( "java/lang/String", "valueOf", 
				    "(" + type.getDescriptor() + ")Ljava/lang/String;" );
		else if( type == classTable.charTypeArray )
		    ; // nothing
		else
		  {
		    if( type != classTable.stringClass )
		      gen.invokevirtual( "java/lang/Object", "toString", "()Ljava/lang/String;" );
		    gen.invokevirtual( "java/lang/String", "toCharArray", "()[C" );
		  }
		gen.dup( "[C" );
		gen.store( extVars[extCell], "[C" );
		gen.arraylength();
		gen.putelement( "I" );
	      }
	  }
	gen.jump( initRefBack );
      }
    
    gen.mark( tMark );
    
    if( tpn.pickList != null )
      pickRegexVars( tpn, jcm, commVar[jcm.I_STRING], commVar[jcm.I_CELLS], false );
    
    gen.loadConst( Boolean.TRUE );
    AbstractMark cont = gen.newMark();
    gen.jump( cont );
    gen.mark( fMark );
    gen.loadConst( Boolean.FALSE );
    gen.mark( cont );
  }

  void pickRegexVars( RegexNode tpn, RJavaClassMachine jcm, 
	int v_buf, int v_cells, boolean ignoreCase ) throws IOException
  {
    Hashtable vars = tpn.vars;
    String[] pickList = tpn.pickList;
    for( int i = 0 ; i < pickList.length ; i++ )
      {
	String var = pickList[i];
	if( ! Character.isLetter(var.charAt(0)) )
	  continue;
	AbstractVariable[] av = (AbstractVariable[])vars.get(var);
	int begin = jcm.getVariableHandle( var, true );
	int end = jcm.getVariableHandle( var, false );
	contextState = STATE_BASE;
	evalVarList( av, tpn, true );
	AbstractClass type = ((CompilerVariable)av[av.length-1]).getType();
	if( type == classTable.anyClass )
	  type = classTable.charStringClass;
	if( type != classTable.stringClass && type != classTable.charStringClass )
	  throw new CompilerException( ERR_C_PICKEDNONSTR, tpn.getFilePos() );
	String pcn = type.getJVMName();
	String pcd = type.getDescriptor();
	gen.newobject( pcn );
	gen.dup( pcd );
	if( ignoreCase )
	  {
	    AbstractClass circ = classTable.caseInsensitiveRegexClass;
	    gen.load( 0, "L" + jcm.getClassName() + ";" );
	    gen.getfield( jcm.getClassName(), "frontEnd", circ.getDescriptor() );
	    gen.dup( circ.getDescriptor() );
	    gen.loadConst( new Integer(-1) );
	    gen.invokevirtual( circ.getJVMName(), "getCharBuffer", "(I)[C" );
	    gen.swap();
	    gen.loadConst( new Integer( begin ) );
	    gen.invokevirtual( circ.getJVMName(), "getIndex", "(I)I" );
	    gen.dup( "I" );
	    gen.load( 0, "L" + jcm.getClassName() + ";" );
	    gen.getfield( jcm.getClassName(), "frontEnd", circ.getDescriptor() );
	    gen.loadConst( new Integer( end ) );
	    gen.invokevirtual( circ.getJVMName(), "getIndex", "(I)I" );
	  }
	else
	  {
	    gen.load( v_buf, "[C" );
	    gen.load( v_cells, "[I" );
	    gen.loadConst( new Integer( begin ) );
	    gen.getelement( "I" );
	    gen.dup( "I" );
	    gen.load( v_cells, "[I" );
	    gen.loadConst( new Integer( end ) );
	    gen.getelement( "I" );
	  }
	gen.swap();
	gen.op( '-', "I" );
	gen.invokespecial( pcn, "<init>", "([CII)V" );
	contextState = STATE_STORE;
	evalVarList( av, tpn, true );
	contextState = STATE_LOAD;
      }
  }

  void loadFormat( String formStr, boolean bound ) throws IOException
  {
    String fnm = classTable.formatterClass.getJVMName();
    String fds = classTable.formatterClass.getDescriptor();
    String name = "F$" + formatCount++;
    gen.addField( gen.ACC_STATIC|gen.ACC_PRIVATE, name, fds, null );
    gen.getstatic( currentClass.getJVMName(), name, fds );
    gen.dup( fds );
    AbstractMark nonNul = gen.newMark();
    gen.jumpIf( true, TOKEN_NE, fds, nonNul );
    gen.pop( fds, 1 );
    gen.newobject( fnm );
    gen.dup( fds );
    gen.loadConst( formStr );
    gen.loadConst( new Boolean( bound ) );
    gen.invokespecial( fnm, "<init>", "(Ljava/lang/String;Z)V" );
    gen.dup( fds );
    gen.putstatic( currentClass.getJVMName(), name, fds );
    gen.mark( nonNul );
  }

  void genTextProcessing( AbstractClass arg1class, ExpressionNode arg2, boolean subst )
    throws IOException
  {
    if( arg2 instanceof RegexNode && ((RegexNode)arg2).embed )
      {
	// embeded regex
	RegexNode tpn = (RegexNode)arg2;
	EmbedRegexAttachment att = (EmbedRegexAttachment)tpn.attachment;
	
	BitSet savedRegisterMask = registerMask;
	try
	  {
	    if( arg1class == classTable.charStringClass || arg1class == classTable.anyClass )
	      {
		gen.dup( "Ljava/lang/Object;" );
		gen.instnceof( classTable.charStringClass.getJVMName() );
		AbstractMark cont = gen.newMark();
		AbstractMark fin = gen.newMark();
		AbstractMark fin1 = gen.newMark();
		gen.jumpIf( true, TOKEN_EE, "Z", cont );
		gen.cast( "Ljava/lang/Object;", classTable.charStringClass.getDescriptor() );
		String cscd = classTable.charStringClass.getDescriptor();
		String cscn = classTable.charStringClass.getJVMName();
		int tmpr = allocRegister( classTable.charStringClass );
		gen.store( tmpr, cscd );
		gen.load( tmpr, cscd );
		gen.getfield( cscn, "last", "I" );
		gen.load( tmpr, cscd );
		gen.getfield( cscn, "first", "I" );
		gen.load( tmpr, cscd );
		gen.getfield( cscn, "buf", "[C" );
		freeRegister( tmpr, classTable.charStringClass );
		gen.jump( fin );
		gen.mark( cont );
		gen.dup( "Ljava/lang/Object;" );
		gen.instnceof( "[C" );
		cont = gen.newMark();
		gen.jumpIf( true, TOKEN_EE, "Z", cont );
		gen.cast( "Ljava/lang/Object;", "[C" );
		gen.jump( fin1 );
		gen.mark( cont );
		gen.invokevirtual( "java/lang/Object", "toString", "()Ljava/lang/String;" );
		gen.invokevirtual( "java/lang/String", "toCharArray", "()[C" );
		gen.mark( fin1 );
		gen.dup( "[C" );
		gen.arraylength();
		gen.swap();
		gen.loadConst( I_ZERO );
		gen.swap();
		gen.mark( fin );
		embedRegexMatch( tpn, att, false );
	      }
	    else
	      {
		if( arg1class == classTable.stringClass )
		  gen.invokevirtual( "java/lang/String", "toCharArray", "()[C" );
		else if( arg1class != classTable.charStringClass )
		  throw new CompilerException( ERR_C_MATCHARG1, -1 );
		embedRegexMatch( tpn, att, true );
	      }
	  }
	finally
	  {
	    registerMask = savedRegisterMask;
	  }
      }
    else
      {
	arg2.eval( this );
	if( arg2.clazz == classTable.anyClass )
	  {
	    if( subst )
	      gen.cast( classTable.objectClass.getDescriptor(), classTable.replacerClass.getDescriptor() );
	    else
	      {
		gen.invokestatic( ANY_UTILITIES, ANY_TEXTSEARCH_METHOD, ANY_TEXTSEARCH_METHOD_SIG );
		return;
	      }
	  }
	gen.swap();
	if( arg2.clazz == classTable.regexClass || arg2.clazz.isSubclassOf(classTable.regexClass) )
	  {
	    gen.invokevirtual( classTable.regexClass.getJVMName(), "searchOnce",
				"(" + arg1class.getDescriptor() + ")Z" );
	  }
	else
	  {
	    gen.invokevirtual( classTable.replacerClass.getJVMName(), "process",
				"(" + arg1class.getDescriptor() + ")" +
				arg1class.getDescriptor() );
	  }
      }
  }

  private void loadArgs( AbstractMethod method, int first, ExpressionNode[] args )
    throws IOException
  {
    AbstractClass[] paramTypes = method.getParameterTypes();
    if( paramTypes != null )
      {
	int last = paramTypes.length - 1;
	if( last >= 0 && paramTypes[last] == classTable.arglistClass &&
	    ( paramTypes.length != args.length || 
	      args[last].clazz != classTable.arglistClass ) )
	  last--; // ArgList parameters
	for( int i = first ; i <= last ; i++ )
	  evalAndWrapOrCast( args[i-first], paramTypes[i] );
	if( last != paramTypes.length - 1 )
	  {
	    gen.newobject( classTable.arglistClass.getJVMName() );
	    gen.dup( classTable.arglistClass.getDescriptor() );
	    int varArgs = args.length - last - 1 + first;
	    genArgArray( args, last+1-first, varArgs );
	    gen.invokespecial( classTable.arglistClass.getJVMName(), "<init>",
			       "([Ljava/lang/Object;)V" );
	  }
      }
  }

  private void evalStatements( StatementNode[] statements )
  {
    if( statements != null )
      {
	for( int i = 0 ; i < statements.length ; i++ )
	  if( ! (statements[i] instanceof MethodNode || statements[i] instanceof ClassNode ) )
	    statements[i].eval( this );
      }
  }

  private void genAnyMethodCall( ExpressionNode func, ExpressionNode call,
     AnyMethod method, ExpressionNode[] args ) throws IOException
  {
    Identifier name = method.getIdentifier();
    String methodSignature;

    if( name == INIT || (method.getModifiers() & STATIC_FLAG) != 0 )
      {
	// func cannot be null for static AnyMethod
	// call cannot be null for "constructor" AnyMethod
	AbstractClass clazz = ( name == INIT ? call.clazz : func.clazz );
	gen.loadConst( null );
	genClass( clazz );
	methodSignature = ANY_CALL_METHOD_FULL_SIG;
      }
    else
      {
	if( func instanceof IdentifierNode && func.attachment == null )
	  loadHiddenNewArg( classTable.anyClass, null, false, true );
	else
	  func.eval( this );
	methodSignature = ANY_CALL_METHOD_SIG;
      }

    gen.loadConst( name.toString() );
    if( args == null )
      {
	gen.loadConst( null );
	gen.loadConst( null );
      }
    else
      {
	genArgClassArray( args, 0, args.length );
	genArgArray( args, 0, args.length );
      }
    genNextCID();
    gen.invokestatic( ANY_UTILITIES, ANY_CALL_METHOD, methodSignature );
  }

  //------------ Context implementation

  public Object evalArrayValue( ArrayValueNode arrayValueNode )
  {
    try
      {
	AbstractClass clazz = arrayValueNode.clazz.getElementClass();
	String descr = (clazz == classTable.anyClass ? 
			"Ljava/lang/Object;" : clazz.getDescriptor());
	ExpressionNode[] vals = arrayValueNode.vals;
	int length = vals.length;
	gen.loadConst( new Integer( length ) );
	gen.newarray( descr );
	String arrdscr = "[" + descr;
	for( int i = 0 ; i < length ; i++ )
	  {
	    gen.dup( arrdscr );
	    gen.loadConst( new Integer( i ) );
	    evalAndWrapOrCast( vals[i], clazz );
	    gen.putelement( descr );
	  }
      }
    catch( Exception e )
      {
	internalError( e, arrayValueNode );
      }
    return null;
  }

  public Object evalBody( BodyNode bodyNode )
  {
    LabelInfo myInfo = nestedLabelsPush( bodyNode );
    try
      {
	evalStatements( bodyNode.statements );
	if( myInfo.finishMark != null )
	  gen.mark( myInfo.finishMark );
      }
    catch( Exception e )
      {
	internalError( e, bodyNode );
      }
    finally
      {
	nestedLabelsPop( bodyNode );
      }
    return null;
  }

  public Object evalBreak( BreakNode breakNode )
  {
    try
      {
	gen.linenumber( breakNode.getLine() );
	LabeledNode labelOwner = (LabeledNode)breakNode.attachment;
	if( labelOwner != null )
	  jumpOut( breakNode, labelOwner.statement );
	else
	  jumpOut( breakNode, null );
      }
    catch( Exception e )
      {
	internalError( e, breakNode );
      }
    return null;
  }

  public Object evalCall( CallNode callNode )
  {
    try
      {
	ExpressionNode[] args = callNode.args;
	
	Object attachment = callNode.attachment;
	
	CallAttachment localCall = null;
	
	if( attachment instanceof CallAttachment )
	  {
	    localCall = (CallAttachment)attachment;
	    attachment = localCall.method;
	  }

	ExpressionNode func = callNode.func;
	
	if( attachment instanceof AnyMethod )
	  {
	    genAnyMethodCall( func, null, (AnyMethod)attachment, args );
	  }
	else
	  {
	    AbstractMethod method = (AbstractMethod)attachment;
	    AbstractClass targetClass = method.getDeclaringClass();
	    boolean special = ( localCall != null && localCall.targetClass != null );
	    int first = 0;
	    if( method.getName() != "<init>" )
	      {
		if( ( method.getModifiers() & DefinitionNode.STATIC_FLAG ) == 0 )
		  {
		    if( func instanceof IdentifierNode && func.attachment == null )
		      {
			// System.out.println( "Calling " + method );
			loadHiddenNewArg( targetClass, null, false, true );
		      }
		    else
		      func.eval( this );
		  }
		else
		  evalAndPop( func );
	      }
	    if( special )
	      {
		//this is this() or super() or super.method() construction
		targetClass = localCall.targetClass;
		if( method.getName() == "<init>" )
		  {
		    gen.load( 0, currentClass.getDescriptor() );
		    if( loadHiddenConstructorArg( currentClass, targetClass ) != null )
		      first = 1;
		  }
	      }
	    loadArgs( method, first, args );
	    String cl = targetClass.getJVMName();
	    String nm = method.getName();
	    String ds = method.getTypeDescriptor();
	    int modifiers = method.getModifiers();
	    if( (modifiers & DefinitionNode.STATIC_FLAG) != 0 && !special )
	      gen.invokestatic( cl, nm, ds );
	    else if( special || 
		    ( ( modifiers & DefinitionNode.PRIVATE_FLAG ) != 0 && 
		        currentClass == targetClass ) )
	      gen.invokespecial( cl, nm, ds );
	    else if( targetClass.isInterface() )
	      gen.invokeinterface( cl, nm, ds );
	    else
	      gen.invokevirtual( cl, nm, ds );
	  }
      }
    catch( Exception e )
      {
	internalError( e, callNode );
      }
    return null;
  }

  public Object evalCase( CaseNode caseNode )
  {
    evalStatements( caseNode.statements );
    return null;
  }

  public Object evalCast( CastNode castNode )
  {
    evalAndWrapOrCast( castNode.arg, castNode.clazz, true );
    return null;
  }

  public Object evalCatch( CatchNode catchNode )
  {
    throw new CompilerException( ERR_C_NEVER, catchNode.getFilePos() );
  }

  public Object evalClass( ClassNode classNode )
  {
    try
      {
	if( classNode.regex == null )
	  {
	    JintClass scope = (JintClass)classNode.attachment;
	    evalScope( scope, classNode.statements, classNode );
	  }
	else
	  {
	    JintClass scope = (JintClass)classNode.regex.attachment;
	    AbstractClass oldClass = currentClass;
	    AbstractClass oldSelfClass = selfClass;
	    int oldSelfRegister = selfRegister;
	    currentClass = scope;
	    selfClass = scope;
	    selfRegister = 0;
	    try
	      {
		// System.out.println( "---------->" + selfClass );
		compileRegexToGivenClass = true;
		classNode.regex.eval( this );
	      }
	    finally
	      {
		compileRegexToGivenClass = false;
		currentClass = oldClass;
		selfClass = oldSelfClass;
		selfRegister = oldSelfRegister;
	      }
	  }
      }
    catch( Exception e )
      {
	internalError( e, classNode );
      }
    return null;
  }

  public Object evalCond( CondNode condNode )
  {
    try
      {
	AbstractClass type = condNode.clazz;
	AbstractMark mark1 = gen.newMark();
	evalCond( condNode.cond, null, mark1 );
	evalAndWrapOrCast( condNode.arg1, type );
	AbstractMark mark2 = gen.newMark();
	gen.jump( mark2 );
	gen.mark( mark1 );
	evalAndWrapOrCast( condNode.arg2, type );
	gen.mark( mark2 );
      }
    catch( Exception e )
      {
	internalError( e, condNode );
      }
    return null;
  }

  public Object evalConst( ConstNode constNode )
  {
    try
      {
	gen.loadConst( constNode.val );
      }
    catch( Exception e )
      {
	internalError( e, constNode );
      }
    return null;
  }

  public Object evalContinue( ContinueNode continueNode )
  {
    try
      {
	gen.linenumber( continueNode.getLine() );
	LabeledNode labelOwner = (LabeledNode)continueNode.attachment;
	if( labelOwner != null )
	  jumpOut( continueNode, labelOwner.statement );
	else
	  jumpOut( continueNode, null );
      }
    catch( Exception e )
      {
	internalError( e, continueNode );
      }
    return null;
  }

  public Object evalDo( DoNode doNode )
  {
    LabelInfo myInfo = nestedLabelsPush( doNode );
    try
      {
	gen.linenumber( doNode.getLine() );
 	AbstractMark mark = gen.newMark();
	gen.mark( mark );
	if( doNode.body != null )
	  doNode.body.eval( this );
	if( myInfo.againMark != null )
	  gen.mark( myInfo.againMark );
	evalCond( doNode.cond, mark, null );
	if( myInfo.finishMark != null )
	  gen.mark( myInfo.finishMark );
      }
    catch( Exception e )
      {
	internalError( e, doNode );
      }
    finally
      {
	nestedLabelsPop( doNode );
      }
    return null;
  }

  public Object evalExpressionStatement( ExpressionStatementNode esNode )
  {
    try
      {
	gen.linenumber( esNode.getLine() );
      }
    catch( Exception e )
      {
	internalError( e, esNode );
      }
    if( esNode.printResult )
      {
	try
	  {
	    AbstractClass clazz = esNode.expr.clazz;
	    if( clazz != classTable.voidType )
	      {
		gen.getstatic( "java/lang/System", "out", "Ljava/io/PrintStream;" );
		esNode.expr.eval( this );
		String dscr;
		if( clazz != null && ( clazz.isPrimitive() ||
				       clazz == classTable.stringClass ) )
		  dscr = "(" + clazz.getDescriptor() + ")V";
		else 
		  dscr = "(Ljava/lang/Object;)V";
		gen.invokevirtual( "java/io/PrintStream", "println", dscr );
	      }
	    else
	      evalExpr( esNode.expr );
	  }
	catch( Exception e )
	  {
	    internalError( e, esNode );
	  }
      }
    else
      evalAndPop( esNode.expr );
    return null;
  }

  public Object evalFor( ForNode forNode )
  {
    LabelInfo myInfo = nestedLabelsPush( forNode );
    try
      {
	gen.linenumber( forNode.getLine() );
	if( forNode.init != null )
	  forNode.init.eval( this );
	AbstractMark mark = gen.newMark();
	if( forNode.body != null || forNode.next != null )
	  {
	    AbstractMark fmark = gen.newMark();
	    gen.jump( fmark );
	    gen.mark( mark, 0 );
	    if( forNode.body != null )
	      forNode.body.eval( this );
	    if( myInfo.againMark != null )
	      gen.mark( myInfo.againMark );
	    gen.linenumber( forNode.getLine() );
	    evalAndPop( forNode.next );
	    gen.mark( fmark );
	  }
	else
	  {
	    gen.linenumber( forNode.getLine() );
	    gen.mark( mark );
	  }
	if( forNode.cond == null )
	  gen.jump( mark );
	else
	  evalCond( forNode.cond, mark, null );
	if( myInfo.finishMark != null )
	  gen.mark( myInfo.finishMark );
      }
    catch( Exception e )
      {
	internalError( e, forNode );
      }
    finally
      {
	nestedLabelsPop( forNode );
      }
    return null;
  }

  public Object evalFormat( FormatNode formatNode )
  {
    try
      {

	loadFormat( formatNode.formStr, formatNode.bound );

	if( formatNode.bound )
	  {
	    AbstractVariable[][] vars = (AbstractVariable[][])formatNode.attachment;
	    gen.newobject( classTable.arglistClass.getJVMName() );
	    gen.dup( classTable.arglistClass.getDescriptor() );
	    if( vars == null )
	      {
		gen.loadConst( null );
	      }
	    else
	      {
		gen.loadConst( new Integer( vars.length ) );
		gen.newarray( "Ljava/lang/Object;" );
		for( int i = 0 ; i < vars.length ; i++ )
		  {
		    gen.dup( "[Ljava/lang/Object;" );
		    gen.loadConst( new Integer( i ) );
		    IdentifierNode psi = new IdentifierNode( formatNode.getFilePos(), 
							     (Identifier)null );
		    AbstractVariable[] varsi = vars[i];
		    psi.attachment = varsi;
		    AbstractVariable last = varsi[varsi.length-1];
		    psi.clazz = last.getType();
		    evalAndWrapOrCast( psi, classTable.anyClass );
		    gen.putelement( "Ljava/lang/Object;" );
		  }
	      }
	    gen.invokespecial( classTable.arglistClass.getJVMName(), "<init>",
			       "([Ljava/lang/Object;)V" );
	    gen.invokevirtual( classTable.formatterClass.getJVMName(), "sprintf", 
	      "(" + classTable.arglistClass.getDescriptor() + ")Ljava/lang/String;" );
	  }
      }
    catch( Exception e )
      {
	internalError( e, formatNode );
      }
    return null;
  }

  public Object evalIdentifier( IdentifierNode identifierNode )
  {
    AbstractVariable[] iavars = (AbstractVariable[])identifierNode.attachment;

    try
      {
	evalVarList( iavars, identifierNode, true );
      }
    catch( Exception e )
      {
	internalError( e, identifierNode );
      }
    return null;
  }

  public Object evalIf( IfNode ifNode )
  {
    try
      {
	gen.linenumber( ifNode.getLine() );
	AbstractMark mark1 = gen.newMark();
	evalCond( ifNode.cond, null, mark1 );
	if( ifNode.thenPart != null )
	  ifNode.thenPart.eval( this );
	if( ifNode.elsePart != null )
	  {
	    AbstractMark mark2 = gen.newMark();
	    gen.jump( mark2 );
	    gen.mark( mark1 );
	    ifNode.elsePart.eval( this );
	    gen.mark( mark2 );
	  }
	else
	  gen.mark( mark1 );
      }
    catch( Exception e )
      {
	internalError( e, ifNode );
      }
    return null;
  }

  public Object evalImportStatement( ImportStatementNode importStatement )
  {
    return null;
  }

  public Object evalInfix( InfixNode infixNode )
  {
    boolean trash = trashTheValue;
    trashTheValue = false;
    AbstractClass common = (AbstractClass)infixNode.attachment;
    int op = infixNode.op;
    ExpressionNode arg1 = infixNode.arg1;
    int storeState = contextState;
    try
      {
	boolean commonAny = common == classTable.anyClass;
	switch( op )
	  {
	    case '~' :
	      arg1.eval( this );
	      genTextProcessing( arg1.clazz, infixNode.arg2, false );
	      break;
	    case TOKEN_TE :
	      {
		contextState = STATE_BASE;
		arg1.eval( this );
		contextState = STATE_LOAD;
		String baseType = baseType( arg1 );
		if( baseType != "V" )
		  gen.dup( baseType );
		contextState = STATE_LOAD2;
		arg1.eval( this );
		contextState = STATE_LOAD;
		genTextProcessing( arg1.clazz, infixNode.arg2, true );
		if( !trash )
		  {
		    if( baseType == "V" )
		      gen.dup( arg1.clazz.getDescriptor() );
		    else
		      gen.insert( arg1.clazz.getDescriptor(), baseType );
		  }
		contextState = STATE_STORE;
		arg1.eval( this );
		contextState = STATE_LOAD;
		if( trash )
		  return TRASHED;
	      }
	      break;
	    case TOKEN_INSTANCEOF :
	      {
		arg1.eval( this );
		gen.instnceof( infixNode.arg2.clazz.getJVMName() );
	      }
	      break;
	    case '=' :
	      {
		contextState = STATE_BASE;
		arg1.eval( this );
		contextState = STATE_LOAD;
		AbstractClass varType = infixNode.clazz;
		evalAndWrapOrCast( infixNode.arg2, varType );
		if( ! trash )
		  {
		    String under = baseType( arg1 );
		    if( under == "V" )
		      gen.dup( varType.getDescriptor() );
		    else
		      gen.insert( varType.getDescriptor(), under );
		  }
		contextState = STATE_STORE;
		arg1.eval( this );
		if( trash )
		  return TRASHED;
	      }
	      break;
	    case '.' :
	      {
		if( contextState == STATE_LOAD || contextState == STATE_BASE )
		  {
		    contextState = STATE_LOAD;
		    arg1.eval( this );
		    contextState = storeState;
		  }
		try
		  {
		    baseType = arg1.clazz;
		    infixNode.arg2.eval( this );
		  }
		finally
		  {
		    baseType = null;
		  }
	      }
	      break;
	    case '[' :
	      {
		AbstractClass arrClass = arg1.clazz;
		if( contextState == STATE_STORE )
		  genGetOrPutElement( false, arrClass );
		else if( contextState == STATE_LOAD2 )
		  genGetOrPutElement( true, arrClass );
		else
		  {
		    contextState = STATE_LOAD;
		    arg1.eval( this );
		    if( arrClass.getDimension() > 0 || 
			( classTable.abstractListClass != null ?
			     arrClass == classTable.abstractListClass || 
			  arrClass.isSubclassOf( classTable.abstractListClass ) :
			     arrClass == classTable.vectorClass || 
			     arrClass.isSubclassOf( classTable.vectorClass )) )
		      evalAndWrapOrCast( infixNode.arg2, classTable.intType );
		    else
		      evalAndWrapOrCast( infixNode.arg2, classTable.objectClass );
		    contextState = storeState;
		    if( contextState == STATE_LOAD )
		      genGetOrPutElement( true, arrClass );
		  }
	      }
	      break;
	    case TOKEN_AND :
	      {
		AbstractMark mark1 = gen.newMark();
		AbstractMark mark2 = gen.newMark();
		evalCond( infixNode, null, mark1 );
		gen.loadConst( I_ONE );
		gen.jump( mark2 );
		gen.mark( mark1 );
		gen.loadConst( I_ZERO );
		gen.mark( mark2 );
	      }
	      break;
	    case TOKEN_NE :
	    case TOKEN_EE :
	    case TOKEN_GE :
	    case TOKEN_LE :
	    case '>' :
	    case '<' :
	    case TOKEN_OR :
	      {
		if( (common == classTable.longType || common == classTable.floatType ||
		     common == classTable.doubleType ) )
		  {
		    // op cannot be TOKEN_OR because common would be booleanType
		    evalAndWrapOrCast( arg1, common );
		    evalAndWrapOrCast( infixNode.arg2, common );
		    gen.compare( false, common.getDescriptor(), 
				 ( op == '<' || op == TOKEN_LE ) ); // -1 0 1
		    switch( op )
		      {
		        case TOKEN_NE :
			  gen.loadConst( I_ONE );
			  gen.op( '&', "I" ); // 1 0 1
			  break;
		        case TOKEN_EE :
			  gen.loadConst( I_ONE );
			  gen.op( '+', "I" ); // 0 1 2
			  gen.loadConst( I_ONE );
			  gen.op( '&', "I" ); // 0 1 0
			  break;
		        case '>' :
			  gen.loadConst( I_ONE );
			  gen.op( '+', "I" ); // 0 1 2
			  gen.loadConst( I_ONE );
			  gen.op( TOKEN_GG, "I" ); // 0 0 1
			  break;
		        case '<' :
			  gen.loadConst( I_ONE );
			  gen.op( TOKEN_GG, "I" ); // -1 0 0
			  gen.loadConst( I_ONE );
			  gen.op( '&', "I" ); // 1 0 0
			  break;
		        case TOKEN_GE :
			  gen.loadConst( I_ONE );
			  gen.op( TOKEN_GG, "I" ); // -1 0 0
			  gen.loadConst( I_ONE );
			  gen.op( '+', "I" ); // 0 1 1
			  break;
		        case TOKEN_LE :
			  gen.op( UNARY_MINUS, "I" ); // 1 0 -1
			  gen.loadConst( I_ONE );
			  gen.op( TOKEN_GG, "I" ); // 0 0 -1
			  gen.loadConst( I_ONE );
			  gen.op( '+', "I" ); // 1 1 0
			  break;
		        default :
			  throw new CompilerException( ERR_C_NEVER, infixNode.getFilePos() );
		      }
		  }
		else
		  {
		    AbstractMark mark1 = gen.newMark();
		    AbstractMark mark2 = gen.newMark();
		    evalCond( infixNode, mark1, null );
		    gen.loadConst( I_ZERO );
		    gen.jump( mark2 );
		    gen.mark( mark1 );
		    gen.loadConst( I_ONE );
		    gen.mark( mark2 );
		  }
	      }
	      break;
	    case ',' :
	      {
		evalAndPop( arg1 );
		infixNode.arg2.eval( this );
	      }
	      break;
	    case TOKEN_GG :
	    case TOKEN_LL :
	    case TOKEN_GGG :
	      if( commonAny )
		gen.loadConst( new Integer( op ) );
	      evalAndWrapOrCast( arg1, common );
	      evalAndWrapOrCast( infixNode.arg2, classTable.intType );
	      if( commonAny )
		gen.invokestatic( ANY_UTILITIES, ANY_SHIFT_METHOD, ANY_SHIFT_METHOD_SIG );
	      else
		gen.op( op, common.getDescriptor() );
	      break;
	    case '|' :
	    case '&' :
	    case '^' :
	    case '+' :
	    case '-' :
	    case '*' :
	    case '/' :
	    case '%' :
	      {
		if( common == classTable.stringClass )
		  {
		    Vector sbNodes = new Vector();
		    InfixNode infn = infixNode;
		    do
		      {
			sbNodes.addElement( infn );
			if( infn.arg1 instanceof InfixNode )
			  {
			    InfixNode infn1 = (InfixNode)infn.arg1;
			    if( infn1.op != '+' )
			      break;
			    infn = infn1;
			  }
			else
			  break;
		      }
		    while( infn.op == '+' && infn.attachment == classTable.stringClass );

		    gen.newobject( "java/lang/StringBuffer" );
		    gen.dup( "Ljava/lang/StringBuffer;" );
		    gen.invokespecial( "java/lang/StringBuffer", "<init>", "()V" );
		    int i = sbNodes.size() - 1;
		    evalAndAppend( infn.arg1 );
		    infn = (InfixNode)sbNodes.elementAt(i);
		    for( ; i >= 0 ; i-- )
		      {
			infn = (InfixNode)sbNodes.elementAt(i);
			evalAndAppend( infn.arg2 );
		      }
		    gen.invokevirtual( "java/lang/StringBuffer", "toString",
				       "()Ljava/lang/String;" );
		  }
		else
		  {
		    if( commonAny )
		      gen.loadConst( new Integer( op ) );
		    evalAndWrapOrCast( arg1, common );
		    evalAndWrapOrCast( infixNode.arg2, common );
		    if( commonAny )
		      gen.invokestatic( ANY_UTILITIES, ANY_OP2_METHOD, ANY_OP2_METHOD_SIG );
		    else
		      gen.op( op, common.getDescriptor() );
		  }
	      }
	      break;
	    case TOKEN_GGE :
	    case TOKEN_GGGE :
	    case TOKEN_LLE :
	    case TOKEN_ORE :
	    case TOKEN_ANDE :
	    case TOKEN_XE :
	    case TOKEN_PE :
	    case TOKEN_ME :
	    case TOKEN_AE :
	    case TOKEN_SE :
	    case TOKEN_PCE :
	      return opEq( infixNode.op, arg1, infixNode.arg2, common, true, trash );
	    default:
	      {
		Object[] addInfo = { "'" + (char)infixNode.op + "'" };
		throw new CompilerException( ERR_C_BADOP, infixNode.getFilePos(), addInfo );
	      }
	  }
      }
    catch( Exception e )
      {
	internalError( e, infixNode );
      }
    finally
      {
	contextState = storeState;
      }
    return null;
  }

  public Object evalLabeled( LabeledNode labeledNode )
  {
    labeledNode.statement.eval( this );
    return null;
  }

  public Object evalMember( MemberNode memberNode )
  {
    // just eval initializations
    ExpressionNode[] defs = memberNode.defs;
    for( int i = 0 ; i < defs.length ; i++ )
      if( defs[i] instanceof InfixNode )
        {
	  try
	    {
	      gen.linenumber( defs[i].getLine() );
	    }
	  catch( IOException e )
	    {
	    }
	  evalAndPop( defs[i] );
	}
      else if( defs[i] instanceof IdentifierNode )
	{
	  try
	    {
	      AbstractVariable[] iavars = 
		(AbstractVariable[])((IdentifierNode)defs[i]).attachment;
	      ((CompilerVariable)iavars[0]).genInit( gen, this );
	    }
	  catch( Exception e )
	    {
	      internalError( e, memberNode );
	    }
	}
    return null;
  }

  public Object evalMethod( MethodNode methodNode )
  {
    boolean hasError = false;
    BitSet oldRegisterMask = registerMask;
    int oldSelfRegister = selfRegister;
    AbstractClass oldSelfClass = selfClass;
    JintMethod method = (JintMethod)methodNode.name.attachment;
    boolean noBody = (methodNode.getFlags() & (ABSTRACT_FLAG|NATIVE_FLAG)) != 0;
    try
      {
	registerMask = new BitSet();
	AbstractClass[] exc = method.getExceptionTypes();
	String[] exceptionList;
	if( exc == null )
	  exceptionList = null;
	else
	  {
	    exceptionList = new String[exc.length];
	    for( int i = 0 ; i < exc.length ; i++ )
	      exceptionList[i] = exc[i].getJVMName();
	  }
	if( methodNode.name.idList[0] != CLINIT )
	  gen.startMethod( toJVMFlags(method.getModifiers()),
			 method.getName(), method.getTypeDescriptor(),
			 exceptionList );
	if( noBody )
	  return null;
	// TODO: implicit chained constructor call for constructors
	int maxRegister = 0;
	if( (method.getModifiers() & STATIC_FLAG) == 0 )
	  {
	    registerMask.set(0);
	    maxRegister = 1;
	  }

	if( method.getName() == INIT.toString() && currentClass.getOuter() != null )
	  {
	    // hidden parameter for the constructor of inner class
	    registerMask.set(1);
	    maxRegister = 2;
	    constructorOuterClass = currentClass.getOuter();
	  }

	int firstParamRegister = maxRegister;

	// inhibit register re-assignment during parameter variable initialization
	reassignRegisterFlag = false;

	ParameterNode[] params = methodNode.params;

 	if( params != null )
	  for( int i = 0 ; i < params.length ; i++ )
	    {
	      IdentifierNode pname = params[i].name;
	      CompilerVariable var = (CompilerVariable)pname.attachment;
	      AbstractClass type = pname.clazz;
	      int size = type.getNCells();
	      var.getBase( this ); // initialize variable
	      if( var.getRegister() >= 0 )
		{
		  if( var.getRegister() != maxRegister )
		    {
		      // System.out.println( methodNode );
		      throw new CompilerException( ERR_C_NEVER,
						   methodNode.name.getFilePos() );
		    }
		  registerMask.set( maxRegister );
		  if( size > 1 )
		    registerMask.set( maxRegister+1 );
		}
	      maxRegister += size;
	    }
	
	reassignRegisterFlag = true;
 
	JintClass scope = (JintClass)methodNode.attachment;
	String scopeDescriptor = scope.getDescriptor();
	StatementNode[] stmts = methodNode.body;

	boolean hasSelfClass = scope.mustBeClass();

	if( hasSelfClass )
	  {
	    selfClass = scope;
	    selfRegister = maxRegister;
	    registerMask.set( selfRegister );
	    gen.newobject( scope.getJVMName() );
	    gen.dup( scopeDescriptor );
	    Enumeration initList;
	    if( (method.getModifiers() & STATIC_FLAG) == 0 )
	      {
		gen.load( 0, currentClass.getDescriptor() );
		initList = scope.getMethods( INIT, 1 );
	      }
	    else
	      initList = scope.getMethods( INIT, 0 );
	    AbstractMethod init = (AbstractMethod)initList.nextElement();
	    gen.invokespecial( scope.getJVMName(), "<init>", init.getTypeDescriptor() );
	    gen.store( selfRegister, scopeDescriptor );
	    evalScope( scope, stmts, methodNode );
	  }
	else 
	  {
	    if( (method.getModifiers() & STATIC_FLAG) == 0 )
	      {
		selfClass = currentClass;
		selfRegister = 0;
	      }
	    else
	      {
		selfClass = null;
		selfRegister = -1;
	      }

	    // in the other branch evalScope takes care of that
	    Vector regexList = new Vector();
	    genSubscopes( stmts, regexList );
	    genRegexesInit( registerMask, gen, regexList.elements() );
	  }

	maxRegister = firstParamRegister;

	if( params != null )
	  for( int i = 0 ; i < params.length ; i++ )
	    {
	      IdentifierNode pname = params[i].name;
	      CompilerVariable var = (CompilerVariable)pname.attachment;
	      AbstractClass type = pname.clazz;
	      if( var.getRegister() < 0 )
		{
		  gen.load( selfRegister, scopeDescriptor );
		  gen.load( maxRegister, type.getDescriptor() );
		  var.genPut( gen, this );
		  // register(s) available for re-use
		}
	      maxRegister += type.getNCells();
	    }

	int firstStmt = 0;

	if( method.getName() == INIT.toString() )
	  {
	    boolean hasChainConstructorCall = false;
	    if( stmts != null && stmts.length > 0 && 
		stmts[0] instanceof ExpressionStatementNode )
	      {
		ExpressionStatementNode esn = (ExpressionStatementNode)(stmts[0]);
		if( esn.expr instanceof CallNode )
		  {
		    CallNode callNode = (CallNode)esn.expr;
		    if( callNode.attachment instanceof CallAttachment )
		      {
			CallAttachment localCall = (CallAttachment)callNode.attachment;
			if( localCall.targetClass != null )
			  {
			    hasChainConstructorCall = true;
			    firstStmt = 1;
			    esn.eval( this );
			  }
		      }
		  }
	      }

	    if( ! hasChainConstructorCall )
	      genDefaultConstructorCall( methodNode );

	    genInit();
	  }

	if( stmts != null )
	  {
	    for( int i = firstStmt ; i < stmts.length ; i++ )
	      {
		StatementNode s = stmts[i];
		if( ! ( s instanceof MethodNode || s instanceof ClassNode ) )
		  s.eval( this );
	      }
	  }
      }
    catch( Exception e )
      {
	hasError = true;
	internalError( e, methodNode );
      }
    finally
      {
	constructorOuterClass = null;
	try
	  {
	    if( gen.isReachable() && !noBody )
	      {
		AbstractClass t = method.getType();
		if( t == classTable.anyClass )
		  {
		    gen.loadConst( null );
		    gen.retrn ( "Ljava/lang/Object;" );
		  }
		else
		  {
		    if( t != classTable.voidType && !hasError )
		      storeError( ERR_C_RETURNREQ, methodNode );
		    gen.retrn( "V" );
		  }
	      }
	  }
	catch( Exception e )
	  {
	    internalError( e, methodNode );
	  }
	try
	  {
	    if( methodNode.name.idList[0] != CLINIT )
	      gen.endMethod();
	  }
 	catch( Exception e )
	  {
	    internalError( e, methodNode );
	  }
	registerMask = oldRegisterMask;
	selfRegister = oldSelfRegister;
	selfClass = oldSelfClass;
	reassignRegisterFlag = true;
      }
    return null;
  }

  public Object evalNew( NewNode newNode )
  {
    AbstractClass aClass = newNode.clazz;
    try
      {
	if( aClass.getDimension() == 0 )
	  {
	    NewAttachment ina = (NewAttachment)newNode.attachment;
	    ExpressionNode[] args = newNode.args;
	    AbstractMethod method = ina.method;
	    if( aClass == classTable.anyClass )
	      {
		// well, there is no Any class per se, but we can produce
		// DispatchObject, I guess
		gen.newobject( DISPATCH_OBJECT );
		gen.dup( "L" + DISPATCH_OBJECT + ";" );
		gen.invokespecial( DISPATCH_OBJECT, "<init>", "()V" );
	      }
	    else if( method instanceof AnyMethod )
	      {
		genAnyMethodCall( null, newNode, (AnyMethod)method, args );
	      }
	    else
	      {
		gen.newobject( aClass.getJVMName() );
		gen.dup( aClass.getDescriptor() );
		if( newNode.base == null )
		  {
		    // System.out.println( "Calling " + method );
		    loadHiddenNewArg( aClass.getOuter(), null, true, true );
		  }
		else
		  {
		    newNode.base.eval( this );
		    loadHiddenNewArg( aClass.getOuter(), newNode.base.clazz, true, true );
		  }
		int first = ( aClass.getOuter() == null ? 0 : 1 );
		loadArgs( ina.method, first, args );
		gen.invokespecial( aClass.getJVMName(),
				   "<init>", method.getTypeDescriptor() );
	      }
	    if( ina.addMethod != null )
	      {
		gen.dup( aClass.getDescriptor() );
		gen.load( selfRegister, selfClass.getDescriptor() );
		gen.swap();
		gen.invokevirtual( ina.addMethod.getDeclaringClass().getJVMName(),
				   ina.addMethod.getName(),
				   ina.addMethod.getTypeDescriptor() );
		if( ina.addMethod.getType() != classTable.voidType )
		  gen.pop( ina.addMethod.getType().getDescriptor(), 1 );
	      }
	  }
	else
	  {
	    ExpressionNode[] indexes = newNode.type.indexes;
	    
	    if( indexes.length >= 2 && indexes[1] != null )
	      {
		int i;
		for( i = 0 ; i < indexes.length && indexes[i] != null ; i++ )
		  indexes[i].eval( this );
		gen.newmultiarray( aClass.getDescriptor(), i );
	      }
	    else
	      {
		indexes[0].eval( this );
		gen.newarray( aClass.getElementClass().getDescriptor() );
	      }
	  }
      }
    catch( Exception e )
      {
	internalError( e, newNode );
      }
    return null;
  }

  public Object evalPackageStatement( PackageStatementNode packageStatement )
  {
    return null;
  }

  public Object evalParameter( ParameterNode parameterNode )
  {
    return null;
  }

  public Object evalPostfix( PostfixNode postfixNode )
  {
    boolean trash = trashTheValue;
    trashTheValue = false;
    try
      {
	switch( postfixNode.op )
	  {
	    case TOKEN_PP :
	    case TOKEN_MM :
	      return opEq( postfixNode.op, postfixNode.arg, null,
			   postfixNode.clazz, false, trash );
	    default:
	      {
		Object[] addInfo = { "'" + (char)postfixNode.op + "'" };
		throw new CompilerException( ERR_C_BADOP, postfixNode.getFilePos(), addInfo );
	      }
	  }
      }
    catch( Exception e )
      {
	internalError( e, postfixNode );
      }
    return null;
  }

  public Object evalPrefix( PrefixNode prefixNode )
  {
    boolean trash = trashTheValue;
    trashTheValue = false;
    try
      {
	AbstractClass clazz = prefixNode.clazz;
	int op = prefixNode.op;
	ExpressionNode arg = prefixNode.arg;
	if( clazz == classTable.anyClass && op != TOKEN_MM && op != TOKEN_PP )
	  {
	    gen.loadConst( new Integer( op ) );
	    arg.eval( this );
	    gen.invokestatic( ANY_UTILITIES, ANY_OP1_METHOD, ANY_OP1_METHOD_SIG );
	    return null;
	  }
	switch( op )
	  {
	    case TOKEN_MM :
	    case TOKEN_PP :
	      return opEq( op, arg, null, prefixNode.clazz, true, trash );
	    case '-' :
	      arg.eval( this );
	      gen.op( gen.UNARY_MINUS, clazz.getDescriptor() );
	      break;
	    case '~' :
	      arg.eval( this );
	      gen.op( '~', clazz.getDescriptor() );
	      break;
	    case '+' :
	      arg.eval( this );
	      break;
	    case '!' :
	      {
		AbstractMark mark1 = gen.newMark();
		AbstractMark mark2 = gen.newMark();
		evalCond( arg, mark1, null ); 
		gen.loadConst( I_ONE );
		gen.jump( mark2 );
		gen.mark( mark1 );
		gen.loadConst( I_ZERO );
		gen.mark( mark2 );
	      }
	      break;
	    default:
	      {
		Object[] addInfo = { "'" + (char)prefixNode.op + "'" };
		throw new CompilerException( ERR_C_BADOP, prefixNode.getFilePos(), addInfo );
	      }
	  }
      }
    catch( Exception e )
      {
	internalError( e, prefixNode );
      }
    return null;
  }

  public Object evalReturn( ReturnNode returnNode )
  {
    try
      {
	gen.linenumber( returnNode.getLine() );
	AbstractClass methodType = null; 
	if( needToJumpOut() )
	  {
	    if( returnNode.result != null && returnNode.result.clazz != null )
	      {
		CompilerVariable cv = (CompilerVariable)returnNode.attachment;
		methodType = cv.getType();
		evalAndWrapOrCast( returnNode.result, methodType );
		cv.genPut( gen, this );
		jumpOut( returnNode, null );
		cv.genGet( gen, this );
	      }
	    else
	      {
		methodType = classTable.objectClass; // works for "return null"
		jumpOut( returnNode, null );
		gen.loadConst( null );
	      }
	  }
	else
	  {
	    if( returnNode.result != null )
	      {
		if( returnNode.attachment instanceof AbstractVariable )
		  methodType = ((CompilerVariable)returnNode.attachment).getType();
		else
		  methodType = (AbstractClass)returnNode.attachment;
		evalAndWrapOrCast( returnNode.result, methodType );
	      }
	  }
	String descr;
	if( returnNode.result == null )
	  descr = "V";
	else
	  descr = methodType.getDescriptor();
	gen.retrn( descr );
      }
    catch( Exception e )
      {
	internalError( e, returnNode );
      }
    return null;
  }

  public Object evalSwitch( SwitchNode switchNode )
  {
    LabelInfo myInfo = nestedLabelsPush( switchNode );
    try
      {
	gen.linenumber( switchNode.getLine() );
	CaseNode[] cases = switchNode.cases;
	evalExpr( switchNode.expr );
	AbstractClass type = switchNode.expr.clazz;
	if( type == classTable.intType || type == classTable.charType ||
	     type == classTable.shortType || type == classTable.byteType )
	  {
	    CaseNode defaultCase = null;
	    boolean allConst = true;
	    int minValue = Integer.MAX_VALUE;
	    int maxValue = Integer.MIN_VALUE;
	    for( int i = 0 ; i < cases.length ; i++ )
	      {
		ExpressionNode expr = cases[i].expr;
		if( expr == null )
		  {
		    if( defaultCase != null )
		      throw new CompilerException( ERR_C_TWODEFAULT,
						   defaultCase.getFilePos() );
		    defaultCase = cases[i];
		    continue;
		  }
		AbstractClass exprType = expr.clazz;
		if( exprType != classTable.intType && exprType != classTable.charType &&
		   exprType != classTable.shortType && exprType != classTable.byteType )
		  storeError( ERR_C_CASEMISMATCH, expr );
		if( expr instanceof ConstNode )
		  {
		    int label = intValue( (ConstNode)expr );
		    if( minValue > label )
		      minValue = label;
		    if( maxValue < label )
		      maxValue = label;
		  }
		else
		  allConst = false;
	      }
	    int tableSize = cases.length;
	    if( defaultCase != null )
	      tableSize--;
	    if( allConst && tableSize > 1 )
	      {
		// all are integer constants
		if( maxValue - minValue < 2*tableSize + 10 )
		  {
		    // go for table switch
		    tableSize = maxValue - minValue + 1;
		    AbstractMark[] markTable = new AbstractMark[ tableSize ];
		    AbstractMark[] markTable1 = new AbstractMark[ tableSize ];
		    for( int i = 0 ; i < tableSize ; i++ )
		      markTable[i] = gen.newMark();
		    System.arraycopy( markTable, 0, markTable1, 0, tableSize);
		    for( int i = 0 ; i < cases.length ; i++ )
		      {
			ConstNode cn = (ConstNode)cases[i].expr;
			if( cn != null )
			  {
			    int index = intValue( cn ) - minValue;
			    if( markTable1[index] == null )
			      throw new CompilerException( ERR_C_DUPCASE, cases[i].getFilePos() );
			    markTable1[index] = null;
			  }
		      } 

		    AbstractMark defaultMark = gen.newMark();
		    gen.tableswitch( minValue, markTable, defaultMark );
		    for( int i = 0 ; i < cases.length ; i++ )
		      {
			ConstNode cn = (ConstNode)cases[i].expr;
			if( cn == null )
			  {
			    if( defaultMark == null )
			      throw new CompilerException( ERR_C_TWODEFAULT, cases[i].getFilePos() );
			    gen.mark( defaultMark );
			    for( int p = 0 ; p < tableSize ; p++ )
			      if( markTable1[p] != null )
				gen.mark( markTable1[p] );
			    defaultMark = null;
			  }
			else
			  {
			    int index = intValue( cn ) - minValue;
			    gen.mark( markTable[index] );
			  }
			cases[i].eval( this );
		      }
		    if( defaultMark != null )
		      {
			gen.mark( defaultMark );
			for( int i = 0 ; i < tableSize ; i++ )
			  if( markTable1[i] != null )
			    gen.mark( markTable1[i] );
		      }
		    return null;
		  }
		else
		  {
		    // go for lookup switch
		    AbstractMark[] markTable = new AbstractMark[ tableSize ];
		    AbstractMark[] labelTable = new AbstractMark[ tableSize ];
		    int[] labels = new int[ tableSize ];
		    int j = 0;
		    for( int i = 0 ; i < cases.length ; i++ )
		      {
			ConstNode cn = (ConstNode)cases[i].expr;
			if( cn == null )
			  continue;
			AbstractMark m = gen.newMark();
			markTable[j] = m;
			labelTable[j] = m;
			labels[j] = intValue( cn );
			j++;
		      }

		    for( int i = 0 ; i < tableSize ; i++ )
		      for( j = i+1 ; j < tableSize ; j++ )
			{
			  if( labels[j] < labels[i] )
			    {
			      int t = labels[i];
			      labels[i] = labels[j];
			      labels[j] = t;
			      AbstractMark m = labelTable[i];
			      labelTable[i] = labelTable[j];
			      labelTable[j] = m;
			    }
			}

		    AbstractMark defaultMark = gen.newMark();
		    gen.lookupswitch( labels, labelTable, defaultMark );

		    j = 0;
		    for( int i = 0 ; i < cases.length ; i++ )
		      {
			if( cases[i].expr == null )
			  {
			    gen.mark( defaultMark );
			    defaultMark = null;
			  }
			else
			  gen.mark( markTable[j++] );
			cases[i].eval( this );
		      }
		    if( defaultMark != null )
		      gen.mark( defaultMark );
		    return null;
		  }
	      }
	  }

	// some labels are not constants or not an integer switch
	if( type == classTable.charStringClass )
	  {
	    type = classTable.stringClass;
	    gen.invokevirtual( "java/lang/Object", "toString", "()Ljava/lang/String;" );
	  }
	int last = cases.length-1;
	String dscr = type.getDescriptor();
	boolean primitive = type.isPrimitive();
	boolean string = type == classTable.stringClass;
	AbstractMark next = null;
	AbstractMark fallThrough = null;
	AbstractMark defaultMark = null;
	for( int i = 0 ; i <= last ; i++ )
	  {
	    ExpressionNode expr = cases[i].expr;
	    if( expr == null )
	      {
		if( defaultMark != null )
		  throw new CompilerException( ERR_C_TWODEFAULT, cases[i].getFilePos() );
		if( i == last )
		  {
		    if( next != null )
		      {
			gen.mark( next );
			next = null;
		      }
		  }
		else
		  {
		    // don't touch 'next' variable
		    defaultMark = gen.newMark();
		    gen.mark( defaultMark, 0 );
		  }
		if( fallThrough != null )
		  gen.mark( fallThrough );
		fallThrough = ( i != last ? gen.newMark() : null );
		cases[i].eval( this );
		if( fallThrough != null )
		  gen.jump( fallThrough );
		continue;
	      }
	    if( next != null )
	      gen.mark( next );
	    if( i != last )
	      {
		if( i != last-1 || cases[i+1].expr != null )
		  gen.dup( dscr );
		next = gen.newMark();
	      }
	    else
	      if( defaultMark != null )
		next = defaultMark;
	      else
		{
		  if( myInfo.finishMark == null )
		    myInfo.finishMark = gen.newMark();
		  next = myInfo.finishMark;
		}
	      
	    if( primitive )
	      {
		evalAndWrapOrCast( expr, type );
		gen.jumpIf( false, TOKEN_NE, dscr, next );
	      }
	    else
	      {
		if( expr.clazz == classTable.regexClass )
		  {
		    genTextProcessing( type, expr, false );
		  }
		else if( string )
		  {
		    evalAndWrapOrCast( expr, type );
		    gen.invokevirtual( "java/lang/String", "equals", "(Ljava/lang/Object;)Z" );
		  }
		else
		  {
		    gen.loadConst( new Integer( TOKEN_EE ) );
		    gen.swap();
		    evalAndWrapOrCast( expr, classTable.anyClass );
		    gen.invokestatic( ANY_UTILITIES, ANY_COND_METHOD, ANY_COND_METHOD_SIG );
		  }
		gen.jumpIf( true, TOKEN_EE, "Z", next );
	      }
	    if( i != last && ( i != last-1 || cases[i+1].expr != null ) )
	      gen.pop( dscr, 1 );
	    if( fallThrough != null )
	      gen.mark( fallThrough );
	    fallThrough = ( i != last ? gen.newMark() : null );
	    cases[i].eval( this );
	    if( fallThrough != null )
	      gen.jump( fallThrough );
	  }
	if( next != null )
	  gen.mark( next );
      }
    catch( Exception e )
      {
	internalError( e, switchNode );
      }
    finally
      {
	if( myInfo.finishMark != null )
	  try
	    {
	      gen.mark( myInfo.finishMark );
	    }
	  catch( Exception e )
	    {
	      internalError( e, switchNode );
	    }
	nestedLabelsPop( switchNode );
      }
    return null;    
  }

  public Object evalSynchronized( SynchronizedNode synchronizedNode )
  {
    AbstractClass type = synchronizedNode.lock.clazz;
    int tmpRegister = allocRegister( type );
    LabelInfo myInfo = nestedLabelsPush( synchronizedNode );
    myInfo.syncLock = tmpRegister;
    try
      {
	gen.linenumber( synchronizedNode.getLine() );
	String dscr = type.getDescriptor();
	evalExpr( synchronizedNode.lock );
	gen.dup( dscr );
	gen.store( tmpRegister, "Ljava/lang/Object;" );
	gen.monitorenter();
	AbstractMark begin = gen.newMark();
	gen.mark( begin );
	synchronizedNode.body.eval( this );
	AbstractMark end = gen.newMark();
	gen.mark( end, true ); // may be unreachable
	boolean endReachable = gen.isReachable();
	AbstractMark fin = null;
	if( endReachable )
	  {
	    fin = gen.newMark();
	    gen.load( tmpRegister, "Ljava/lang/Object;" );
	    gen.monitorexit();
	    gen.jump( fin );
	  }
	AbstractMark handler = gen.newMark();
	gen.addExceptionHandler( begin, end, handler, "java/lang/Throwable" );
	gen.mark( handler );
	gen.load( tmpRegister, "Ljava/lang/Object;" );
	gen.monitorexit();
	gen.athrow();
	if( endReachable )
	  gen.mark( fin );
      }
    catch( Exception e )
      {
	internalError( e, synchronizedNode );
      }
    finally
      {
	nestedLabelsPop( synchronizedNode );
	registerMask.clear( tmpRegister );
      }
    return null;    
  }

  public Object evalRegex( RegexNode regexNode )
  {
    final RegexNode tpn = regexNode;
    if( tpn.embed )
      throw new CompilerException( ERR_C_NEVER, regexNode.getFilePos() );
    CodeGenerator oldGen = gen;

    try
      {
	final boolean ignoreCase = (regexNode.flags & TOK_OPT_IGNORE) != 0;
	final AbstractClass circ = classTable.caseInsensitiveRegexClass;

	boolean hasPicks1 = false;
	final String[] pickList = tpn.pickList;
	final String[] refList = tpn.refList;
	final Hashtable vars = tpn.vars;

	final boolean compileClass = compileRegexToGivenClass;

	if( pickList != null )
	  {
	    if( compileClass )
	      hasPicks1 = true;
	    else
	      {
		for( int i = 0 ; i < pickList.length ; i++ )
		  {
		    if( vars.get( pickList[i] ) != null )
		      {
			hasPicks1 = true;
			break;
		      }
		  }
	      }
	  }
	
	final boolean hasPicks = hasPicks1;
	final String className = regexNode.clazz.getJVMName();
	final String regexType = "L" + className + ";";
	final String constrDscr = ( !hasPicks || (tpn.flags & TOK_OPT_STANDALONE) != 0 ? "()V" :
				        "(" + tpn.clazz.getOuter().getDescriptor() + ")V" );
	int modifiers = toJVMFlags( regexNode.clazz );
	if( regexNode.clazz.getLexicallyOuter() == null )
	  modifiers &= ~gen.ACC_STATIC;
	gen = factory.createCodeGenerator( modifiers,
				   className, classTable.regexStubClass.getJVMName() );
	final RJavaClassMachine jcm = new RJavaClassMachine();
	jcm.setCodeGenerator( gen );
	jcm.setLoadClass( false );
	jcm.setClassName( className );
	jcm.setSaveBytecode( false );
	jcm.setReportPosition( true );
	jcm.setNoRefiller( (regexNode.flags & TOK_OPT_BUFFERONLY) != 0 );
	RJavaClassCustomizer customizer = new RJavaClassCustomizer() {
          public void customMembers( CodeGenerator gen ) throws IOException
	  {
	    if( (tpn.flags & TOK_OPT_STANDALONE) != 0 )
	      {
		AbstractField[] fields = ((JintClass)tpn.clazz).getFields();
		for( int i = 0 ; i < fields.length ; i++ )
		  {
		    JintField f = (JintField)fields[i];
		    int accessFlags = toJVMFlags(f.getModifiers());
		    String dscr = f.getType().getDescriptor();
		    String fldname = f.getName();
		    gen.addField( accessFlags, fldname, dscr, f.getConstValue() );
		  }
	      }
	    if( hasPicks )
	      {
		if( (tpn.flags & TOK_OPT_STANDALONE) == 0 )
		  gen.addField( 0, tpn.clazz.getOuterField().getName(),
				tpn.clazz.getOuter().getDescriptor(), null );
		if( ignoreCase )
		  gen.addField( 0, "frontEnd", circ.getDescriptor(), null );
	      }
	  }
          public void customSuccessAction( CodeGenerator gen, int v_buf, int v_cells )
	    throws IOException
	  {
            //System.out.println( "selfClass = " + selfClass );
	    if( tpn.pickList != null )
	      pickRegexVars( tpn, jcm, v_buf, v_cells, ignoreCase );
	  }
          public String customConstructorSignature()
	  {
	    return constrDscr;
	  }
          public void customConstructorAction( CodeGenerator gen ) throws IOException
          {
	    if( !hasPicks || (tpn.flags & TOK_OPT_STANDALONE) != 0 )
	      return;
	    String myDscr = tpn.clazz.getDescriptor();
	    String outerDscr = tpn.clazz.getOuter().getDescriptor();
	    gen.load( 0, myDscr );
	    gen.load( 1, outerDscr );
	    gen.putfield( className, tpn.clazz.getOuterField().getName(), outerDscr );	    
          }
	};
	jcm.setCustomizer( customizer );
	RCompiler rcomp = new RCompiler( jcm );

	AbstractClass saveSelf = selfClass;
	AbstractClass saveCurrent = currentClass;
	int saveRegister = selfRegister;
	try
	  {
	    selfClass = regexNode.clazz;
	    currentClass = regexNode.clazz;
	    selfRegister = 0;
	    rcomp.compile( tpn.regexNode, tpn.str );
	    factory.close( gen, className );
	  }
	finally
	  {
	    gen = oldGen;
	    selfClass = saveSelf;
	    currentClass = saveCurrent;
	    selfRegister = saveRegister;
	  }

	if( compileClass )
	  {
	    if( ignoreCase )
	      throw new RuntimeException( "Not implemented" );
	    return null;
	  }
	
	if( ignoreCase )
	  {
	    gen.newobject( circ.getJVMName() );
	    gen.dup( circ.getDescriptor() );
	  }

	gen.newobject( className );
	gen.dup( regexType );
	if( hasPicks && (tpn.flags & TOK_OPT_STANDALONE) == 0 )
	  loadHiddenNewArg( tpn.clazz.getOuter(), null, false, false );
	gen.invokespecial( className, "<init>", constrDscr );

	if( ignoreCase )
	  {
	    if( hasPicks )
	      gen.dup( "2" );
	    gen.invokespecial( circ.getJVMName(), "<init>", 
			      "(" + classTable.regexClass.getDescriptor() + ")V" );
	    if( hasPicks )
	      {
		gen.swap();
		// now we have on stack: ignoreCaseRegex customRegex ignoreCaseRegex
		gen.putfield( className, "frontEnd", circ.getDescriptor() );
	      }
	  }

	// TODO: move it to constructor
	if( refList != null )
	  {
	    for( int i = 0 ; i < refList.length ; i++ )
	      {
		String vname = refList[i];
		AbstractVariable[] av = (AbstractVariable[])vars.get( vname );
		if( av != null && av.length > 0 )
		  {
		    gen.dup( regexType );
		    gen.loadConst( new Integer( jcm.getExtVariableHandle( vname ) ) );
		    gen.loadConst( new Integer( jcm.getVariableHandle( vname, true ) ) );
		    gen.loadConst( new Integer( jcm.getVariableHandle( vname, false ) ) );
		    evalVarList( av, tpn, true );
		    AbstractClass t = av[av.length-1].getType();
		    if( t.isPrimitive() )
		      {
			gen.invokestatic( "java/lang/String", "valueOf", 
					 "(" + t.getDescriptor() + ")Ljava/lang/String;" );
			t = classTable.stringClass;
		      }
		    else if( t != classTable.stringClass && t != classTable.charStringClass
			    && t != classTable.charTypeArray )
		      t = classTable.objectClass;
		    gen.invokevirtual( classTable.regexClass.getJVMName(), "setExtVariable", 
				      "(III" + t.getDescriptor() + ")V" );
		  }
	      }
	  }
      }
    catch( Exception e )
      {
	internalError( e, tpn );
      }
    finally
      {
	gen = oldGen;
      }
    return null;
  }

  public Object evalReplacer( ReplacerNode replacerNode )
  {
    try
      {
	gen.newobject( classTable.replacerClass.getJVMName() );
	gen.dup( classTable.replacerClass.getDescriptor() );
	replacerNode.regex.eval( this );
	loadFormat( replacerNode.formStr, true );
	gen.loadConst( new Integer( replacerNode.flags ) );
	gen.invokespecial( classTable.replacerClass.getJVMName(), "<init>",
			  "(" + classTable.regexClass.getDescriptor() +
			  classTable.formatterClass.getDescriptor() + "I)V" );
      }
    catch( Exception e )
      {
	internalError( e, replacerNode );
      }
    return null;
  }

  public Object evalThrow( ThrowNode throwNode )
  {
    try
      {
	gen.linenumber( throwNode.getLine() );
	evalExpr( throwNode.exception );
	gen.athrow();
      }
    catch( Exception e )
      {
	internalError( e, throwNode );
      }
    return null;
  }

  public Object evalTry( TryNode tryNode )
  {
    LabelInfo myInfo;
    CatchNode[] catches = tryNode.catches;
    CatchNode finallyPart = catches[catches.length-1];
    if( finallyPart.type != null ) // finally part has type==null, so it is last catch clause
      {
	finallyPart = null;
	myInfo = null;
      }
    else
      myInfo = nestedLabelsPush( tryNode );
    try
      {
	gen.linenumber( tryNode.getLine() );
	if( myInfo != null )
	  myInfo.finishMark = gen.newMark();
	AbstractMark begin = gen.newMark();
	gen.mark( begin );
	tryNode.body.eval( this );
	AbstractMark end = gen.newMark();
	boolean rch = gen.isReachable();
	gen.mark( end, true );
	AbstractMark fin = null;
	if( rch )
	  {
	    if( myInfo != null )
	      gen.jsr( myInfo.finishMark, 0 );
	    fin = gen.newMark();
	    gen.jump( fin );
	  }
	for( int i = 0 ; i < catches.length ; i++ )
	  {
	    CatchNode cn = catches[i];
	    if( cn.type == null )
	      continue;
	    String type = cn.name.clazz.getJVMName();
	    AbstractMark handler = gen.newMark();
	    gen.addExceptionHandler( begin, end, handler, type );
	    gen.mark( handler );
	    ((CompilerVariable)cn.name.attachment).genPut( gen , this );
	    cn.body.eval( this );
	    if( gen.isReachable() )
	      {
		if( myInfo != null )
		  gen.jsr( myInfo.finishMark, 0 );
		if( fin == null )
		  fin = gen.newMark();
		gen.jump( fin );
	      }
	  }
	if( finallyPart != null )
	  {
	    gen.mark( myInfo.finishMark );
	    myInfo.finishMark = null; // finally part is not active anymore
	    CompilerVariable tmp = (CompilerVariable)finallyPart.attachment;
	    tmp.genPut( gen, this );
	    finallyPart.body.eval( this );
	    tmp.genRet( gen, this );
	  }
	if( fin != null )
	  gen.mark( fin );
      }
    catch( Exception e )
      {
	internalError( e, tryNode );
      }
    finally
      {
	if( myInfo != null )
	  nestedLabelsPop( tryNode );
      }
    return null;
  }

  public Object evalWhile( WhileNode whileNode )
  {
    LabelInfo myInfo = nestedLabelsPush( whileNode );
    try
      {
	ExpressionNode cond = whileNode.cond;
	boolean infiniteLoop = cond instanceof ConstNode &&
	  ((ConstNode)cond).val == Boolean.TRUE;
 	AbstractMark mark = gen.newMark();
	if( infiniteLoop )
	  myInfo.againMark = mark;	
	if( whileNode.body != null )
	  {
	    AbstractMark fmark = gen.newMark();
	    gen.jump( fmark );
	    gen.mark( mark, 0 );
	    whileNode.body.eval( this );
	    gen.mark( fmark );
	  }
	else
	  gen.mark( mark );
	gen.linenumber( whileNode.getLine() );
	if( infiniteLoop )
	  gen.jump( mark );
	else
	  {
	    if( myInfo.againMark != null )
	      gen.mark( myInfo.againMark );
	    evalCond( whileNode.cond, mark, null );
	  }
	if( myInfo.finishMark != null )
	  gen.mark( myInfo.finishMark );
      }
    catch( Exception e )
      {
	internalError( e, whileNode );
      }
    finally
      {
	nestedLabelsPop( whileNode );
      }
    return null;
  }

}
