/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.compiler;

import java.util.*;

import kmy.jint.typer.*;
import kmy.jint.reflect.*;
import kmy.jint.gen.CodeGenerator;
import kmy.jint.parser.*;
import kmy.jint.tree.*;
import kmy.jint.constants.DefinitionConst;

import java.io.IOException;

/**
 *
 */
public class CompilerHelper extends AbstractTyperHelper implements DefinitionConst
{

  JintClass scope;

  class Variable extends CompilerVariable
  {

    JintClass      scope;
    AbstractClass  type;
    boolean        finalFlag;
    Identifier     name;
    BasicNode      dcl;
    int            offset;
    JintField      field;
    boolean        virgin = true;
    boolean        fieldMode = true;

    Variable( JintClass scope, boolean fin, AbstractClass type,
	      Identifier name, BasicNode dcl, int offset )
    {
      this.scope = scope;
      this.type = type;
      finalFlag = fin;
      this.name = name;
      this.dcl = dcl;
      this.offset = offset;
    }

    private void init( CompilerContext context )
    {
      // if scope uses "self" variable (or it is a class), we might need
      // to revert to field
      virgin = false;
      if( field == null && scope.selfUsed() )
	field = (JintField)scope.getField( name );
      if( field == null && name != context.THIS && name != context.SUPER )
	context.assignRegister( this );
    }

    void setFieldMode( boolean fieldMode )
    {
      // sometimes we need access reverted-to-field variable before
      // its holder object is created; in such cases fieldMode will be set to false
      this.fieldMode = fieldMode;
    }

    void genGet( CodeGenerator gen, CompilerContext context ) throws IOException
    {
      if( virgin )
	init( context );
      if( field == null || !fieldMode )
	gen.load( offset, type.getDescriptor() ); 
      else if( (field.getModifiers() & STATIC_FLAG) != 0 )
	gen.getstatic( scope.getJVMName(), field.getName().toString(), type.getDescriptor() );
      else
	gen.getfield( scope.getJVMName(), field.getName().toString(), type.getDescriptor() );
    }

    void genPut( CodeGenerator gen, CompilerContext context ) throws IOException
    {
      if( virgin )
	init( context );
      if( field == null || !fieldMode )
	gen.store( offset, type.getDescriptor() ); 
      else if( (field.getModifiers() & STATIC_FLAG) != 0 )
	gen.putstatic( scope.getJVMName(), field.getName().toString(), type.getDescriptor() );
      else
	gen.putfield( scope.getJVMName(), field.getName().toString(), type.getDescriptor() );
    }

    void genRet( CodeGenerator gen, CompilerContext context ) throws IOException
    {
      if( virgin )
	init( context );
      gen.ret( offset ); 
    }

    void genInit( CodeGenerator gen, CompilerContext context ) throws IOException
    {
      if( virgin )
	init( context );
      if( field != null && fieldMode )
	return;
      String dscr = type.getDescriptor();
      if( type.isPrimitive() )
	{
	  switch( dscr.charAt(0) )
	    {
	      case 'I' :
	      case 'C' :
	      case 'B' :
	      case 'Z' :
	      case 'S' :
		gen.loadConst( CompilerContext.I_ZERO );
		break;
	      case 'F' :
		gen.loadConst( CompilerContext.F_ZERO );
		break;
	      case 'J' :
		gen.loadConst( CompilerContext.L_ZERO );
		break;
	      case 'D' :
		gen.loadConst( CompilerContext.D_ZERO );
		break;
	      default:
		throw new RuntimeException( "Unexpected primitive type descriptor: " + dscr );
	    }
	}
      else
	gen.loadConst( null );
      gen.store( offset, dscr );
    }

    void setRegister( int reg )
    {
      if( field == null )
	offset = reg;
      else
	throw new RuntimeException( "Not a local variable!" );
    }

    int getRegister()
    {
      return ( field == null ? offset : -1 );
    }

    public int getFilePos()
    {
      return ( dcl == null ? -1 : dcl.getFilePos() );
    }

    public AbstractClass getType()
    {
      return type;
    }
    
    public String toString()
    {
      if( field == null )
	return "Local[" + name + ", " + offset + "]";
      else
	return "Variable[" + scope + ", " + name + "]";
    }

    public boolean isStatic()
    {
      return field != null && (field.getModifiers() & STATIC_FLAG) != 0;
    }

    public Object getConstValue()
    {
      if( field != null && field.isSimpleConst() )
	return field.getConstValue();
      return NOT_CONST;
    }

    public void mustBeField()
    {
      if( field == null )
	field = scope.findLocalField( name, type );
    }

    AbstractClass getBase( CompilerContext context )
    {
      if( virgin )
	init( context );
      return ( field == null || !fieldMode || (field.getModifiers() & STATIC_FLAG) != 0
	       ? scope.getClassTable().voidType : scope );
    }

  }

  class FieldAccess extends CompilerVariable
  {

    AbstractClass objClass;
    AbstractField field;

    FieldAccess( AbstractClass c, AbstractField f )
    {
      objClass = c;
      field = f;
    }
    
    void genGet( CodeGenerator gen, CompilerContext context ) throws IOException
    {
      String dc = getDeclaringName();
      if( dc == null )
	return;
      String nm = field.getName();
      if( field instanceof AnyField )
	{
	  gen.loadConst( nm );
	  context.genNextCID();
	  gen.invokestatic( context.ANY_UTILITIES,
	      context.ANY_FLDGET_METHOD, context.ANY_FLDGET_METHOD_SIG );
	}
      else
	{
	  String tp = field.getType().getDescriptor();
	  int modifiers = field.getModifiers();
	  if( ( modifiers & STATIC_FLAG ) != 0 )
	    gen.getstatic( dc, nm, tp );
	  else
	    gen.getfield( dc, nm, tp );
	}
    }

    void genPut( CodeGenerator gen, CompilerContext context ) throws IOException
    {
      String dc = getDeclaringName();
      if( dc == null )
	return;
      String nm = field.getName();
      if( field instanceof AnyField )
	{
	  gen.loadConst( nm );
	  context.genNextCID();
	  gen.invokestatic( context.ANY_UTILITIES,
	      context.ANY_FLDPUT_METHOD, context.ANY_FLDPUT_METHOD_SIG );
	}
      else
	{
	  String tp = field.getType().getDescriptor();
	  int modifiers = field.getModifiers();
	  if( (modifiers & STATIC_FLAG) != 0 )
	    gen.putstatic( dc, nm, tp );
	  else
	    gen.putfield( dc, nm, tp );
	}
    }

    String getDeclaringName()
    {
      AbstractClass d = field.getDeclaringClass();
      if( d instanceof JintClass )
	{
	  JintClass jd = (JintClass)d;
	  if( ! jd.mustBeClass() )
	    {
	      if( jd.getOuterField() != field )
		throw new RuntimeException( "Internal error: access of non-class field: " + field );
	      return null;
	    }
	}
      return d.getJVMName();
    }

    AbstractClass getBase( CompilerContext context )
    {
      return ( field.getModifiers() & STATIC_FLAG) != 0 ? 
	         objClass.getClassTable().voidType : objClass;
    }

    public int getFilePos()
    {
      return -1;
    }

    public AbstractClass getType()
    {
      return field.getType();
    }

    public boolean isStatic()
    {
      return (field.getModifiers() & STATIC_FLAG) != 0;
    }

    public Object getConstValue()
    {
      if( field.isSimpleConst() )
	return field.getConstValue();
      return NOT_CONST;
    }

    public String toString()
    {
      return "Field[" + field + "]";
    }
  }

  class ArrayLengthAccess  extends CompilerVariable
  {

    AbstractClass intType;

    ArrayLengthAccess( JintClassTable ct )
    {
      intType = ct.intType;
    }

    void genGet( CodeGenerator gen, CompilerContext context ) throws IOException
    {
      gen.arraylength();
    }

    void genPut( CodeGenerator gen, CompilerContext context ) throws IOException
    {
      throw new RuntimeException( "Cannot set array length!" );
    }

    AbstractClass getBase( CompilerContext context )
    {
      return context.classTable.voidType;
    }

    public int getFilePos()
    {
      return -1;
    }

    public AbstractClass getType()
    {
      return intType;
    }

    public String toString()
    {
      return "ArrayLength[]";
    }
  }

  class ClassAccess  extends CompilerVariable
  {

    AbstractClass classClass;
    AbstractClass clazz;

    ClassAccess( JintClassTable ct, AbstractClass clazz )
    {
      classClass = ct.classClass;
      this.clazz = clazz;
    }

    void genGet( CodeGenerator gen, CompilerContext context ) throws IOException
    {
      context.genClass( clazz );
    }

    void genPut( CodeGenerator gen, CompilerContext context ) throws IOException
    {
      throw new RuntimeException( "Cannot set array length!" );
    }

    AbstractClass getBase( CompilerContext context )
    {
      return context.classTable.voidType;
    }

    public int getFilePos()
    {
      return -1;
    }

    public AbstractClass getType()
    {
      return classClass;
    }

    public String toString()
    {
      return "ArrayLength[]";
    }
  }

  public void enterScope( JintClass clazz )
  {
    scope = clazz;
  }

  public void leaveScope( JintClass clazz )
  {
    scope = clazz;
  }

  public AbstractVariable newVariable( boolean finalFlag, AbstractClass type,
				Identifier name, BasicNode node, int offset )
  {
    return new Variable( scope, finalFlag, type, name, node, offset );
  }

  public AbstractVariable newVariable( AbstractClass objClass, AbstractField field )
  {
    return new FieldAccess( objClass, field );
  }

  public AbstractVariable arrayLength( JintClassTable ct, AbstractClass array )
  {
    return new ArrayLengthAccess( ct );
  }
    
  public AbstractVariable classAccess( JintClassTable ct, AbstractClass clazz )
  {
    return new ClassAccess( ct, clazz );
  }
}
