/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.gen;

import java.io.*;

public class JVMClassReader
{

  static final int R_FIELD  = 1;
  static final int R_METHOD = 2;
  static final int R_CLASS  = 3;

  static final String[] EMPTY = {};

  int version;
  int subversion;

  JVMConstantPool constantPool;

  JVMClassConsumer consumer;

  public JVMClassReader( JVMClassConsumer consumer )
  {
    this.consumer = consumer;
  }

  private void readInterfaces( DataInputStream in ) throws IOException
  {
    int nInterfaces = in.readShort();
    String[] iNames = ( nInterfaces == 0 ? EMPTY : new String[nInterfaces] );
    for( int i = 0 ; i < nInterfaces ; i++ )
      {
	int nameIndex = in.readShort();
	JVMConstantPool.CP_Class_Entry name =
	  (JVMConstantPool.CP_Class_Entry)constantPool.getEntry( nameIndex );	
	iNames[i] = (String)name.getData();
      }
    consumer.tellInterfaces( iNames );
  }

  private void readAttributes( DataInputStream in, int what ) throws IOException
  {
    int nattr = in.readShort();
    for( int j = 0 ; j < nattr ; j++ )
      {
	int attrNameIndex = in.readShort();
	int size = in.readInt();
	JVMConstantPool.CP_UTF8_Entry attrNameEntry = 
	  (JVMConstantPool.CP_UTF8_Entry)constantPool.getEntry( attrNameIndex );
	String attrName = (String)attrNameEntry.getData();
	boolean want;
	if( what == R_FIELD )
	  want = consumer.wantFieldAttribute( attrName );
	else
	  want = consumer.wantMethodAttribute( attrName );

	if( want )
	  {
	    Object value;
	    if( attrName == "ConstantValue" && what == R_FIELD )
	      {
		if( size != 2 )
		  throw new IOException(
			 "Field value is not constant pool reference!" );
		JVMConstantPool.CP_Entry e = constantPool.getEntry( in.readShort() );
		value = e.getData();
	      }
	    else if( attrName == "JintType" && size == 2 )
	      {
		int index = in.readShort();
		value = constantPool.getEntry( index ).getData();
	      }
	    else if( attrName == "Exceptions" && what == R_METHOD )
	      {
		int nExceptions = in.readShort();
		if( size != nExceptions*2 + 2 )
		  throw new IOException( "Invalid exception list size!" );
		String[] exc = new String[nExceptions];
		for( int i = 0 ; i < nExceptions ; i++ )
		  exc[i] = (String)constantPool.getEntry( in.readShort() ).getData();
		value = exc;
	      }
	    else
	      {
		byte[] arr = new byte[size];
		int count = 0;
		while( count < size )
		  {
		    int add = in.read( arr, count, size-count );
		    if( add < 0 )
		      throw new IOException( "End of file" );
		    count += add;
		  }
		value = arr;
	      }
	    if( what == R_FIELD )
	      consumer.tellFieldAttribute( attrName, value );
	    else
	      consumer.tellMethodAttribute( attrName, value );
	  }
	else
	  {
	    // somehow skip does not work!!! JDK bug?
	    // in.skip( size );
	    byte[] arr = new byte[size];
	    int count = size;
	    while( count > 0 )
	      {
		int add = in.read( arr, 0, count );
		if( add < 0 )
		  throw new IOException( "End of file" );
		count -= add;
	      }
	  }
      }
  }


  private void readFieldsOrMethods( DataInputStream in, int what ) throws IOException
  {
    int nEntries = in.readShort();
    for( int i = 0 ; i < nEntries ; i++ )
      {
	int flags = in.readShort();
	int nameIndex = in.readShort();
	int typeIndex = in.readShort();

	JVMConstantPool.CP_UTF8_Entry nameEntry =
	  (JVMConstantPool.CP_UTF8_Entry)constantPool.getEntry( nameIndex );
	JVMConstantPool.CP_UTF8_Entry typeEntry = 
	  (JVMConstantPool.CP_UTF8_Entry)constantPool.getEntry( typeIndex );

	String name = (String)nameEntry.getData();
	String type = (String)typeEntry.getData();
	if( what == R_FIELD )
	  consumer.startField( flags, name, type );
	else
	  consumer.startMethod( flags, name, type );

	readAttributes( in, what ); 
	
	if( what == R_FIELD )
	  consumer.endField();
	else
	  consumer.endMethod();
      }
  }

  public void parse( InputStream in ) throws IOException
  {
    parse( new DataInputStream( new BufferedInputStream( in ) ) );
  }

  public void parse( DataInputStream in ) throws IOException
  {
    int magic = in.readInt();
    if( magic != 0xCAFEBABE )
      throw new IOException( "Not a class file, header is " + Integer.toHexString( magic ) );
    version = in.readShort();
    if( version != 3 )
      throw new IOException( "Version is " + version + ", 3 expected" );
    subversion = in.readShort();
    if( subversion != 45 )
      throw new IOException( "Subversion is " + subversion + ", 45 expected" );
    consumer.tellVersion( version, subversion );
    constantPool = new JVMConstantPool();
    constantPool.readConstantPool( in );
    int flags      = in.readShort();
    int classIndex = in.readShort();
    int superIndex = in.readShort();
    JVMConstantPool.CP_Class_Entry className =
	  (JVMConstantPool.CP_Class_Entry)constantPool.getEntry( classIndex );
    JVMConstantPool.CP_Class_Entry superName = ( superIndex == 0 ? null :
	  (JVMConstantPool.CP_Class_Entry)constantPool.getEntry( superIndex ) );
    consumer.startClass( flags, (String)className.getData(),
			(String)( superName == null ? null : superName.getData() ) );
    readInterfaces( in );
    readFieldsOrMethods( in, R_FIELD );
    readFieldsOrMethods( in, R_METHOD );
    consumer.endClass();
  }

}
