/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.io;

import java.io.*;

import kmy.jint.lang.*;
import kmy.regex.util.*;

public class JintReader extends Reader
{

  Reader in;

  char[] buffer = new char[bufSize];
  int    head;    // index of first used character in the buffer
  int    tail;    // index of first free character in the buffer, head <= tail

  int    markCount; // while mark is active (markCount>0) buffer cannot be reallocated

  boolean dirty;
  boolean eof; // end of stream in underlying Reader (in) 
  
  private Regex ignoreRegex;

  static final boolean debug = false;

  static int bufSize = ( debug ? 80 : 2048 );

  public class Mark
  {
    int markHead;

    Mark()
    {
      synchronized( JintReader.this )
	{
	  markHead = head;
	  markCount++;
	}
    }

    void trash()
    {
      synchronized( JintReader.this )
	{
	  if( markHead < 0 )
	    return;
	  markHead = -1;
	  markCount--;
	}
    }

    protected void finalize()
    {
      if( markHead >= 0 )
	trash();
    }
  }

  static class IOExceptionWrapper extends RuntimeException
  {
    IOException e;

    IOExceptionWrapper( IOException e )
      {
	this.e = e;
      }
  }

  String bufRep()
  {
    int len = buffer.length - head > 20 ? 20 : buffer.length - head;
    if( tail > head && head + len > tail )
      len = tail - head;
    return "'" + new String( buffer, head, len ) + "'";
  }

  RegexRefiller refiller = new RegexRefiller() {
    public int refill( Regex regex, int boundary )
      {
	char[] oldBuf = buffer;
	int oldHead = head;
	if( boundary != tail )
	  throw new RuntimeException( "Internal error: boundary != tail" );
	try
	  {
	    fillBufferIfHaveTo( tail );
	  }
	catch( IOException e )
	  {
	    throw new IOExceptionWrapper( e );
	  }
	if( head != oldHead )
	  return -1; // restart matching
	if( tail == boundary )
	  regex.setRefiller( null );
	if( oldBuf != buffer )
	  regex.setRefilledBuffer( buffer );
	return tail;
      }
  };

  int fillBufferIfHaveTo( int h ) throws IOException
  {
    if( debug )
      System.out.println( "fillBuffer " + head + " " + tail + " " + bufRep() );

    if( h < tail )
      return h;
    if( buffer == null )
      buffer = new char[bufSize];
    if( tail >= buffer.length )
      {
	if( head == 0 || markCount > 0 || (dirty && head < buffer.length/4) )
	  {
	    char[] newBuf = new char[buffer.length*2];
	    System.arraycopy( buffer, 0, newBuf, 0, tail );
	    buffer = newBuf;
	  }
	else if( dirty )
	  {
	    char[] newBuf = new char[buffer.length];
	    System.arraycopy( buffer, head, newBuf, 0, tail-head );
	    buffer = newBuf;
	    tail -= head;
	    h -= head;
	    head = 0;
	  }
	else
	  {	    
	    for( int i = head ; i < tail ; i++ )
	      buffer[i-head] = buffer[i];
	    tail -= head;
	    h -= head;
	    head = 0;
	  }
      }

    if( debug )
      System.out.println( "trying to read " + head + " " + tail + " " + bufRep() );

    int r;
    while( true )
      {
	r = in.read( buffer, tail, buffer.length - tail );
	if( r != 0 )
	  break;
	Thread.yield();
      }
    if( r < 0 )
      {
	eof = true;
	return h;
      }

    tail += r;

    if( debug )
      System.out.println( "read " + r + " chars " + head + " " + tail + " " + bufRep() );

    return h;
  }

  public boolean matchToken( Regex regex ) throws IOException
  {
    dirty = true;
    int h = head;

    if( debug )
      System.out.println( "Match: " + regex + " " + bufRep() );

    try
      {
	regex.setRefiller( refiller );
	while( true )
	  {
	    regex.init( buffer, h, tail-h );
	    try
	      {
		if( regex.match() )
		  return true;
	      }
	    catch( IOExceptionWrapper wrapper )
	      {
		throw wrapper.e;
	      }
	    if( head == h )
	      return false;
	    h = head;
	  }
      }
    finally
      {
	regex.setRefiller( null );
      }
  }

  void skipIgnore() throws IOException
  {
    if( ignoreRegex == null )
      return;
    while( matchToken( ignoreRegex ) )
      {
	int h = ignoreRegex.getMatchEnd();
	if( h == head )
	  break;
	head = h;
      }
  }

  //--------- public stuff -------------

  public JintReader( Reader r ) throws IOException
  {
    in = r;
  }

  public JintReader( String fileName ) throws IOException
  {
    in = new FileReader( fileName );
  }

  public void setIgnore( Regex r )
  {
    ignoreRegex = r;
  }

  public Regex getIgnore()
  {
    return ignoreRegex;
  }

  public CharString readDelimitedToken( Regex regex ) throws IOException
  {    
    skipIgnore();
    if( matchToken( regex ) )
      {
	int start = head;
	head = regex.getMatchEnd();
	return new CharString( buffer, start, regex.getMatchStart()-start );
      }
    if( eof && head < tail )
      {
	CharString res = new CharString( buffer, head, tail-head );
	head = tail;
	return res;
      }
    return null;
  }

  public boolean read( Regex regex ) throws IOException
  {    
    skipIgnore();
    if( matchToken( regex ) )
      {
	int start = regex.getMatchStart();
	head = regex.getMatchEnd();
	return true;
      }
    return false;
  }

  public boolean read( String string ) throws IOException
  {    
    skipIgnore();
    int len = string.length();
    for( int i = 0 ; i < len ; i++ )
      {
	if( head+i > tail )
	  {
	    fillBufferIfHaveTo( head+i );
	    if( head+i > tail )
	      return false;
	  }
	// System.out.println( "---> '" + buffer[head+i] + "' '" + string.charAt(i) + "'" );
	if( buffer[head+i] != string.charAt(i) )
	  return false;
      }
    head += len;
    return true;
  }

  //--------- Reader implementation ----

  public int read() throws IOException
  {
    fillBufferIfHaveTo( head+1 );
    if( head == tail )
      return -1;
    return buffer[head++];
  }

  public int read( char[] buf, int off, int len ) throws IOException
  {
    int length;
    if( head >= tail )
      length = 0; // buffer is already empty
    else
      {
	length = len;
	if( head + length > tail )
	  {
	    length = tail - head;
	    len -= length;
	  }
	else
	  len = 0;

	System.arraycopy( buffer, head, buf, off, length );
	head += length;
	if( length == len || !in.ready() )
	  return len;
	off += length;
      }

    // at this point buffer is empty
    if( markCount == 0 )
      {
	if( dirty )
	  {
	    buffer = new char[buffer.length];
	    dirty = false;
	  }
	
	head = 0;
	tail = 0;
      }

    fillBufferIfHaveTo( tail );

    if( head + len > tail )
      len = tail - head;

    System.arraycopy( buffer, head, buf, off, len );
    head += len;

    return length + len;
  }

  public long skip(long nChars) throws IOException
  {
    while( true )
      {
	int nCharsInBuffer = head - tail;
	if( nChars < nCharsInBuffer )
	  {
	    head += (int)nChars;
	    return nChars;
	  }
	if( markCount == 0 )
	  {
	    head = tail;
	    return nCharsInBuffer + in.skip(nChars-nCharsInBuffer);
	  }
	fillBufferIfHaveTo( tail );
      }
  }

  public boolean ready() throws IOException
  {
    return head < tail || in.ready();
  }

  public boolean eof()
  {
    return head == tail && eof;
  }

  public Mark mark()
  {
    return new Mark();
  }

  public void reset( Mark m )
  {
    head = m.markHead;
  }

  public boolean markSupported()
  {
    return false;
  }

  public void mark(int nChars)
  {
  }

  public void reset()
  {
  }

  public void close() throws IOException
  {
    buffer = null;
    head = 0;
    tail = 0;
    in.close();
  }
}
