/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.parser;

import kmy.jint.constants.TokenConst;
import kmy.jint.constants.ErrorCodes;

/**
 *  This class represents a stream of Java/Jint tokens. Method nextToken()
 *  can be used to fill in next token's tType and possibly other fields.
 *  Method close() frees resources used by this stream.
 */
public abstract class AbstractTokenizer implements TokenConst, ErrorCodes
{
  /**
   *  Token type. Values are defined in TokenConst interface as TOKEN_*.
   *  Also valid values are single-character operator signs, such as +, -, etc. -
   *  they ase represented by character numerical values.
   *  Negative values represent keywords and identifiers
   */ 
  public int         tType;

  /**
   *  Integer/long token value. TOKEN_V_LONG, TOKEN_V_INT, TOKEN_V_CHAR return
   *  their numerical value in this field.
   */
  public long        lVal;

  /**
   *  Float/double token value. TOKEN_V_FLOAT, TOKEN_V_DOUBLE return
   *  their numerical value in this field.
   */
  public double      dVal;

  /**
   *  String token value. TOKEN_STRING, TOKEN_REGEX, TOKEN_TRANS, TOKEN_PATT,
   *  TOKEN_FORM, TOKEN_SUBST, and TOKEN_BOUND set this field.
   */
  public String      sVal;

  /**
   *  Additional String token value. TOKEN_TRANS and TOKEN_SUBST set this field.
   */
  public String      sVal2;

  /**
   *  Identifier token value. If tType is negative, this field contains corresponding
   *  Identifier object.
   */
  public Identifier  ident;

  /**
   *  Integer that represents token position inside the stream. It is not byte count! Use
   *  getLine, getColumn, and makeFilePos to work with stream positions.
   */
  public int         filePos = makeFilePos( 1, 0 );

  /**
   *  Field that stores a NameTable that was passed to constructor. NameTable holds
   *  all Identifiers that this stream returns and they are inserted in the table as
   *  needed.
   */
  public NameTable nameTable;

  /**
   *  Reads new token from the stream. Stores token type into tType field and
   *  possibly initializes other fields (depending on tType value).
   */
  public abstract void nextToken();

  /**
   *  Frees resources held by this stream. Method nextToken() must not be called after
   *  call to close.
   */
  public abstract void close();

  /**
   *  Create and initialize an AbstractTokenizer instance.
   */
  protected AbstractTokenizer( NameTable nameTable )
  {
    this.nameTable = nameTable;
  }

  /**
   *  Creates representation of stream position based on line
   *  and column values. If column value is more then a 1000 it
   *  might not get represented properly.
   */
  public static int makeFilePos( int line, int col )
  {
    if( col > 0x3FF || col < 0 )
      col = 0x3FF;
    return (line << 10)|col; 
  }

  /**
   *  Extracts line number in a stream given the stream position.
   */
  public static int getLine( int fpos )
  {
    return fpos >>> 10;
  }

  /**
   *  Extracts column number in a stream given the stream position.
   */
  public static int getColumn( int fpos )
  {
    return fpos & 0x3FF;
  }

}

