/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.parser;

import java.util.*;

/**
 *  This class keeps record of all identifiers and keywords.
 *  It maintains a single unique, copy of every identifier, and
 *  assigns every identifier a unique integer id.
 *  One can look up an identifier by its string representation.
 */
public class NameTable
{
  final static String[] reserved =
  {
    null,
    "abstract",
    "boolean",
    "break",
    "byte",
    "byvalue",
    "case",
    "catch",
    "char",
    "class",
    "const",
    "continue",
    "default",
    "do",
    "double",
    "else",
    "extends",
    "false",
    "final",
    "finally",
    "float",
    "for",
    "goto",
    "if",
    "implements",
    "import",
    "instanceof",
    "int",
    "interface",
    "long",
    "native",
    "new",
    "null",
    "package",
    "private",
    "protected",
    "public",
    "return",
    "short",
    "static",
    "super",
    "switch",
    "synchronized",
    "this",
    "threadsafe",
    "throw",
    "throws",
    "transient",
    "true",
    "try",
    "void",
    "volatile",
    "while"
  };

  final static String[] privateReserved =
  {
    "where"
  };

  private Hashtable nameTable = new Hashtable();

  public final Identifier ADD      = new Identifier( this, "_add", -1 );
  public final Identifier ARGV     = new Identifier( this, "argv", -1 );
  public final Identifier ARGLIST  = new Identifier( this, "ArgList", -1 );
  public final Identifier BLOCK    = new Identifier( this, "Block", -1 );
  public final Identifier EVAL     = new Identifier( this, "eval", -1 );
  public final Identifier SELF     = new Identifier( this, "self", -1 );
  public final Identifier INIT     = new Identifier( this, "<init>", -1 );
  public final Identifier CLINIT   = new Identifier( this, "<clinit>", -1 );
  public final Identifier LANG     = new Identifier( this, "lang", -1 );
  public final Identifier JAVA     = new Identifier( this, "java", -1 );
  public final Identifier JINT     = new Identifier( this, "jint", -1 );
  public final Identifier ANY      = new Identifier( this, "Any", -1 );
  public final Identifier KMY      = new Identifier( this, "kmy", -1 );
  public final Identifier LENGTH   = new Identifier( this, "length", -1 );
  public final Identifier RUN      = new Identifier( this, "run", -1 );
  public final Identifier RUNNABLE = new Identifier( this, "Runnable", -1 );
  public final Identifier VAR      = new Identifier( this, "var", -1 );

  public final Identifier THIS   = new Identifier( this, "this", AbstractTokenizer.TOKEN_THIS );
  public final Identifier SUPER  = new Identifier( this, "super", AbstractTokenizer.TOKEN_SUPER );
  public final Identifier TRUE   = new Identifier( this, "true", AbstractTokenizer.TOKEN_TRUE );
  public final Identifier FALSE  = new Identifier( this, "false", AbstractTokenizer.TOKEN_FALSE );
  public final Identifier NULL   = new Identifier( this, "null", AbstractTokenizer.TOKEN_NULL );
  public final Identifier CLASS  = new Identifier( this, "class", AbstractTokenizer.TOKEN_CLASS );
  public final Identifier VOID   = new Identifier( this, "void", AbstractTokenizer.TOKEN_VOID );

  /**
   *  Create a NameTable. Normally only one NameTable is needed.
   */
  public NameTable()
  {
    for( int i = 1 ; i < reserved.length ; i++ )
      if( reserved[i] == "this" )
	nameTable.put( reserved[i], THIS );
      else if( reserved[i] == "super" )
	nameTable.put( reserved[i], SUPER );
      else if( reserved[i] == "true" )
	nameTable.put( reserved[i], TRUE );
      else if( reserved[i] == "false" )
	nameTable.put( reserved[i], FALSE );
      else if( reserved[i] == "null" )
	nameTable.put( reserved[i], NULL );
      else if( reserved[i] == "class" )
	nameTable.put( reserved[i], CLASS );
      else if( reserved[i] == "void" )
	nameTable.put( reserved[i], VOID );
      else
	nameTable.put( reserved[i], new Identifier( this, reserved[i], -i ) );

    for( int i = 0 ; i < privateReserved.length ; i++ )
      nameTable.put( privateReserved[i], 
          new Identifier( this, privateReserved[i], -(reserved.length+i) ) );

    Identifier[] toInit = { ADD, ARGV, ARGLIST, BLOCK, EVAL,
			    SELF, INIT, LANG, JAVA, JINT,
			    ANY, KMY, LENGTH, RUN, RUNNABLE, VAR };
    for( int i = 0 ; i < toInit.length ; i++ )
      {
	Identifier idnt = toInit[i];
	idnt.nameTable = this;
	idnt.id = -(reserved.length+privateReserved.length+i);
	nameTable.put( idnt.name, idnt );
      }
 }

  /**
   *  Returns Identifier by its String representation. If no such identifier is
   *  yet defined, null is returned.
   */
  public Identifier lookup( String name )
  {
    return (Identifier)nameTable.get( name );
  }

  /**
   *  Returns Identifier by its String representation. If no such identifier is
   *  defined it gets inserted, so this method always returns an Identifier.
   */
  public Identifier lookupAndInsert( String name )
  {
    synchronized( nameTable )
      {
	Identifier id = (Identifier)nameTable.get( name );
	if( id == null )
	  {
	    name = name.intern();
	    id = new Identifier( this, name, -(nameTable.size()+1) );
	    nameTable.put( name, id );
	  }
	return id;
      }
  }

}
