/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.parser;

import kmy.jint.constants.*;
import kmy.jint.tree.*;
import kmy.jint.util.CompilerException;
import kmy.jint.util.ErrorStorage;
import kmy.jint.util.ErrorKeeper;

import java.util.*;
import java.io.*;

/**
 *  This class can parse Jint (or Java) token stream and produce a parsing
 *  tree. It also calls ClassConsumer methods as it encounters corresponding
 *  constructions. Generally, one should create this class (giving it a token
 *  stream), set some properties (like errorStorage) and call parseStatement
 *  in a loop till the end of token stream.
 */
public class Parser
  implements ErrorStorage, DefinitionConst, TokenConst, ErrorCodes
{

  private AbstractTokenizer in;
  private boolean           processed = true;
  private boolean           fresh     = true;
  private ClassConsumer     classConsumer;
  private int               nextInlineClassName = 0;
  private Object            upperLevelClassHandle = null;
  private Object            currentClassHandle = null;
  private Identifier        currentClassName; // to catch constructors
  private IdentifierNode[]  commonImports;

  private ErrorStorage errorStorage;

  private boolean inInterface = false;

  private Object publicScope = new Object(); // all subscopes of this scope are public

  private boolean looseExceptionCheck = false;

  static final Boolean TRUE = Boolean.TRUE;
  static final Boolean FALSE = Boolean.FALSE;

  /**
   *  Creates a new parser that will parse the given token stream.
   */
  public Parser( AbstractTokenizer t )
  {
    this( t, new ClassConsumer() );
  }

  /**
   *  Creates a new parser that will parse the given token stream and call
   *  the given ClassConsumer as it parses.
   */
  public Parser( AbstractTokenizer t, ClassConsumer cc )
  {
    in = t;
    classConsumer = cc;
  }

  /**
   *  Exceptions should be treated loosely.
   */
  public void setLooseExceptionCheck( boolean t )
  {
    looseExceptionCheck = t;
  }

  public boolean getLooseExceptionCheck()
  {
    return looseExceptionCheck;
  }

  /**
   *  Assigns an ErrorStorage that this parser will use to report errors.
   *  Both parsing and lexer errors are caught and sent to the given ErrorStorage.
   */
  public void setErrorStorage( ErrorStorage errorStorage )
  {
    this.errorStorage = errorStorage;
  }

  /**
   *  This method initializes outermost scope handle, so the code parser gets is
   *  treated in the context of the given scope. This scope handle will be passed
   *  to the first ClassConsumer.startScope method call as outerHandle parameter.
   */
  public void setClassHandle( Object handle )
  {
    if( upperLevelClassHandle == publicScope )
      publicScope = handle;
    upperLevelClassHandle = handle;
    currentClassHandle = handle;
  }

  /**
   *  Returns current class handle. Class handle is an object that corresponds to
   *  current scope (method or class); it is used by ClassConsumer.
   */
  public Object getClassHandle()
  {
    return currentClassHandle;
  }

  /**
   *  Report all future subscopes of outermost scope as public.
   */
  public void forcePublic( boolean force )
  {
    if( force )
      publicScope = upperLevelClassHandle;
    else
      publicScope = new Object(); // not null!
  }

  /**
   *  This is bogus expression node that expression parser returns
   *  when it encounters labeled statement.
   */
  static class LabelExpression extends ExpressionNode
  {

    IdentifierNode label;

    LabelExpression( IdentifierNode label )
    {
      super( -1 );
      this.label = label;
    }

    public Object eval( Context context )
    {
      return null;
    }

    public String toString()
    {
      return "<<<LabelExpression>>>";
    }
  }

  /**
   *  This is bogus expression node that expression parser returns
   *  when it encounters method definition.
   */
  static class LocalMethodExpression extends ExpressionNode
  {

    public MethodNode method;

    public LocalMethodExpression( int filePos, MethodNode method )
    {
      super( filePos );
      this.method = method;
      method.setFilePos( filePos );
    }

    public Object eval( Context context )
    {
      return null;
    }

    public String toString()
    {
      return "<<<LocalMethodExpression>>>";
    }
  }

  /**
   *  This is bogus expression node that expression parser returns
   *  when it encounters variable/member definition.
   */
  static class LocalVariableExpression extends ExpressionNode
  {

    MemberNode memb;

    public LocalVariableExpression( int filePos, MemberNode memb )
    {
      super( filePos );
      this.memb = memb;
      memb.setFilePos( filePos );
    }

    public Object eval( Context context )
    {
      return null;
    }

    public String toString()
    {
      return "<<<LocalVariableExpression>>>";
    }
  }

  /**
   *  Generates a new name for inner class. Normally it just increments an internal
   *  name counter and converts it into Identifier. Number by itself is not a valid name,
   *  but full (JVM) inner class name is parent$inner, so there is no problem. So
   *  for upper-level inner classes (which are nested in "anonymous" class), it adds 
   *  ___C_ prefix, so we still have a valid name (although now we can only hope it's
   *  unique).
   */
  private Identifier nextInnerClassIdentifier()
  {
    String nextName = Integer.toString( ++nextInlineClassName );
    if( currentClassHandle == upperLevelClassHandle )
      nextName = "___C_" + nextName;
    return in.nameTable.lookupAndInsert( nextName );
  }

  /**
   *  The same as nextInnerClassIdentifier() without args, but appends "m_" + given
   *  string to the newly generated name; also uses ___F_ instead of ___C_.
   *  Such class names are used for (hidden) method scopes.
   */
  private Identifier nextInnerClassIdentifier( String patt )
  {
    String nextName = Integer.toString( ++nextInlineClassName );
    if( currentClassHandle == upperLevelClassHandle )
      nextName = "___F_" + nextName;
    return in.nameTable.lookupAndInsert( nextName + "m_" + patt );
  }

  /**
   *  The same as nextInnerClassIdentifier() without args, but appends "r"
   *  to the newly generated name; also uses ___R_ instead of ___C_.
   *  Such class names are used for (hidden) regex classes.
   */
  private Identifier nextRegexClassIdentifier()
  {
    String nextName = Integer.toString( ++nextInlineClassName );
    if( currentClassHandle == upperLevelClassHandle )
      nextName = "___R_" + nextName;
    return in.nameTable.lookupAndInsert( nextName + "r" );
  }

  /**
   *  Reads new token from the stream if needed. This method won't read the
   *  new token until it is consumed by calling consumeToken(), so it is safe
   *  to call it many times in a row.
   */
  private final void nextToken()
  {
    if( processed )
      {
	while( true )
	  try
	    {
	      in.nextToken();
	      break;
	    }
  	  catch( CompilerException e )
	    {
	      storeError( e );
	      if( in.tType != in.TOKEN_NONE )
		break;
	    }
	processed = false;
      }
  }

  /**
   *  Marks token as consumed, so next call to nextToken() will read a new one.
   */
  private final void consumeToken()
  {
    processed = true;
    fresh = true;
  }

  private void storeError( int code, int filePos )
  {
    storeError( code, filePos, null );
  }

  private void storeError( int code, int filePos, Object info )
  {
    CompilerException e;
    if( info != null )
      {
	Object[] add = { info };
	e = new CompilerException( code, filePos, add );
      }
    else
      e = new CompilerException( code, filePos );
    storeError( e );
  }

  public void storeError( CompilerException e )
  {
    fresh = false;
    if( errorStorage != null && errorStorage != this )
      errorStorage.storeError( e );
    else
      System.err.println( "Line " + e.getLine() + ", pos " + 
			e.getColumn() + ": " + e.getMessage() );
  }

  /**
   *  Asserts that next token is of type tType.
   *  @param tType expected token type
   *  @param vis string representation of expected token type for error reporting
   */
  void assertToken( int tType, String vis )
  {
    nextToken();
    if( in.tType != tType && fresh )
      storeError( ERR_P_TOKENEXPECTED, in.filePos, vis );
    else
      consumeToken();
  }

  /**
   *  Asserts that next token is of type tType.
   *  @param tType expected token type
   */
  void assertToken( char tType )
  {
    nextToken();
    if( in.tType != tType && fresh )
      storeError( ERR_P_TOKENEXPECTED, in.filePos, "'" + tType + "'" );
    else
      consumeToken();
  }

  /**
   *  Parses class or interface. This method expects to see
   *  either keyword 'class' or 'interface'
   *  as the first token, so things like 'public' or 'abstract'
   *  should be taken care of by caller.
   *  @param flags modifiers (parsed by caller and passed as flags)
   *  @returns parsing tree for class or interface
   */
  ClassNode parseClass( int flags )
  {
    IdentifierNode className = null;

    Vector whereList = new Vector();

    IdentifierNode[] parentArray = null;
    IdentifierNode[] interfaceArray = null;

    boolean header = true;

    nextToken();
    switch( in.tType )
      {
        case TOKEN_INTERFACE :
	  flags |= INTERFACE_FLAG;
	  consumeToken();
	  break;
        case TOKEN_CLASS :
	  consumeToken();
	  break;
        default :
	  storeError( ERR_P_CLASSKEYWORD, in.filePos );
      }

    className = parseIdentifier( false, true );
  
    while( header )
      {
	nextToken();
	switch( in.tType )
	  {
	    case '=' :
	      {
		consumeToken();
		nextToken();
		RegexNode regex;
		try
		  {
		    if( in.tType != in.TOKEN_V_REGEX && in.tType != in.TOKEN_V_PATT )
		      throw new CompilerException( in.filePos ); 
		    consumeToken();
		    regex = new RegexNode( in.filePos, in.sVal, (int)in.lVal, 
					  in.tType == in.TOKEN_V_REGEX );
		  }
		catch( CompilerException e )
		  {
		    storeError( e );
		    regex = new RegexNode( in.filePos, ".*", 0, true );
		  }
		if( currentClassHandle == null && (regex.flags & in.TOK_OPT_STANDALONE) == 0 )
		  {
		    storeError( ERR_P_NOTSTANDALONE, in.filePos );
		    regex.flags |= in.TOK_OPT_STANDALONE;
		  }
		if( (regex.flags & in.TOK_OPT_STANDALONE) != 0 )
		  flags |= STATIC_FLAG;
		if( publicScope == currentClassHandle )
		  flags |= PUBLIC_FLAG;
		Object handle = classConsumer.startScope( currentClassHandle, className.idList[0],
				  flags, null, null, RegexNode.class );
		classConsumer.finishScope( handle, regex );
		ClassNode res = new ClassNode( className, regex );
		res.addFlags( flags );
		assertToken( ';' );
		return res;
	      }
	    case TOKEN_EXTENDS :
	      consumeToken();
	      if( parentArray != null )
		{
		  storeError( ERR_P_DUPEXTENDS, in.filePos );
		  parseIdentifierList();
		}
	      else
		parentArray = parseIdentifierList();
	      break;
	    case TOKEN_IMPLEMENTS :
	      consumeToken();
	      if( interfaceArray != null )
		{
		  storeError( ERR_P_DUPIMPLEMENTS, in.filePos );
		  parseIdentifierList();
		}
	      else
		interfaceArray = parseIdentifierList();
	      break;
	    case TOKEN_WHERE : // 'where' stuff is not implemented!!! 
	      {
		consumeToken();
		IdentifierNode whereName = parseIdentifier();
		ClassNode whereClass = parseClassBody( whereName,
					  null, null, null, WHERE_FLAG );
		whereList.addElement( whereClass );
	      }
	      break;
	    default :
	      header = false;
	  }
      }

    ClassNode[] whereArray;
    if( whereList.size() == 0 )
      whereArray = null;
    else
      {
	whereArray = new ClassNode[whereList.size()];
	whereList.copyInto( whereArray );
      }

    ClassNode res = parseClassBody( className, parentArray, interfaceArray, whereArray, flags );

    return res;
  }

  /**
   *  Parses comma-separated identifier list.
   *  @returns array of parsed identifiers
   */
  IdentifierNode[] parseIdentifierList()
  {
    Vector list = new Vector();
    list.addElement( parseIdentifier( false, true ) );
    nextToken();
    while( in.tType == ',' )
      {
	consumeToken();
	list.addElement( parseIdentifier( false, true ) );
      }
    IdentifierNode[] array = new IdentifierNode[list.size()];
    list.copyInto( array );
    return array;
  }

  /**
   *  Parses class body starting with '{'.
   */
  ClassNode parseClassBody( IdentifierNode className,
				   IdentifierNode[] parentArray,
				   IdentifierNode[] interfaceArray,
				   ClassNode[] whereArray, 
				   int flags )
  {
    if( className == null )
      {
	Identifier[] id = { nextInnerClassIdentifier() };
	className = new IdentifierNode( 0, id );
      }

    classConsumer.tellErrorStorage( errorStorage );
    
    if( publicScope == currentClassHandle )
      flags |= PUBLIC_FLAG;
    Object handle = classConsumer.startScope( currentClassHandle, className.idList[0],
					      flags, parentArray, interfaceArray, ClassNode.class );
    boolean oldInInterface = inInterface;
    inInterface = (flags & INTERFACE_FLAG) != 0;
    Object prevHandle = currentClassHandle;
    Identifier prevClassName = currentClassName;
    try
      {
	currentClassHandle = handle;
	currentClassName = className.idList[0];
	nextToken();
	int filePos = in.filePos;
	Vector stList = new Vector();
	assertToken( '{' );
	while( true )
	  {
	    nextToken();
	    if( in.tType == '}' || in.tType == in.TOKEN_EOF )
	      break;
	    stList.addElement( parseStatement() );
	  }
	assertToken( '}' );
	StatementNode[] stArray = new StatementNode[stList.size()];
	stList.copyInto( stArray );
	ClassNode res = new ClassNode( className, parentArray, interfaceArray,
				       whereArray, stArray );
	res.addFlags( flags );
	classConsumer.finishScope( handle, res );
	return res;
      }
    finally
      {
	currentClassHandle = prevHandle;
	currentClassName = prevClassName;
	inInterface = oldInInterface;
      }
  }

  /**
   *  Parses method, field or class definition, starting with modifiers (if any).
   */
  DefinitionNode parseDefinition( int flags )
  {
    boolean header = true;
    nextToken();
    while( in.tType == ';' )
      {
	consumeToken();
	nextToken();
      }
    int filePos = in.filePos;
    while( header )
      {
	nextToken();
	switch( in.tType )
	  {
	    case TOKEN_PUBLIC :
	      flags |= PUBLIC_FLAG;
	      consumeToken();
	      break;
	    case TOKEN_PRIVATE :
	      flags |= PRIVATE_FLAG;
	      consumeToken();
	      break;
	    case TOKEN_PROTECTED :
	      flags |= PROTECTED_FLAG;
	      consumeToken();
	      break;
	    case TOKEN_STATIC :
	      flags |= STATIC_FLAG;
	      consumeToken();
	      break;	
	    case TOKEN_ABSTRACT :
	      flags |= ABSTRACT_FLAG;
	      consumeToken();
	      break;
	    case TOKEN_FINAL :
	      flags |= FINAL_FLAG;
	      consumeToken();
	      break;	
	    case TOKEN_SYNCHRONIZED :
	      flags |= SYNCHRONIZED_FLAG;
	      consumeToken();
	      break;
	    case TOKEN_NATIVE :
	      flags |= NATIVE_FLAG;
	      consumeToken();
	      break;
	    case TOKEN_VOLATILE :
	      flags |= VOLATILE_FLAG;
	      consumeToken();
	      break;
	    case TOKEN_TRANSIENT :
	      flags |= TRANSIENT_FLAG;
	      consumeToken();
	      break;
	    default :
	      header = false;
	  }
      }

    DefinitionNode res;
    if( in.tType == in.TOKEN_CLASS || in.tType == in.TOKEN_INTERFACE )
      res = parseClass( flags );
    else if( in.tType == '{' )
      {
	// probably, "static { }" construction 
	IdentifierNode clinit = new IdentifierNode( in.filePos, in.nameTable.CLINIT );
	IdentifierNode clinitType = new IdentifierNode( in.filePos, in.nameTable.VOID );
	consumeToken();
	Object handle = classConsumer.startScope( currentClassHandle,
		       nextInnerClassIdentifier( "clinit" ), 0, null, null, MethodNode.class );
	Object prevHandle = currentClassHandle;
	boolean oldInInterface = inInterface;
	inInterface = false;
	try
	  {
	    StatementNode[] body = parseStatementList();
	    assertToken( '}' );
	    res = new MethodNode( clinitType, clinit, null, null, body );
	    res.addFlags( flags );
	    classConsumer.finishScope( handle, res );
	  }
	finally
	  {
	    currentClassHandle = prevHandle;
	    inInterface = oldInInterface;
	  }
      }
    else
      {
	IdentifierNode type = parseIdentifier( false, true );
	IdentifierNode name;
	nextToken();
	if( in.tType < 0 )
	  {
	    name = parseIdentifier( false, true ); // field can have []
	    nextToken();
	  }
	else
	  {
	    name = type;
	    type = null; // constructor or omitted type
	  }
	
	if( in.tType == '(' )
	  res = parseOpenMethod( type, name, flags );
	else
	  {
	    res = parseOpenMember( type, name );
	  }
      }

    res.addFlags( flags );
    res.setFilePos( filePos );
    return res;
  }

  /**
   *  Parses method definition starting with '(' that follows method name.
   */
  MethodNode parseOpenMethod( IdentifierNode type, IdentifierNode name, int flags )
  {
    assertToken( '(' );
    nextToken();
    Vector paramList = new Vector();
    try
      {
	if( in.tType != ')' )
	  while( true )
	    {
	      paramList.addElement( parseParameter() );
	      nextToken();
	      if( in.tType != ',' )
		break;
	      consumeToken();
	    }
	assertToken( ')' );
      }
    catch( CompilerException e )
      {
	storeError( e );
      }
    ParameterNode[] paramArray = new ParameterNode[ paramList.size() ];
    paramList.copyInto( paramArray );
    
    nextToken();
    
    return parseMethodBody( type, name, paramArray, flags );
  }
  
  /**
   *  Parses method body starting with token that follows ')' that closes
   *  formal parameter list. It can handle the case of abstract or native
   *  methods (that do not have body) as well.
   */
  MethodNode parseMethodBody( IdentifierNode type, IdentifierNode name,
			      ParameterNode[] paramArray, int flags )
  {
    int arrCounter = 0;
    IdentifierNode[] exceptionArray = null;

    while( in.tType == '[' )
      {
	consumeToken();
	arrCounter++;
	assertToken( ']' );
	nextToken();
      }

    if( arrCounter > 0 && type != null )
      type = new IdentifierNode( type, new ExpressionNode[arrCounter] );

    if( type == null )
      flags |= SCRIPT_STYLE_FLAG;
    
    if( name.idList[0] == currentClassName && !looseExceptionCheck )
      flags &= ~SCRIPT_STYLE_FLAG;
    
    if( in.tType == in.TOKEN_THROWS )
      {
	consumeToken();
	exceptionArray = parseIdentifierList();
	nextToken();
      }
	
    if( (flags & (PRIVATE_FLAG|PROTECTED_FLAG|PACKAGE_PRIVATE_FLAG)) == 0 )
      {
	if( type == null || publicScope == currentClassHandle )
	  flags |= PUBLIC_FLAG;
      }

    MethodNode res;    
    Object handle = classConsumer.startScope( currentClassHandle,
	nextInnerClassIdentifier( name.idList[0].toString() ), flags, null, null, MethodNode.class );
    Object prevHandle = currentClassHandle;
    boolean oldInInterface = inInterface;
    inInterface = false;
    try
      {
	currentClassHandle = handle;

	if( in.tType == ';' )
	  {
	    // empty body
	    consumeToken();
	    res = new MethodNode( type, name, paramArray, exceptionArray, null );
	  }
	else
	  {
	    assertToken( '{' );
	    StatementNode[] body = parseStatementList();
	    assertToken( '}' );
	    res = new MethodNode( type, name, paramArray, exceptionArray, body );
	  }
	res.addFlags( flags );
	classConsumer.finishScope( handle, res );
	return res;
      }
    finally
      {
	currentClassHandle = prevHandle;
	inInterface = oldInInterface;
      }
  }

  /**
   *  Parses single formal parameter.
   */
  ParameterNode parseParameter()
  {
    nextToken();
    int filePos = in.filePos;
    boolean finalFlag = ( in.tType == in.TOKEN_FINAL );
    if( finalFlag )
      {
	consumeToken();
	nextToken();
      }
    if( in.tType >= 0 )
      throw new CompilerException( ERR_P_TYPEEXPECTED, in.filePos );
	
    IdentifierNode type = parseIdentifier( false, true );
    IdentifierNode name;

    nextToken();
    if( in.tType >= 0 )
      {
	// in java - error, we assume Any type
	name = type;
	type = null;
      }
    else
      name = parseIdentifier( false, false );

    nextToken();
    int arrCounter = 0;
    while( in.tType == '[' )
      {
	consumeToken();
	arrCounter++;
	assertToken( ']' );
	nextToken();
      }

    if( arrCounter > 0 )
      type = new IdentifierNode( type, new ExpressionNode[arrCounter] );

    return new ParameterNode( filePos, finalFlag, type, name );
  }

  /**
   *  Parses field (or variable) definition starting with token that follows variable name.
   */
  MemberNode parseOpenMember( IdentifierNode type, IdentifierNode name )
  {
    if( type != null && type.idList.length == 1 && type.idList[0] == in.nameTable.VAR )
      type = null;
    ExpressionNode[] vars = parseOpenVariableList( name );
    if( in.tType == '{' && vars.length == 1 && vars[0] instanceof IdentifierNode )
      {
	int filePos = in.filePos;
	IdentifierNode[] parents = { type };
	ClassNode inlineClass = parseClassBody( null, parents, null, null, PUBLIC_FLAG );
	type = inlineClass.name;
	vars[0] = new InfixNode( filePos, '=', vars[0], 
		      new NewNode( filePos, null, type, new ExpressionNode[0], inlineClass ) );
      }
    else
      assertToken( ';' );

    MemberNode m = new MemberNode( type, vars );
    if( type == null )
      m.addFlags( SCRIPT_STYLE_FLAG );
    if( publicScope == currentClassHandle )
      m.addFlags( PUBLIC_FLAG );
    if( inInterface )
      m.addFlags( PUBLIC_FLAG | FINAL_FLAG | STATIC_FLAG ); // interface members are always public final static
    return m;
  }

  /**
   *  Parses comma-separated variable name or initialization list
   *  starting with the first name.
   */
  ExpressionNode[] parseOpenVariableList( IdentifierNode name )
  {
    Vector variableList = new Vector();
    while( true )
      {
	nextToken();
	if( in.tType == '=' )
	  {
	    int pos1 = in.filePos;
	    consumeToken();
	    ExpressionNode init = parseExpression( 20 /* '=' priority */ );
	    variableList.addElement( opInfix( pos1, '=', name, init ) );
	    nextToken();
	    if( in.tType == ',' )
	      consumeToken();
	    else
	      break;
	  }
	else if( in.tType == ';' || in.tType == '{' )
	  {
	    variableList.addElement( name );
	    break;
	  }
	else if( in.tType == ',' )
	  {
	    consumeToken();
	    variableList.addElement( name );
	  }
	else
	  {
	    variableList.addElement( name );
	    break;
	  }
	name = parseIdentifier( false, true );
      }
    ExpressionNode[] variableArray = new ExpressionNode[variableList.size()];
    variableList.copyInto( variableArray );
    return variableArray;
  }

  /**
   *  Parses startement list inside '{...}' (but not brackets themselves).
   */
  BodyNode parseOpenBody()
  {
    int filePos = in.filePos;
    nextToken();
    Identifier prevClassName = currentClassName;
    currentClassName = null;
    try
      {
	return new BodyNode( filePos, parseStatementList() );
      }
    finally
      {
	currentClassName = prevClassName;
      }
  }

  /**
   *  Parsers a list of statements. Reads as much statements as it can. The following tokens
   *  cause it to stop: '}', ']', 'case', 'default', EOF.
   */
  StatementNode[] parseStatementList()
  {
    Vector statementList = new Vector();
    while( true )
      {
	nextToken();
	if( in.tType == '}' || in.tType == ']' || in.tType == in.TOKEN_CASE || 
	    in.tType == in.TOKEN_EOF || in.tType == in.TOKEN_DEFAULT )
	  break;
	try
	  {
	    StatementNode elem = parseStatement();
	    if( elem != null )
	      statementList.addElement( elem );
	  }
	catch( CompilerException e )
	  {
	    storeError( e );
	  }
      }
    StatementNode[] statementArray = new StatementNode[statementList.size()];
    statementList.copyInto( statementArray );
    return statementArray;
  }

  /**
   *  Parses a single statement. Jint statements include package declarations, imports,
   *  and class/interface definitions as well as regular Java statements.
   *  @returns parsed statement or null when end of file token reached.
   */
  public StatementNode parseStatement()
  {
    nextToken();
    int filePos = in.filePos;
    switch( in.tType )
      {
        case TOKEN_IF :
	  consumeToken();
	  return parseOpenIf();
        case TOKEN_WHILE :
	  consumeToken();
	  return parseOpenWhile();
        case TOKEN_DO :
	  consumeToken();
	  return parseOpenDo();
        case TOKEN_SYNCHRONIZED :
	  consumeToken();
	  nextToken();
	  if( in.tType == '(' )
	    return parseOpenSynchronized();
	  else
	    {
	      DefinitionNode def = parseDefinition(0);
	      def.addFlags( def.SYNCHRONIZED_FLAG );
	      return def;
	    }
        case TOKEN_FOR :
	  consumeToken();
	  return parseOpenFor();
        case TOKEN_TRY :
	  consumeToken();
	  return parseOpenTry();
	case TOKEN_RETURN :
	  consumeToken();
	  return parseOpenReturn();
	case TOKEN_THROW :
	  consumeToken();
	  return parseOpenThrow();
	case TOKEN_BREAK :
	  consumeToken();
	  return parseOpenBreak();
	case TOKEN_CONTINUE :
	  consumeToken();
	  return parseOpenContinue();
	case TOKEN_SWITCH :
	  consumeToken();
	  return parseOpenSwitch();
	case TOKEN_IMPORT :
	  consumeToken();
	  return parseOpenImport();
	case TOKEN_PACKAGE :
	  consumeToken();
	  nextToken();
	  if( in.tType == in.TOKEN_PRIVATE )
	    {
	      consumeToken();
	      return parseDefinition( PACKAGE_PRIVATE_FLAG );
	    }
	  else
	    return parseOpenPackage();
        case TOKEN_PUBLIC :
        case TOKEN_PRIVATE :
        case TOKEN_PROTECTED :
        case TOKEN_FINAL :
        case TOKEN_NATIVE :
        case TOKEN_STATIC :
        case TOKEN_ABSTRACT :
        case TOKEN_VOLATILE :
	case TOKEN_TRANSIENT :
	case TOKEN_CLASS :
	case TOKEN_INTERFACE :
	  return parseDefinition(0);
	case ';' :
	  consumeToken();
	  return new BodyNode( filePos, null );
	case '{' :
	  {
	    consumeToken();
	    BodyNode body = parseOpenBody();
	    assertToken( '}' );
	    return body;
	  }
	case TOKEN_EOF :
	  return null;
	default :
	  return parseExpressionStatement();
      }
  }

  /**
   *  Parses import statement (not including 'import' keyword).
   */
  ImportStatementNode parseOpenImport()
  {
    int filePos = in.filePos;
    IdentifierNode imp = parseIdentifier(true);
    classConsumer.tellImport( imp );
    ImportStatementNode n = new ImportStatementNode( filePos, imp );
    assertToken( ';' );
    return n;
  }

  /**
   *  Parses package statement (not including 'package' keyword).
   */
  PackageStatementNode parseOpenPackage()
  {
    int filePos = in.filePos;
    IdentifierNode pack = parseIdentifier();
    classConsumer.tellPackage( pack );
    PackageStatementNode n = new PackageStatementNode( filePos, pack );
    assertToken( ';' );
    return n;
  }

  /**
   *  Parses return statement (not including 'return' keyword).
   */
  ReturnNode parseOpenReturn()
  {
    int filePos = in.filePos;
    nextToken();
    if( in.tType == ';' )
      {
	consumeToken();
	return new ReturnNode( filePos, null );
      }
    else
      {
	ExpressionNode expr = parseExpression();
	assertToken( ';' );
	return new ReturnNode( filePos, expr );
      }
  }

  /**
   *  Parses throw statement (not including 'throw' keyword).
   */
  ThrowNode parseOpenThrow()
  {
    int filePos = in.filePos;
    nextToken();
    ExpressionNode expr = parseExpression();
    assertToken( ';' );
    return new ThrowNode( filePos, expr );
  }

  /**
   *  Parses break statement (not including 'break' keyword).
   */
  BreakNode parseOpenBreak()
  {
    int filePos = in.filePos;
    nextToken();
    if( in.tType == ';' )
      {
	consumeToken();
	return new BreakNode( filePos, null );
      }
    else if( in.tType < 0 && !in.ident.isReserved() )
      {
	IdentifierNode label = parseIdentifier();
	assertToken( ';' );
	return new BreakNode( filePos, label );
      }
    else
      {
	storeError( ERR_P_TOKENINSERTED, in.filePos, "';'" );
	return new BreakNode( filePos, null );
      }
  }

  /**
   *  Parses continue statement (not including 'continue' keyword).
   */
  ContinueNode parseOpenContinue()
  {
    int filePos = in.filePos;
    nextToken();
    if( in.tType == ';' )
      {
	consumeToken();
	return new ContinueNode( filePos, null );
      }
    else if( in.tType < 0 && !in.ident.isReserved() )
      {
	IdentifierNode label = parseIdentifier();
	assertToken( ';' );
	return new ContinueNode( filePos, label );
      }
    else
      {
	storeError( ERR_P_TOKENINSERTED, in.filePos, "';'" );
	return new ContinueNode( filePos, null );
      }
  }

  /**
   *  Parses if statement (not including 'if' keyword).
   */
  IfNode parseOpenIf()
  {
    nextToken();
    int filePos = in.filePos;
    assertToken( '(' );
    ExpressionNode cond = parseExpression();
    assertToken( ')' );
    StatementNode thenPart = parseStatement();
    StatementNode elsePart = null;
    nextToken();
    if( in.tType == in.TOKEN_ELSE )
      {
	consumeToken();
	elsePart = parseStatement();
      }
    return new IfNode( filePos, cond, thenPart, elsePart );
  }
  
  /**
   *  Parses while statement (not including 'while' keyword).
   */
  WhileNode parseOpenWhile()
  {
    nextToken();
    int filePos = in.filePos;
    assertToken( '(' );
    ExpressionNode cond = parseExpression();
    assertToken( ')' );
    StatementNode body = parseStatement();
    return new WhileNode( filePos, cond, body );
  }
  
  /**
   *  Parses do statement (not including 'do' keyword).
   */
  DoNode parseOpenDo()
  {
    nextToken();
    int filePos = in.filePos;
    StatementNode body = parseStatement();
    assertToken( in.TOKEN_WHILE, "while" );
    assertToken( '(' );
    ExpressionNode cond = parseExpression();
    assertToken( ')' );
    assertToken( ';' );
    return new DoNode( filePos, body, cond );
  }
  
  /**
   *  Parses synchronized statement (not including 'synchronized' keyword).
   */
  SynchronizedNode parseOpenSynchronized()
  {
    nextToken();
    int filePos = in.filePos;
    assertToken( '(' );
    ExpressionNode lock = parseExpression();
    assertToken( ')' );
    StatementNode body = parseStatement();
    return new SynchronizedNode( filePos, lock, body );
  }
  
  /**
   *  Parses switch statement (not including 'switch' keyword).
   */
  SwitchNode parseOpenSwitch()
  {
    nextToken();
    int filePos = in.filePos;
    assertToken( '(' );
    ExpressionNode expr = parseExpression();
    assertToken( ')' );
    assertToken( '{' );
    Vector caseList = new Vector();
    while( true )
      {
	nextToken();
	int line1 = filePos;
	if( in.tType == '}' || in.tType == in.TOKEN_EOF )
	  break;
	if( in.tType != in.TOKEN_CASE && in.tType != in.TOKEN_DEFAULT )
	  {
	    storeError( ERR_P_CASEEXPECTED, in.filePos );
	    while( in.tType != '}' && in.tType != in.TOKEN_EOF && 
		   in.tType != in.TOKEN_CASE && in.tType != in.TOKEN_DEFAULT )
	      {
		consumeToken();
		nextToken();
	      }
	    continue;
	  }
	ExpressionNode cexpr = null;
	if( in.tType == in.TOKEN_CASE )
	  {
	    consumeToken();
	    cexpr = parseExpression( 30 /* ':' priority */ );
	  }
	else
	  consumeToken();
	assertToken( ':' );
	StatementNode[] body = parseStatementList();
	caseList.addElement( new CaseNode( line1, cexpr, body ) );
      }
    assertToken( '}' );
    CaseNode[] caseArray = new CaseNode[caseList.size()];
    caseList.copyInto( caseArray );
    return new SwitchNode( filePos, expr, caseArray );
  }
  
  /**
   *  Parses <i>try</i> statement (not including 'try' keyword).
   */
  TryNode parseOpenTry()
  {
    int filePos = in.filePos;
    StatementNode body = parseStatement();
    Vector catchList = new Vector();
    while( true )
      {
	nextToken();
	int catchLine = in.filePos;
	if( in.tType == in.TOKEN_CATCH )
	  {
	    consumeToken();
	    assertToken( '(' );
	    IdentifierNode type = parseIdentifier();
	    IdentifierNode name = parseIdentifier();
	    assertToken( ')' );
	    StatementNode catchBody = parseStatement();
	    catchList.addElement( new CatchNode( filePos, type, name, catchBody ) );
	  }
	else if( in.tType == in.TOKEN_FINALLY )
	  {
	    consumeToken();
	    StatementNode catchBody = parseStatement();
	    catchList.addElement( new CatchNode( filePos, catchBody ) );
	    break;
	  }
	else
	  {
	    if( catchList.size() == 0 )
	      storeError( ERR_P_CATCHEXPECTED, in.filePos );
	    break;
	  }
      }
    CatchNode[] catchArray = new CatchNode[ catchList.size() ];
    catchList.copyInto( catchArray );
    return new TryNode( filePos, body, catchArray );
  }
  
  /**
   *  Parses <i>for</i> statement (not including 'for' keyword).
   */
  ForNode parseOpenFor()
  {
    nextToken();
    int filePos = in.filePos;
    assertToken( '(' );

    StatementNode init;
    ExpressionNode cond, next; 

    nextToken();
    if( in.tType == ';' )
      {
	init = null;
	consumeToken();
      }
    else
      init = parseStatement();

    nextToken();
    if( in.tType == ';' )
      cond = null;
    else
      cond = parseExpression();

    assertToken( ';' );

    nextToken();
    if( in.tType == ')' )
      next = null;
    else
      next = parseExpression();
    assertToken( ')' );

    StatementNode body = parseStatement();
    return new ForNode( filePos, init, cond, next, body );
  }
  
  /**
   *  Parses expression statement and certain other types of statements.
   *  In addition to expressions it can recognize labeled statement, variable
   *  and method definitions that do not start with a keyword.
   */
  StatementNode parseExpressionStatement()
  {
    nextToken();
    int filePos = in.filePos;
    ExpressionNode expr = parseExpression();
    if( expr instanceof LocalMethodExpression )
      return ((LocalMethodExpression)expr).method;
    else if( expr instanceof LabelExpression )
      return new LabeledNode( filePos, ((LabelExpression)expr).label, parseStatement() );
    else if( expr instanceof LocalVariableExpression )
      {
	LocalVariableExpression lv = (LocalVariableExpression)expr;
	MemberNode memb = lv.memb;
	memb.setFilePos( filePos );
	/*
	if( lv.finalFlag )
	  memb.addFlags( memb.FINAL_FLAG );
	*/
	return memb;
      }
    else
      {
	if( in.tType == '{' || in.tType == in.TOKEN_THROWS )
	  {
	    if( expr instanceof CallNode )
	      {
		CallNode callNode = (CallNode)expr;
		if( callNode.func instanceof IdentifierNode )
		  {
		    ExpressionNode[] args = callNode.args;
		    ParameterNode[] params = null;
		    boolean good = true;
		    if( args != null )
		      {
			params = new ParameterNode[args.length];
			for( int i = 0 ; i < args.length ; i++ )
			  {
			    ExpressionNode arg = args[i];
			    if( arg instanceof IdentifierNode )
			      {
				params[i] = new ParameterNode( arg.getFilePos(),
					  false, null, (IdentifierNode)arg );
				continue;
			      }
			    else if( arg instanceof InfixNode )
			      {
				InfixNode infixNode = (InfixNode)arg;
				if( infixNode.op == ' ' )
				  {
				    ExpressionNode type = infixNode.arg1;
				    ExpressionNode name = infixNode.arg2;

				    boolean finalFlag = false;
				    if( type instanceof InfixNode )
				      {
					InfixNode t = (InfixNode)type;
					type = t.arg2;
					// check for 'final TYPE'
					if( t.op == ' ' && t.arg1 instanceof IdentifierNode && 
					    ((IdentifierNode)t.arg1).idList.length == 1 &&
					    ((IdentifierNode)t.arg1).idList[0].id == TOKEN_FINAL &&
					    (type instanceof IdentifierNode || type instanceof InfixNode) )
					  {
					    type = toType(type);
					    finalFlag = true;
					  }
					// check for 'TYPE[]'
					else if( t.op == '[' )
					  {
					    type = toType( t );
					  }
					else
					  {
					    storeError( ERR_P_TYPEEXPECTED, t.getFilePos() );
					    type = null;
					  }
				      }
				    if( name instanceof InfixNode )
				      name = toType( name );
				    if( name != null )
				      {
					IdentifierNode n = (IdentifierNode)name;
					IdentifierNode t = (IdentifierNode)type;
					if( n.indexes != null )
					  {
					    if( t != null )
					      t = new IdentifierNode( t, n.indexes );
					    n.indexes = null;
					  }
					params[i] = new ParameterNode( infixNode.arg1.getFilePos(),
							   finalFlag, t, n );
					continue;
				      }
				  }
			      }
			    good = false;
			    break;
			  }
		      }
		    if( good )
		      {
			MethodNode node = parseMethodBody( null, (IdentifierNode)callNode.func,
							   params, SCRIPT_STYLE_FLAG );
			node.setFilePos( filePos );
			return node;
		      }
		  }
	      }
	  }
	assertToken( ';' );
	return new ExpressionStatementNode( filePos, expr );
      }
  }

  /**
   *  Returns infix operator priority. Higher priority means tighter binding,
   *  so priopity of '*' is higher then '+'. Brackets also treated somewhat like
   *  operators, so they also have a priopity.
   */
  private final int getPriority( int token )
  {
    switch( token )
      {
        case '.' :
        case TOKEN_SS :
	  return Integer.MAX_VALUE;
        case '(' :
        case '[' :
        case '{' :
        case TOKEN_PP :
        case TOKEN_MM :
	  return Integer.MAX_VALUE-1;
	case '*' :
	case '/' :
	case '%' :
	  return 130;
	case '+' :
	case '-' :
	  return 120;
	case TOKEN_GG :
	case TOKEN_GGG :
	case TOKEN_LL :
	  return 110;
	case '<' :
	case '>' :
	case TOKEN_GE :
	case TOKEN_LE :
	case TOKEN_INSTANCEOF :
	  return 100;
	case TOKEN_EE :
	case TOKEN_NE :
	case '~' :
	  return 90;
	case '&' :
	case '^' :
        case '|' : // all have the same priority in Java
	  return 60;
	case TOKEN_AND :
	  return 50;
	case TOKEN_OR :
	  return 40;
	case '?' :
	case ':' :
	  return 30;
	case '=' :
	case TOKEN_TE :
	case TOKEN_PE :
	case TOKEN_ME :
	case TOKEN_SE :
	case TOKEN_AE :
	case TOKEN_ANDE :
	case TOKEN_ORE :
	case TOKEN_XE :
	case TOKEN_PCE :
	case TOKEN_LLE :
	case TOKEN_GGE :
	case TOKEN_GGGE :
	  return 21; // odd - right associativity
	case ',' :
	  return 10;
      }
    return 0;
  }

  /**
   *  Parses an expression. Expression can be delimited by either
   *  closing bracket, ';' or EOF.
   */
  ExpressionNode parseExpression()
  {
    return parseExpression( -1 );
  }

  /**
   *  Parses a subexpression that contains operators of priority less
   *  or equal to <i>finPriority</i>.
   */
  ExpressionNode parseExpression( int finPriority )
  {
    boolean inBrackets = false;
    boolean finalFlag = false;
    ExpressionNode head = null;
    while( true )
      {
	nextToken();
	int filePos = in.filePos;
	if( in.tType == in.TOKEN_NEW )
	  {
	    consumeToken();
	    IdentifierNode type = parseIdentifier( false, true );
	    ExpressionNode[] args = null;
	    ClassNode inlineClass = null;
	    nextToken();
	    if( in.tType == '(' )
	      {
		consumeToken();
		nextToken();
		if( in.tType == ')' )
		  {
		    consumeToken();
		    nextToken();
		    args = new ExpressionNode[0];
		    if( in.tType == '{' )
		      {
			IdentifierNode[] parents = { type };
			inlineClass = parseClassBody( null, parents, null, null, 0 );
		      }
		  }
		else
		  {
		    args = parseArgs();
		    assertToken( ')' );
		  }
	      }
	    head = new NewNode( filePos, head, type, args, inlineClass );
	    continue;
	  }
	if( head == null )
	  {
	    // expect operand
	    switch( in.tType )
	      {
	        case '(' :
		  {
		    consumeToken();
		    nextToken();
		    switch( in.tType )
		      {
		        case TOKEN_DOUBLE :
		        case TOKEN_FLOAT :
		        case TOKEN_INT :
		        case TOKEN_SHORT :
		        case TOKEN_BYTE :
		        case TOKEN_CHAR :
		        case TOKEN_LONG :
			  {
			    IdentifierNode type = parseIdentifier( false, true );
			    assertToken( ')' );
			    head = parseExpression( Integer.MAX_VALUE-2 );
			    head = opCast( filePos, type, head );
			  }
			  break;
			default :
			  head = parseExpression( 0 );
			  assertToken( ')' );
			  inBrackets = true;
			  break;
		      }
		  }
		  break;
	        case '{' :
		  head = parseArrayValue( null );
		  break;
	        case '[' :
		  head = parseBlock();
		  break;
		case TOKEN_V_FLOAT :
		  head = new ConstNode( filePos, new Float( (float)in.dVal ) );
		  consumeToken();
		  break;
		case TOKEN_V_DOUBLE :
		  head = new ConstNode( filePos, new Double( in.dVal ) );
		  consumeToken();
		  break;
		case TOKEN_V_LONG :
		  head = new ConstNode( filePos, new Long( in.lVal ) );
		  consumeToken();
		  break;
		case TOKEN_V_INT :
		  if( Integer.MIN_VALUE <= in.lVal && 
		      0xFFFFFFFFL >= in.lVal )
		    head = new ConstNode( filePos, new Integer( (int)in.lVal ) );
		  else
		    head = new ConstNode( filePos, new Long( in.lVal ) );
		  consumeToken();
		  break;
		case TOKEN_V_STRING :
		  head = new ConstNode( filePos, in.sVal );
		  consumeToken();
		  break;
		case TOKEN_V_CHAR :
		  head = new ConstNode( filePos, new Character( (char)in.lVal ) );
		  consumeToken();
		  break;
		case TOKEN_V_REGEX :
		case TOKEN_V_PATT :
		  {
		    consumeToken();
		    try
		      {
			head = new RegexNode( filePos, in.sVal, (int)in.lVal, 
					      in.tType == in.TOKEN_V_REGEX );
		      }
		    catch( CompilerException e )
		      {
			storeError( e );
			head = new RegexNode( filePos, ".*", 0, true );
		      }
		    int flags;
		    if( (in.lVal & in.TOK_OPT_STANDALONE) != 0 )
		      flags = STATIC_FLAG; 
		    else
		      flags = 0;
		    flags |= PUBLIC_FLAG;
		    Object handle = classConsumer.startScope( currentClassHandle,
					nextRegexClassIdentifier(), flags, null, null, RegexNode.class );
		    classConsumer.finishScope( handle, head );
		  }
		  break;
		case TOKEN_V_SUBST :
		  {
		    consumeToken();
		    ReplacerNode rn;
		    try
		      {
			rn = new ReplacerNode( filePos, in.sVal,
						  in.sVal2, (int)in.lVal );
		      }
		    catch( CompilerException e )
		      {
			storeError( e );
			rn = new ReplacerNode( filePos, ".*", "", 0 );
		      }
		    head = rn;
		    RegexNode regex = rn.regex;
		    Object handle = classConsumer.startScope( currentClassHandle,
			      nextRegexClassIdentifier(), PUBLIC_FLAG, null, null, RegexNode.class );
		    classConsumer.finishScope( handle, regex );
		  }
		  break;
		case TOKEN_V_FORM :
		case TOKEN_V_BOUND :
		  consumeToken();
		  try
		    {
		      head = new FormatNode( filePos, in.tType == in.TOKEN_V_BOUND,
					 in.sVal, (int)in.lVal );
		    }
		  catch( CompilerException e )
		    {
		      storeError( e );
		      head = new FormatNode( filePos, in.tType == in.TOKEN_V_BOUND,
					     "<<<BAD_ FORMAT>>>", 0 );
		    }		  
		  break;
		case TOKEN_V_TRANS :
		  storeError( ERR_P_NOTIMPL, filePos, "t`..`" );
		  break;
		case '-' :
		case '+' :
		case '~' :
		case '!' :
		case TOKEN_PP :
		case TOKEN_MM :
		case TOKEN_SS :
		  {
		    int op = in.tType;
		    consumeToken();
		    head = parseExpression( Integer.MAX_VALUE-2 );
		    head = opPrefix( filePos, op, head );
		  }
		  break;
		/*
	        case TOKEN_FINAL :
		  consumeToken();
		  if( finPriority != -1 )
		    storeError( ERR_E_UNEXPECTED, filePos, "'final'" );
		  else
		    {
		      head = parseIdentifier( false, true );
		      nextToken();
		      if( in.tType >= 0 )
			storeError( ERR_VAREXPECTED, in.filePos );
		      finalFlag = true;
		    }
		  break;
		*/
	        case TOKEN_TRUE :
		  head = new ConstNode( filePos, TRUE );
		  consumeToken();
		  break;
	        case TOKEN_FALSE :
		  head = new ConstNode( filePos, FALSE );
		  consumeToken();
		  break;
	        case TOKEN_NULL :
		  head = new ConstNode( filePos, null );
		  consumeToken();
		  break;
		default :
		  if( in.tType < 0 )
		    {
		      head = parseIdentifier( false, false );
		      nextToken();
		      if( in.tType == ':' && finPriority == -1 )
			{
			  consumeToken();
			  nextToken();
			  if( in.tType == '{' || 
			      ( in.tType < 0 && in.ident.isReserved() &&
			        ( in.tType == in.TOKEN_DO ||
				  in.tType == in.TOKEN_FOR ||
				  in.tType == in.TOKEN_TRY ||
				  in.tType == in.TOKEN_SWITCH ||
				  in.tType == in.TOKEN_SYNCHRONIZED ||
				  in.tType == in.TOKEN_WHILE ) ) )
			    return new LabelExpression( (IdentifierNode)head );
			  else
			    {
			      ExpressionNode arg = parseExpression( 20 /* '=' priority */ );
			      ExpressionNode[] argList = { arg };
			      Identifier[] idList = ((IdentifierNode)head).idList;
			      String last = idList[idList.length-1].toString();
			      idList[idList.length-1] = 
				in.nameTable.lookupAndInsert( "set" + 
					Character.toUpperCase(last.charAt(0)) +
						  last.substring(1) );
			      head = opCall( filePos, head, argList );
			    }
			}
		    }
		  else
		    {		      
		      head = new ConstNode( filePos, new Integer(0) );
		      storeError( ERR_P_VALUEEXPECTED, filePos );
		    }
		  break;
	      }
	  }
	else
	  {
	    // "expect operator" state
	    int pri;
	    if( ( in.tType > 0 && in.tType < 0xFFFF ) ||
		in.tType == in.TOKEN_INSTANCEOF )
	      {
		pri = getPriority( in.tType );
		if( pri <= finPriority )
		  return head;
	      }
	    else
	      pri = 0;
	    switch( in.tType )
	      {
	        case TOKEN_EOF :
	        case ')' :
	        case ']' :
	        case ';' :
	        case '}' :
		  return head;
	        case '(' :
		  if( inBrackets )
		    {
		      // Cast
		      ExpressionNode arg = parseExpression( Integer.MAX_VALUE-2 );
		      head = opCast( filePos, toType(head), arg );
		    }
		  else
		    {
		      consumeToken();
		      nextToken();
		      if( in.tType == ')' )
			{
			  consumeToken();
			  nextToken();
			  head = opCall( filePos, head, null );
			}
		      else
			{
			  ExpressionNode[] args = parseArgs();
			  assertToken( ')' );
			  head = opCall( filePos, head, args );
			}
		    }
		  break;
	        case '[' :
		  {
		    // Note: syntax (...)[..] is always parsed as index op, not cast
		    // so no check for inBrackets here
		    consumeToken();
		    nextToken();
		    if( in.tType == ']' )
		      {
			consumeToken();
			nextToken();
			head = opInfix( filePos, '[', head, null );
		      }
		    else
		      {
			ExpressionNode expr = parseExpression( 0 );
			assertToken( ']' );
			head = opInfix( filePos, '[', head, expr );
		      }
		  }
		  break;
	        case '{' :
		  if( head instanceof InfixNode )
		    head = parseArrayValue( toType(head) );
		  else
		    return head;  // to parse function definition 
		  break;
	        case TOKEN_THROWS :
		  return head;  // to parse function definition 
		case TOKEN_PP :
		case TOKEN_MM :
		  head = opPostfix( filePos, (char)in.tType, head );
		  consumeToken();
		  break;
		case '?' :
		  {
		    consumeToken();
		    ExpressionNode arg1 = parseExpression( pri );
		    assertToken( ':' );
		    ExpressionNode arg2 = parseExpression( pri );
		    head = opCond( filePos, head, arg1, arg2 );
		  }
		  break;
		default :
		  if( pri > 0 )
		    {
		      // infix operator
		      int op = in.tType;
		      consumeToken();
		      if( op == '.' )
			{
			  nextToken();
			  if( in.tType == in.TOKEN_NEW )
			    {
			      // just act as if there were no '.'
			      continue;
			    }
			}
		      pri -= (pri%2); // this takes care of associativity
		      ExpressionNode arg = parseExpression( pri );
		      head = opInfix( filePos, op, head, arg );
		    }
		  else if( (head instanceof IdentifierNode) ||
			   (head instanceof InfixNode ) )
		    {
		      if( inBrackets )
			{
			  // Cast
			  ExpressionNode arg = 
			    parseExpression( Integer.MAX_VALUE-2 );
			  head = opCast( filePos, toType(head), arg );
			}
		      else if( finPriority == -1 )
			{
			  // Declaration
			  IdentifierNode name = parseIdentifier( false, true );
			  nextToken();
			  if( in.tType != '(' )
			    return new LocalVariableExpression( filePos,
			            parseOpenMember( toType(head), name ) );
			  else
			    return new LocalMethodExpression( filePos,
				    parseOpenMethod( toType(head), name, 0 ) );
			}
		      else
			{
			  // recover by inserting empty operator
			  ExpressionNode arg = parseExpression( 10 /* ',' priority */ );
			  head = opInfix( filePos, ' ', head, arg );
			}
		    }
		  else
		    {
		      storeError( ERR_P_OPEXPECTED, filePos );
		      consumeToken();
		    }
		  break;
	      }
	    inBrackets = false;
	  }
	  
      }
  }

  /**
   *  Parses comma-separated list of expressions.
   */
  private ExpressionNode[] parseArgs()
  {
    Vector argList = new Vector();
    while( true )
      {
	argList.addElement( parseExpression( 10 /* ',' priority */ ) );
	nextToken();
	if( in.tType != ',' )
	  break;
	consumeToken();
      }
    ExpressionNode[] args = new ExpressionNode[argList.size()];
    argList.copyInto( args );
    return args;
  }

  /**
   *  Converts indexed expression with empty indexes (array type name)
   *  to identifier node.
   */
  private IdentifierNode toType( ExpressionNode t )
  {
    if( t instanceof IdentifierNode )
      return (IdentifierNode)t;

    int size = 0;

    while( t instanceof InfixNode )
      {
	InfixNode n = (InfixNode)t;
	if( n.op != '[' )
	  {
	    storeError( ERR_P_TYPEEXPECTED, t.getFilePos() );
	    return null;
	  }
	t = n.arg1;
	size++;
	if( n.arg2 != null )
	  storeError( ERR_P_ARRNOTEMPTY, n.getFilePos() );
      }

    if( ! ( t instanceof IdentifierNode ) )
      {
	storeError( ERR_P_TYPEEXPECTED, t.getFilePos() );
	return null;
      }

    IdentifierNode tid = (IdentifierNode)t;

    if( size > 0 )
      tid.indexes = new ExpressionNode[size];

    return tid;
  }

  /**
   *  Parses Jint block (statement list inside '[ ]').
   */
  private ExpressionNode parseBlock()
  {
    nextToken();
    int filePos = in.filePos;
    NameTable nameTable = in.nameTable;
    Identifier name = nextInnerClassIdentifier();
    Identifier[] blockSpec = { nameTable.KMY, nameTable.JINT, nameTable.LANG, nameTable.BLOCK };
    IdentifierNode[] classes = { new IdentifierNode( filePos, blockSpec ) };
    Identifier[] argvSpec = { nameTable.ARGV };
    IdentifierNode argv =  new IdentifierNode( filePos, argvSpec );
    Identifier[] arglistSpec = { nameTable.KMY, nameTable.JINT, nameTable.LANG, nameTable.ARGLIST };
    IdentifierNode arglist =  new IdentifierNode( filePos, arglistSpec );
    ParameterNode[] params = { new ParameterNode( filePos, true, arglist, argv ) };
    Object classHandle = classConsumer.startScope( currentClassHandle,
				 name, PUBLIC_FLAG, classes, null, ClassNode.class );
    Object handle = classConsumer.startScope( classHandle,
		nameTable.EVAL, PUBLIC_FLAG|SCRIPT_STYLE_FLAG, null, null, MethodNode.class );
    Object prevHandle = currentClassHandle;
    boolean oldInInterface = inInterface;
    inInterface = false;
    try
      {
	currentClassHandle = handle;
	IdentifierNode nameId = new IdentifierNode( filePos, name );
	IdentifierNode eval = new IdentifierNode( filePos, nameTable.EVAL );
	assertToken( '[' );
	StatementNode[] statements = parseStatementList();
	assertToken( ']' );
	MethodNode method = new MethodNode( null, eval, params, null, statements );
	method.setFilePos( filePos );
	method.addFlags( PUBLIC_FLAG|SCRIPT_STYLE_FLAG );
	classConsumer.finishScope( handle, method );
	StatementNode[] stmts = { method };
	ClassNode clazz = new ClassNode( nameId, classes, null, null, stmts );
	clazz.setFilePos( filePos );
	classConsumer.finishScope( classHandle, clazz );
	return new NewNode( filePos, null, classes[0], new ExpressionNode[0], clazz );
      }
    finally
      {
	currentClassHandle = prevHandle;
	inInterface = oldInInterface;
      }
  }

  /**
   *  Parses array expression (comma-separated expression list inside '{ }').
   */
  private ArrayValueNode parseArrayValue( IdentifierNode type )
  {
    // Array initializer, take into account hanging comma: { a, b, }
    nextToken();
    int filePos = in.filePos;
    assertToken( '{' );
    nextToken();
    Vector valList = new Vector();
    while( in.tType != '}' )
      {
	valList.addElement( parseExpression( 10 /* ',' priority */ ) );
	nextToken();
	if( in.tType != ',' )
	  break;
	consumeToken();
	nextToken();
      }
    assertToken( '}' );
    ExpressionNode[] vals = new ExpressionNode[valList.size()];
    valList.copyInto( vals );
    return new ArrayValueNode( filePos, type, vals );
  }

  /**
   *  Parses qualified identifier.
   */
  private IdentifierNode parseIdentifier()
  {
    return parseIdentifier( false, false );
  }

  /**
   *  Parses qualified identifier.
   *  If <i>starAllowed</i> is true, in addition, '*' can be a part of it (as in import statement).
   *  If <i>starAllowed</i> is false, in addition, a list of '[...]' (either empty or not)
   *  is allowed at the end.
   */
  private IdentifierNode parseIdentifier( boolean starAllowed )
  {
    return parseIdentifier( starAllowed, !starAllowed );
  }

  /**
   *  Parses qualified identifier.
   *  If <i>starAllowed</i> is true, in addition, '*' can be a part of it (as in import statement).
   *  If <i>indexAllowed</i> is true, in addition, a list of '[...]' (either empty or not)
   *  is allowed at the end.
   */
  private IdentifierNode parseIdentifier( boolean starAllowed, boolean indexAllowed )
  {
    int filePos = in.filePos;
    nextToken();
    Vector idList = new Vector();
    while( true )
      {
	if( in.tType == in.TOKEN_NEW )
	  break;
	if( in.tType >= 0 && !( starAllowed && in.tType == '*' ) )
	  {
	    storeError( ERR_P_IDENTEXPECTED, in.filePos );
	    break;
	  }
	if( in.tType == '*' )
	  idList.addElement( null );
	else
	  idList.addElement( in.ident );
	consumeToken();
	nextToken();
	if( in.tType != '.' )
	  break;
	consumeToken();
	nextToken();
      }

    Identifier[] ids = new Identifier[idList.size()];
    idList.copyInto( ids );

    if( in.tType != '[' || !indexAllowed )
      return new IdentifierNode( filePos, ids );

    Vector indexList = new Vector();

    while( in.tType == '[' )
      {
	consumeToken();
	nextToken();
	if( in.tType == ']' )
	  {
	    consumeToken();
	    indexList.addElement( null );
	  }
	else
	  {
	    indexList.addElement( parseExpression( 0 ) );
	    assertToken( ']' );
	  }
	nextToken();
      }

    ExpressionNode[] indexArray = new ExpressionNode[ indexList.size() ];
    indexList.copyInto( indexArray );
    
    return new IdentifierNode( filePos, ids, indexArray );
  }

  ExpressionNode opCall( int filePos, ExpressionNode expr, 
			ExpressionNode[] args )
  {
    return new CallNode( filePos, expr, args );
  }

  ExpressionNode opInfix( int filePos, int op,
			 ExpressionNode head, ExpressionNode arg )
  {
    if( op == TOKEN_INSTANCEOF )
      arg = toType( arg );
    return new InfixNode( filePos, op, head, arg );
  }
  
  ExpressionNode opPrefix( int filePos, int op, ExpressionNode arg )
  {
    return new PrefixNode( filePos, op, arg );
  }
  
  ExpressionNode opPostfix( int filePos, int op, ExpressionNode arg )
  {
    return new PostfixNode( filePos, op, arg );
  }
  
  ExpressionNode opCond( int filePos,
		 ExpressionNode head, ExpressionNode arg1, ExpressionNode arg2 )
  {
    return new CondNode( filePos, head, arg1, arg2 );
  }
  
  ExpressionNode opCast( int filePos, IdentifierNode type, ExpressionNode arg )
  {
    return new CastNode( filePos, type, arg );
  }  
}
