/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.parser;

import kmy.jint.reflect.AbstractClass;

/**
 *  This is a representation of a class or array name, that
 *  is something like java.lang.Hashtable or java.lang.Object[].
 *  QualifiedIdentifier consists of a list of Identifiers and a
 *  number of dimensions.
 *  A qualified identifier can be resolved by a JintClassTable
 *  to an AbstractClass.
 *  <p>
 *  Node: arrays that are passed to QualifiedIdentifier constructors
 *  are not copied and should not be modified afterwards.
 */
public class QualifiedIdentifier
{
  Identifier[] arr;
  int off;
  int len;
  int dim;

  AbstractClass type;

  /**
   *  Creates a qualified identifier from a list of identifiers.
   */
  public QualifiedIdentifier( Identifier[] arr )
  {
    this( arr, 0, arr.length, 0 );
  }

  /**
   *  Creates a qualified identifier from a list of identifiers, taking
   *  only identifiers from index off to off+len-1 inclusively.
   */
  public QualifiedIdentifier( Identifier[] arr, int off, int len )
  {
    this( arr, off, len, 0 );
  }

  /**
   *  Creates array name: qualified identifier with number of dimensions.
   *  Creates a qualified identifier from a list of identifiers, taking
   *  only identifiers from index off to off+len-1 inclusively.
   */
  public QualifiedIdentifier( Identifier[] arr, int off, int len, int dim )
  {
    this.arr = arr;
    this.len = len;
    this.off = off;
    this.dim = dim;
  }

  /**
   *  Creates a new QualifiedIdentifer by appending Identifier id
   *  to a qualified identifier qi.
   */
  public QualifiedIdentifier( QualifiedIdentifier qi, Identifier id )
  {
    arr = new Identifier[qi.length()+1];
    qi.copyInto( arr );
    len = arr.length;
    off = 0;
    arr[arr.length-1] = id;
  }

  /**
   *  Creates array name: qualified identifier with number of dimensions.
   *  Takes qualified identifier (that already can have some number of
   *  dimensions) and additional number of dimensions. 
   */
  public QualifiedIdentifier( QualifiedIdentifier qi, int dim )
  {
    this.arr = qi.arr;
    this.off = qi.off;
    this.len = qi.len;
    this.dim = qi.dim + dim;
  }

  /**
   *  Creates QualifiedIdentifier by dropping the last identifier in
   *  original QualifiedIdentifier's identifier list.
   */
  public QualifiedIdentifier dropOne()
  {
    return new QualifiedIdentifier( arr, off, len-1 );
  }

  /**
   *  Returns QualifiedIdentifier produced by stripping
   *  all dimensions from the original QualifiedIdentifer.
   */
  public QualifiedIdentifier base()
  {
    if( dim == 0 )
      return this;
    else
      return new QualifiedIdentifier( arr, off, len );    
  }

  /**
   *  Returns true if the other object is a QualifiedIdentifier equal
   *  to this one.
   */
  public boolean equals( Object obj )
  {
    if( obj instanceof QualifiedIdentifier )
      {
	QualifiedIdentifier other = (QualifiedIdentifier)obj;
	return same( other );
      }
    return false;
  }

  /**
   *  Returns true if the other QualifiedIdentifier equal to this one.
   *  Qualified identifiers are considered to be equal if they consist of
   *  the same identifiers going in the same order and have the same number
   *  of dimensions.
   */
  public boolean same( QualifiedIdentifier other )
  {
    if( other.len != len || other.dim != dim )
      {
	return false;
      }
    for( int i = len-1 ; i >= 0 ; i-- )
      if( other.arr[other.off+i] != arr[off+i] )
	{
	  return false;
	}
    return true;
  }

  /**
   *  Returns true if this QualifiedIdentifier's identifier list
   *  starts with the other QualifiedIdentifier's identifier list.
   */
  public boolean startsWith( QualifiedIdentifier other )
  {
    if( other.len > len )
      {
	return false;
      }
    for( int i = other.len-1 ; i >= 0 ; i-- )
      if( other.arr[other.off+i] != arr[off+i] )
	{
	  return false;
	}
    return true;
  }

  /**
   *  Hash code is calculated by the last identifier in QualifiedIdentifier's
   *  identifier list.
   */
  public int hashCode()
  {
    return arr[off+len-1].hashCode();
  }

  /**
   *  Returns last identifier in this QualifiedIdentifier's identifier list.
   */
  public Identifier lastIdentifier()
  {
    return arr[off+len-1];
  }

  /**
   *  Returns identifier at position <i>i</i> in this QualifiedIdentifier's
   *  identifier list. The first identifier's position is 0. 
   */
  public Identifier getIdentifier( int i )
  {
    return arr[off+i];
  }

  /**
   *  Returns number of identifiers in this QualifiedIdentifier's identifier list.
   */
  public int length()
  {
    return len;
  }

  /**
   *  Returns String representation of this QualifiedIdentifier: dot-separated
   *  list of identifiers plus '[]' for every dimension.
   */
  public String toString()
  {
    return toString( '.' );
  }

  /**
   * Returns JVM class name for non-inner, non-array classes.
   */
  public String toSimpleJVMName()
  {
    return toString( '/' );
  }

  /**
   *  The same as toString(), but uses <i>sep</i> instead of '.' as separator.
   */
  public String toString( char sep )
  {
    StringBuffer sb = new StringBuffer();
    for( int i = 0 ; i < len ; i++ )
      {
	if( i != 0 )
	  sb.append( sep );
	sb.append( arr[off+i].name );
      }
    for( int i = 0 ;i < dim ; i++ )
      sb.append( "[]" );
    return sb.toString();
  }

  /**
   *  Writes class descriptor to a given StringBuffer.
   */
  public void writeDescriptor( StringBuffer buf )
  {
    if( type != null )
      buf.append( type.getDescriptor() );
    else
      {
	// should not really happen
	for( int i = 0 ; i < dim ; i++ )
	  buf.append( '[' );
	buf.append( 'L' );
	for( int i = 0 ; i < len ; i++ )
	  {
	    if( i != 0 )
	      buf.append( '/' );
	    buf.append( arr[off+i].name );
	  }
	buf.append( ';' );
	(new RuntimeException( "Hacked descriptor: " + this )).printStackTrace();
      }
  }

  /**
   *  This method is called when QualifiedIdentifier is resolved to
   *  an AbstractClass, so the next time it can be resolved without
   *  any lookups.
   */
  public void setType( AbstractClass type )
  {
    this.type = type;
  }

  /**
   *  Returns an AbstractClass to which this QualifiedIdentifier was resolved
   *  or null if it was not yet resolved or resolved unsuccessfully.
   */
  public AbstractClass getType()
  {
    return type;
  }

  /**
   *  Copies this QualifiedIdentifier's identifier list into a given array.
   */
  public void copyInto( Identifier[] idArray )
  {
    System.arraycopy( arr, off, idArray, 0, len );
  }

  /**
   *  Returns number of dimensions of this QualifiedIdentifier.
   */
  public int getDimension()
  {
    return dim;
  }
}
