/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.parser;

import java.util.*;
import java.io.*;

import kmy.jint.util.CompilerException;
import kmy.jint.constants.ErrorCodes;

/**
 *  Implementation of AbstractTokenizer that reads Jint tokens from a Reader.
 *  Jint tokens are basically the same as Java tokens, except for text
 *  processing tokens (like regular expression), '~=' token and '::'
 *  token (which is not currently used).
 */
public class ReaderTokenizer extends AbstractTokenizer implements ErrorCodes
{

  final static int SUPER_BIT = 0x10000;

  final static int BUF_SIZE = 8*1024;
  final static int UNGET_SIZE = 4;

  Reader    in;
  char[]    buffer = new char[BUF_SIZE+UNGET_SIZE];
  int       index;
  int       size;

  int       rFilePos = makeFilePos( 1, 0 );

  boolean   eofDetected = false;

  boolean   nestedCommentsEnabled = false;

  /**
   *  If this flag is set and the next character is '/' it will be treated
   *  as an opening character of regex.
   */
  boolean   slashWillBeRegex = true;

  /**
   *  Creates a new ReaderTokenizer given a Reader stream and a NameTable. Most of
   *  the time one has to create a single NameTable and pass it to all ReaderTokenizers.
   */
  public ReaderTokenizer( Reader in, NameTable nt )
  {
    super( nt );
    this.in = in;
    nameTable = nt;
  }

  final private void error( int code, Object info )
  {
    if( info != null )
      {
	Object[] add = { info };
	throw new CompilerException( code, rFilePos, add );
      }
    throw new CompilerException( code, rFilePos );
  }

  void fillBuffer()
  {
    index = UNGET_SIZE;
    if( eofDetected )
      size = UNGET_SIZE;
    else
      {
	try
	  {
	    size = UNGET_SIZE + in.read( buffer, UNGET_SIZE, BUF_SIZE );
	  }
	catch( IOException e )
	  {
	    tType = TOKEN_EOF;
	    error( ERR_L_IOERROR, e.getMessage() );
	  }
	eofDetected = size == index;
      }
  }

  private final void nextLine()
  {
    rFilePos &= ~0x3FF;
    rFilePos += 0x400;
  }

  private void prevLine()
  {
    rFilePos -= 0x400;
    rFilePos |= 0x3FF;
  }

  void ungetChar( int c )
  {
    if( c == -1 )
      return;
    buffer[--index] = (char)c;
    if( c == '\n' )
      prevLine();
    else
      rFilePos--;
    // generally, ungetted char loses its SUPER_BIT, with exception of backslash
    if( c == ( SUPER_BIT | '\\' ) )
      {
	buffer[--index] = (char)c;
	rFilePos--;
      }
  }

  final private int nextRawChar()
  {
    if( index >= size )
      {
	fillBuffer();
	if( index >= size )
	  return -1;
      }
	
    char r = buffer[index++];
    if( r == '\n' )   // Mac: sorry
      nextLine();
    else
      rFilePos++; // line numbering will be screwed for >1000 char long lines
    return r;
  }

  final private int readHexChar( int max )
  {
    int acc = 0;
    for( int i = 0 ; i < max ; i++ )
      {
	int c = nextRawChar();
	if( '0' <= c && c <= '9' )
	  c -= '0';
	else if( 'A' <= c && c <= 'F' )
	  c -= 'A' - 10;
	else if( 'a' <= c && c <= 'f' )
	  c -= 'a' - 10;
	else
	  {
	    ungetChar( c );
	    return acc;
	  }
	acc = (acc << 4) | c;
      }
    return acc;
  }

  final private int readOctalChar()
  {
    int acc = 0;
    for( int i = 0 ; i < 3 ; i++ )
      {
	int c = nextRawChar();
	if( '0' <= c && c <= '7' )
	  c -= '0';
	else
	  {
	    ungetChar( c );
	    return acc | SUPER_BIT;
	  }
	acc = (acc << 3) | c;
      }
    return acc | SUPER_BIT;
  }

  final private int nextChar()
  {
    while( true )
      {
	int c = nextRawChar();
	if( c != '\\' )
	  return c;
	
	c = nextRawChar();
	switch( c )
	  {
	    case -1 :
	      tType = TOKEN_EOF;
	      error( ERR_L_BACKSLASHEOF, null );
	    case '\n' :
	      continue;
	    case '\\' :
	      return SUPER_BIT | '\\';
	    case 'n' :
	      return SUPER_BIT | '\n';
	    case 'b' :
	      return SUPER_BIT | '\b';
	    case 'r' :
	      return SUPER_BIT | '\r';
	    case 't' :
	      return SUPER_BIT | '\t';
	    case 'a' :
	      return SUPER_BIT | '\7';
	    case 'x' :
	      return readHexChar( 2 ) | SUPER_BIT;
	    case 'u' :
	      return readHexChar( 4 );
	    default :
	      if( '0' <= c && c <= '7' )
		{
		  ungetChar( c );
		  return readOctalChar();
		}
	      return SUPER_BIT | c;
	  }
      }
  }

  /**
   *  Read char in regex
   */
  final private int nextRegChar()
  {
    while( true )
      {
	int c = nextRawChar();
	if( c != '\\' )
	  return c;
	
	c = nextRawChar();
	switch( c )
	  {
	    case -1 :
	      tType = TOKEN_EOF;
	      error( ERR_L_BACKSLASHEOF, null );
	    case '\n' :
	      continue;
	    case '`' :
	    case '/' :
	      return SUPER_BIT | c;
	    case 'u' :
	      return readHexChar( 4 );
	    default:
	      ungetChar( c );
	      return '\\';
	  }
      }
  }

  private void readString()
  {
    StringBuffer sb = null;
    while( true )
      {
	int startIndex = index;
	char c = 0;
	while( index < size )
	  {
	    c = buffer[index++];
	    if( c == '\\' || c == '"' || c == '\n' )
	      break;
	  }

	int len = index - startIndex;
	rFilePos += len;

	if( c == '"' || c == '\n' )
	  {
	    tType = TOKEN_V_STRING;
	    if( sb == null )
	      sVal = new String( buffer, startIndex, len - 1 );
	    else
	      {
		sb.append( buffer, startIndex, len - 1 );
		sVal = sb.toString();
	      }
	    if( c == '\n' )
	      {
		index--;
		rFilePos--;
		error( ERR_L_STRINGEND, null );
	      }
	    return;
	  }

	if( c == '\\' )
	  {
	    rFilePos--;
	    index--;
	    len--;
	  }

	if( sb == null )
	  sb = new StringBuffer();
	sb.append( buffer, startIndex, len );

	int cSuper = nextChar();
	if( cSuper == -1 )
	  {
	    tType = TOKEN_EOF;
	    error( ERR_L_EOFINSTRING, null );
	  }

	if( cSuper != '\n' && cSuper != '"' )
	  sb.append( (char)cSuper );
	else
	  ungetChar( cSuper );
      }
  }

  private void readCharacter()
  {
    int c = nextChar();
    if( c == -1 )
      {
	tType = TOKEN_EOF;
	error( ERR_L_EOFINCHAR, null );
      }
    if( c == '\'' )
      {
	tType = TOKEN_V_CHAR;
	lVal = 0;
	error( ERR_L_EMPTYCHAR, null );
      }
    lVal = c & 0xFFFF; // possibly reset SUPER_BIT
    tType = TOKEN_V_CHAR;
    c = nextChar();
    if( c != '\'' )
      {
	ungetChar( c );
	error( ERR_L_CHAREND, null );
      }
  }

  void skipSlashSlashComment()
  {
    while( true )
      {
	while( index < size && buffer[index] != '\n' )
	  {
	    index++; // don't increment rFilePos, rely on line part
	  }
	int c = nextRawChar();
	if( c == '\n' || c == -1 )
	  break;
      }
  }

  void skipSlashStarComment()
  {
    int nesting = 1;
    while( nesting > 0 )
      {
	int state = 0;  /* 1 - saw '*', 2 - saw '/' */
	while( index < size )
	  {
	    char c0 = buffer[index];
	    if( c0 == '\n' )
	      nextLine();
	    else
	      rFilePos++;
	    index++;
	    if( c0 == '*' )
	      {
		if( index < size && buffer[index] == '/' )
		  {
		    nesting--;
		    index++;
		    rFilePos++;
		    if( nesting == 0 )
		      return;
		    continue;
		  }
		state = 1;
		break;
	      }
	    if( c0 == '/' )
	      {
		if( index < size && buffer[index] == '*' )
		  {
		    if( nestedCommentsEnabled )
		      nesting++;
		    index++;
		    rFilePos++;
		    continue;
		  }
		state = 2;
		break;
	      }
	  }
	int c = nextRawChar();
	if( c == -1 )
	  {
	    tType = TOKEN_EOF;
	    error( ERR_L_COMMENTEND, null );
	  }
	if( state == 1 && c == '/' )
	  nesting--;
	else if( state == 2 && c == '*' )
	  nesting++;
	else
	  ungetChar( c );
      }
  }

  private void readSpecial( char c0 )
  {
    int c1Super = nextChar();
    char c1 = (char)c1Super;
    switch( c0 )
      {
        case '.' :
	  ungetChar( c1Super );
	  if( '0' <= c1 && c1 <= '9' )
	    readNumber( '.' );
	  else
	    tType = '.';
	  break;
        case '/' :
	  switch( c1 )
	    {
	      case '/' :
	        skipSlashSlashComment();
		break;
	      case '*' :
		skipSlashStarComment();
		break;
	      case '=' :
		tType = TOKEN_SE;
		break;
	      default :
		tType = '/';
		ungetChar( c1Super );
	    }
	  break;
	case '+' :
	  switch( c1 )
	    {
	      case '=' :
	        tType = TOKEN_PE;
		break;
	      case '+' :
	        tType = TOKEN_PP;
		break;
	      default :
		tType = '+';
		ungetChar( c1Super );
	    }
	  break;
	case '-' :
	  switch( c1 )
	    {
	      case '=' :
	        tType = TOKEN_ME;
		break;
	      case '-' :
	        tType = TOKEN_MM;
		break;
	      default :
		tType = '-';
		ungetChar( c1Super );
	    }
	  break;
	case '|' :
	  switch( c1 )
	    {
	      case '=' :
	        tType = TOKEN_ORE;
		break;
	      case '|' :
	        tType = TOKEN_OR;
		break;
	      default :
		tType = '|';
		ungetChar( c1Super );
	    }
	  break;
	case '&' :
	  switch( c1 )
	    {
	      case '=' :
	        tType = TOKEN_ANDE;
		break;
	      case '&' :
	        tType = TOKEN_AND;
		break;
	      default :
		tType = '&';
		ungetChar( c1Super );
	    }
	  break;
	case '<' :
	  switch( c1 )
	    {
	      case '=' :
	        tType = TOKEN_LE;
		break;
	      case '<' :
		c1Super = nextChar();
		c1 = (char)c1Super;
		if( c1 == '=' )
		  tType = TOKEN_LLE;
		else
		  {
		    tType = TOKEN_LL;
		    ungetChar( c1Super );
		  }
		break;
	      default :
		tType = '<';
		ungetChar( c1Super );
	    }
	  break;
	case '>' :
	  switch( c1 )
	    {
	      case '=' :
	        tType = TOKEN_GE;
		break;
	      case '>' :
		c1Super = nextChar();
		c1 = (char)c1Super;
		if( c1 == '=' )
		  tType = TOKEN_GGE;
		else if( c1 == '>' )
		  {
		    c1Super = nextChar();
		    c1 = (char)c1Super;
		    if( c1 == '=' )
		      tType = TOKEN_GGGE;
		    else
		      {
			tType = TOKEN_GGG;
			ungetChar( c1Super );
		      }
		  }
		else
		  {
		    tType = TOKEN_GG;
		    ungetChar( c1Super );
		  }
		break;
	      default :
		tType = '>';
		ungetChar( c1Super );
	    }
	  break;
	case '*' :
	  switch( c1 )
	    {
	      case '=' :
	        tType = TOKEN_AE;
		break;
	      default :
		tType = '*';
		ungetChar( c1Super );
	    }
	  break;
	case '=' :
	  switch( c1 )
	    {
	      case '=' :
	        tType = TOKEN_EE;
		break;
	      default :
		tType = '=';
		ungetChar( c1Super );
	    }
	  break;
	case '^' :
	  switch( c1 )
	    {
	      case '=' :
	        tType = TOKEN_XE;
		break;
	      default :
		tType = '^';
		ungetChar( c1Super );
	    }
	  break;
	case '!' :
	  switch( c1 )
	    {
	      case '=' :
	        tType = TOKEN_NE;
		break;
	      default :
		tType = '!';
		ungetChar( c1Super );
	    }
	  break;
	case '%' :
	  switch( c1 )
	    {
	      case '=' :
	        tType = TOKEN_PCE;
		break;
	      default :
		tType = '%';
		ungetChar( c1Super );
	    }
	  break;
	case '~' :
	  switch( c1 )
	    {
	      case '=' :
	        tType = TOKEN_TE;
		break;
	      default :
		tType = '~';
		ungetChar( c1Super );
	    }
	  break;
        case ':' :
	  switch( c1 )
	    {
	      case ':' :
	        tType = TOKEN_SS;
		break;
	      default :
		tType = ':';
		ungetChar( c1Super );
	    }
	  break;
	default :
	  tType = c0;
	  ungetChar( c1Super );
      }
  }

  private void readNumber( char c0 )
  {
    boolean doubleFlag = false;
    boolean exponentFlag = false;
    boolean exponentSignFlag = false;
    boolean hexFlag = false;

    ungetChar( c0 );
    StringBuffer sb = null;
    while( true )
      {
	int startIndex = index;
	char c = 0;
	boolean problemChar = false;

	while( index < size )
	  {
	    c = buffer[index++];
	    if( '0' <= c && c <= '9' )
	      {
		if( exponentFlag )
		  exponentSignFlag = true;
		continue;
	      }
	    if( hexFlag && ( ( 'A' <= c && c <= 'F' ) ||
			     'a' <= c && c <= 'f' ) )
	      continue;
	    if( ( c == 'x' || c == 'X' ) && 
	       ! ( hexFlag || exponentFlag || doubleFlag ) )
	      {
		hexFlag = true;
		continue;
	      }
	    if( c == '.' )
	      {
		if( !(doubleFlag || hexFlag) )
		  {
		    doubleFlag = true;
		    continue;
		  }
		problemChar = true;
		break;
	      }
	    if( c == 'e' || c == 'E' )
	      {
		if( exponentFlag || hexFlag )
		  {
		    problemChar = true;
		    break;
		  }
		doubleFlag = true;
		exponentFlag = true;
		continue;
	      }
	    if( c == 'l' || c == 'L' )
	      {
		if( exponentFlag || doubleFlag )
		  problemChar = true;
		break; // can be only the last char!
	      }
	    if( c == 'd' || c == 'D' )
	      {
		if( hexFlag )
		  problemChar = true;
		doubleFlag = true;
		break; // can be only the last char!
	      }
	    if( c == 'f' || c == 'F' )
	      {
		if( hexFlag )
		  problemChar = true;
		doubleFlag = true;
		break; // can be only the last char!
	      }
	    if( c == '+' || c == '-' )
	      {
		if( ! exponentFlag )
		  {
		    problemChar = true;
		    break;
		  }
		if( exponentSignFlag )
		  {
		    problemChar = true;
		    break;
		  }
		exponentSignFlag = true;
		continue;
	      }
	    problemChar = true;
	    break;
	  }

	int len = index - startIndex;
	rFilePos += len;

	if( problemChar )
	  {
	    rFilePos--;
	    index--;
	    len--;
	  }

	if( problemChar && c != '\\' )
	  {
	    if( sb == null )
	      sVal = new String( buffer, startIndex, len );
	    else
	      {
		sb.append( buffer, startIndex, len );
		sVal = sb.toString();
	      }
	    try
	      {
		String s;
		boolean lFlag = sVal.endsWith( "L" ) || sVal.endsWith( "l" );
		boolean dFlag = !hexFlag && (sVal.endsWith( "D" ) || sVal.endsWith( "d" ));
		boolean fFlag = !hexFlag && (sVal.endsWith( "F" ) || sVal.endsWith( "f" ));
		if( lFlag || dFlag || fFlag )
		  s = sVal.substring(0,sVal.length()-1);
		else
		  s = sVal;
		if( doubleFlag )
		  {
		    tType = ( fFlag ? TOKEN_V_FLOAT : TOKEN_V_DOUBLE );
		    dVal = (new Double( sVal )).doubleValue();
		  }
		else if( hexFlag && 
			 ( s.startsWith( "0x" ) || s.startsWith( "0X" ) ) )
		  {
		    tType = ( lFlag ? TOKEN_V_LONG : TOKEN_V_INT );
		    s = s.substring(2);
		    if( s.length() == 16 )
		      {
			// a simple way to avoid
			// non-negativity requirement of Long.parseLong
			lVal = Long.parseLong( s.substring(0, 8), 16 );
			lVal <<= 32;
			lVal |= Long.parseLong( s.substring(8), 16 );
		      }
		    else
		      lVal = Long.parseLong( s, 16 );
		  }
		else if( s.startsWith( "0" ) )
		  {
		    tType = ( lFlag ? TOKEN_V_LONG : TOKEN_V_INT );
		    lVal = Long.parseLong( s, 8 );
		  }
		else
		  {
		    tType = ( lFlag ? TOKEN_V_LONG : TOKEN_V_INT );
		    lVal = Long.parseLong( s );
		  }
	      }
	    catch( NumberFormatException e )
	      {
		dVal = 0;
		lVal = 0;
		error( ERR_L_NUMFORMAT, sVal );
	      }
	    return;
	  }

	if( sb == null )
	  sb = new StringBuffer();
	sb.append( buffer, startIndex, index-startIndex );

	int cSuper = nextChar();
	if( Character.isDigit( (char)cSuper ) )
	  sb.append( (char)cSuper );
	else
	  ungetChar( cSuper );
      }
  }

  private void readWord( char c0 )
  {
    ungetChar( c0 );
    StringBuffer sb = null;
    while( true )
      {
	int startIndex = index;
	char c = 0;
	boolean problemChar = false;

	while( index < size )
	  {
	    c = buffer[index++];
	    if( 'a' <= c && c <= 'z' )
	      continue;
	    if( 'A' <= c && c <= 'Z' )
	      continue;
	    if( '0' <= c && c <= '9' )
	      continue;
	    if( c == '_' )
	      continue;
	    if( c <= 0x177 || !Character.isLetter(c) )
	      {
		problemChar = true;
		break;
	      }
	  }

	int len = index - startIndex;
	rFilePos += len;

	if( problemChar )
	  {
	    rFilePos--;
	    index--;
	    len--;
	  }

	if( problemChar && c != '\\' )
	  {
	    if( sb == null )
	      sVal = new String( buffer, startIndex, len );
	    else
	      {
		sb.append( buffer, startIndex, len );
		sVal = sb.toString();
	      }
	    ident = nameTable.lookupAndInsert( sVal );
	    tType = ident.id;
	    return;
	  }

	if( sb == null )
	  sb = new StringBuffer();
	sb.append( buffer, startIndex, len );

	int cSuper = nextChar();
	c = (char)cSuper;
	if( Character.isLetterOrDigit( c ) || c == '_' )
	  sb.append( c );
	else
	  ungetChar( cSuper );
      }
  }

  private int readRegOpt()
  {
    int c;
    int res = 0;
    while( true )
      {
	c = nextChar();
	switch( (char)c )
	  {
	    case 'g' :
	      res |= TOK_OPT_GLOBAL;
	      break;
	    case 'i' :
	      res |= TOK_OPT_IGNORE;
	      break;
	    case 'm' :
	      res |= TOK_OPT_MULTILINE;
	      break;
	    case 's' :
	      res |= TOK_OPT_SINGLELINE;
	      break;
	    case 'B' :
	      res |= TOK_OPT_BUFFERONLY;
	      break;
	    case 'D' :
	      res |= TOK_OPT_DECLARE;
	      break;
	    case 'O' :
	      res |= TOK_OPT_OFFLINE;
	      break;
	    case 'R' :
	      res |= TOK_OPT_RETAINALL;
	      break;
	    case 'S' :
	      res |= TOK_OPT_STANDALONE | TOK_OPT_OFFLINE;
	      break;
	    default:
	      ungetChar( c );
	      return res;
	  }
      }
  }

  private String readRegSpan( char delim )
  {
    StringBuffer sb = new StringBuffer();
    while( true )
      {
	int cSuper = nextRegChar();
	if( cSuper == -1 )
	  error( ERR_L_EOFINREGEX, null );
	if( cSuper == '\n' )
	  error( ERR_L_REGEXEND, null );
	if( cSuper == delim )
	  break;
	if( (cSuper&SUPER_BIT) != 0 && cSuper != (delim|SUPER_BIT) )
	  sb.append( '\\' );
	sb.append( (char)cSuper );
      }
    return sb.toString();
  }

  /**
   *  Read regex, substitution, transliteration or file pattern
   */
  private void readRegex( char t, char delim )
  {
    switch( t )
      {
        case 'm' :
	  tType = TOKEN_V_REGEX;
	  sVal = readRegSpan( delim );
	  break;
        case 'p' :
	  tType = TOKEN_V_PATT;
	  sVal = readRegSpan( delim );
	  break;
        case 'f' :
	  tType = TOKEN_V_FORM;
	  sVal = readRegSpan( delim );
	  break;
        case 'b' :
	  tType = TOKEN_V_BOUND;
	  sVal = readRegSpan( delim );
	  break;
        case 'r' :
        case 's' :
	  tType = TOKEN_V_SUBST;
	  sVal = readRegSpan( delim );
	  sVal2 = readRegSpan( delim );
	  break;
        case 't' :
        case 'y' :
	  tType = TOKEN_V_TRANS;
	  sVal = readRegSpan( delim );
	  sVal2 = readRegSpan( delim );
	  break;
        default :
	  error( ERR_L_NEVER, null );
      }
    lVal = readRegOpt();
  }

  public void nextToken()
  {
    if( tType == TOKEN_EOF )
      return;

    int cSuper;
    char c;

    tType = TOKEN_NONE;

    while( tType == TOKEN_NONE )
      {
	cSuper = nextChar();

	if( cSuper == -1 )
	  {
	    tType = TOKEN_EOF;
	    filePos = rFilePos;
	    return;
	  }

	c = (char)cSuper;

	if( c <= ' ' )
	  continue;

	filePos = rFilePos;

	if( c < 'A' )
	  {
	    if( c < '0' )
	      {
		if( c == '"' )
		  readString();
		else if( c == '\'' )
		  readCharacter();
		else if( c == '/' && slashWillBeRegex )
		  {
		    // if we use nextChar here, backslashes like in /\w+/ will be lost
		    int cs = nextRawChar();
		    ungetChar( cs );
		    if( cs != '*' && cs != '/' )
		      readRegex( 'm', '/' );
		    else
		      readSpecial( c );
		  }
		else
		  readSpecial( c );
	      }
	    else if( c <= '9' )
	      readNumber( c );
	    else
	      readSpecial( c );
	  }
	else if( c > 'z' )
	  {
	    if( c >= 0x177 )
	      {
		if( Character.isLetter( c ) )
		  readWord( c );
		else
		  {
		    tType = TOKEN_NONE;
		    error( ERR_L_ILLEGALCHAR, new Character(c) );
		  }
	      }
	    else
	      readSpecial( c );
	  }
	else if( c > 'Z' && c < 'a' && c != '_' )
	  {
	    if( c == '`' )
	      readRegex( 'b', c );
	    else
	      readSpecial( c );
	  }
	else
	  {
	    if( c == 's' || c == 'p' || c == 'm' || c == 'r' ||
		c == 'y' || c == 't' || c == 'f' || c == 'b' )
	      {
		cSuper = nextChar();
		if( (char)cSuper == '`' )
		  {
		    readRegex( c, '`' );
		    break;
		  }
		ungetChar( cSuper );
		readWord( c );
	      }
	    else
	      readWord( c );
	  }
      }

    slashWillBeRegex = ( tType > 0 && tType < 0x10000 && tType != TOKEN_V_REGEX &&
      tType != TOKEN_V_SUBST && tType != TOKEN_V_TRANS && tType != TOKEN_V_PATT &&
      tType != TOKEN_V_FORM && tType != TOKEN_V_BOUND && tType != ']' &&
      tType != ')' && tType != TOKEN_PP && tType != TOKEN_MM ) || 
      (tType == TOKEN_CASE || tType == TOKEN_RETURN || tType == TOKEN_THROW);
  }

  public void close()
  {
    try
      {
	in.close();
      }
    catch( IOException e )
      {
	// ignore
      }
  }

  public static void main( String[] args )
  {
    try
      {
	AbstractTokenizer t = 
	   new ReaderTokenizer( new InputStreamReader( System.in ),
	     new NameTable() );
	while( true )
	  {
	    t.nextToken();
	    System.out.print( "" + getLine(t.filePos) + ":" +
			      getColumn(t.filePos) + "  " );
	    if( t.tType == t.TOKEN_EOF )
	      break;
	    if( t.tType == t.TOKEN_V_STRING )
	      System.out.println( "String: " + t.sVal );
	    else if( t.tType == t.TOKEN_WORD )
	      System.out.println( "Word: " + t.sVal );
	    else if( t.tType == t.TOKEN_V_CHAR )
	      System.out.println( "Char: " + (char)t.lVal );
	    else if( t.tType == t.TOKEN_V_INT )
	      System.out.println( "Int: " + t.lVal );
	    else if( t.tType == t.TOKEN_V_DOUBLE )
	      System.out.println( "Double: " + t.dVal );
	    else if( t.tType == t.TOKEN_V_REGEX )
	      System.out.println( "Regex: " + t.sVal + " " + t.lVal );
	    else if( t.tType == t.TOKEN_V_SUBST )
	      System.out.println( "Subst: " + t.sVal + " " + t.sVal2 + " " + t.lVal );
	    else if( t.tType == t.TOKEN_V_TRANS )
	      System.out.println( "Trans: " + t.sVal + " " + t.sVal2 + " " + t.lVal );
	    else if( t.tType == t.TOKEN_V_PATT )
	      System.out.println( "Patt: " + t.sVal + " " + t.lVal);
	    else if( t.tType > 0 )
	      System.out.println( "Token: " + (char)t.tType );
	    else if ( t.ident != null )
	      System.out.println( "Ident: " + t.ident.name + " = " + t.tType );
	    else
	      System.out.println( "Token: Type" + (-t.tType) );
	  }
      }
    catch( Exception e )
      {
	e.printStackTrace();
      }
  }

}
