/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.reflect;

import kmy.jint.parser.QualifiedIdentifier;
import kmy.jint.parser.Identifier;
import java.util.*;
import java.lang.reflect.Array;

import kmy.jint.constants.DefinitionConst;

public class BasicClass extends AbstractClass implements DefinitionConst
{

  JintClassTable classTable;
  QualifiedIdentifier className;
  QualifiedIdentifier packageName;

  int modifiers;
  String descriptor;
  String name;

  Hashtable members = new Hashtable();

  AbstractClass superclass;
  AbstractClass[] interfaces;

  AbstractClass[] inner;

  AbstractClass outer;

  AbstractField outerField = null;

  public BasicClass( JintClassTable ct, QualifiedIdentifier className,
		    AbstractClass outer )
  {
    classTable = ct;
    this.className = className; // can be null for CompiledClass
    this.outer = outer;
    initDescriptor();
  }

  public QualifiedIdentifier getPackage()
  {
    return packageName;
  }

  protected void initDescriptor()
  {
    if( className != null )
      {
	if( outer == null || outer.isWorld() )
	  packageName = (className == null ? null : className.dropOne() );
	else
	  packageName = outer.getPackage();
	name = className.toSimpleJVMName();
	if( outer == null || outer.isWorld() )
	  {
	    descriptor = "L" + name + ";";
	  }
	else
	  descriptor = "L" + outer.getJVMName() + "$" + className.lastIdentifier() + ";";
	className.setType( this );
      }
  }

  public String getName()
  {
    return name;
  }

  public QualifiedIdentifier getFullName()
  {
    return className;
  }

  public int getModifiers()
  {
    return modifiers;
  }

  public AbstractField getField( Identifier name )
  {
    try
      {
	return (AbstractField)(((Object[])members.get(name))[FIELD_INDEX]);
      }
    catch( NullPointerException e )
      {
	return null;
      }
  }

  public AbstractField[] getFields()
  {
    Vector acc = new Vector();
    Enumeration list = members.elements();
    while( list.hasMoreElements() )
      {
	Object[] m = (Object[])list.nextElement();
	Object fld = m[FIELD_INDEX];
	if( fld != null )
	  acc.addElement( fld );
      }
    AbstractField[] res = new AbstractField[acc.size()];
    acc.copyInto( res );
    return res;
  }

  public Enumeration getMethods( Identifier name, int nparam )
  {
    nparam += METHOD_INDEX;
    try
      {
	return ((Hashtable)(((Object[])members.get(name))[nparam])).elements();
      }
    catch( NullPointerException e )
      {
	// method does not exist
	return null;
      }    
    catch( ArrayIndexOutOfBoundsException e )
      {
	// method does not exist
	return null;
      }    
  }

  public AbstractClass[] getInnerClasses()
  {
    return inner;
  }

  public AbstractClass[] getImplementedInterfaces()
  {
    return interfaces;
  }

  public AbstractClass getSuperclass()
  {
    if( superclass == null && this != classTable.objectClass )
      superclass = classTable.objectClass;
    return superclass;
  }

  public boolean isSubclassOf( AbstractClass cls )
  {
    if( cls == null )
      return false;
    getSuperclass(); // to initialize superclass if needed
    // System.out.println( this + " is subclass of " + cls + " " + superclass );
    if( cls.isInterface() )
      {
	if( interfaces != null )
	  for( int i = 0 ; i < interfaces.length ; i++ )
	    if( interfaces[i] == cls || interfaces[i].isSubclassOf( cls ) )
	      return true;
      }
    else
      {
	// superclass is never an interface, so skip this check for interfaces
	if( cls == superclass )
	  return true;
      }
    if( superclass == null )
      return false;
    return superclass.isSubclassOf( cls );
  }

  public boolean isInterface()
  {
    return (modifiers & INTERFACE_FLAG) != 0;
  }

  public boolean isPrimitive()
  {
    return false;
  }

  public String getDescriptor()
  {
    return descriptor;
  }

  public JintClassTable getClassTable()
  {
    return classTable;
  }

  public AbstractField getOuterField()
  {
    return outerField;
  }

  public AbstractClass getOuter()
  {
    return ( (modifiers & STATIC_FLAG) == 0 ? outer : null);
  }

  public AbstractClass getLexicallyOuter()
  {
    return outer;
  }

  void doneAdding()
  {
  }

  //----------------

  public void dumpMethods( Identifier name )
  {
    System.err.println( "Objects named " + name + " in class " + this + ": " );
    Object[] arr = (Object[])members.get(name);
    int count = 0;
    if( arr != null )
      {
	for( int i = 0 ; i < arr.length ; i++ )
	  {
	    Object sub = arr[i];
	    if( sub == null )
	      continue;
	    if( sub instanceof Hashtable )
	      {
		Enumeration list = ((Hashtable)sub).elements();
		while( list.hasMoreElements() )
		  {
		    count++;
		    System.err.println( "\t" + list.nextElement() );
		  }
	      }
	    else
	      {
		count++;
		System.err.println( "\t" + sub );
	      }
	  }
      }
    if( count == 0 )
      System.err.println( "\t<-- No objects: " + arr + " -->" );
  }

  void addMember( Identifier name, Object member, int nparams )
  {
    Object[] arr = (Object[])members.get( name );
    if( arr == null || arr.length <= nparams )
      {
	Object[] newArr = new Object[nparams+1];
	if( arr != null )
	  System.arraycopy( arr, 0, newArr, 0, arr.length );
	arr = newArr;
	members.put( name, arr );
      }
    if( arr[nparams] == null )
      if( nparams == 0 )
	{
	  arr[nparams] = member;
	  return;
	}
      else
	arr[nparams] = new Hashtable();
  
    if( nparams == 0 )
      throw new IllegalArgumentException( "Field " + name + " already exists in " + this );
    Hashtable table = (Hashtable)(arr[nparams]);
    table.put( ((AbstractMethod)member).getTypeDescriptor(), member );
  }  

}

