/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.reflect;

import java.lang.reflect.*;
import java.util.*;

import kmy.jint.parser.*;
import kmy.jint.lang.ArgList;

import kmy.jint.tree.DefinitionNode;

public class JVMClass extends BasicClass
{
  Class                clazz;

  int ncells = 1;

  /**
   *  Called only by JintClassTable.
   */
  JVMClass( JintClassTable classTable, QualifiedIdentifier fullName, Class clazz )
  {
    super( classTable, fullName, null );

    // System.out.println( "Creating JVMClass: " + fullName );

    this.clazz = clazz;

    if( clazz.isPrimitive() )
      {
	name = clazz.getName();
	if( clazz == Double.TYPE )
	  {
	    descriptor = "D";
	    ncells = 2;
	  }
	else if( clazz == Long.TYPE )
	  {
	    descriptor = "J";
	    ncells = 2;
	  }
	else if( clazz == Integer.TYPE )
	  descriptor = "I";
	else if( clazz == Boolean.TYPE )
	  descriptor = "Z";
	else if( clazz == Byte.TYPE )
	  descriptor = "B";
	else if( clazz == Short.TYPE )
	  descriptor = "S";
	else if( clazz == Character.TYPE )
	  descriptor = "C";
	else if( clazz == Float.TYPE )
	  descriptor = "F";
	else if( clazz == Void.TYPE )
	  {
	    descriptor = "V";
	    ncells = 0;
	  }
	else
	  throw new RuntimeException( "Unknown primitive type: " + clazz );
      }
    else
      {
	name = clazz.getName().replace( '.', '/' );
	descriptor = "L" + name + ";";
      }
    fullName.setType( this );
    
    classTable.registerClass( this );

    if( ! classTable.isStartUp() )
      finish();
  }

  void finish()
  {

    // System.out.println( "Finishing JVMClass: " + getFullName() );

    Class scl = clazz.getSuperclass();
    if( scl == null && clazz != Object.class )
      scl = Object.class;
    if( scl != null )
      superclass = classTable.getClass( scl );
      
    setImplementedInterfaces();

    addMethods( clazz );

    Constructor[] constructorList = clazz.getDeclaredConstructors();
    for( int i = 0 ; i < constructorList.length ; i++ )
      {
	Constructor constructor = constructorList[i];
	Class[] argTypes = constructor.getParameterTypes();
	int narg = argTypes.length;
	if( narg > 0 && argTypes[narg-1] == ArgList.class )
	  {
	    addMember( constructor, ARGLIST_INDEX );
	  }
	else
	  addMember( constructor, narg + METHOD_INDEX );
      }

    Field[] fieldList = clazz.getDeclaredFields();
    for( int i = 0 ; i < fieldList.length ; i++ )
      {
	addMember( fieldList[i], FIELD_INDEX );
      }

    modifiers = DefinitionNode.fromJVMModifiers( clazz.getModifiers() );
  }

  private void addMethods( Class clazz )
  {
    Method[] methodList = clazz.getDeclaredMethods();
    for( int i = 0 ; i < methodList.length ; i++ )
      {
	Method method = methodList[i];
	Class[] argTypes = method.getParameterTypes();
	int narg = argTypes.length;
	if( narg > 0 && argTypes[narg-1] == ArgList.class )
	  addMember( method, ARGLIST_INDEX );
	else
	  addMember( method, narg + METHOD_INDEX );
      }
  }


  private void addMember( Member member, int nparams )
  {
    NameTable nameTable = classTable.getNameTable();
    Identifier name;
    try
      {
	if( member instanceof Constructor )
	  {
	    name = nameTable.INIT;
	    addMember( name, new JVMMethod( classTable, this, (Constructor)member ), nparams );
	  }
	else
	  {
	    name = nameTable.lookupAndInsert( member.getName() );
	    if( member instanceof Field )
	      addMember( name, new JVMField( classTable, this, (Field)member ), nparams );
	    else
	      addMember( name, new JVMMethod( classTable, this, (Method)member ), nparams );
	  }
      }
    catch( Exception e )
      {
	e.printStackTrace();
      }
  }
  
  private void setImplementedInterfaces()
  {
    Class[] classes = clazz.getInterfaces();
    AbstractClass[] res = new AbstractClass[classes.length];
    for( int i = 0 ;i < res.length ; i++ )
      res[i] = classTable.getClass( classes[i] );
    interfaces = res;
  }

  public boolean isInterface()
  {
    return clazz.isInterface();
  }

  public boolean isPrimitive()
  {
    return clazz.isPrimitive();
  }

  public int getNCells()
  {
    return ncells;
  }

  public AbstractClass getSuperclass()
  {
    return superclass;
  }

  // ---- JVMClass specific:

  public Class getRealClass()
  {
    return clazz;
  }

}

