/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.reflect;

import java.lang.reflect.*;

import kmy.jint.tree.DefinitionNode;

public class JVMMethod extends AbstractMethod
{

  String          name;

  Member          method; // can be Method or Constructor
  AbstractClass   declaring;

  AbstractClass   type;
  AbstractClass[] params;
  AbstractClass[] exceptions;

  int             modifiers;

  JVMMethod( JintClassTable classTable, JVMClass declaring, Method method )
  {
    this.method = method;
    this.declaring = declaring;

    modifiers = DefinitionNode.fromJVMModifiers( method.getModifiers() );

    name = method.getName();
  }

  JVMMethod( JintClassTable classTable, JVMClass declaring, Constructor method )
  {
    this.method = method;
    this.declaring = declaring;

    this.type = classTable.voidType;
    
    modifiers = DefinitionNode.fromJVMModifiers( method.getModifiers() );
    modifiers |= DefinitionNode.STATIC_FLAG;

    name = classTable.getNameTable().INIT.name;
  }

  public String getName()
  {
    return name;
  }

  public int getModifiers()
  {
    return modifiers;
  }

  public AbstractClass getDeclaringClass()
  {
    return declaring;
  }

  public AbstractClass getType()
  {
    if( type == null )
      {
	if( method instanceof Method )
	  {
	    Method m = (Method)method;
	    synchronized( this )
	      {
		if( type == null )
		  {
		    JintClassTable classTable = declaring.getClassTable();
		    type = classTable.getClass( m.getReturnType() );
		    if( type == null )
		      throw new RuntimeException( "Class " + 
						  m.getReturnType().getName() + " not found" );
		  }
	      }
	  }
	else
	  type = declaring.getClassTable().voidType;
      }
    return type;
  }

  public AbstractClass[] getExceptionTypes()
  {
    if( exceptions == null )
      {
	synchronized( this )
	  {
	    if( exceptions == null )
	      {
		Class[] classArr;
		if( method instanceof Method )
		  classArr = ((Method)method).getExceptionTypes();
		else
		  classArr = ((Constructor)method).getExceptionTypes();
		JintClassTable classTable = declaring.getClassTable();
		exceptions = new AbstractClass[classArr.length];
		for( int i = 0 ; i < exceptions.length ; i++ )
		  {
		    exceptions[i] = classTable.getClass( classArr[i] );
		    if( exceptions[i] == null )
		      throw new RuntimeException( "Class " + classArr[i].getName() + " not found" );
		  }
	      }
	  }
      }
    return exceptions;
  }

  public AbstractClass[] getParameterTypes()
  {
    if( params == null )
      {
	synchronized( this )
	  {
	    if( params == null )
	      {
		Class[] classArr;
		if( method instanceof Method )
		  classArr = ((Method)method).getParameterTypes();
		else
		  classArr = ((Constructor)method).getParameterTypes();
		JintClassTable classTable = declaring.getClassTable();
		params = new AbstractClass[classArr.length];
		for( int i = 0 ; i < params.length ; i++ )
		  {
		    params[i] = classTable.getClass( classArr[i] );
		    if( params[i] == null )
		      throw new RuntimeException( "Class " + classArr[i].getName() + " not found" );
		  }
	      }
	  }
      }
    return params;
  }
      
  public Member getMethod()
  {
    return method;
  }

  static String makeDescriptor( Class clazz )
  {
    if( clazz.isPrimitive() )
      {
	if( clazz == Double.TYPE )
	  return "D";
	else if( clazz == Long.TYPE )
	  return "J";
	else if( clazz == Integer.TYPE )
	  return "I";
	else if( clazz == Boolean.TYPE )
	  return "Z";
	else if( clazz == Byte.TYPE )
	  return "B";
	else if( clazz == Short.TYPE )
	  return "S";
	else if( clazz == Character.TYPE )
	  return "C";
	else if( clazz == Float.TYPE )
	  return "F";
	else if( clazz == Void.TYPE )
	  return "V";
	else
	  throw new RuntimeException( "Unknown primitive type: " + clazz );
      }
    else
      {
	String s = clazz.getName().replace( '.', '/' );
	if( s.startsWith( "[" ) )
	  return s;
	else
	  return "L" + s + ";";
      }
  }

  static String makeDescriptor( Class[] params, Class type )
  {
    StringBuffer sb = new StringBuffer();
    sb.append( '(' );
    if( params != null )
      {
	for( int i = 0 ; i < params.length ; i++ )
	  {
	    sb.append( makeDescriptor( params[i] ) );
	  }
      }
    sb.append( ')' );
    sb.append( makeDescriptor( type ) );
    return sb.toString();
  }

  protected void initDescriptor()
  {
    if( descriptor == null )
      if( method instanceof Method )
	{
	  Method m = (Method)method;
	  descriptor = makeDescriptor( m.getParameterTypes(), m.getReturnType() ).intern();
	}
      else
	{
	  Constructor c = (Constructor)method;
	  descriptor = makeDescriptor( c.getParameterTypes(), Void.TYPE ).intern();
	}
    if( longDescriptor == null )
      longDescriptor = (getName().toString() + descriptor).intern();
  }


}
