/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.reflect;

import kmy.jint.tree.MethodNode;
import kmy.jint.tree.MemberNode;
import kmy.jint.tree.ClassNode;
import kmy.jint.tree.StatementNode;
import kmy.jint.tree.DefinitionNode;
import kmy.jint.tree.BasicNode;
import kmy.jint.tree.IdentifierNode;
import kmy.jint.util.ErrorStorage;
import kmy.jint.util.CompilerException;
import kmy.jint.parser.Identifier;
import kmy.jint.parser.QualifiedIdentifier;
import kmy.jint.parser.QualifiedIdentifier;
import kmy.jint.parser.NameTable;

import java.util.*;
import java.lang.reflect.Array;

import kmy.jint.constants.DefinitionConst;
import kmy.jint.constants.ErrorCodes;

public class JintClass extends BasicClass implements ErrorStorage, ErrorCodes
{

  private static int staticCounter = 1;

  private BasicNode body;
  private StatementNode[] statements;

  private Vector innerList;

  private Hashtable importCache;

  private Hashtable methodBodies; // by descriptor

  private IdentifierNode[] imports;

  private boolean incomplete = true;

  private ErrorStorage errorStorage;

  private int accessorCount = 1;
  
  public JintClass( JintClassTable ct, QualifiedIdentifier className,
		    AbstractClass outer, int modifiers )
  {
    this( ct, className, outer );
    this.modifiers = modifiers;
    if( outer instanceof ErrorStorage )
      errorStorage = (ErrorStorage)outer;
    //System.out.println( "Created: " + this + " " + Integer.toHexString( modifiers ) );
  }

  public JintClass( JintClassTable ct, QualifiedIdentifier className,
		    AbstractClass outer )

  {
    super( ct, className, outer );

    if( outer instanceof JintClass )
      importCache = ((JintClass)outer).getImportCache();
    else
      importCache = new Hashtable();
  }

  public AbstractClass getSuperclass()
  {
    if( superclass == null )
      throw new IllegalStateException( "Class " + this + " not fully created yet!" );
    return superclass;
  }

  public AbstractField getField( Identifier name )
  {
    if( incomplete )
      {
	((JintClassTable)getClassTable()).completeScope( body );
	incomplete = false;
      }
    return super.getField( name );
  }

  public Enumeration getMethods( Identifier name, int nparams )
  {
    if( incomplete )
      {
	((JintClassTable)getClassTable()).completeScope( body );
	incomplete = false;
      }
    return super.getMethods( name, nparams );
  }

  //-------- specific

  public BasicNode getBody()
  {
    return body;
  }

  public MethodNode getMethodBody( JintMethod method )
  {
    if( methodBodies == null )
      return null;
    return (MethodNode)methodBodies.get( method.getDescriptor() );
  }

  public MethodNode getMethodBody( String method )
  {
    if( methodBodies == null )
      return null;
    return (MethodNode)methodBodies.get( method );
  }

  public StatementNode[] getStatements()
  {
    return statements;
  }

  public Hashtable getImportCache()
  {
    return importCache;
  }

  public void setImports( IdentifierNode[] imports )
  {
    this.imports = imports;
  }

  public IdentifierNode[] getImports()
  {
    return imports;
  }

  void setMethodBody( JintMethod method, MethodNode body)
  {
    if( methodBodies == null )
      methodBodies = new Hashtable();
    methodBodies.put( method.getDescriptor(), body );
  }

  void setBody( BasicNode body )
  {
    if( this.body != null )
      throw new IllegalStateException( "Body already assigned: " + this );
    //(new Exception("setBody: " + this )).printStackTrace();
    this.body = body;
    if( body instanceof DefinitionNode )
      {
	//modifiers = ((DefinitionNode)body).getFlags();
      }
    // System.out.println( "SetBody " + this + " " + Integer.toHexString( modifiers ) );
    /*
    if( (modifiers & STATIC_FLAG) == 0 )
      fieldCounter = 1;
     */
    if( body instanceof ClassNode )
      {
	statements = ((ClassNode)body).statements;
	//modifiers |= MUST_BE_CLASS_FLAG;
      }
    else if( body instanceof MethodNode )
      {
	statements = ((MethodNode)body).body;
	//modifiers |= METHOD_SCOPE_FLAG;
      }
  }
  
  void inherit( QualifiedIdentifier[] ext, QualifiedIdentifier[] impl )
  {
    //this.modifiers = modifiers;
    Vector inh = new Vector();
    inherit( ext, inh );
    inherit( impl, inh );
    if( inh != null && inh.size() > 0 )
      {
	interfaces = new AbstractClass[ inh.size() ];
	inh.copyInto( interfaces );
      }
    if( superclass == null )
      superclass = classTable.objectClass;
  }

  void inherit( AbstractClass clazz )
  {
    if( superclass != null )
      throw new IllegalStateException( "Superclass already assigned" );
    superclass = clazz;
  }

  JintMethod addMethod( int modifiers, Identifier name,
			 QualifiedIdentifier type,
			 QualifiedIdentifier[] params,
			 QualifiedIdentifier[] exceptions )
  {
    JintMethod m = new JintMethod( classTable, this, modifiers,  name,
				   type, params, exceptions );
    if( params != null && params.length > 0 && params[params.length-1] != null &&
	params[params.length-1].getType() == classTable.arglistClass )
      {
	addMember( name, m, ARGLIST_INDEX );
	m.varargFlag = true;
      }
    else
      addMember( name, m, (params == null ? 0 : params.length) + METHOD_INDEX );
    return m;
  }

  JintField addField( int modifiers, Identifier name,
			 QualifiedIdentifier type )
  {
    JintField f = new JintField( classTable, this, modifiers,  name,
				   type, 0 /*fieldCounter*/ );
    AbstractClass t = f.getType();
    // fieldCounter += classTable.getSize( t );
    addMember( name, f, FIELD_INDEX );
    return f;
  }

  public JintField findLocalField( Identifier name, AbstractClass type )
  {
    modifiers |= MUST_BE_CLASS_FLAG;
    Object[] entry = (Object[])members.get( name );
    if( entry != null && entry[FIELD_INDEX] != null )
      return (JintField)(entry[FIELD_INDEX]);
    NameTable nameTable = classTable.getNameTable();
    for( int count = 0 ; ; count++ )
      {
	Identifier tmpName = nameTable.lookupAndInsert( "l" + count + "$" + name );
	entry = (Object[])members.get( tmpName );
	if( entry == null )
	  {
	    entry = new Object[FIELD_INDEX+1];
	    members.put( tmpName, entry );
	  }
	if( entry[FIELD_INDEX] == null )
	  {
	    JintField f = new JintField( classTable, this, 0, tmpName.toString(), type, -1 );
	    entry[FIELD_INDEX] = f;
	    return f;
	  }
      }
  }

  void addInnerClass( AbstractClass clazz )
  {
    if( innerList == null )
      innerList = new Vector();
    innerList.addElement( clazz );
  }

  void doneAdding()
  {
    super.doneAdding();
    incomplete = false;
    if( innerList != null )
      {
	inner = new AbstractClass[innerList.size()];
	innerList.copyInto( inner );
	innerList = null;
      }
    if( outer != null && (modifiers & STATIC_FLAG) == 0 )
      {
	if( outerField == null )
	  {
	    outerField = new JintField( classTable, this, FINAL_FLAG,
				  "outer$0", outer, 0 );
	    addMember( (JintField)outerField );
	  }
      }
  }

  private void inherit( QualifiedIdentifier[] classes, Vector inh )
  {
    if( classes == null || classes.length == 0 )
      return;
    for( int i = 0 ; i < classes.length ; i++ )
      {
	if( classes[i] == null )
	  {
	    // unresolved class, error reported in JintClassTable; now just skip
	    continue;
	  }

	AbstractClass aClass = classTable.getClass( classes[i] );

	/*
	  System.out.println( "inherit " + this + ": from "
	  + aClass + " " + aClass.isInterface() );
	*/
	if( aClass.isInterface() )
	  inh.addElement( aClass );
	else
	  if( superclass != null )
	    {
	      Object[] addInfo = { toString(), superclass.toString(), aClass.toString() };
	      throw new CompilerException( ERR_T_MULTIINHERIT, -1, addInfo );
	    }
	  else
	    {
	      superclass = aClass;
	      if( superclass instanceof JintClass )
		{
		  JintClass jintSuper = (JintClass)superclass;
		  //fieldCounter = jintSuper.getNFields();
		  if( outer != null && (modifiers & STATIC_FLAG) == 0 )
		    {
		      AbstractClass superOuter = jintSuper.getOuter();
		      if( superOuter != null && outer == superOuter )
			{
			  outerField = jintSuper.getOuterField();
			}
		      else
			{
			  //int outerIndex = fieldCounter;
			  int counter = staticCounter++;
			  outerField = new JintField( classTable, this, FINAL_FLAG,
				  "outer$" + counter, outer, 0 /*outerIndex*/ );
			  //fieldCounter++;
			  addMember( (JintField)outerField );
			}
		    }
		}
	    }
      }
  }

  private void addMember( JintField f )
  {
    addMember( classTable.getNameTable().lookupAndInsert( f.getName() ), 
	       f, FIELD_INDEX );
  }

  public void reportSelfUsage()
  {
    modifiers |= (SELF_USED_FLAG | MUST_BE_CLASS_FLAG);
  }

  public boolean mustBeClass()
  {
    return (modifiers & METHOD_SCOPE_FLAG) == 0 || (modifiers & MUST_BE_CLASS_FLAG) != 0;
  }
    
  public boolean selfUsed()
  {
    return (modifiers & METHOD_SCOPE_FLAG) == 0 || (modifiers & SELF_USED_FLAG) != 0;
  }

  public void markAsClass()
  {
    modifiers |= MUST_BE_CLASS_FLAG;
  }
    
  public void markAsWorld()
  {
    modifiers |= WORLD_FLAG;
  }
    
  public boolean isWorld()
  {
    return (modifiers & WORLD_FLAG) != 0;
  }

  public void squeezeOuterChain()
  {
    if( outerField == null )
      return;
    AbstractClass prevOuter = outer;
    while( outer instanceof JintClass && !((JintClass)outer).mustBeClass() )
      {
	outer = outer.getOuter();
	if( outer == null )
	  outer = classTable.objectClass;
      }
    if( prevOuter != outer )
      {
	//System.out.println( "Squeezed " + this + " " + outer );
	((JintField)outerField).type = outer;
	outerField.initDescriptor();
	Object[] constrList = (Object[])members.get( classTable.INIT );
	if( constrList != null )
	  {
	    for( int i = METHOD_INDEX-1 ; i < constrList.length ; i++ )
	      {
		Hashtable tb = (Hashtable)constrList[i];
		//System.out.println( "\t" + i + " " + tb );
		if( tb == null )
		  continue;
		Enumeration clist = tb.elements();
		Hashtable newTb = new Hashtable();
		if( i == METHOD_INDEX )
		  throw new IllegalStateException( "Unexpected zero-arg constructor: "
						   + clist.nextElement() );
		while( clist.hasMoreElements() )
		  {
		    JintMethod constr = (JintMethod)clist.nextElement();
		    if( constr.params == null )
		      constr.getParameterTypes(); // give it a chance to initialize
		    if( constr.params.length == 0 )
		      {
			throw new RuntimeException( "Internal error: constructor with no params? " + constr ); 
		      }
		    constr.params[0] = outer;
		    constr.paramTypeNames[0] = outer.getFullName();
		    Object body = null;
		    if( methodBodies != null )
		      body = methodBodies.remove( constr.getDescriptor() );
		    constr.descriptor = null;
		    constr.longDescriptor = null;
		    constr.initDescriptor();
		    if( body != null )
		      methodBodies.put( constr.getDescriptor(), body );
		    newTb.put( constr.getTypeDescriptor(), constr );
		  }
		constrList[i] = newTb;
		//System.out.println( "\t" + i + " " + newTb );
	      }
	  }
      }
    if( inner != null )
      {
	for( int i = 0 ; i < inner.length ; i++ )
	  ((JintClass)inner[i]).squeezeOuterChain();
      }
  }

  public void storeError( CompilerException e )
  {
    if( errorStorage != null )
      errorStorage.storeError( e );
    else
      System.err.println( "Line " + e.getLine() + ", pos " + 
			e.getColumn() + ": " + e.getMessage() );
  }

  public void setErrorStorage( ErrorStorage errorStorage )
  {
    this.errorStorage = errorStorage;
  }

}

