/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.reflect;

import java.lang.reflect.*;
import java.util.Vector;
import java.util.Dictionary;
import java.util.Hashtable;
import java.util.Stack;
import java.util.StringTokenizer;
import java.util.Enumeration;
import java.io.Serializable;

import kmy.jint.parser.*;
import kmy.jint.tree.*;
import kmy.jint.lang.*;
import kmy.jint.util.*;
import kmy.regex.util.*;
import kmy.regex.jvm.JavaClassRegexStub;
import kmy.jint.tree.IdentifierNode;
import kmy.jint.constants.DefinitionConst;
import kmy.jint.constants.TokenConst;
import kmy.jint.constants.ErrorCodes;

public class JintClassTable implements DefinitionConst, ErrorCodes
{

  Hashtable table = new Hashtable();
  NameTable nameTable;

  boolean noLongTypes = false;
  
  private static Object NOT_THERE = "NOT_THERE";

  private Object lock = new Object();

  private boolean jvmClassesEnabled = true;
  private boolean canListPackages;
  private boolean javaOnlyClasses;

  private boolean looseExceptionCheck = false;

  /**
   * maps import IdentifierNodes to their cache tables
   */
  private Hashtable importCaches = new Hashtable();
  private Hashtable javaLangCache = new Hashtable();

  private boolean startUp = true;

  public Identifier INIT;

  public AbstractClass intType;
  public AbstractClass byteType;
  public AbstractClass charType;
  public AbstractClass shortType;
  public AbstractClass doubleType;
  public AbstractClass floatType;
  public AbstractClass longType;
  public AbstractClass booleanType;
  public AbstractClass voidType;
  public AbstractClass numberClass;
  public AbstractClass integerClass;
  public AbstractClass byteClass;
  public AbstractClass characterClass;
  public AbstractClass shortClass;
  public AbstractClass doubleClass;
  public AbstractClass floatClass;
  public AbstractClass longClass;
  public AbstractClass booleanClass;
  public AbstractClass objectClass;
  public AbstractClass cloneableClass;
  public AbstractClass serializableClass;
  public AbstractClass stringClass;
  public AbstractClass throwableClass;
  public AbstractClass exceptionClass;
  public AbstractClass errorClass;
  public AbstractClass runtimeExceptionClass;
  public AbstractClass arglistClass;
  public AbstractClass classClass;
  public AbstractClass anyClass;
  public AbstractClass regexClass;
  public AbstractClass regexStubClass;
  public AbstractClass replacerClass;
  public AbstractClass formatterClass;
  public AbstractClass charStringClass;
  public AbstractClass charTypeArray;
  public AbstractClass abstractContainerClass;
  public AbstractClass caseInsensitiveRegexClass;
  public AbstractClass abstractListClass;
  public AbstractClass vectorClass;
  public AbstractClass dictionaryClass;

  public AbstractClass[] exceptionClassArr = { null };

  public ClassUtilities utilities;

  SourceFactory sourceFactry;

  boolean scopeMode = true;

  AbstractClass worldClass;

  SourceFactory sourceFactory;

  Hashtable classNodeTable = new Hashtable();
    
  Stack startStack = new Stack();

  IdentifierNode javaLangImport;
  IdentifierNode kmyJintLangImport;

  public JintClassTable( NameTable nameTable )
  {
    this( nameTable, false, false, null );
  }

  public JintClassTable( NameTable nameTable, boolean canListPackages,
			 boolean javaOnlyClasses, SourceFactory sourceFactory )
  {
    this.nameTable = nameTable;
    this.canListPackages = canListPackages;
    this.javaOnlyClasses = javaOnlyClasses;
    this.sourceFactory = sourceFactory;
  }

  public void init()
  {
    Identifier[] javaLangIds = { nameTable.JAVA, nameTable.LANG, null };
    Identifier[] kmyJintLangIds = { nameTable.KMY, nameTable.JINT, nameTable.LANG, null };

    javaLangImport = new IdentifierNode( -1, javaLangIds );
    kmyJintLangImport = new IdentifierNode( -1, kmyJintLangIds );

    INIT = nameTable.INIT;

    intType        = getClass( Integer.TYPE );
    byteType       = getClass( Byte.TYPE );
    charType       = getClass( Character.TYPE );
    shortType      = getClass( Short.TYPE );
    booleanType    = getClass( Boolean.TYPE );
    floatType      = getClass( Float.TYPE );
    doubleType     = getClass( Double.TYPE );
    longType       = getClass( Long.TYPE );
    voidType       = getClass( Void.TYPE );
    objectClass    = getClass( Object.class );
    cloneableClass = getClass( Cloneable.class );
    serializableClass = getClass( Serializable.class );
    numberClass    = getClass( Number.class );
    integerClass   = getClass( Integer.class );
    byteClass      = getClass( Byte.class );
    characterClass = getClass( Character.class );
    shortClass     = getClass( Short.class );
    doubleClass    = getClass( Double.class );
    floatClass     = getClass( Float.class );
    longClass      = getClass( Long.class );
    booleanClass   = getClass( Boolean.class );
    stringClass    = getClass( String.class );
    throwableClass = getClass( Throwable.class );
    exceptionClass = getClass( Exception.class );
    errorClass     = getClass( Error.class );
    runtimeExceptionClass = getClass( RuntimeException.class );
    dictionaryClass = getClass( Dictionary.class );
    vectorClass    = getClass( Vector.class );
    classClass     = getClass( Class.class );

    try
      {
	// Java 1.2
	abstractListClass    = getClass( Class.forName( "java.util.AbstractList" ) );
      }
    catch( ClassNotFoundException e )
      {
      }

    charTypeArray = getClass( "[C" );
    exceptionClassArr[0] = exceptionClass;

    if( ! javaOnlyClasses )
      {
	anyClass = new AnyClass( this );
	registerClass( anyClass );

	arglistClass   = getClass( ArgList.class );
	regexClass     = getClass( Regex.class );
	regexStubClass = getClass( JavaClassRegexStub.class );
	formatterClass = getClass( Formatter.class );
	replacerClass  = getClass( Replacer.class );
	caseInsensitiveRegexClass = getClass( CaseInsensitiveRegex.class );

	charStringClass = getClass( CharString.class );
	abstractContainerClass = getClass( AbstractContainer.class );
      }

    utilities = new ClassUtilities( new JintClassMethods( this ) );
  }

  public void setLooseExceptionCheck( boolean t )
  {
    looseExceptionCheck = t;
  }

  public boolean getLooseExceptionCheck()
  {
    return looseExceptionCheck;
  }

  public boolean isStartUp()
  {
    return startUp;
  }

  public void setJVMClassesEnabled( boolean jvmClassesEnabled )
  {
    this.jvmClassesEnabled = jvmClassesEnabled;
  }

  public void finishStartUp()
  {
    startUp = false;

    Enumeration classList = table.elements();
    Vector toDo = new Vector();
    while( classList.hasMoreElements() )
      toDo.addElement( classList.nextElement() );

    classList = toDo.elements();
    while( classList.hasMoreElements() )
      {
	Object[] t = (Object[])classList.nextElement();
	for( int i = 0 ; i < t.length ; i++ )
	  if( t[i] instanceof JVMClass )
	    {
	      ((JVMClass)t[i]).finish();
	    }
      }
  }

  public void setNoLongTypes( boolean noLongTypes )
  {
    this.noLongTypes = noLongTypes;
  }

  public boolean getNoLongTypes()
  {
    return noLongTypes;
  }

  public int getSize( AbstractClass ac )
  {
    if( noLongTypes )
      return 1;
    if( ac == longType || ac == doubleType )
      return 2;
    return 1;
  }

  public NameTable getNameTable()
  {
    return nameTable;
  }

  public boolean canAssign( AbstractClass value, AbstractClass var )
  {
    if( value == null )
      return !var.isPrimitive();
    if( value == var || var == anyClass || value == anyClass )
      return true;

    if( var.isPrimitive() )
      {
	if( ! value.isPrimitive() )
	  return false;
	if( var == booleanType || var == charType || 
	    var == byteType || value == booleanType )
	  return false;
	if( var == doubleType || value == byteType )
	  return true;
	if( value == doubleType )
	  return false;
	if( value == longType )
	  return var == floatType;
	if( value == charType )
	  return var != shortType;
	if( value == shortType )
	  return var != charType;
	if( var == intType || var == shortType )
	  return false;
	if( var == longType )
	  return value == intType;
	if( var != floatType || value != intType)
	  throw new RuntimeException( "Unknown primitive type: " + var + " or " + value );
	return true;
      }

    boolean res = value.isSubclassOf( var );
    // System.out.println( "CanAssign: " + value + " " + var + " : " + res );
    return res;

  }

  public AbstractClass getClass( Class clazz )
  {
    return getClass( (String)null, clazz );
  }

  public AbstractClass getClass( String descriptor )
  {
    return getClass( descriptor, null );
  }

  public AbstractClass getClass( QualifiedIdentifier name )
  {
    return getClass( name, null );
  }

  public AbstractClass getClass( String descriptor, Class clazz )
  {
    int dim = 0;
    String name;
    if( clazz != null && !clazz.isArray() && descriptor == null )
      name = clazz.getName();
    else
      {
	if( descriptor == null )
	  descriptor = clazz.getName();
	while( descriptor.charAt( dim ) == '[' )
	  dim++;
	char t = descriptor.charAt( dim );
	switch( t )
	  {
 	    case 'L' :
	      name = descriptor.substring( dim+1, descriptor.length()-1 );
	      clazz = null;
	      break;
 	    case 'A' :
	      return anyClass; 
 	    case 'B' :
	      name = "byte";
	      clazz = Byte.TYPE;
	      break;
	    case 'C' :
	      name = "char";
	      clazz = Character.TYPE;
	      break;
	    case 'D' :
	      name = "double";
	      clazz = Double.TYPE;
	      break;
	    case 'F' :
	      name = "float";
	      clazz = Float.TYPE;
	      break;
	    case 'I' :
	      name = "int";
	      clazz = Integer.TYPE;
	      break;
	    case 'J' :
	      name = "long";
	      clazz = Long.TYPE;
	      break;
	    case 'S' :
	      name = "short";
	      clazz = Short.TYPE;
	      break;
	    case 'Z' :
	      name = "boolean";
	      clazz = Boolean.TYPE;
	      break;
	    case 'V' :
	      name = "void";
	      clazz = Void.TYPE;
	      break;
	    default :
	      throw new RuntimeException( "Unexpected class descriptor: " + descriptor );
	  }
      }
    StringTokenizer strTok = new StringTokenizer( name, "/.$" );
    Identifier[] idList = new Identifier[ strTok.countTokens() ];
    for( int i = 0 ; i < idList.length ; i++ )
      idList[i] = nameTable.lookupAndInsert( strTok.nextToken() );
    return getClass( new QualifiedIdentifier( idList, 0, idList.length, dim ), clazz );
  }

  public AbstractClass getClass( QualifiedIdentifier name, Class clazz )
  {
    if( name == null )
      throw new RuntimeException( "JintClassTable.getClass(" + name + "," + clazz + ")" );

    AbstractClass[] t = (AbstractClass[])table.get( name.lastIdentifier() );
    if( t != null )
      {
	if( anyClass != null && anyClass.getFullName().same( name ) )
	  return anyClass; // ignore dim
	int k;
	for( k = 0 ; k < t.length ; k++ )
	  if( t[k].getFullName().same( name ) )
	    return t[k];
      }
    AbstractClass aClass;
    if( name.getDimension() > 0 )
      {
	aClass = getClass( name.base(), clazz );
	if( aClass == null )
	  return null;
	aClass = new ArrayClass( this, aClass, name );
	registerClass( aClass );
      }
    else
      aClass = createClass( name, clazz );

    return aClass;
  }

  public AbstractClass getClass( Identifier shortName, IdentifierNode[] imports )
  {
    if( shortName.isReserved() )
      {
	Identifier[] arr = { shortName };
	AbstractClass aClass = getClass( new QualifiedIdentifier( arr ) );
	if( aClass != null )
	  return aClass;
	return null;
      }
    if( imports != null )
      {
	for( int i = 0 ; i < imports.length ; i++ )
	  {
	    IdentifierNode imp = imports[i];
	    Identifier lastId = imp.idList[imp.idList.length-1];
	    if( lastId == shortName )
	      {
		QualifiedIdentifier qi = new QualifiedIdentifier( imp.idList );
		AbstractClass aClass = getClass( qi );
		if( aClass != null )
		  return aClass;
		return null;
	      }
	  }
	for( int i = 0 ; i < imports.length ; i++ )
	  {
	    IdentifierNode imp = imports[i];
	    // System.out.println( "Trying " + shortName + " in " + imp );
	    Identifier lastId = imp.idList[imp.idList.length-1];
	    if( lastId == null )
	      {
		AbstractClass clazz = lookupInImport( imp, shortName );
		if( clazz != null )
		  return clazz;
	      }
	  }
      }
    
    AbstractClass clazz = lookupInImport( javaLangImport, shortName );
    if( clazz != null )
      return clazz;

    if( shortName == nameTable.ANY )
      return anyClass;

    clazz = lookupInImport( kmyJintLangImport, shortName );
    if( clazz != null )
      return clazz;

    return null;
  }

  public AbstractClass getClass( AbstractClass scope, Identifier[] idList,
				 IdentifierNode[] imports )
  {
    return getClass( scope, idList, 0, idList.length, imports );
  }

  public AbstractClass getClass( AbstractClass scope, Identifier[] idList,
				 int off, int length, IdentifierNode[] imports )
  {
    AbstractClass aClass = null;

    Hashtable importCache = null;

    if( length == 1 && scope != null )
      {
	importCache = scope.getImportCache();
	if( importCache != null )
	  {
	    aClass = (AbstractClass)importCache.get( idList[off] );
	    if( aClass != null )
	      return aClass;
	  }
      }

    try
      {

	if( scope != null )
	  {
	    AbstractClass scp = scope;
	    Identifier[] testNameList = null;
	    if( length > 1 || !idList[off].isReserved() )
	      {
		do
		  {
		    AbstractClass scp1 = scp;
		    do
		      {
			QualifiedIdentifier currentName = scp1.getFullName();
			int len = currentName.length();
		   
			if( testNameList == null || testNameList.length < len + length )
			  testNameList = new Identifier[ len + length ];

			currentName.copyInto( testNameList );
			System.arraycopy( idList, off, testNameList, len, length );
			QualifiedIdentifier testName = new QualifiedIdentifier( testNameList,
								     0, length+len );
			// System.out.println( "Trying " + testName );
			aClass = getClass( testName );
			if( aClass != null )
			  return aClass;
			// System.out.println( "class " + scp1 + ": super " + scp1.getSuperclass() );
			scp1 = scp1.getSuperclass();
		      }		
		    while( scp1 != null && scp1 != objectClass );
		    scp = scp.getLexicallyOuter();
		  }
		while( scp != null );
	      }
	  }

	for( int i = length ; i > 0 ; i-- )
	  {
	    if( i == 1 )
	      aClass = getClass( idList[off], imports );
	    else
	      aClass = getClass( new QualifiedIdentifier( idList, off, i ) );
	    for( int j = i ; j < length && aClass != null ; j++ )
	      {
		QualifiedIdentifier nm = new QualifiedIdentifier( aClass.getFullName(), idList[off+j] );
		aClass = getClass( nm );
	      }
	    if( aClass != null )
	      return aClass;
	  }
	
	for( int i = 1 ; i < length && aClass != null ; i++ )
	  {
	    QualifiedIdentifier nm = new QualifiedIdentifier( aClass.getFullName(), idList[off+i] );
	    aClass = getClass( nm );
	  }

	return aClass;
      }
    finally
      {
	if( importCache != null && aClass != null )
	  importCache.put( idList[off], aClass );
      }
  }
 
  protected AbstractClass lookupClass( QualifiedIdentifier name )
  {
    AbstractClass[] t = (AbstractClass[])table.get( name.lastIdentifier() );
    if( t != null )
      {
	int k;
	for( k = 0 ; k < t.length ; k++ )
	  if( t[k].getFullName().same( name ) && t[k].getDimension() == 0 )
	    return t[k];
      }
    return null;
  }

  public void registerClass( AbstractClass aClass )
  {
    registerClass( aClass, false );
  }

  protected void registerClass( AbstractClass aClass, boolean force )
  {
    QualifiedIdentifier name = aClass.getFullName();
    if( name == null )
      throw new IllegalArgumentException( "Class " + aClass + ": null name" );
    int dim = aClass.getDimension();
    Identifier key = name.lastIdentifier();
    AbstractClass[] t = (AbstractClass[])table.get( key );
    if( t != null )
      {
	int k;
	for( k = 0 ; k < t.length ; k++ )
	  if( t[k].getFullName().same( name ) && t[k].getDimension() == dim )
	    {
	      if( force )
		{
		  t = null;
		  break;
		}
	      else
		throw new RuntimeException( "Class " + name + " is already defined" );
	    }
      }
    if( canListPackages && dim == 0 )
      {
	Identifier[] arr = new Identifier[name.length()];
	name.copyInto( arr );
	Identifier last = arr[arr.length-1];
	arr[arr.length-1] = null;
	IdentifierNode importIdent = new IdentifierNode( -1, arr );
	Hashtable cache = getImportCache( importIdent );
	cache.put( last, aClass );
      }
    if( t == null )
      {
	AbstractClass[] t0 = { aClass };
	t = t0;
      }
    else
      {
	AbstractClass[] t0 = new AbstractClass[t.length+1];
	System.arraycopy( t, 0, t0, 0, t.length );
	t0[t.length] = aClass;
	t = t0;
      }
    table.put( key, t );
  }

  protected AbstractClass createJVMClass( QualifiedIdentifier name, 
				       Class clazz )
  {
    if( clazz == null )
      {
	try
	  {
	    if( name.length() == 1 )
	      {
		Identifier ni = name.lastIdentifier();
		if( ni.isReserved() )
		  {
		    switch( ni.id )
		      {
		        case AbstractTokenizer.TOKEN_INT :
			  clazz = Integer.TYPE;
			  break;
		        case AbstractTokenizer.TOKEN_BOOLEAN :
			  clazz = Boolean.TYPE;
			  break;
		        case AbstractTokenizer.TOKEN_BYTE :
			  clazz = Byte.TYPE;
			  break;
		        case AbstractTokenizer.TOKEN_CHAR :
			  clazz = Character.TYPE;
			  break;
		        case AbstractTokenizer.TOKEN_SHORT :
			  clazz = Short.TYPE;
			  break;
		        case AbstractTokenizer.TOKEN_LONG :
			  clazz = Long.TYPE;
			  break;
		        case AbstractTokenizer.TOKEN_FLOAT :
			  clazz = Float.TYPE;
			  break;
		        case AbstractTokenizer.TOKEN_DOUBLE :
			  clazz = Double.TYPE;
			  break;
		        case AbstractTokenizer.TOKEN_VOID :
			  clazz = Void.class;
			  break;
		      }
		  }
	      }
	    if( clazz == null )
	      {
		String cName = name.toString();
		if( cName.startsWith( ":" ) )
		  return null;
		clazz = Class.forName( cName );
	      }
	  }
	catch( ClassNotFoundException e )
	  {
	    return null;
	  }
	catch( IllegalArgumentException e )
	  {
	    return null;
	  }
      }
    JVMClass res = new JVMClass( this, name, clazz );
    return res;
  }

  public AbstractClass getWrapper( AbstractClass clazz )
  {
    if( clazz == intType )
      return integerClass;
    if( clazz == doubleType )
      return doubleClass;
    if( clazz == longType )
      return longClass;
    if( clazz == floatType )
      return floatClass;
    if( clazz == shortType )
      return shortClass;
    if( clazz == charType )
      return characterClass;
    if( clazz == booleanType )
      return booleanClass;
    if( clazz == byteType )
      return byteClass;
    throw new IllegalArgumentException( "No wrapper for " + clazz );
  }

  public AbstractClass getWrappedType( AbstractClass clazz )
  {
    if( clazz == integerClass )
      return intType;
    if( clazz == doubleClass )
      return doubleType;
    if( clazz == longClass )
      return longType;
    if( clazz == floatClass )
      return floatType;
    if( clazz == shortClass )
      return shortType;
    if( clazz == characterClass )
      return charType;
    if( clazz == booleanClass )
      return booleanType;
    if( clazz == byteClass )
      return byteType;
    throw new IllegalArgumentException( "No wrapped type in " + clazz );
  }

  public void setSourceFactory( SourceFactory sf )
  {
    sourceFactory = sf;
    setJVMClassesEnabled( false );
  }

  public void setWorld( AbstractClass world )
  {
    this.worldClass = world;
  }

  public AbstractClass getWorld()
  {
    return worldClass;
  }

  public ClassConsumer getClassConsumer()
  {
    return new JintClassConsumer();
  }

  /**
   *  Called when someone wants to find some new class
   */
  protected AbstractClass createClass( QualifiedIdentifier name, 
				       Class clazz )
  {
    if( sourceFactory != null )
      {
	boolean sourceLoaded = sourceFactory.loadSource( name, new JintClassConsumer() );
	if( sourceLoaded )
	  return lookupClass( name );
      /*
	else
	  System.out.println( "----> " + name );
      */
      }
    if( jvmClassesEnabled )
      return createJVMClass( name, clazz );
    return null;
  }

  public JintClass createEmptyClass( AbstractClass outer,
				     QualifiedIdentifier fullName )
  {
    if( fullName == null )
      {
	Identifier[] init = { nameTable.lookupAndInsert( "WORLD" ) };
	fullName = new QualifiedIdentifier( init );
      }
    JintClass newClass = new JintClass( this, fullName, outer, 0 );
    registerClass( newClass );
    newClass.addMethod( 0, INIT, null, null, null );
    newClass.doneAdding();
    return newClass;
  }

  void throwCantResolve( IdentifierNode ident, AbstractClass scope )
  {
    Object[] addInfo = { ident, scope };
    throw new CompilerException( ERR_T_CANTRESOLVE, ident.getFilePos(), addInfo );
  }

  class JintClassConsumer extends ClassConsumer implements ErrorStorage
  {
    IdentifierNode packageName;
    Vector importVector;
    IdentifierNode[] imports;
    ErrorStorage errorStorage;

    public void tellImport( IdentifierNode imp )
    {
      if( importVector == null )
	{
	  importVector = new Vector();
	  // insert implicit "import *;"
	  importVector.addElement( new IdentifierNode( -1, new Identifier[1] ) );
	}
      importVector.addElement( imp );
      imports = null;
      // System.out.println( "Import " + imp + " " + this );
    }

    public void tellPackage( IdentifierNode packagePath )
    {
      this.packageName = packagePath;
      Identifier[] idList = new Identifier[packagePath.idList.length+1];
      System.arraycopy( packagePath.idList, 0, idList, 0, packagePath.idList.length );
      IdentifierNode implicitImport = new IdentifierNode( packagePath.getFilePos(), idList );
      // override implicit "import *;"
      importVector = new Vector();
      tellImport( implicitImport );
    }

    public void tellErrorStorage( ErrorStorage errorStorage )
    {
      this.errorStorage = errorStorage;
    }

    public void storeError( CompilerException e )
    {
      if( errorStorage != null )
	errorStorage.storeError( e );
      else
	{
	  // should not happen!
	  System.err.println( "Line " + e.getLine() + ", pos " + 
			    e.getColumn() + ": " + e.getMessage() );
	  e.printStackTrace();
	}
    }

    public Object startScope( Object outer, Identifier className, int modifiers, IdentifierNode[] ext,
			     IdentifierNode[] impl, Class what )
    {
      //System.out.println( "Start scope: " + className + " " + (ext==null?0:ext.length )
      //		  + " " + (impl==null?0:impl.length ) );
      if( imports == null )
	{
	  if( importVector != null )
	    {
	      imports = new IdentifierNode[importVector.size()];
	      importVector.copyInto( imports );
	      /*
		for( int i = 0 ; i < imports.length ; i++ )
		  System.out.println( "\timport: " + imports[i] );
	      */
	    }
	  else
	    {
	      imports = new IdentifierNode[1];
	      // insert implicit "import *;"
	      imports[0] = new IdentifierNode( -1, new Identifier[1] );
	    }
	  /*
	  System.out.println( "Start scope " + className + " " + this );
	  for( int i = 0; i < imports.length ; i++ )
	    System.out.println( "~~~~~~> " + imports[i] );
	  */
	}

      AbstractClass outerClass = (AbstractClass)outer;

      if( outerClass == null )
	outerClass = worldClass; // can be null

      Identifier[] idList;
      if( outerClass == null || outerClass == anyClass )
	{
	  if( packageName == null )
	    idList = new Identifier[1];
	  else
	    {
	      idList = new Identifier[ packageName.idList.length + 1 ];
	      System.arraycopy( packageName.idList, 0, idList, 0, packageName.idList.length );
	    }
	}
      else
	{
	  QualifiedIdentifier outerName = outerClass.getFullName();
	  idList = new Identifier[ outerName.length() + 1 ];
	  outerName.copyInto( idList );
	}

      idList[idList.length-1] = className;
      QualifiedIdentifier fullName = new QualifiedIdentifier( idList );

      try
	{
	  startStack.push( fullName );
	  
	  if( what == MethodNode.class )
	    modifiers |= METHOD_SCOPE_FLAG;
	  else
	    modifiers |= MUST_BE_CLASS_FLAG;
	  JintClass newClass = new JintClass( JintClassTable.this, fullName, outerClass, modifiers );
	  Hashtable importCache = newClass.getImportCache();	  	  
	  newClass.setImports( imports );
	  
	  try
	    {
	      registerClass( newClass );
	    }
	  catch( RuntimeException e )
	    {
	      // kinda hack!!! 
	      // TODO: define special exception type
	      if( e.getMessage() == null || e.getMessage().indexOf( " is already defined" ) < 0 )
		throw e;
	      System.err.println( "Warning: class " + fullName + " NOT shadowed: " + e.getMessage() );
	      System.err.println( "Previous class " + JintClassTable.this.getClass( fullName ) );
	      //registerClass( newClass, true );
	    }
	  
	  QualifiedIdentifier[] extended = 
	    locateClassList( outerClass, imports, importCache, ext, null, this );
	  QualifiedIdentifier[] implemented = 
	    locateClassList( outerClass, imports, importCache, impl, null, this );
	  try
	    {
	      if( what != RegexNode.class )
	        {
		  newClass.inherit( extended, implemented );
		  // System.out.println( "Inherited: " + newClass );
		}
	    }
	  catch( CompilerException e )
	    {
	      if( ext != null && ext.length > 0 )
		e.setFilePos( ext[0].getFilePos() );
	      else if( impl != null && impl.length > 0 )
		e.setFilePos( impl[0].getFilePos() );
	      storeError( e );
	    }
	  /*
	     if( fullName.toString().indexOf( "ArrayValueNode" ) > 0 )
	     {
	     for( int i = 0 ; i < startStack.size() ; i++ )
	     System.err.println( "\t" + startStack.elementAt(i) );
	     }
	     */

	  if( outerClass instanceof JintClass )
	    ((JintClass)outerClass).addInnerClass( newClass );
	  newClass.setErrorStorage( errorStorage );

	  return newClass;
	}
      finally
	{
	  startStack.pop();
	}
    }

    public void finishScope( Object handle, BasicNode baseNode )
    {
      try
	{
	  JintClass newClass = (JintClass)handle;
	  if( baseNode instanceof RegexNode )
	    {
	      if( (((RegexNode)baseNode).flags & TokenConst.TOK_OPT_IGNORE) != 0 )
		newClass.inherit( caseInsensitiveRegexClass );
	      else
		newClass.inherit( regexStubClass );
	    }
	  newClass.setBody( baseNode );
	  baseNode.attachment = newClass;
	  classNodeTable.put( newClass.getFullName(), baseNode );
	}
      catch( Exception e )
	{
	  Object[] addInfo = { handle == null ? "null" : handle.toString(), e.getMessage() };
	  storeError( new CompilerException( ERR_T_ONFINISH, baseNode.getFilePos(), e, addInfo ) );
	}
    }

  }


    private boolean processMethod( Identifier className, JintClass newClass, Identifier methodName,
		      MethodNode methodNode, Hashtable importCache, IdentifierNode[] imports )
    {
      if( newClass.isInterface() && (methodNode.getFlags()&STATIC_FLAG) == 0 )
	methodNode.addFlags( ABSTRACT_FLAG );
      int modifiers = methodNode.getFlags();
      QualifiedIdentifier type = null;
      if( methodNode.type != null )
	{
	  try
	    {
	      type = locateClass( newClass, imports, importCache, methodNode.type, null );
	    }
	  catch( CompilerException e )
	    {
	      newClass.storeError( e );
	    }
	}

      QualifiedIdentifier[] params = locateClassList( 
			newClass, imports, importCache, methodNode.params, null, newClass );
      QualifiedIdentifier[] exceptions = locateClassList( 
			newClass, imports, importCache, methodNode.exceptions, null, newClass );

      if( methodName == className && type == null )
	{
	  methodName = INIT;
	  AbstractClass outer = newClass.getOuter();
	  if( outer != null )
	    {
	      // add hidden argument (for the outer class) to the constructor arguments
	      QualifiedIdentifier[] p;
	      if( params != null )
		{
		  p = new QualifiedIdentifier[ params.length + 1 ];
		  System.arraycopy( params, 0, p, 1, params.length );
		}
	      else
		p = new QualifiedIdentifier[ 1 ];
	      p[0] = outer.getFullName();
	      params = p;
	    }
	}

      JintMethod resolved = newClass.addMethod( modifiers,
		     methodName, type, params, exceptions );
      if( resolved != null )
	{
	  newClass.setMethodBody( resolved, methodNode );
	  if( methodNode.type != null )
	    {
	      AbstractClass rclazz = resolved.getType();
	      methodNode.type.clazz = rclazz;
	    }
	  methodNode.name.attachment = resolved;
	  AbstractClass[] paramTypes = resolved.getParameterTypes();
	  if( paramTypes != null && methodNode.params != null )
	    for( int j = 0 ; j < methodNode.params.length ; j++ )
	      {
		AbstractClass pclazz = paramTypes[j-methodNode.params.length+params.length];
		methodNode.params[j].name.clazz = pclazz;
	      }
	  AbstractClass[] excepts = resolved.getExceptionTypes();
	  if( excepts != null && methodNode.exceptions != null )
	    for( int j = 0 ; j < excepts.length ; j++ )
	      {
		AbstractClass eclazz = excepts[j];
		methodNode.exceptions[j].clazz = eclazz;
	      }
	}

      return methodName == INIT;
    }

    private void processDefs( JintClass newClass, int modifiers, 
			     QualifiedIdentifier type, ExpressionNode[] defs )
    {
      for( int j = 0 ; j < defs.length ; j++ )
	{
	  ExpressionNode expr = defs[j];
	  ExpressionNode def = null;
	  if( expr instanceof InfixNode )
	    {
	      InfixNode infix = (InfixNode)expr;
	      expr = infix.arg1;
	      def  = infix.arg2;
	    }
	  if( expr instanceof IdentifierNode )
	    {
	      try
		{
		  QualifiedIdentifier t = type;
		  IdentifierNode id = (IdentifierNode)expr;
		  if( id.indexes != null )
		    {
		      int dim = id.indexes.length; // additional dimensions
		      if( dim > 0 )
			t = new QualifiedIdentifier( t, dim );
		    }
		  int m = modifiers;
		  JintField field = newClass.addField( m, id.idList[0], t );
		  id.clazz = field.getType();
		  if( (m & STATIC_FLAG) != 0 && (m & FINAL_FLAG) != 0 )
		    try
		      {
			Object v = def.getConstValue();
			if( v != def.NOT_CONST )
			  {
			    // TODO: run through typer!!!
			    if( t.getIdentifier(0).isReserved() )
			      {
				if( t.getType() == booleanType )
				  field.setConstValue( v );
				else
				  {
				    Class realClass = ((JVMClass)t.getType()).getRealClass();
				    field.setConstValue( AnyUtilities.convertNumber( v, realClass ) );
				  }
			      }
			    else
			      field.setConstValue( (String)v );
			  }
		      }
		    catch( Exception e )
		      {
			e.printStackTrace();
		      }
		}
	      catch( Exception e )
		{
		  e.printStackTrace();
		  System.err.println( expr );
		  newClass.storeError( new CompilerException( ERR_T_INTERNAL,
							      expr.getFilePos(), e ) );
		}
	    }
	  else
	    {
	      // error reported elsewhere (?)
	    }
	}
    }



    public void completeScope( BasicNode baseNode )
    {

      boolean noConstructor = true;

      final JintClass newClass = (JintClass)baseNode.attachment;

      classNodeTable.remove( newClass.getFullName() );

      final IdentifierNode[] imports = newClass.getImports();

      AbstractClass outer = newClass.getLexicallyOuter();
      if( outer instanceof JintClass )
	{
	  QualifiedIdentifier outerName = outer.getFullName();
	  BasicNode outerNode = (BasicNode)classNodeTable.remove( outerName );
	  if( outerNode != null )
	    completeScope( outerNode );
	}

      try
	{
	  final Identifier className = newClass.getFullName().lastIdentifier();
	  final Hashtable importCache = newClass.getImportCache();
	  
	  StatementNode[] statements;

	  if( baseNode instanceof ClassNode )
	    {
	      ClassNode classNode = (ClassNode)baseNode;
	      statements = classNode.statements;
	    }
	  else if( baseNode instanceof RegexNode )
	    {
	      RegexNode rnode = (RegexNode)baseNode;
	      if( (rnode.flags & TokenConst.TOK_OPT_STANDALONE) != 0 )
		{
		  Hashtable dcl = new Hashtable();
		  String[] picks = rnode.pickList;
		  if( picks != null )
		    {
		      for( int i = 0 ; i < picks.length ; i++ )
			{
			  if( ! Character.isLetter(picks[i].charAt(0)) )
			    continue;
			  Identifier name = nameTable.lookupAndInsert(picks[i]);
			  dcl.put( name, name );
			  //System.out.println( "Adding field " + name + " to " + newClass );
			  newClass.addField( PUBLIC_FLAG, name, charStringClass.getFullName() );
			}
		    }
		  String[] refs = rnode.refList;
		  if( refs != null )
		    {
		      for( int i = 0 ; i < refs.length ; i++ )
			{
			  if( ! Character.isLetter(refs[i].charAt(0)) )
			    continue;
			  Identifier name = nameTable.lookupAndInsert(refs[i]);
			  if( dcl.get( name ) != null )
			    continue;
			  newClass.addField( PUBLIC_FLAG, name, charStringClass.getFullName() );
			}
		    }
		}
	      statements = null;
	    }
	  else if( baseNode instanceof MethodNode )
	    {
	      //newClass.setBody( baseNode );
	      MethodNode methodNode = (MethodNode)baseNode;
	      ParameterNode[] params = methodNode.params;
	      // if no body, we do not have to add any members to this method's scope!!!
	      if( params != null && methodNode.body != null )
		{
		  for( int i = 0 ; i < params.length ; i++ )
		    {
		      ParameterNode p = params[i];
		      int modifiers = ( p.finalFlag ? MemberNode.FINAL_FLAG : 0 );
		      JintField f = null;
		      QualifiedIdentifier type = null;
		      try
			{
			  try
			    {
			      type = locateClass( newClass, imports, importCache, p.type, null );
			    }
			  catch( CompilerException e )
			    {
			      newClass.storeError( e );
			    }
			  f = newClass.addField( modifiers, p.name.idList[0], type );
			}
		      catch( RuntimeException e )
			{
			  newClass.storeError( new CompilerException( ERR_T_INTERNAL,
								      p.getFilePos(), e ) );
			}
		      if( f != null && p.type != null )
			p.type.clazz = f.getType();
		    }
		}
	      statements = methodNode.body;
	    }
	  else
	    throw new IllegalArgumentException( "baseNode: " + baseNode.getClass() );

	  TraverserContext traverser = new TraverserContext() {
	     int mnCount = 0;
             public Object evalClass( ClassNode classNode )
	     {
	       // never go in inner classes
	       return classNode;
	     }
             public Object evalMethod( MethodNode methodNode )
             {
	       if( methodNode.name.attachment == null )
	         {
		   Identifier incName = nameTable.lookupAndInsert( 
		       "i$" + methodNode.name.idList[0] + "$" + (mnCount++) );
		   processMethod( className, newClass, incName,
				 methodNode, importCache, imports );
	         }
	       return methodNode;
	     }
	  };	  

	  if( statements != null )
	    {
	      for( int i = 0 ; i < statements.length ; i++ )
		{
		  StatementNode dn = statements[i];
		  if( dn instanceof MethodNode )
		    {
		      if( processMethod( className, newClass, ((MethodNode)dn).name.idList[0],
					 (MethodNode)dn, importCache, imports ) )
			noConstructor = false;
		    }
		  else if( dn instanceof MemberNode )
		    {
		      MemberNode memberNode = (MemberNode)dn;
		      int modifiers = memberNode.getFlags();
		      QualifiedIdentifier type = null;
		      if( memberNode.type == null )
			{
			  memberNode.addFlags( SCRIPT_STYLE_FLAG );
			  if( (modifiers & (PROTECTED_FLAG|PRIVATE_FLAG|PACKAGE_PRIVATE_FLAG)) == 0 )
			    memberNode.addFlags( PUBLIC_FLAG );
			  modifiers = memberNode.getFlags();
			}
		      else
			{
			  try
			    {
			      type = locateClass( newClass,
						  imports, importCache, memberNode.type, null );
			    }
			  catch( CompilerException e )
			    {
			      newClass.storeError( e );
			    }
			}
		      processDefs( newClass, modifiers, type, memberNode.defs );
		    }
		  else if( dn instanceof ClassNode )
		    {
		      // already reported by corresponding startClass
		    }
		  else
		    dn.eval( traverser );
		}
	    }
	  if( newClass.getOuter() == null )
	    {
	      if( noConstructor && (newClass.getModifiers() & INTERFACE_FLAG) == 0 )
		{
		  newClass.addMethod( newClass.getModifiers() & PUBLIC_FLAG,
				     INIT, null, null, null );
		}
	    }
	  else
	    {
	      if( noConstructor && (newClass.getModifiers() & INTERFACE_FLAG) == 0 )
		{
		  QualifiedIdentifier[] params = { outer.getFullName() }; 
		  newClass.addMethod( newClass.getModifiers() & PUBLIC_FLAG, INIT, null, params, null );
		}
	    }
	  
	  newClass.doneAdding();
	  //System.out.println( "Completed scope: " + newClass );
	}
      catch( RuntimeException e )
	{
	  e.printStackTrace();
	  Object[] addInfo = { newClass, e.toString() };
	  newClass.storeError( new CompilerException( ERR_T_ONFINISH, baseNode.getFilePos(), e, addInfo ) );
	}
    }

  boolean alreadyCompleting = false;

  public void completeAllScopes()
  {
    if( alreadyCompleting )
      return;
    alreadyCompleting = true;
    while( classNodeTable.size() > 0 )
      {
	BasicNode node = (BasicNode)classNodeTable.elements().nextElement();
	try
	  {
	    completeScope( node );
	  }
	catch( Exception e )
	  {
	    System.err.println( "In completeAllScopes: " );
	    e.printStackTrace();
	  }
      }
    alreadyCompleting = false;
  }
  
    
    QualifiedIdentifier[] locateClassList( AbstractClass scope, IdentifierNode[] imports,
		Hashtable importCache, IdentifierNode[] ident, JintClassConsumer consumer,
					       ErrorStorage errorStorage )
    {
      if( ident == null )
	return null;
      QualifiedIdentifier[] qlist = new QualifiedIdentifier[ident.length];
      for( int i = 0 ; i < qlist.length ; i++ )
	{
	  try
	    {
	      qlist[i] = locateClass( scope, imports, importCache, ident[i], consumer );
	    }
	  catch( CompilerException e )
	    {
	      errorStorage.storeError( e );
	    }
	}
      return qlist;
    }

    QualifiedIdentifier[] locateClassList( AbstractClass scope, IdentifierNode[] imports,
	     		      Hashtable importCache, ParameterNode[] params,
					       JintClassConsumer consumer, ErrorStorage errorStorage )
    {
      if( params == null )
	return null;
      QualifiedIdentifier[] qlist = new QualifiedIdentifier[params.length];
      for( int i = 0 ; i < qlist.length ; i++ )
	{
	  try
	    {
	      qlist[i] = locateClass( scope, imports, importCache, params[i].type, consumer );
	    }
	  catch( CompilerException e )
	    {
	      errorStorage.storeError( e );
	    }
	}
      return qlist;
    }



  QualifiedIdentifier locateSimpleClass( AbstractClass scope, IdentifierNode[] imports,
					Hashtable importCache, IdentifierNode ident,
					JintClassConsumer consumer )
    {
      if( ident == null )
	return null;
      AbstractClass aClass = JintClassTable.this.getClass( scope, ident.idList, imports );
      if( aClass == null )
	throwCantResolve( ident, scope );
      QualifiedIdentifier res = aClass.getFullName();
      if( consumer != null )
	{
	  BasicNode node = (BasicNode)classNodeTable.remove( res );
	  if( node != null )
	    completeScope( node );
	}
      return res;
    }
 
    QualifiedIdentifier locateClass( AbstractClass scope, IdentifierNode[] imports,
	     Hashtable importCache, IdentifierNode ident, JintClassConsumer consumer )
      {
	if( ident == null )
	  return null;
	QualifiedIdentifier primary = locateSimpleClass( scope, imports, importCache, ident, consumer );
	ExpressionNode[] indexes = ident.indexes;
	if( indexes == null || indexes.length == 0 )
	  return primary;
	int dim = indexes.length;
	for( int i = 1 ; i < indexes.length ; i++ )
	  if( indexes[i] != null )
	    throw new CompilerException( ERR_T_ARRNOTEMPTY, indexes[i].getFilePos() );
	return getClass( new QualifiedIdentifier( primary, dim ) ).getFullName();
      }
    
  AbstractClass getClass( AbstractClass scope, IdentifierNode[] imports,
			 Hashtable importCache, IdentifierNode ident )
    {
      if( ident == null )
	return anyClass;
      return getClass( locateClass( scope, imports, importCache, ident, null ) );
    }

  private Hashtable getImportCache( IdentifierNode imp )
  {
    Hashtable cache = (Hashtable)imp.attachment;
    if( cache == null )
      {
	cache = (Hashtable)importCaches.get( imp );
	if( cache == null )
	  synchronized( lock )
	    {
	      cache = (Hashtable)importCaches.get( imp );
	      if( cache == null )
		{
		  cache = new Hashtable();
		  importCaches.put( imp, cache );
		  if( canListPackages )
		    {
		      QualifiedIdentifier packageName = 
			new QualifiedIdentifier( imp.idList, 0, imp.idList.length-1 );
		      sourceFactory.listPackage( packageName, cache );
		    }
		}
	    }
	imp.attachment = cache;
      }
    return cache;
  }

  private AbstractClass lookupInImport( IdentifierNode imp, Identifier shortName )
  {
    Hashtable cache = getImportCache( imp );
    // System.out.println( "Lookup in import: " + imp + " " + shortName + " " + cache );
    Object res = cache.get( shortName );
    if( canListPackages )
      {
	if( res == null || res instanceof AbstractClass )
	  return (AbstractClass)res;
      }
    else
      {
	if( res == NOT_THERE )
	  return null;
	if( res != null )
	  return (AbstractClass)res;
      }
	
    Identifier[] arr = new Identifier[imp.idList.length];
    System.arraycopy( imp.idList, 0, arr, 0, arr.length-1 );
    arr[arr.length-1] = shortName;
    QualifiedIdentifier qi = new QualifiedIdentifier( arr );
    AbstractClass aClass = getClass( qi );
    if( aClass != null )
      {
	cache.put( shortName, aClass );
	return aClass;
      }
    else
      cache.put( shortName, NOT_THERE );

    return null;
  }

}

