/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.tools;

import java.io.*;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Dictionary;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import kmy.jint.util.AnyUtilities;
import kmy.jint.util.DispatchObject;
import kmy.jint.util.WrapperException;
import kmy.jint.parser.Parser;
import kmy.jint.parser.ReaderTokenizer;
import kmy.jint.parser.AbstractTokenizer;
import kmy.jint.parser.NameTable;
import kmy.jint.util.ErrorKeeper;
import kmy.jint.util.ErrorStorage;
import kmy.jint.parser.Identifier;
import kmy.jint.parser.QualifiedIdentifier;
import kmy.jint.parser.ClassConsumer;
import kmy.jint.tree.ClassNode;
import kmy.jint.tree.MethodNode;
import kmy.jint.tree.MemberNode;
import kmy.jint.tree.DefinitionNode;
import kmy.jint.tree.ExpressionNode;
import kmy.jint.tree.InfixNode;
import kmy.jint.tree.ConstNode;
import kmy.jint.tree.StatementNode;
import kmy.jint.tree.PackageStatementNode;
import kmy.jint.tree.ExpressionStatementNode;
import kmy.jint.tree.ImportStatementNode;
import kmy.jint.tree.IdentifierNode;
import kmy.jint.tree.PrintContext;
import kmy.jint.gen.CodeGenerator;
import kmy.jint.gen.JVMClassGenerator;
import kmy.jint.reflect.JintClassTable;
import kmy.jint.reflect.SourceFactory;
import kmy.jint.reflect.CompiledClass;
import kmy.jint.typer.TyperContext;
import kmy.jint.compiler.CodeGeneratorFactory;
import kmy.jint.compiler.CompilerContext;
import kmy.jint.compiler.CompilerHelper;

/**
 *  This class is used to invoke Jint compiler. It parses command-line arguments,
 *  extracts or reads Jint program sources and then calls all parts of compiler in
 *  the proper order.
 */
public class Compiler
{

  static final String METHOD_SIG = "___M_";
  static final String STATEMENT_SIG = "___S_";

  static final int G_LINES   = 0x01;
  static final int G_VARS    = 0x02;
  static final int G_SOURCE  = 0x04;

  String outputPath = null;
  Object[] classPathListIn = { "." };
  Object[] sourcePathListIn = { "." };
  Object[] bootClassPathListIn = initBootClassPath();
  Object[] extDirs = null;

  Object[] classPathList;
  Object[] sourcePathList;
  Object[] bootClassPathList;

  String encoding = null;
  String target = null;

  int debugFlags = 0;
  int optimizeLevel = 0;
  int warningLevel = 0xFFFFFFFF;
  int verboseLevel = 0;
  int deprecationLevel = 1;

  boolean dumpCode = false;
  boolean printJavaCode = false;
  boolean printDebugInfo = false;
  boolean stackTrace = false;
  boolean relativeSources = false;
  boolean lineByLine = true;

  boolean wholeOption = false;
  boolean lineByLineOption = false;

  String[] files;

  public NameTable nameTable;
  public JintClassTable classTable;

  Hashtable workSet;

  Hashtable loadedList;

  Vector workToDo;

  int sourceLoadDepth = 0;

  Identifier JAVA;
  Identifier JAVAX;

  boolean interpMode = false;
  boolean looseExceptionCheck = false;

  // for interpMode==true :
  int adapterCount = 0;
  final static boolean saveByteCode = false; // true to debug

  public DispatchObject theWorld = null;
  public Loader loader;

  public String prompt = "jint: ";

  static final String[] extList = { ".jint", ".j", ".java" };
  static final String[] cextList = { ".class" };
  static final String[] noextList = { "" };

  boolean canListPackages = File.separatorChar == '/'; // for case-sensitive filesystems only!
  boolean javaOnlyClasses = false;
  boolean chain = true;

  static class ExtensionFilenameFilter implements FilenameFilter
  {

    String[] extList;

    ExtensionFilenameFilter( String[] ext )
    {
      extList = ext;
    }

    public boolean accept( File f, String name )
    {
      if( name.indexOf( '$' ) >= 0 )
	return false; // inner classes
      for( int i = 0 ; i < extList.length ; i++ )
	if( name.endsWith( extList[i] ) )
	  return true;
      return false;
    }
  }

  private static Object[] parseList( String pathList )
  {
    StringTokenizer tok = new StringTokenizer( pathList, File.pathSeparator );
    Object[] res = new Object[tok.countTokens()];
    int i = 0;
    while( tok.hasMoreTokens() )
      {
	res[i] = tok.nextToken();
	i++;
      }
    return res;
  }

  private static Object[] initBootClassPath()
  {
    String cp = System.getProperty("java.class.path");
    String cp1 = System.getProperty("java.home") + 
	  File.separatorChar + "lib" + File.separatorChar + "rt.jar" +
	  File.pathSeparatorChar + "." ;
    if( cp == null )
      cp = cp1;
    else
      cp = cp + File.pathSeparatorChar + cp1;
    return parseList( cp );
  }

  boolean alreadyLoaded( String name )
  {
    int index = name.lastIndexOf( '.' );
    if( index >= 0 )
      name = name.substring( 0, index );
    if( loadedList.get( name ) != null )
      return true;
    loadedList.put( name, "" );
    return false;
  }

  static class ZipTreeElement
  {
    Hashtable table;
    String    shortName;

    ZipTreeElement( Hashtable table, String shortName )
    {
      this.table = table;
      this.shortName = shortName;
    }

    String getName()
    {
      return shortName;
    }

    ZipEntry getZipEntry()
    {
      Object t = table.get( shortName );
      if( t instanceof ZipEntry )
	return (ZipEntry)t;
      else
	return null;
    }

    String[] list( FilenameFilter f )
    {
      Object t = table.get( shortName );
      Vector acc = new Vector();
      if( t instanceof Hashtable )
	{
	  Enumeration keys = ((Hashtable)t).keys();
	  while( keys.hasMoreElements() )
	    {
	      String key = (String)keys.nextElement();
	      if( f.accept( null, key ) )
		acc.addElement( key );
	    }
	}
      if( acc.size() == 0 )
	return null;
      String[] names = new String[acc.size()];
      acc.copyInto( names );
      return names;
    }

    Enumeration listSiblingNames()
    {
      return table.keys();
    }

    ZipTreeElement getSibling( String name )
    {
      return new ZipTreeElement( table, name );
    }
  }

  static class ZipFileEx extends ZipFile
  {

    Hashtable tree;
    String zipName;

    ZipFileEx( File f ) throws IOException
    {
      super( f );
      zipName = "Zip[" + f.getAbsolutePath() + "]";;
      tree = new Hashtable();
      Enumeration list = entries();
      while( list.hasMoreElements() )
	{
	  ZipEntry entry = (ZipEntry)list.nextElement();
	  String name = entry.getName();
	  if( name.endsWith( "/" ) || name.endsWith( "\\" ) || name.endsWith( ":" ) )
	    continue;
	  StringTokenizer tok = new StringTokenizer( name, "\\/:" );
	  String pathElement = null;
	  if( !tok.hasMoreTokens() )
	    continue;
	  Hashtable t = tree;
	  while( true )
	    {
	      pathElement = ((String)tok.nextToken()).intern();
	      if( ! tok.hasMoreTokens() )
		break;
	      Object subtree = t.get( pathElement );
	      if( subtree == null || !(subtree instanceof Hashtable) )
		{
		  if( subtree != null )
		    System.err.println( f.getName() + ": duplicate entry: " + name );
		  Hashtable tt = new Hashtable();
		  tt.put( "..", t );
		  tt.put( ".", tt );
		  t.put( pathElement, tt );
		  t = tt;
		}
	      else
		t = (Hashtable)subtree;
	    }
	  if( t.get( pathElement ) != null )
	    {
	      System.err.println( f.getName() + ": duplicate entry: " + name );
	      continue;
	    }
	  t.put( pathElement, entry );
	}
    }

    public String toString()
    {
      return zipName;
    }

    ZipTreeElement findInTree( String name, String[] ext )
    {
      StringTokenizer tok = new StringTokenizer( name, "/\\:" );
      Hashtable t = ( tok.hasMoreTokens() ? tree : null );
      while( t != null )
	{
	  String pathElement = tok.nextToken();
	  Object subTree = t.get( pathElement );
	  if( ! tok.hasMoreTokens() )
	    {
	      if( ext != null )
		{
		  for( int i = 0 ; i < ext.length ; i++ )
		    {
		      String extPathElement = pathElement + ext[i];
		      Object f = t.get( extPathElement );
		      if( f instanceof ZipEntry )
			return new ZipTreeElement( t, extPathElement );
		    }
		}
	      else
		{
		  // even if subTree is null - "new entry"
		  return new ZipTreeElement( t, pathElement );
		}
	      break;
	    }
	  try
	    {
	      t = (Hashtable)subTree;
	    }
	  catch( ClassCastException e )
	    {
	      break;
	    }
	}
      return null;
    }

  }

  abstract class AbstractFile
  {

    AbstractTokenizer tok;

    abstract String getName();
    abstract InputStream getInputStream() throws IOException;

    long lastModified()
    {
      return System.currentTimeMillis();
    }

    boolean ready() throws IOException
    {
      return true;
    }

    AbstractFile[] getSiblings()
    {
      return null;
    }

    AbstractTokenizer getTokenizer() throws IOException
    {
      if( tok == null )
	tok = new ReaderTokenizer( new InputStreamReader( getInputStream() ), nameTable );
      return tok;
    }

  }

  class StdinFile extends AbstractFile
  {

    boolean used;

    String getName()
    {
      return "<stdin>";
    }

    boolean ready() throws IOException
    {
      return System.in.available() != 0;
    }

    InputStream getInputStream() throws IOException
    {
      if( used )
	return null;
      used = true;
      return System.in;
    }
  }

  class DiskFile extends AbstractFile
  {

    String name;
    File file;

    DiskFile( String name, File file )
    {
      this.name = name;
      this.file = file;
    }

    String getName()
    {
      return name;
    }

    InputStream getInputStream() throws IOException
    {
      return new FileInputStream( file );
    }

    AbstractFile[] getSiblings()
    {
      String sname = file.getName();
      if( !sname.endsWith( ".class" ) )
	return null;
      File dir = new File( file.getParent() );
      String baseName;
      baseName = name.substring( 0, name.replace(File.separatorChar,'/').lastIndexOf( '/' )+1 );
      final String truncName = sname.substring( 0, sname.length()-6 ) + "$";
      String[] list = dir.list( new FilenameFilter() {
	int len = truncName.length();
	public boolean accept(File d, String name)
	  {
	    return name.startsWith( truncName ) && name.endsWith( ".class" ) &&
	      name.substring(len).indexOf( '$' ) < 0 && !Character.isDigit( name.charAt(len) );
	  }
      } );
      if( list == null || list.length == 0 )
        return null;

      DiskFile[] arr = new DiskFile[list.length];
      for( int i = 0 ; i < list.length ; i++ )
        {
	  File f = new File( dir, list[i] );
	  arr[i] = new DiskFile( baseName + list[i], f );
	}
      return arr;
    }

    long lastModified()
    {
      return file.lastModified();
    }
  }

  class ZipEntryFile extends AbstractFile
  {
    ZipFileEx zip;
    ZipEntry entry;
    ZipTreeElement treeElement;

    ZipEntryFile( ZipTreeElement treeElement, ZipFileEx zip )
    {
      this.treeElement = treeElement;
      this.zip = zip;
      entry = treeElement.getZipEntry();
    }

    String getName()
    {
      return entry.getName();
    }

    AbstractFile[] getSiblings()
    {
      String sname = treeElement.getName();
      if( !sname.endsWith( ".class" ) )
        return null;
      int len = sname.length()-5;
      String truncName = sname.substring( 0, len-1 ) + "$";
      Enumeration list = treeElement.listSiblingNames();
      Vector entries = new Vector();
      while( list.hasMoreElements() )
        { 
	  String nm = (String)list.nextElement();
	  if( nm.startsWith( truncName ) && nm.endsWith( ".class" ) &&
	      nm.substring(len).indexOf( '$' ) < 0 && !Character.isDigit( nm.charAt(len) ) )
	    {
	      entries.addElement( new ZipEntryFile( treeElement.getSibling( nm ), zip ) );
	    }
	}

      int sz = entries.size();
      if( sz == 0 )
        return null;

      ZipEntryFile[] arr = new ZipEntryFile[sz];
      entries.copyInto( arr );
      return arr;
    }

    InputStream getInputStream() throws IOException
    {
      return zip.getInputStream( entry );
    }
  }

  static class WorkUnit
  {
    WorkUnit( AbstractFile in, StatementNode[] statements,
	      ErrorKeeper errors, boolean interp )
    {
      this.source = in;
      this.statements = statements;
      this.errors = errors;
      this.interp = interp;
    }

    AbstractFile     source;
    StatementNode[]  statements;
    boolean          interp;
    ErrorKeeper      errors;
  }

  Object[] preprocessPaths( Object[] pathList ) throws IOException
  {
    Object[] res = new Object[pathList.length];
    for( int i = 0 ; i < pathList.length ; i++ )
      {
	Object pl = pathList[i];
	if( pl == null )
	  continue;
	ZipFileEx zip = null;
	File folder = null;
	if( pl instanceof String )
	  {
	    folder = new File( (String)pathList[i] );
	    if( folder.isDirectory() )
	      res[i] = folder;
	    else if( folder.canRead() )
	      try
		{
		  zip = new ZipFileEx( folder );
		  folder = null;
		  res[i] = zip;
		}
	      catch( Exception e )
		{
		  continue;
		}
	    else
	      {
		continue;
	      }
	  }
      }
    return res;
  }

  AbstractFile openFile( Object[] pathList, String fname, String[] ext )
    throws IOException
  {
    //System.out.println( "Looking for " + fname );
    for( int i = 0 ; i < pathList.length ; i++ )
      {
	Object pl = pathList[i];
	if( pl == null )
	  continue;

	if( pl instanceof File )
	  {
	    File folder = (File)pl;
	    String fnamext = findInDirectoryTree( folder, fname, ext );
	    if( fnamext == null )
	      continue;
	    File f = new File( folder, fnamext );
	    if( f.canRead() )
	      return new DiskFile( fnamext, f );
	  }
	else
	  {
	    ZipFileEx zip = (ZipFileEx)pl;
	    ZipTreeElement fnamext = zip.findInTree( fname, ext );
	    if( fnamext == null )
	      continue;
	    return new ZipEntryFile( fnamext, zip );
	  }
      }
    throw new IOException( "File not found" );
  }

  void listPackage( QualifiedIdentifier qname, Dictionary list ) throws IOException
  {
    //System.out.println( "List package: " + qname );
    String name = qname.toString( '/' );
    Identifier first = ( qname.length() == 0 ? null : qname.getIdentifier(0) );
    boolean sys = first == JAVA || first == JAVAX;
    listPackage( sourcePathList, name, list, extList );
    listPackage( (sys?bootClassPathList:classPathList), name, list, cextList );
  }

  void listPackage( Object[] pathList, String fname, Dictionary list, String[] extList )
    throws IOException
  {
    
    FilenameFilter filter = new ExtensionFilenameFilter( extList );

    for( int i = 0 ; i < pathList.length ; i++ )
      {
	Object pl = pathList[i];
	if( pl == null )
	  continue;

	// System.out.println( "Listing " + pl );

	String[] names = null;

	if( pl instanceof File )
	  {
	    File folder = (File)pl;
	    String fnamext = findInDirectoryTree( folder, fname, null );
	    if( fnamext == null )
	      continue;
	    File f = new File( folder, fnamext.replace( '/', File.separatorChar ) );
	    if( f.isDirectory() && f.canRead() )
	      names = f.list( filter );
	  }
	else
	  {
	    ZipFileEx zip = (ZipFileEx)pl;
	    ZipTreeElement fnamext = zip.findInTree( fname, null );
	    if( fnamext == null )
	      continue;
	    names = fnamext.list( filter );
	  }

	if( names != null )
	  {
	    for( int p = 0 ; p < names.length ; p++ )
	      {
		String name = names[p];
		name = name.substring( 0, name.lastIndexOf('.') );
		Identifier ident = nameTable.lookupAndInsert( name );
		list.put( ident, ident );
	      }
	  }
      }
  }

  private String findInDirectoryTree( File folder, String name, String[] ext )
  {
    StringTokenizer tok = new StringTokenizer( name, "/\\:" );
    if( !tok.hasMoreTokens() )
      return null;

    StringBuffer path = new StringBuffer();
    String pathElement = null;

    do
      {
	pathElement = tok.nextToken();
	path.append( "/" );
	path.append( pathElement );
	File subTree = new File( folder, pathElement );
	if( !subTree.isDirectory() || !subTree.canRead() )
	  break;
	folder = subTree;
      }
    while( tok.hasMoreTokens() );

    if( tok.hasMoreTokens() )
      return null;

    if( ext == null )
      return path.toString().substring(1);

    for( int i = 0 ; i < ext.length ; i++ )
      {
	String extPathElement = pathElement + ext[i];
	File f = new File( folder, extPathElement );
	if( f.isFile() && f.canRead() )
	  return path.toString().substring(1) + ext[i];
      }

    return null;
  }

  class SourceLoader extends SourceFactory
  {
    
    public boolean loadSource( QualifiedIdentifier qi, ClassConsumer classConsumer )
    {
      // System.out.println( "--->" + qi );
      sourceLoadDepth++;
      try
	{
	  String stub = qi.toString( '/' ); // was: File.separatorChar );
	  AbstractFile in;
	  try
	    {
	      in = openFile( sourcePathList, stub, extList );
	    }
	  catch( IOException e )
	    {
	      in = null;
	    }
	  AbstractFile cin;
	  if( in != null && workSet.get( in.getName() ) != null )
	    cin = null;
	  else
	    {
	      Object[] pathList;
	      Identifier first = qi.getIdentifier(0);
	      if( first == JAVA || first == JAVAX )
		pathList = bootClassPathList;
	      else
		pathList = classPathList;
	      try
		{
		  cin = openFile( pathList, stub, cextList );
		}
	      catch( IOException e )
		{
		  cin = null;
		}
	    }
	  if( in != null && ( cin == null || cin.lastModified() < in.lastModified() ) )
	    {
	      if( alreadyLoaded( in.getName() ) )
		return false;
	      parse( in, classConsumer, false );
	    }
	  else if( cin != null )
	    {
	      if( alreadyLoaded( cin.getName() ) )
		return false;
	      if( !loadCompiledClassAndSiblings( cin, null ) )
		return false;
	    }
	  else
	    return false;
	  return true;
	}
      finally
	{
	  sourceLoadDepth--;
	}
    }

    private boolean loadCompiledClassAndSiblings( AbstractFile cin, CompiledClass outer )
    {
      try
	{
	  CompiledClass cc = new CompiledClass( classTable, outer, cin.getInputStream() );
	  /*
	  if( verboseLevel != 0 )
	    System.out.println( "Registering " + cin.getName() );
	  */
	  classTable.registerClass( cc );
	  if( verboseLevel != 0 )
	    System.out.println( "Loaded " + cin.getName() + " : " + cc.getFullName() );
	  AbstractFile[] siblings = cin.getSiblings();
	  if( siblings != null )
	    {
	      for( int i = 0 ; i < siblings.length ; i++ )
		if( ! loadCompiledClassAndSiblings( siblings[i], cc ) )
		  return false;
	    }
	}
      catch( RuntimeException e )
	{
	  System.err.println( "Error while loading " + cin.getName() + ":" ); 
	  e.printStackTrace();
	  return false;
	}
      catch( IOException e )
	{
	  e.printStackTrace();
	  return false;
	}
      return true;
    }

    public void listPackage( QualifiedIdentifier packageName, Dictionary packageList )
    {
      try
	{
	  Compiler.this.listPackage( packageName, packageList );
	}
      catch( IOException e )
	{
	  throw new RuntimeException( "IOException: " + e.getMessage() );
	}
    }

  }

  class DumpFactory extends CodeGeneratorFactory
  {
    public CodeGenerator createCodeGenerator( int accFlags, String name, String ext )
      {
	try
	  {
	    return new CodeGenerator( accFlags, name, ext );
	  }
	catch( IOException e )
	  {
	    throw new RuntimeException( e.toString() );
	  }
      }
    
    public void close( CodeGenerator gen, String name )
      {
	try
	  {
	    gen.flush();
	  }
	catch( Exception e )
	  {
	  }
      }
  }

  class ClassGeneratorFactory extends CodeGeneratorFactory
  {

    String sourceFile = "[UNKNOWN]";

    ClassGeneratorFactory( String sourceFile )
      {
	this.sourceFile = sourceFile;
      }

    public CodeGenerator createCodeGenerator( int accFlags, String name, String ext )
      {
	JVMClassGenerator gen = new JVMClassGenerator( accFlags, name, ext );
	gen.setSourceFile( sourceFile ); 
	return gen;
      }
	      
    public void close( CodeGenerator gen, String name )
      {
	String fname = ( outputPath == null ? "" : outputPath + File.separatorChar ) 
	  + name + ".class";
	try
	  {
	    File outputFile = new File( fname );
	    String dir = outputFile.getParent();
	    if( dir != null )
	      {
		File dirFile = new File( dir );
		if( ! dirFile.exists() )
		  dirFile.mkdirs();
	      }
	    BufferedOutputStream out = new BufferedOutputStream( 
				       new FileOutputStream( outputFile ) );
	    DataOutputStream dout = new DataOutputStream( out );
	    ((JVMClassGenerator)gen).writeTo( dout );
	    dout.close();
	  }
	catch( IOException e )
	  {
	    System.err.println( "Could not write " + fname + ": " + e.getMessage() );
	  }
	catch( Exception e )
	  {
	    System.err.println( "Internal error for " + fname );
	    e.printStackTrace();
	  }
      }
  }

  class ClassLoaderFactory extends CodeGeneratorFactory
  {

    String sourceFile = "[UNKNOWN]";

    ClassLoaderFactory( String sourceFile )
      {
	this.sourceFile = sourceFile;
      }

    public CodeGenerator createCodeGenerator( int accFlags, String name, String ext )
      {
	if( dumpCode )
	  {
	    try
	      {
		return new CodeGenerator( accFlags, name, ext );
	      }
	    catch( IOException e )
	      {
		throw new RuntimeException( e.toString() );
	      }
	  }

	JVMClassGenerator gen = new JVMClassGenerator( accFlags, name, ext );
	gen.setSourceFile( sourceFile ); 
	return gen;
      }
	      
    public void close( CodeGenerator gen, String name )
      {
	if( dumpCode )
	  {
	    try
	      {
		gen.flush();
	      }
	    catch( Exception e )
	      {
		e.printStackTrace();
	      }
	    return;
	  }

	try
	  {
	    //System.out.println( "Creating class: " + name );
	    ByteArrayOutputStream barr = new ByteArrayOutputStream();
	    BufferedOutputStream out = new BufferedOutputStream( barr );
	    DataOutputStream dout = new DataOutputStream( out );
	    ((JVMClassGenerator)gen).writeTo( dout );
	    dout.close();
	    byte[] bytecode = barr.toByteArray();
	    if( saveByteCode )
	      {
		String fname = ( outputPath == null ? "" : outputPath + File.separatorChar ) 
		  + name + ".class";
		if( verboseLevel != 0 )
		  System.err.println( "Writing " + fname );
		File outputFile = new File( fname );
		String dir = outputFile.getParent();
		if( dir != null )
		  {
		    File dirFile = new File( dir );
		    if( ! dirFile.exists() )
		      dirFile.mkdirs();
		  }
		FileOutputStream fo = new FileOutputStream( outputFile );
		fo.write( bytecode );
		fo.close();
	      }
	    loader.storeCode( name, bytecode );		     
	  }
	catch( Exception e )
	  {
	    System.err.println( "Internal error for " + name );
	    e.printStackTrace();
	  }
      }
  }

  class Loader extends ClassLoader
  {

    ClassLoader chain;
    Hashtable code = new Hashtable();

    Loader()
    {
      chain = getClass().getClassLoader();
    }

    void storeCode( String name, byte[] body )
    {
      name = name.replace( '/', '.' );
      code.put( name, body );
    }

    Class makeClass( String name )
    {
      try
	{
	  name = name.replace( '/', '.' );
	  byte[] body = (byte[])code.get( name );
	  if( body == null )
	    return null;
	  Class result = defineClass( name, body, 0, body.length);
	  return result;
	}
      catch( ClassFormatError e )
	{
	  throw new RuntimeException( "Internal error: class format error: " + e.getMessage() );
	}
    }

    public Class loadClass( String name, boolean resolveIt )
      throws ClassNotFoundException
    {
      Class clazz = makeClass( name );
      if( clazz != null )
	{
	  if( resolveIt )
	    resolveClass(clazz);
	  return clazz;
	}
      if( chain != null )
	return chain.loadClass( name );
      else
	return super.findSystemClass( name );
    }

    public Class loadClass(String name)
       throws ClassNotFoundException
    {
      return loadClass( name, true );
    }
    
    public java.io.InputStream getResourceAsStream( String resource )
    {
      if( chain != null )
        return chain.getResourceAsStream( resource );
      else
        return getSystemResourceAsStream( resource );
    }

    public java.net.URL getResource(String name)
    {
      if( chain != null )
        return chain.getResource( name );
      else
        return getSystemResource( name );
    }

    protected void finalize() throws Throwable
    {
      if( verboseLevel != 0 )
	System.err.println( "Finalizing class loader " + this );
      super.finalize();
    }

  }

  public static void main( String[] args )
  {
    Compiler compiler = new Compiler();
    compiler.init( args );
    int status = compiler.run();
    if( status > 0 )
      System.exit( status );
  }

  public void init( String[] args )
  {
    int i;
    for( i = 0 ; i < args.length && args[i].startsWith( "-" ) ; i++ )
      {
	String opt = args[i];
	String val = ( i+1 < args.length ? args[i+1] : null );
	if( opt.equals( "-d" ) )
	  {
	    outputPath = val;
	    i++;
	  }
	else if( opt.equals( "-classpath" ) )
	  {
	    classPathListIn = parseList( val );
	    i++;
	  }
	else if( opt.equals( "-sourcepath" ) )
	  {
	    sourcePathListIn = parseList( val );
	    i++;
	  }
	else if( opt.equals( "-bootclasspath" ) )
	  {
	    bootClassPathListIn = parseList( val );
	    i++;
	  }
	else if( opt.equals( "-extdirs" ) )
	  {
	    extDirs = parseList( val );
	    i++;
	  }
	else if( opt.equals( "-g" ) )
	  {
	    debugFlags = G_LINES | G_VARS | G_SOURCE;
	  }
	else if( opt.startsWith( "-g:" ) )
	  {
	    val = opt.substring(2);
	    debugFlags = 0;
	    if( val.indexOf( "lines" ) >= 0 )
	      debugFlags |= G_LINES;
	    if( val.indexOf( "vars" ) >= 0 )
	      debugFlags |= G_VARS;
	    if( val.indexOf( "source" ) >= 0 )
	      debugFlags |= G_SOURCE;
	  }
	else if( opt.equals( "-O" ) )
	  {
	    optimizeLevel = 1;
	  }	
	else if( opt.equals( "-nowarn" ) )
	  {
	    warningLevel = 0;
	  }
	else if( opt.equals( "-verbose" ) )
	  {
	    verboseLevel = 0xFFFFFFFF;
	  }
	else if( opt.equals( "-deprecation" ) )
	  {
	    deprecationLevel = 2;
	  }
	else if( opt.equals( "-encoding" ) )
	  {
	    encoding = val;
	    i++;
	  }
	else if( opt.equals( "-target" ) )
	  {
	    target = val;
	    i++;
	  }
	else if( opt.equals( "-Xwhole" ) )
	  wholeOption = true;
	else if( opt.equals( "-Xlinebyline" ) )
	  lineByLineOption = true;
	else if( opt.equals( "-Xscript" ) )
	  looseExceptionCheck = true;
	else if( opt.equals( "-Xrelative" ) )
	  relativeSources = true;
	else if( opt.equals( "-Xdump" ) )
	  dumpCode = true;
	else if( opt.equals( "-Xstacktrace" ) )
	  stackTrace = true;
	else if( opt.equals( "-Xnochain" ) )
	  chain = false;
	else if( opt.equals( "-Xdumpjava" ) )
	  printJavaCode = true;	
	else if( opt.equals( "-Xdumpjavadebug" ) )
	  {
	    printJavaCode = true;	
	    printDebugInfo = true;
	  }
	else if( opt.equals( "-Xinterp" ) )
	  interpMode = true;
	else if( opt.equals( "-Xjavaclassesonly" ) )
	  javaOnlyClasses = true;
	else if( opt.equals( "-Xcanlistpackages" ) )
	  canListPackages = true;
	else if( opt.equals( "-Xcantlistpackages" ) )
	  canListPackages = false;
	else if( opt.equals( "--" ) )
	  {
	    i++;
	    break;
	  }
	else if( opt.equals( "-" ) )
	  {
	    break;
	  }
	else
	  System.err.println( "Unknown option: " + opt + ", ignored" ); 
      }
    if( args.length == i )
      {
	interpMode = true;
	looseExceptionCheck = true;
	lineByLine = !wholeOption;
	files = new String[1];
	files[0] = "-";
      }
    else
      {
	lineByLine = lineByLineOption;
	files = new String[args.length - i];
	System.arraycopy( args, i, files, 0, args.length-i );
      }

    if( interpMode )
      {
	theWorld = new DispatchObject();
	adapterCount = 0;
	theWorld.setElement( "compiler", this );
      }
    else
      theWorld = null;

    newClassTable();
  }

  public void newClassTable()
  {
    nameTable = new NameTable();

    if( interpMode )
      loader = new Loader();

    try
      {
	classPathList = preprocessPaths( classPathListIn );
	bootClassPathList = preprocessPaths( bootClassPathListIn );
	sourcePathList = preprocessPaths( sourcePathListIn );
      }
    catch( IOException e )
      {
	e.printStackTrace();
	return;
      }

    JAVA = nameTable.lookupAndInsert( "java" );
    JAVAX = nameTable.lookupAndInsert( "javax" );
    workSet = new Hashtable();
    loadedList = new Hashtable();
    workToDo = new Vector();
    classTable = new JintClassTable( nameTable, canListPackages,
		   javaOnlyClasses, ( chain ? new SourceLoader() : null ) );
    classTable.init();
    classTable.setLooseExceptionCheck( looseExceptionCheck );
    classTable.finishStartUp();
  }

  private boolean parse( AbstractFile in, ClassConsumer classConsumer, boolean interp )
  {
    boolean res = true;
    ErrorKeeper errorKeeper = new ErrorKeeper();
    errorKeeper.setStackTrace( stackTrace );
    JintClassTable origClassTable;
    try
      {
	if( (verboseLevel & 1) != 0 )
	  System.err.println( "Parsing: " + in.getName() );
	AbstractTokenizer t = in.getTokenizer();

      ANOTHER_CLASS_TABLE:
	while( true )
	  {
	    t.nameTable = nameTable;
	    Parser p = new Parser( t, ( classConsumer != null ? classConsumer : classTable.getClassConsumer() ) );
	    if( interp )
	      {
		classTable.setWorld( classTable.anyClass );
		p.forcePublic( true );
	      }
	    else
	      classTable.setWorld( null );
	    p.setLooseExceptionCheck( looseExceptionCheck );
	    p.setErrorStorage( (ErrorStorage)errorKeeper );
	    Vector stmtList = new Vector(); 
	    while( true )
	      {
		if( interp && lineByLine && in instanceof StdinFile && !in.ready() )
		  {
		    System.out.print( prompt );
		    System.out.flush();
		  }
		StatementNode n = p.parseStatement();
		if( n == null )
		  break;
		if( !(interp && lineByLine) || n instanceof ImportStatementNode )
		  stmtList.addElement( n );
		else
		  {
		    StatementNode[] statements = new StatementNode[stmtList.size()+1];
		    statements[statements.length-1] = n;
		    stmtList.copyInto( statements );
		    workToDo.addElement( new WorkUnit( in, statements, errorKeeper, interp ) );
		    doCompiling();
		  }
		if( nameTable != t.nameTable )
		  continue ANOTHER_CLASS_TABLE;
	      }
	    if( ! (interp && lineByLine) )
	      {
		StatementNode[] statements = new StatementNode[stmtList.size()];
		stmtList.copyInto( statements );
		workToDo.addElement( new WorkUnit( in, statements, errorKeeper, interp ) );
	      }
	    break;
	  }
      }
    catch( Exception e )
      {
	System.err.println( "Internal error:\n" );
	e.printStackTrace();
	return false;
      }
    return true;
  }
  
  private boolean compile( AbstractFile in, StatementNode[] statements,
			   ErrorKeeper errorKeeper, boolean interp )
  {
    boolean res = true;
    try
      {
	TyperContext typer;
	if( interp )
	  typer = new TyperContext( classTable.anyClass, new CompilerHelper() );
	else
	  typer = new TyperContext( classTable, new CompilerHelper() );
	typer.setErrorStorage( errorKeeper );
	CodeGeneratorFactory factory;
	if( interpMode )
	  factory = new ClassLoaderFactory( in.getName() );
	else if( dumpCode )
	  factory = new DumpFactory();
	else
	  factory = new ClassGeneratorFactory( in.getName() );

	ClassConsumer consumer = (interp?classTable.getClassConsumer():null);

	Vector statementList = null;

	int filePos = 0;

	for( int i = 0 ; i < statements.length ; i++ )
	  {
	    boolean wrapped = false;
	    StatementNode n = statements[i];

	    if( printJavaCode && !printDebugInfo )
	      {
		PrintContext pc = new PrintContext();
		n.eval( pc );
		System.out.println( pc.toString() );
	      }
	    
	    if( (verboseLevel & 1) != 0 && n instanceof ClassNode )
	      System.err.println( "Analyzing: " + in.getName() + " [" + ((ClassNode)n).name + "]" );

	    String instName = null;
	    if( ! interp )
	      {
		if( n instanceof ClassNode || n instanceof PackageStatementNode ||
		    n instanceof ImportStatementNode )
		  n.eval( typer );
	      }
	    else
	      {
		if( n instanceof ClassNode )
		  {
		    // nothing
		  }
		else if( n instanceof ImportStatementNode )
		  {
		    consumer.tellImport( ((ImportStatementNode)n).path );
		  }
		else if( n instanceof PackageStatementNode )
		  {
		    consumer.tellPackage( ((PackageStatementNode)n).path );
		  }
		else
		  {
		    StatementNode[] body;
		    String sig;
		    if( ! lineByLine )
		      {
			if( statementList == null )
			  {
			    statementList = new Vector();
			    filePos = n.getFilePos();
			  }
			statementList.addElement( n );
			StatementNode next = ( i+1 < statements.length ? statements[i+1] : null );
			if( next == null || next instanceof ClassNode ||
			    next instanceof PackageStatementNode || next instanceof ImportStatementNode )
			  {
			    body = new StatementNode[statementList.size()];
			    statementList.copyInto( body );
			    statementList = null;
			    sig = METHOD_SIG; // execute and insert
			  }
			else
			  continue;
		      }
		    else
		      {
			filePos = n.getFilePos();
			body = new StatementNode[1];
			body[0] = n;
			sig = ( n instanceof MethodNode || n instanceof MemberNode ?
				   METHOD_SIG : STATEMENT_SIG );
		      }
		    wrapped = true;
		    instName = sig + (adapterCount++);
		    Identifier nextName = nameTable.lookupAndInsert( instName );
		    IdentifierNode name = new IdentifierNode( filePos, nextName );
		    ClassNode cn = new ClassNode( name, null, null, null, body );
		    cn.addFlags( ClassNode.PUBLIC_FLAG );
		    if( lineByLine && n instanceof ExpressionStatementNode )
		      ((ExpressionStatementNode)n).printResult = true;
		    cn.setFilePos( filePos );
		    n = cn;
		    Object h = consumer.startScope( classTable.anyClass, nextName, ClassNode.PUBLIC_FLAG,
						    null, null, ClassNode.class );
		    consumer.finishScope( h, cn );
		    classTable.completeAllScopes();
		  }
		n.eval( typer );
	      }
	
	    if( printJavaCode && printDebugInfo )
	      {
		PrintContext pc = new PrintContext();
		pc.setFlags( pc.DEBUG_IDENTIFIERS | pc.DEBUG_CALLS );
		n.eval( pc );
		System.out.println( pc.toString() );
	      }
	    
	    if( errorKeeper.hasErrors() )
	      throw new WrapperException( null );

	    if( (verboseLevel & 1) != 0 && n instanceof ClassNode )
	      System.err.println( "Compiling: " + in.getName() + " [" + ((ClassNode)n).name + "]" );

	    if( n instanceof ClassNode )
	      {
		CompilerContext cc = new CompilerContext( classTable, factory );
		cc.setErrorStorage( errorKeeper );
		n.eval( cc );
	      }
	    else if( ! ( n instanceof PackageStatementNode || 
			 n instanceof ImportStatementNode ) )
	      System.err.println( "Can compile only classes on the outermost level, have " +
				 n.getClass().getName() );

	    if( interp && instName != null )
	      {
		Class clazz = loader.loadClass( instName );
		if( instName.startsWith( STATEMENT_SIG ) )
		  {
		    // this was a statement -- execute
		    Class[] objClass = { Object.class };
		    Object[] world = { theWorld };
		    try
		      {
			clazz.getConstructor( objClass ).newInstance( world );
		      }
		    catch( InvocationTargetException e )
		      {
			e.getTargetException().printStackTrace();
		      }
		  }
		else if( instName.startsWith( METHOD_SIG ) )
		  {
		    // this was a declaration -- execute and insert
		    Class[] objClass = { Object.class };
		    Object[] world = { theWorld };
		    try
		      {
			theWorld.addTarget( clazz.getConstructor( objClass ).newInstance( world ) );
		      }
		    catch( InvocationTargetException e )
		      {
			e.getTargetException().printStackTrace();
		      }
		  }
	      }
	  }
      }
    catch( WrapperException e )
      {
	res = false;
      }
    catch( Exception e )
      {
	System.err.println( "Internal error:\n" );
	e.printStackTrace();
	res = false;
      }
    catch( VerifyError e )
      {
	System.err.println( "Internal error:\n" );
	e.printStackTrace();
	res = false;
      }

    if( errorKeeper.hasMessages() )
      {
	BufferedReader freader = null;
	try
	  {
	    InputStream inp = in.getInputStream();
	    if( inp != null )
	      freader = new BufferedReader( new InputStreamReader( inp ) );
	  }
	catch( Exception e )
	  {
	  }
	errorKeeper.report( in.getName(), freader );
	errorKeeper.clean();
      }
    return res;
  }

  public int doParsing()
  {
    int result = 0;
    for( int i = 0; i < files.length ; i++ )
      {
	String file = files[i];
	workSet.put( file, file );
      }

    for( int i = 0; i < files.length ; i++ )
      {
	String file = files[i];
	if( alreadyLoaded( file ) )
	  continue;
	AbstractFile in;
	if( file.equals( "-" ) )
	  in = new StdinFile();
	else
	  {
	    try
	      {
		if( relativeSources )
		  in = openFile( sourcePathList, file, noextList );
		else
		  {
		    File f = new File( file );
		    if( f.canRead() && ! f.isDirectory() )
		      in = new DiskFile( file, f );
		    else
		      throw new IOException( "No such file" );
		  }
	      }
	    catch( IOException e )
	      {
		System.err.println( "Could not open file " + file +
				    ": " + e.getMessage() );
		continue;
	      }
	  }
	
	if( ! parse( in, null, interpMode ) )
	  result = 1;
      }
    return result;
  }

  public int doCompiling()
  {
    int result = 0;
    for( int i = 0 ; i < workToDo.size() ; i++ )
      {
	WorkUnit workUnit = (WorkUnit)workToDo.elementAt(i);
	classTable.completeAllScopes();
	if( ! compile( workUnit.source, workUnit.statements, workUnit.errors, workUnit.interp ) )
	  result = 1;
      }
    workToDo.removeAllElements();
    return result;
  }

  public int run()
  {
    int res1 = doParsing();
    int res2 = doCompiling();
    if( res1 != 0 )
      return res1;
    return res2;
  }

}
