/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.tree;

/**
 *  This class allows one to iterate over the parsing tree. For every
 *  non-abstract class that is subclass of BasicNode, there is a
 *  corresponding eval* method. Every node will call corresponding
 *  method when this node's eval(Context) method is called. So, to
 *  iterate over parsing tree nodes, one can subclass Context
 *  (or one of its subclasses, like StubContext or TraverserContext)
 *  and implement needed functionality in the corresponding eval
 *  method.
 *  <p>
 *  This is, to be sure, is more procedural then object-oriented
 *  approach, but I am not religious about OOP. The main advantage
 *  of this approach is that one does not need to modify every
 *  node class to create new kind of parsing tree processing. Also,
 *  every kind of processing code is kept together, not scattered
 *  among numerous node classes; moreover, this logically connected
 *  code has more intimate access to the context state, because it
 *  leaves in context. The downside is, of course, that all the data
 *  inside node classes has to be either public (as I chose) or
 *  publically accessible. Also, if one decides to extend the language
 *  and invent a new kind of node, he has to go and fix all contexts
 *  (this seems fair to me, though).
 */
public abstract class Context
{
  abstract public Object evalArrayValue( ArrayValueNode arrayValueNode );
  abstract public Object evalBody( BodyNode bodyNode );
  abstract public Object evalBreak( BreakNode breakNode );
  abstract public Object evalCall( CallNode callNode );
  abstract public Object evalCase( CaseNode caseNode );
  abstract public Object evalCast( CastNode castNode );
  abstract public Object evalCatch( CatchNode catchNode );
  abstract public Object evalClass( ClassNode classNode );
  abstract public Object evalCond( CondNode condNode );
  abstract public Object evalConst( ConstNode constNode );
  abstract public Object evalContinue( ContinueNode continueNode );
  abstract public Object evalDo( DoNode doNode );
  abstract public Object evalExpressionStatement( ExpressionStatementNode esn );
  abstract public Object evalFor( ForNode forNode );
  abstract public Object evalFormat( FormatNode forNode );
  abstract public Object evalIdentifier( IdentifierNode identifierNode );
  abstract public Object evalIf( IfNode ifNode );
  abstract public Object evalImportStatement( ImportStatementNode importStatement );
  abstract public Object evalInfix( InfixNode infixNode );
  abstract public Object evalLabeled( LabeledNode labeledNode );
  abstract public Object evalMember( MemberNode memberNode );
  abstract public Object evalMethod( MethodNode methodNode );
  abstract public Object evalNew( NewNode newNode );
  abstract public Object evalPackageStatement( PackageStatementNode packageStatement );
  abstract public Object evalParameter( ParameterNode parameterNode );
  abstract public Object evalPostfix( PostfixNode postfixNode );
  abstract public Object evalPrefix( PrefixNode prefixNode );
  abstract public Object evalRegex( RegexNode regexNode );
  abstract public Object evalReplacer( ReplacerNode replacerNode );
  abstract public Object evalReturn( ReturnNode returnNode );
  abstract public Object evalSwitch( SwitchNode switchNode );
  abstract public Object evalSynchronized( SynchronizedNode synchronizedNode );
  abstract public Object evalThrow( ThrowNode throwNode );
  abstract public Object evalTry( TryNode tryNode );
  abstract public Object evalWhile( WhileNode whileNode );
}
