/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.tree;

import java.io.PrintWriter;
import java.io.CharArrayWriter;
import kmy.jint.parser.*;
import kmy.jint.constants.TokenConst;

public class PrintContext extends Context implements TokenConst
{

  CharArrayWriter buffer;
  PrintWriter out;
  String offset = "";
  boolean newLine = false;
  int pri = 0;

  int flags;

  public static final int DEBUG_IDENTIFIERS = 1;
  public static final int DEBUG_CALLS = 1;

  public PrintContext()
  {
    buffer = new CharArrayWriter();
    out = new PrintWriter( buffer );
  }

  public PrintContext( PrintWriter out )
  {
    this.out = out;
  }

  public String toString()
  {
    if( buffer == null )
      return super.toString();
    out.flush();
    return buffer.toString();
  }

  public void setFlags( int flags )
  {
    this.flags = flags;
  }

  public int getFlags()
  {
    return this.flags;
  }

  private void incOffset()
  {
    offset += "  ";
  }

  private void decOffset()
  {
    offset = offset.substring(2);
  }

  private void println( String s )
  {
    if( newLine )
      out.print( offset );
    out.println( s );
    newLine = true;
  }

  private void println()
  {
    out.println();
    newLine = true;
  }

  private void print( String s )
  {
    if( newLine )
      {
	newLine = false;
	out.print( offset );
      }
    out.print( s );
  }

  private void print( char c )
  {
    if( newLine )
      {
	newLine = false;
	out.print( offset );
      }
    out.print( c );
  }

  private void dumpHeader( DefinitionNode def )
  {
    if( (def.getFlags() & def.PUBLIC_FLAG) != 0 )
      print( "public " );
    if( (def.getFlags() & def.PRIVATE_FLAG) != 0 )
      print( "private " );
    if( (def.getFlags() & def.PROTECTED_FLAG) != 0 )
      print( "protected " );
    if( (def.getFlags() & def.STATIC_FLAG) != 0 )
      print( "static " );
    if( (def.getFlags() & def.FINAL_FLAG) != 0 )
      print( "final " );
    if( (def.getFlags() & def.ABSTRACT_FLAG) != 0 )
      print( "abstract " );
    if( (def.getFlags() & def.SYNCHRONIZED_FLAG) != 0 )
      print( "synchronized " );
    if( (def.getFlags() & def.STRICT_FLAG) != 0 )
      print( "fpstrict " );
  }

  private int getPriority( int token )
  {
    switch( token )
      {
        case '(' :
        case '[' :
        case '{' :
        case AbstractTokenizer.TOKEN_PP :
        case AbstractTokenizer.TOKEN_MM :
        case '.' :
	  return Integer.MAX_VALUE;
	case '*' :
	case '/' :
	case '%' :
	  return 130;
	case '+' :
	case '-' :
	  return 120;
	case AbstractTokenizer.TOKEN_GG :
	case AbstractTokenizer.TOKEN_GGG :
	case AbstractTokenizer.TOKEN_LL :
	  return 110;
	case '<' :
	case '>' :
	case AbstractTokenizer.TOKEN_GE :
	case AbstractTokenizer.TOKEN_LE :
	case AbstractTokenizer.TOKEN_INSTANCEOF :
	  return 100;
	case AbstractTokenizer.TOKEN_EE :
	case AbstractTokenizer.TOKEN_NE :
	  return 90;
	case '&' :
	case '^' :
        case '|' : // all have the same priority in Java
	  return 60;
	case AbstractTokenizer.TOKEN_AND :
	  return 50;
	case AbstractTokenizer.TOKEN_OR :
	  return 40;
	case '?' :
	case ':' :
	  return 30;
	case '=' :
	case AbstractTokenizer.TOKEN_PE :
	case AbstractTokenizer.TOKEN_ME :
	case AbstractTokenizer.TOKEN_SE :
	case AbstractTokenizer.TOKEN_AE :
	case AbstractTokenizer.TOKEN_ANDE :
	case AbstractTokenizer.TOKEN_ORE :
	case AbstractTokenizer.TOKEN_XE :
	case AbstractTokenizer.TOKEN_PCE :
	case AbstractTokenizer.TOKEN_LLE :
	case AbstractTokenizer.TOKEN_GGE :
	case AbstractTokenizer.TOKEN_GGGE :
	  return 21; // odd - right associativity
	case ',' :
	  return 10;
      }
    return 0;
  }

  public void printToken( int token )
  {
    switch( token )
      {
        case AbstractTokenizer.TOKEN_PP :
	  print( "++" );
	  break;
        case AbstractTokenizer.TOKEN_MM :
	  print( "--" );
	  break;
	case AbstractTokenizer.TOKEN_GG :
	  print( ">>" );
	  break;
	case AbstractTokenizer.TOKEN_GGG :
	  print( ">>>" );
	  break;
	case AbstractTokenizer.TOKEN_LL :
	  print( "<<" );
	  break;
	case AbstractTokenizer.TOKEN_GE :
	  print( ">=" );
	  break;
	case AbstractTokenizer.TOKEN_LE :
	  print( "<=" );
	  break;
	case AbstractTokenizer.TOKEN_INSTANCEOF :
	  print( "instanceof" );
	  break;
	case AbstractTokenizer.TOKEN_EE :
	  print( "==" );
	  break;
	case AbstractTokenizer.TOKEN_NE :
	  print( "!=" );
	  break;
	case AbstractTokenizer.TOKEN_AND :
	  print( "&&" );
	  break;
	case AbstractTokenizer.TOKEN_OR :
	  print( "||" );
	  break;
	case AbstractTokenizer.TOKEN_PE :
	  print( "+=" );
	  break;
	case AbstractTokenizer.TOKEN_ME :
	  print( "-=" );
	  break;
	case AbstractTokenizer.TOKEN_SE :
	  print( "/=" );
	  break;
	case AbstractTokenizer.TOKEN_AE :
	  print( "*=" );
	  break;
	case AbstractTokenizer.TOKEN_ANDE :
	  print( "&=" );
	  break;
	case AbstractTokenizer.TOKEN_ORE :
	  print( "|=" );
	  break;
	case AbstractTokenizer.TOKEN_XE :
	  print( "^=" );
	  break;
	case AbstractTokenizer.TOKEN_PCE :
	  print( "%=" );
	  break;
	case AbstractTokenizer.TOKEN_LLE :
	  print( "<<=" );
	  break;
	case AbstractTokenizer.TOKEN_GGE :
	  print( ">>=" );
	  break;
	case AbstractTokenizer.TOKEN_GGGE :
	  print( ">>>=" );
	  break;
        default :
	  print( (char)token );
      }
  }

  //----- Context implementation

  public Object evalArrayValue( ArrayValueNode arrayValueNode )
  {
    IdentifierNode type = arrayValueNode.type;
    if( type != null )
      type.eval( this );
    print( '{' );
    ExpressionNode[] vals = arrayValueNode.vals;
    for( int i = 0 ; vals != null && i < vals.length ; i++ )
      {
        if( i != 0 )
	  print( ", " );
        vals[i].eval( this );
      }
    print( '}' );
    return null;
  }

  public Object evalBody( BodyNode bodyNode )
  {
    StatementNode[] statements = bodyNode.statements;
    if( statements != null )
      {
	print('{');
	println();
	incOffset();
	for( int i = 0 ; i < statements.length ; i++ )
	  {
	    statements[i].eval( this );
	    println();
	  }
	decOffset();
	print( '}' );
      }
    else
      {
	println( "{" );
	print( "}" );
      }
    return null;
  }

  public Object evalBreak( BreakNode breakNode )
  {
    if( breakNode.label == null )
      print( "break;" );
    else
      {
	print( "break " );
	breakNode.label.eval( this );
	print( ';' );
      }
    return null;
  }

  public Object evalCall( CallNode callNode )
  {
    callNode.func.eval( this );
    ExpressionNode[] args = callNode.args;
    if( args == null )
      print( "()" );
    else
      {
	int savePri = pri;
	pri = 11;
	print( "( " );
	for( int i = 0 ; i < args.length ; i++ )
	  {
	    if( i != 0 )
	      print( ", " );
	    args[i].eval( this );
	  }
	print( " )" );
	pri = savePri;
      }
    if( ( flags & DEBUG_CALLS ) != 0 && callNode.attachment != null )
      {
	print( "/*@ method: " );
	print( callNode.attachment.toString() );
	print( " */" );
      }
    return null;
  }

  public Object evalCase( CaseNode caseNode )
  {
    if( caseNode.expr == null )
      println( "default:" );
    else
      {
	print( "case " );
	caseNode.expr.eval( this );
	println( ":" );
      }
    StatementNode[] statements = caseNode.statements;
    if( statements != null )
      {
	incOffset();
	for( int i = 0 ; i < statements.length ; i++ )
	  {
	    statements[i].eval( this );
	    println();
	  }
	decOffset();
      }
    return null;
  }

  public Object evalCast( CastNode castNode )
  {
    print( '(' );
    castNode.type.eval( this );
    print( ')' );
    int savePri = pri;
    pri = Integer.MAX_VALUE;
    castNode.arg.eval( this );
    pri = savePri;
    return null;
  }

  public Object evalCatch( CatchNode catchNode )
  {
    if( catchNode.type == null )
      print( "finally" );
    else
      {
	print( "catch( " );
        catchNode.type.eval( this );
	print( ' ' );
	catchNode.name.eval( this );
	println( " )" );
      }
    incOffset();
    catchNode.body.eval( this );
    decOffset();
    return null;
  }

  public Object evalClass( ClassNode classNode )
  {
    if( classNode.name != null )
      {
	dumpHeader( classNode );
	if( ( classNode.getFlags() & classNode.INTERFACE_FLAG ) != 0 )
	  print( "interface " );
	else if( ( classNode.getFlags() & classNode.WHERE_FLAG ) == 0 )
	  print( "class " );
	classNode.name.eval( this );
	println();
      }
    if( classNode.regex != null )
      {
	print( " = " );
	classNode.regex.eval( this );
	println();
	return null;
      }
    if( classNode.parents != null )
      {
	IdentifierNode[] parents = classNode.parents;
	print( "  extends " );
	for( int i = 0 ; i < parents.length ; i++ )
	  {
	    if( i != 0 )
	      print( ", " );
	    parents[i].eval( this );
	  }
	println();
      }
    if( classNode.interfaces != null )
      {
	IdentifierNode[] interfaces = classNode.interfaces;
	print( "  implements " );
	for( int i = 0 ; i < interfaces.length ; i++ )
	  {
	    if( i != 0 )
	      print( ", " );
	    interfaces[i].eval( this );
	  }
	println();
      }
    if( classNode.whereClauses != null )
      {
	ClassNode[] whereClauses = classNode.whereClauses;
	incOffset();
	for( int i = 0 ; i < whereClauses.length ; i++ )
	  {
	    print( "where " );
	    whereClauses[i].eval( this );
	  }
	decOffset();
      }
    println( "{" );
    incOffset();
    StatementNode[] statements = classNode.statements;
    for( int i = 0 ; i < statements.length ; i++ )
      {
	statements[i].eval( this );
	println();
	println();
      }
    decOffset();
    println( "}" );
    println();
    return null;
  }

  public Object evalCond( CondNode condNode )
  {
    int savePri = pri;
    pri = 31;
    print( "( " );
    condNode.cond.eval( this );
    print( " ? " );
    condNode.arg1.eval( this );
    print( " : " );
    condNode.arg2.eval( this );
    print( " )" );
    pri = savePri;
    return null;
  }

  public Object evalConst( ConstNode constNode )
  {
    Object val = constNode.val;
    if( val == null )
      print( "null" );
    else if( val instanceof String )
      {
	char[] ch = ((String)val).toCharArray();
	print( '"' );
	for( int i = 0 ; i < ch.length ; i++ )
	  {
	    char c = ch[i];
	    if( c < ' ' )
	      {
		if( c == '\n' )
		  print( "\\n" );
		else if( c == '\t' )
		  print( "\\t" );
		else
		  {
		    print( "\\0" );
		    print( (char)('0' + c/8) );
		    print( (char)('0' + c%8) );
		  }
	      }
	    else if( c == '"' )
	      print( "\\\"" );
	    else
	      print( c );
	  }
	print( '"' );
      }
    else if( val instanceof Character )
      {
	char c = ((Character)val).charValue();
	print( '\'' );
	if( c < ' ' )
	  {
	    if( c == '\n' )
	      print( "\\n" );
	    else if( c == '\t' )
	      print( "\\t" );
	    else
	      {
		print( "\\0" );
		print( (char)('0' + c/8) );
		print( (char)('0' + c%8) );
	      }
	  }
	else if( c == '\'' )
	  print( "\\'" );
	else
	  print( c );
	print( '\'' );
      }
    else
      print( val.toString() );
    return null;
  }

  public Object evalContinue( ContinueNode continueNode )
  {
    if( continueNode.label == null )
      print( "continue;" );
    else
      {
	print( "continue " );
	continueNode.label.eval( this );
	print( ';' );
      }
    return null;
  }

  public Object evalDo( DoNode doNode )
  {
    println( "do" );
    incOffset();
    doNode.body.eval( this );
    decOffset();
    print( "while( " );
    doNode.cond.eval( this );
    print( " )" );
    return null;
  }

  public Object evalExpressionStatement( ExpressionStatementNode esNode )
  {
    esNode.expr.eval( this );
    print( ';' );
    return null;
  }

  public Object evalFor( ForNode forNode )
  {
    print( "for( " );
    if( forNode.init != null )
      forNode.init.eval( this );
    print( "; " );
    if( forNode.cond != null )
      forNode.cond.eval( this );
    print( "; " );
    if( forNode.next != null )
      forNode.next.eval( this );
    println( " )" );
    incOffset();
    forNode.body.eval( this );
    decOffset();
    return null;
  }

  public Object evalFormat( FormatNode formatNode )
  {
    print( "f`" );
    print( formatNode.formStr );
    print ( "`" );
    return null;
  }

  public Object evalIdentifier( IdentifierNode identifierNode )
  {
    Identifier[] idList = identifierNode.idList;
    ExpressionNode[] indexes = identifierNode.indexes;
    for( int i = 0 ; i < idList.length ; i++ )
      {
	if( i != 0 )
	  print( '.' );
	if( idList[i] == null )
	  print( '*' );
	else
	  print( idList[i].name.toString() );
      }
    if( indexes != null )
      {
	int savePri = pri;
	pri = 0;
	for( int i = 0 ; i < indexes.length ; i++ )
	  if( indexes[i] == null )
	    print( "[]" );
	  else
	    {
	      print( '[' );
	      indexes[i].eval( this );
	      print( ']' );
	    }
	pri = savePri;
      }
    if( ( flags & DEBUG_IDENTIFIERS ) != 0 ) 
      {
	print( "/*@" );
	if( identifierNode.attachment != null )
	  {
	    print( " attach:{ " );
	    Object att = identifierNode.attachment;
	    if( att instanceof Object[] )
	      {
		Object[] arr = (Object[])att;
		for( int i = 0 ; i < arr.length ; i++ )
		  {
		    if( i != 0 )
		      print( ' ' );
		    print( arr[i].toString() );
		  }
	      }
	    else
	      print( att.toString() );
	    print( " } " );
	  }
	if( identifierNode.clazz != null )
	  {
	    print( " class:" );
	    print( identifierNode.clazz.getFullName().toString() );
	    print( " " );
	  }
	print( "*/" );
      }
    return null;
  }

  public Object evalIf( IfNode ifNode )
  {
    print( "if( " );
    ifNode.cond.eval( this );
    println( " )" );
    incOffset();
    ifNode.thenPart.eval( this );
    decOffset();
    if( ifNode.elsePart != null )
      {
	println();
	println( "else" );
	incOffset();
	ifNode.elsePart.eval( this );
	decOffset();
      }
    return null;
  }

  public Object evalImportStatement( ImportStatementNode importStatement )
  {
    print( "import " );
    importStatement.path.eval( this ); 
    print( ';' );
    return null;
  }

  public Object evalInfix( InfixNode infixNode )
  {
    int savePri = pri;
    pri = getPriority( infixNode.op );
    boolean bracket = savePri > pri;
    if( bracket )
      print( "( " );
    pri += pri%2 ;
    infixNode.arg1.eval( this );
    if( infixNode.op == '[' )
      {
	print( '[' );
	if( infixNode.arg2 != null )
	  infixNode.arg2.eval( this );
	print( ']' );
      }
    else
      {
	print( ' ' );
	printToken( infixNode.op );
	print( ' ' );
	infixNode.arg2.eval( this );
      }
    if( bracket )
      print( " )" );
    pri = savePri;
    return null;
  }

  public Object evalLabeled( LabeledNode labeledNode )
  {
    labeledNode.label.eval( this );
    print( ": " );
    labeledNode.statement.eval( this );
    return null;
  }

  public Object evalMember( MemberNode memberNode )
  {
    dumpHeader( memberNode );
    if( memberNode.type != null )
      {
	memberNode.type.eval( this );
	print( ' ' );
      }
    ExpressionNode[] defs = memberNode.defs;
    for( int i = 0 ; i < defs.length ; i++ )
      {
	if( i != 0 )
	  print( ",  " );
	defs[i].eval( this );
      }
    print( ";" );
    return null;
  }

  public Object evalMethod( MethodNode methodNode )
  {
    dumpHeader( methodNode );
    if( methodNode.type != null )
      {
	methodNode.type.eval( this );
	print( ' ' );
      }
    if( methodNode.name == null )
      print( "static" );
    else
      {
	methodNode.name.eval( this );
	ParameterNode[] params = methodNode.params;
	if( params == null )
	  print( "()" );
	else
	  {
	    print( "( " );
	    for( int i = 0 ; i < params.length ; i++ )
	      {
		if( i != 0 )
		  print( ", " );
		params[i].eval( this );
	      }
	    print( " )" );
	  }
      }
    IdentifierNode[] exceptions = methodNode.exceptions;
    if( exceptions != null )
      {
	println();
	print( "  throws " );
	for( int i = 0 ; i < exceptions.length ; i++ )
	  {
	    if( i != 0 )
	      print( ", " );
	    exceptions[i].eval( this );
	  }
      }
    StatementNode[] body = methodNode.body;
    if( body == null )
      print( ';' );
    else
      {
	println();
	println( "{" );
	incOffset();
	for( int i = 0 ; i < body.length ; i++ )
	  {
	    body[i].eval( this );
	    println();
	  }
	decOffset();
	print( '}' );
	print( " //------- end of method " );
	if( methodNode.name == null )
	  print( "static" );
	else
	  methodNode.name.eval( this );
      }
    return null;
  }

  public Object evalNew( NewNode newNode )
  {
    if( pri >= 130 )
      print( '(' );
    print( "new " );
    newNode.type.eval( this );
    if( newNode.args != null )
      {
	ExpressionNode[] args = newNode.args;
	int savePri = pri;
	pri = 11;
	print( "( " );
	for( int i = 0 ; i < args.length ; i++ )
	  {
	    if( i != 0 )
	      print( ", " );
	    args[i].eval( this );
	  }
	print( " )" );
	pri = savePri;
      }
    if( newNode.inlineClass != null )
      {
	println();
	incOffset();
	newNode.inlineClass.eval( this );
	decOffset();
      }
    if( pri >= 130 )
      print( ')' );
    if( ( flags & DEBUG_CALLS ) != 0 )
      {
	print( "/*@" );
        if( newNode.attachment != null )
	  {
	    print( " method: " );
	    print( newNode.attachment.toString() );
	  }
	if( newNode.clazz != null )
	  {
	    print( " clazz: " );
	    print( newNode.clazz.toString() );
	  }
	print( "*/" );
      }
    return null;
  }

  public Object evalPackageStatement( PackageStatementNode packageStatement )
  {
    print( "package " );
    packageStatement.path.eval( this ); 
    print( ';' );
    return null;
  }

  public Object evalParameter( ParameterNode parameterNode )
  {
    if( parameterNode.finalFlag )
      print( "final " );
    if( parameterNode.type != null )
      {
	parameterNode.type.eval( this );
	print( ' ' );
      }
    parameterNode.name.eval( this );
    return null;
  }

  public Object evalPostfix( PostfixNode postfixNode )
  {
    if( pri >= 130 )
      print( '(' );
    int savePri = pri;
    pri = 150;
    postfixNode.arg.eval( this );
    printToken( postfixNode.op );
    pri = savePri;
    if( pri >= 130 )
      print( ')' );
    return null;
  }

  public Object evalPrefix( PrefixNode prefixNode )
  {
    if( pri >= 130 )
      print( '(' );
    int savePri = pri;
    pri = 150;
    printToken( prefixNode.op );
    prefixNode.arg.eval( this );
    pri = savePri;
    if( pri >= 130 )
      print( ')' );    
    return null;
  }

  public Object evalRegex( RegexNode regexNode )
  {
    print( "m`" );
    print( regexNode.str );
    print ( "`" );
    return null;
  }

  public Object evalReplacer( ReplacerNode replacerNode )
  {
    print( "s`" );
    print( replacerNode.regex.str );
    print ( "`" );
    print( replacerNode.formStr );
    print ( "`" );
    return null;
  }

  public Object evalReturn( ReturnNode returnNode )
  {
    if( returnNode.result == null )
      print( "return;" );
    else
      {
	print( "return " );
	returnNode.result.eval( this );
	print( ';' ); 
      }
    return null;
  }

  public Object evalSwitch( SwitchNode switchNode )
  {
    print( "switch( " );
    switchNode.expr.eval( this );
    println( " )" );
    incOffset();
    println( "{" );
    incOffset();
    CaseNode[] cases = switchNode.cases;
    for( int i = 0 ; i < cases.length ; i++ )
      cases[i].eval( this );
    decOffset();
    print( "}" );
    decOffset();
    return null;
  }

  public Object evalSynchronized( SynchronizedNode synchronizedNode )
  {
    print( "synchronized( " );
    synchronizedNode.lock.eval( this );
    println( " )" );
    incOffset();
    synchronizedNode.body.eval( this );
    decOffset();
    return null;
  }

  public Object evalThrow( ThrowNode throwNode )
  {
    print( "throw " );
    throwNode.exception.eval( this );
    print( ';' );
    return null;
  }

  public Object evalTry( TryNode tryNode )
  {
    println( "try" );
    incOffset();
    tryNode.body.eval( this );
    decOffset();
    CatchNode[] catches = tryNode.catches;
    for( int i = 0 ; i < catches.length ; i++ )
      {
	println();
	catches[i].eval( this );
      }
    return null;
  }

  public Object evalWhile( WhileNode whileNode )
  {
    print( "while( " );
    whileNode.cond.eval( this );
    println( " )" );
    incOffset();
    whileNode.body.eval( this );
    decOffset();
    return null;
  }

}
