/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.tree;

/**
 *  Subclass of Context, implementing all its eval* methods. Every
 *  eval* method calls all corresponding node children's eval methods
 *  and replaces those children with whatever this eval returns.
 *  TraverserContext eval* methods always return their argument.
 */
public class TraverserContext extends Context
{

  private void evalStatements( StatementNode[] statements )
  {
    if( statements != null )
      {
	for( int i = 0 ; i < statements.length ; i++ )
	  statements[i] = (StatementNode)statements[i].eval( this );
      }
  }

  //----- Context implementation

  public Object evalArrayValue( ArrayValueNode arrayValueNode )
  {
    IdentifierNode type = arrayValueNode.type;
    if( type != null )
      arrayValueNode.type = (IdentifierNode)type.eval( this );
    ExpressionNode[] vals = arrayValueNode.vals;
    for( int i = 0 ; vals != null && i < vals.length ; i++ )
      vals[i] = (ExpressionNode)vals[i].eval( this );
    return arrayValueNode;
  }

  public Object evalBody( BodyNode bodyNode )
  {
    evalStatements( bodyNode.statements );
    return bodyNode;
  }

  public Object evalBreak( BreakNode breakNode )
  {
    if( breakNode.label != null )
      breakNode.label = (IdentifierNode)breakNode.label.eval( this );
    return breakNode;
  }

  public Object evalCall( CallNode callNode )
  {
    callNode.func = (ExpressionNode)callNode.func.eval( this );
    ExpressionNode[] args = callNode.args;
    if( args != null )
      {
	for( int i = 0 ; i < args.length ; i++ )
	  args[i] = (ExpressionNode)args[i].eval( this );
      }
    return callNode;
  }

  public Object evalCase( CaseNode caseNode )
  {
    if( caseNode.expr != null )
      caseNode.expr = (ExpressionNode)caseNode.expr.eval( this );
    evalStatements( caseNode.statements );
    return caseNode;
  }

  public Object evalCast( CastNode castNode )
  {
    castNode.type = (IdentifierNode)castNode.type.eval( this );
    castNode.arg = (ExpressionNode)castNode.arg.eval( this );
    return castNode;
  }

  public Object evalCatch( CatchNode catchNode )
  {
    if( catchNode.type != null )
      {
        catchNode.type = (IdentifierNode)catchNode.type.eval( this );
	catchNode.name = (IdentifierNode)catchNode.name.eval( this );
      }
    catchNode.body = (StatementNode)catchNode.body.eval( this );
    return catchNode;
  }

  public Object evalClass( ClassNode classNode )
  {
    if( classNode.name != null )
      classNode.name = (IdentifierNode)classNode.name.eval( this );
    IdentifierNode[] parents = classNode.parents;
    if( parents != null )
      {
	for( int i = 0 ; i < parents.length ; i++ )
	  parents[i] = (IdentifierNode)parents[i].eval( this );
      }
    parents = classNode.interfaces;
    if( parents != null )
      {
	for( int i = 0 ; i < parents.length ; i++ )
	  parents[i] = (IdentifierNode)parents[i].eval( this );
      }
    if( classNode.whereClauses != null )
      {
	ClassNode[] whereClauses = classNode.whereClauses;
	for( int i = 0 ; i < whereClauses.length ; i++ )
	  whereClauses[i] = (ClassNode)whereClauses[i].eval( this );
      }
    evalStatements( classNode.statements );
    return classNode;
  }

  public Object evalCond( CondNode condNode )
  {
    condNode.cond = (ExpressionNode)condNode.cond.eval( this );
    condNode.arg1 = (ExpressionNode)condNode.arg1.eval( this );
    condNode.arg2 = (ExpressionNode)condNode.arg2.eval( this );
    return condNode;
  }

  public Object evalConst( ConstNode constNode )
  {
    return constNode;
  }

  public Object evalContinue( ContinueNode continueNode )
  {
    if( continueNode.label != null )
      continueNode.label = (IdentifierNode)continueNode.label.eval( this );
    return continueNode;
  }

  public Object evalDo( DoNode doNode )
  {
    doNode.body = (StatementNode)doNode.body.eval( this );
    doNode.cond = (ExpressionNode)doNode.cond.eval( this );
    return doNode;
  }

  public Object evalExpressionStatement( ExpressionStatementNode esNode )
  {
    esNode.expr = (ExpressionNode)esNode.expr.eval( this );
    return esNode;
  }

  public Object evalFor( ForNode forNode )
  {
    if( forNode.init != null )
      forNode.init = (StatementNode)forNode.init.eval( this );
    if( forNode.cond != null )
      forNode.cond = (ExpressionNode)forNode.cond.eval( this );
    if( forNode.next != null )
      forNode.next = (ExpressionNode)forNode.next.eval( this );
    forNode.body = (StatementNode)forNode.body.eval( this );
    return forNode;
  }

  public Object evalFormat( FormatNode formatNode )
  {
    return formatNode;
  }

  public Object evalIdentifier( IdentifierNode identifierNode )
  {
    ExpressionNode[] indexes = identifierNode.indexes;
    if( indexes != null )
      {
	for( int i = 0 ; i < indexes.length ; i++ )
	  if( indexes[i] != null )
	    indexes[i] = (ExpressionNode)indexes[i].eval( this );
      }
    return identifierNode;
  }

  public Object evalIf( IfNode ifNode )
  {
    ifNode.cond = (ExpressionNode)ifNode.cond.eval( this );
    ifNode.thenPart = (StatementNode)ifNode.thenPart.eval( this );
    if( ifNode.elsePart != null )
      ifNode.elsePart = (StatementNode)ifNode.elsePart.eval( this );
    return ifNode;
  }

  public Object evalImportStatement( ImportStatementNode importStatement )
  {
    importStatement.path = (IdentifierNode)importStatement.path.eval( this ); 
    return importStatement;
  }

  public Object evalInfix( InfixNode infixNode )
  {
    if( infixNode.arg1 != null ) // should not really happen...
      infixNode.arg1 = (ExpressionNode)infixNode.arg1.eval( this );
    if( infixNode.arg2 != null ) // should not really happen...
      infixNode.arg2 = (ExpressionNode)infixNode.arg2.eval( this );
    return infixNode;
  }

  public Object evalLabeled( LabeledNode labeledNode )
  {
    labeledNode.label = (IdentifierNode)labeledNode.label.eval( this );
    labeledNode.statement = (StatementNode)labeledNode.statement.eval( this );
    return labeledNode;
  }

  public Object evalMember( MemberNode memberNode )
  {
    if( memberNode.type != null )
      memberNode.type = (IdentifierNode)memberNode.type.eval( this );
    ExpressionNode[] defs = memberNode.defs;
    for( int i = 0 ; i < defs.length ; i++ )
      defs[i] = (ExpressionNode)defs[i].eval( this );
    return memberNode;
  }

  public Object evalMethod( MethodNode methodNode )
  {
    if( methodNode.type != null )
      methodNode.type = (IdentifierNode)methodNode.type.eval( this );
    if( methodNode.name != null )
      methodNode.name = (IdentifierNode)methodNode.name.eval( this );
    ParameterNode[] params = methodNode.params;
    if( params != null )
      {
	for( int i = 0 ; i < params.length ; i++ )
	  params[i] = (ParameterNode)params[i].eval( this );
      }
    IdentifierNode[] exceptions = methodNode.exceptions;
    if( exceptions != null )
      {
	for( int i = 0 ; i < exceptions.length ; i++ )
	  exceptions[i] = (IdentifierNode)exceptions[i].eval( this );
      }
    evalStatements( methodNode.body );
    return methodNode;
  }

  public Object evalNew( NewNode newNode )
  {
    newNode.type = (IdentifierNode)newNode.type.eval( this );
    if( newNode.args != null )
      {
	ExpressionNode[] args = newNode.args;
	for( int i = 0 ; i < args.length ; i++ )
	  args[i] = (ExpressionNode)args[i].eval( this );
      }
    if( newNode.inlineClass != null )
      newNode.inlineClass = (ClassNode)newNode.inlineClass.eval( this );
    return newNode;
  }

  public Object evalPackageStatement( PackageStatementNode packageStatement )
  {
    packageStatement.path = (IdentifierNode)packageStatement.path.eval( this ); 
    return packageStatement;
  }

  public Object evalParameter( ParameterNode parameterNode )
  {
    if( parameterNode.type != null )
      parameterNode.type = (IdentifierNode)parameterNode.type.eval( this );
    parameterNode.name = (IdentifierNode)parameterNode.name.eval( this );
    return parameterNode;
  }

  public Object evalPostfix( PostfixNode postfixNode )
  {
    postfixNode.arg = (ExpressionNode)postfixNode.arg.eval( this );
    return postfixNode;
  }

  public Object evalPrefix( PrefixNode prefixNode )
  {
    prefixNode.arg = (ExpressionNode)prefixNode.arg.eval( this );
    return prefixNode;
  }

  public Object evalReturn( ReturnNode returnNode )
  {
    if( returnNode.result != null )
      returnNode.result = (ExpressionNode)returnNode.result.eval( this );
    return returnNode;
  }

  public Object evalSwitch( SwitchNode switchNode )
  {
    switchNode.expr = (ExpressionNode)switchNode.expr.eval( this );
    CaseNode[] cases = switchNode.cases;
    for( int i = 0 ; i < cases.length ; i++ )
      cases[i] = (CaseNode)cases[i].eval( this );
    return switchNode;
  }

  public Object evalRegex( RegexNode regexNode )
  {
    return regexNode;
  }

  public Object evalReplacer( ReplacerNode replacerNode )
  {
    replacerNode.regex = (RegexNode)replacerNode.regex.eval( this );
    return replacerNode;
  }

  public Object evalSynchronized( SynchronizedNode synchronizedNode )
  {
    synchronizedNode.lock = (ExpressionNode)synchronizedNode.lock.eval( this );
    synchronizedNode.body = (StatementNode)synchronizedNode.body.eval( this );
    return synchronizedNode;
  }

  public Object evalThrow( ThrowNode throwNode )
  {
    throwNode.exception = (ExpressionNode)throwNode.exception.eval( this );
    return throwNode;
  }

  public Object evalTry( TryNode tryNode )
  {
    tryNode.body = (StatementNode)tryNode.body.eval( this );
    CatchNode[] catches = tryNode.catches;
    for( int i = 0 ; i < catches.length ; i++ )
      catches[i] = (CatchNode)catches[i].eval( this );
    return tryNode;
  }

  public Object evalWhile( WhileNode whileNode )
  {
    whileNode.cond = (ExpressionNode)whileNode.cond.eval( this );
    whileNode.body = (StatementNode)whileNode.body.eval( this );
    return whileNode;
  }

}
