/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.typer;

import java.util.*;

import kmy.jint.reflect.*;
import kmy.jint.parser.*;
import kmy.jint.util.*;
import kmy.regex.util.*;
import kmy.jint.tree.*;

import kmy.jint.constants.TokenConst;
import kmy.jint.constants.DefinitionConst;
import kmy.jint.constants.ErrorCodes;

/**
 *  This context does most of the semantic analysis of
 *  the parsing tree (including figuring out types of all expressions).
 *  It also calculates constant expressions. Its eval* methods
 *  for expression nodes return optimized version of the expression;
 *  return value for all other nodes is not important. It also
 *  fills in nodes' attachment fields; this attachments are normally
 *  used later for things like code generation.
 */
public class TyperContext extends Context implements TokenConst, DefinitionConst, ErrorCodes
{
  AbstractClass     topOuterClass;
  JintClass         currentClass;
  JintClassTable    classTable;
  IdentifierNode[]  imports;

  /**
   *  The closest enclosing method.
   */
  MethodNode        currentMethod;

  Identifier THIS;
  Identifier SUPER;
  Identifier INIT;
  Identifier LENGTH;
  Identifier ADD;
  Identifier SELF;
  Identifier CLASS;

  /**
   *  Variable by Identifier dictionary.
   */
  Hashtable variables;

  /**
   *  Variable stack.
   */
  Stack variableAcc;

  /**
   *  Field cache for the outermost class
   */
  Hashtable fields = new Hashtable();

  /**
   *  Local functions
   */
  Hashtable functions = new Hashtable();

  /**
   *  Label by Identifier dictionary.
   */
  Hashtable labels = new Hashtable();

  /**
   *  First free local variable JVM stack cell number.
   */
  int currVarOffset;

  /**
   *  Determines if we are currently inside method (rather then class) body.
   */
  boolean methodFlag;

  /**
   *  Determines if local variable cells can be reused after their definition
   *  scope ends. Normally it is true.
   */
  boolean cellReuse = true;

  /**
   *  Determines if current expression node should be processed as lvalue
   *  (variable, or field, or array element that is being assigned to).
   */
  boolean lvalue = false;

  /**
   *  If lvalue flag is true, determines if this lvalue if newly assigned
   *  by ('=' operator) or is modified (by, say, '++' operator).
   */
  boolean freshFlag = false;

  /**
   *  This flag is true when statement that is being evaluated is an
   *  initialization of a static field.
   */
  boolean staticInitFlag = false;

  /**
   *  If set - the identifier is the right side of '.' operator.
   */
  AbstractClass fieldOwner = null;

  /**
   *  This is a 'helper' object that generates data attached to the
   *  variable/field access nodes (AbstractVariable).
   */
  AbstractTyperHelper helper;

  /**
   *  This is a stack of enclosing scopes (class or method bodies) for the node
   *  that is being processed. Every element of this stack is of type
   *  TyperContext.LevelStackItem.
   */
  Stack levelStack = new Stack();

  /**
   *  If we need to generate temporary local variables for return,
   *  try and synchronized statements (we normally do).
   */
  boolean tmpVariable = true;

  /**
   *  Counter to generate temporary local variable names.
   */
  int tmpVariableCount = 1;

  /**
   *  ErrorStorage that is used to store generated errors.
   */
  ErrorStorage errorStorage;

  /**
   *  List of exceptions that are thrown (but not caught) in current method.
   */
  Vector exceptionList;

  /**
   *  Create TyperContext that has outermost ("top-level") class.
   *  The code (being represented by a parsing tree) is veiwed
   *  as if it were executing in the context of that top-level class.
   *  Useful for interpreter-like execution in some given context.
   *  @param topLevel class that determines the outermost scope
   *  @param helper AbstractVariable generator
   */
  public TyperContext( AbstractClass topLevel, AbstractTyperHelper helper )
  {
    topOuterClass = topLevel;
    this.classTable = topLevel.getClassTable();
    this.helper = helper;
    init();
  }

  /**
   *  Create TyperContext without outermost ("top-level") class.
   *  Useful for independet (like in Java) class compilation.
   *  @param ct class table to be used by this typer
   *  @param helper AbstractVariable generator
   */
  public TyperContext( JintClassTable ct, AbstractTyperHelper helper )
  {
    this.classTable = ct;
    this.helper = helper;
    init();
  }

  /**
   *  Initialize certain fields.
   */
  private void init()
  {
    LENGTH = classTable.getNameTable().LENGTH;
    THIS = classTable.getNameTable().THIS;
    INIT = classTable.getNameTable().INIT;
    SUPER = classTable.getNameTable().SUPER;
    ADD = classTable.getNameTable().ADD;
    SELF = classTable.getNameTable().SELF;
    CLASS = classTable.getNameTable().CLASS;
    // insert implicit "import *;"
    imports = new IdentifierNode[1];
    imports[0] = new IdentifierNode( -1, new Identifier[1] );
  }

  public void setErrorStorage( ErrorStorage errorStorage )
  {
    this.errorStorage = errorStorage;
  }
 
  public void setCellReuse( boolean cellReuse )
  {
    this.cellReuse = cellReuse;
  }

  public boolean getCellReuse()
  {
    return cellReuse;
  }

  /**
   *  Sets a list of imports to be used for class look-up.
   */
  public void setImports( IdentifierNode[] imports )
  {
    this.imports = imports;
  }

  /**
   *  Returns the list of imports that are currently used in class look-up.
   */
  public IdentifierNode[] getImports()
  {
    return imports;
  }

  /**
   *  Returns "full name" for a method in form method-name(argType,argType,...).
   *  Used to report errors.
   */
  private String fullName( Identifier nm, AbstractClass[] argTypes )
  {
    StringBuffer dscr = new StringBuffer();
    if( nm != null )
      {
	dscr.append( nm );
	dscr.append( '(' );
      }
    for( int i = 0 ; argTypes != null && i < argTypes.length ; i++ )
      {
	if( i != 0 )
	  dscr.append( ',' );
	dscr.append( argTypes[i] );
      }
    if( nm != null )
      dscr.append( ')' );
    return dscr.toString();
  }

  private Object constValue( ExpressionNode node )
  {
    try
      {
	return node.getConstValue();
      }
    catch( CompilerException e )
      {
	storeError( e );
	return ExpressionNode.NOT_CONST;
      }
  }

  private Object constValue( AbstractVariable av )
  {
    try
      {
	return av.getConstValue();
      }
    catch( CompilerException e )
      {
	storeError( e );
	return AbstractVariable.NOT_CONST;
      }
  }

  //-------------- private stuff ---------

  /**
   *  A class representing enclosing scope (class or method body). Such
   *  objects are created to save current scope information in levelStack.
   */
  static class LevelStackItem
  {
    JintClass scope;
    Hashtable variables;
    Stack variableAcc;
    int currVarOffset;
    Hashtable labels;
    Hashtable functions;
    boolean staticInitFlag;
    Vector exceptionList;
  }

  /**
   *  Move one level up in the enclosing scopes stack. Restores all
   *  saved scope information. Called when leaving a scope.
   */
  private void popLevel()
  {
    //System.out.println( "PopLevel " + currentClass );
    LevelStackItem item = (LevelStackItem)levelStack.pop();
    currentClass = item.scope;
    variables = item.variables;
    variableAcc = item.variableAcc;
    labels = item.labels;
    currVarOffset = item.currVarOffset;
    functions = item.functions;
    staticInitFlag = item.staticInitFlag;
    exceptionList = item.exceptionList;
    helper.leaveScope( currentClass );
  }

  /**
   *  Move one level down in the enclosing scopes stack. Saves all
   *  necessary scope information and initializes context fields for
   *  the new scope. Called when entering a new scope.
   */
  private void pushLevel( JintClass scope )
  {
    LevelStackItem item = new LevelStackItem();
    item.scope = currentClass;
    item.variables = variables;
    item.variableAcc = variableAcc;
    item.currVarOffset = currVarOffset;
    item.labels = labels;
    item.functions = functions;
    item.staticInitFlag = staticInitFlag;
    item.exceptionList = exceptionList;
    levelStack.push( item );
    helper.enterScope( scope );
    currVarOffset = 0; //scope.getNFields();
    currentClass = scope;
    variables = new Hashtable();
    variableAcc = new Stack();
    labels = new Hashtable();
    functions = new Hashtable();
    exceptionList = null;
    staticInitFlag = false;
    //System.out.println( "PushLevel " + currentClass );
  }

  /**
   *  Get number of JVM cells that local variable
   *  of type <i>clazz</i> would occupy (2 for long and double, 1 otherwise).
   */
  private int getNCells( AbstractClass clazz )
  {
    if( clazz == null )
      return 1;
    return clazz.getNCells();
  }

  /**
   *  Add given exception to the exception list thrown by this method.
   */
  private void addException( AbstractClass exception )
  {
    if( exception == null || exception.isSubclassOf( classTable.runtimeExceptionClass )
	|| exception.isSubclassOf( classTable.errorClass ) )
      return;
    if( exceptionList == null )
      {
	exceptionList = new Vector();
	// System.out.println( "ADDING: " + exception );
	exceptionList.addElement( exception );
	return;
      }
    Enumeration list = exceptionList.elements();
    Vector toRemove = null;
    while( list.hasMoreElements() )
      {
	AbstractClass ex = (AbstractClass)list.nextElement();
	if( ex == exception || exception.isSubclassOf(ex) )
	  {
	    // System.out.println( "NOT ADDING: " + exception + " {" + ex + "}" );
	    return;
	  }
	if( ex.isSubclassOf(exception) )
	  {
	    if( toRemove == null )
	      toRemove = new Vector();
	    toRemove.addElement( ex );
	    // System.out.println( "REMOVING: " + ex + " {" + exception + "}" );
	  }
      }
    if( toRemove != null )
      {
	list = toRemove.elements();
	while( list.hasMoreElements() )
	  exceptionList.removeElement( list.nextElement() );
      }
    // System.out.println( "ADDING: " + exception );
    exceptionList.addElement( exception );
  }

  /**
   *  Add all exceptions thrown by the given method
   *  to the exception list thrown by this method.
   */
  private void addExceptions( AbstractMethod method )
  {
    AbstractClass[] exceptions = method.getExceptionTypes();
    if( exceptions == null )
      return;
    for( int i = 0 ; i < exceptions.length ; i++ )
      addException( exceptions[i] );
  }

  /**
   *  Remove all exceptions that are the same or subclasses of the given one
   *  from the exception list thrown by this method.
   *  @returns true if some exception were removed from the list.
   */
  private boolean catchException( Vector elist, AbstractClass exception )
  {
    // System.out.println( "CATCHING: " + exception );
    boolean canHappenFlag = false;
    if( elist != null )
      {
	Enumeration list = elist.elements();
	Vector toRemove = null;
	while( list.hasMoreElements() )
	  {
	    AbstractClass ex = (AbstractClass)list.nextElement();
	    // System.out.println( "    > " + ex );
	    if( ex == exception || ex.isSubclassOf(exception) )
	      {
		if( toRemove == null )
		  toRemove = new Vector();
		toRemove.addElement( ex );
		canHappenFlag = true;
	      }
	    else
	      {
		// System.out.println( " %%% " + exception + " cannot catch " + ex );
		if( !canHappenFlag && exception.isSubclassOf(ex) )
		  canHappenFlag = true; // "partially caught" exception
	      }
	  }
	if( toRemove != null )
	  {
	    list = toRemove.elements();
	    while( list.hasMoreElements() )
	      {
		Object ex = list.nextElement();
		if( ! elist.removeElement( ex ) )
		  {
		    System.err.println( "Internal error: " + ex + " not caught" );
		  }
		// System.out.println( " %%% " + exception + " catches " + ex + " " + elist.size() );
	      }
	    return true;
	  }
      }
    return canHappenFlag || exception == classTable.runtimeExceptionClass ||
      exception.isSubclassOf(classTable.runtimeExceptionClass) ||
      exception == classTable.errorClass ||
      exception.isSubclassOf(classTable.errorClass) ||
      exception == classTable.exceptionClass;
  }

  /**
   *  Defines new local variable.
   *  @param finalFlag true is variable is final
   *  @param type variable type
   *  @param name variable name
   *  @param node BasicNode where this variable is defined (mostly for error reporting)
   *  @returns a (helper-dependent) object that contains all necessary information
   *           for this variable (to be used by code generator)
   */
  private AbstractVariable addVariable( boolean finalFlag, AbstractClass type,
					Identifier name, BasicNode node )
  {
    AbstractVariable old = ( name == null ? null : (AbstractVariable)variables.get( name ) );
    // System.out.println( "add variable: " + name + " " + currentClass );
    if( old != null )
      {
	storeError( ERR_A_VARREDECL, node, new Integer(AbstractTokenizer.getLine(old.getFilePos())) );
	return null;
      }
    else
      {
	int size = getNCells( type );
	int offset;
	if( name == THIS || name == SUPER )
	  {
	    offset = 0;
	    if( currVarOffset == 0 )
	      currVarOffset = 1;
	  }
	else
	  {
	    offset = currVarOffset;
	    currVarOffset += size;
	  }
	if( name == null )
	  return null;

	AbstractVariable var = helper.newVariable( finalFlag, type, name, node, offset );
	variables.put( name, var );
	variableAcc.addElement( name );
	return var;
      }
  }

  /**
   *  Find field/variable with the given name, going through the scopes upward.
   *  @param name variable name
   *  @param simple value important only when outermost scope is Any; if simple is true
   *         search is done normally, if it is false, we never resolve name through outermost
   *         scope
   */
  private AbstractVariable findVariable( Identifier name, boolean simple )
  {
    Hashtable vars = variables; // never null
    int level = levelStack.size();
    LevelStackItem item = null;
    while( true )
      {
	AbstractVariable var;
	if( vars == null && (item != null && item.scope == null ) )
	  {
	    var = (AbstractVariable)fields.get( name );
	    if( var == null && topOuterClass != null 
		&& ( simple || topOuterClass != classTable.anyClass ) )
	      {
		AbstractField field = findField( topOuterClass, name );
		if( field != null )
		  {
		    var = helper.newVariable( topOuterClass, field );
		    fields.put( name, var );
		  }
	      }
	  }
	else
	  {
	    var = (AbstractVariable)vars.get( name );
	    if( var == null )
	      {
		AbstractClass scope = ( item == null ? currentClass : item.scope );
		// only classes (not method scopes) resolve variables through fields
		// because in method it should not possible to refer to variable before
		// it is declared!
		if( scope != null && (scope.getModifiers() & METHOD_SCOPE_FLAG) == 0 )
		  {
		    AbstractField f = findField( scope, name );
		    if( f != null )
		      var = helper.newVariable( scope, f );
		  }
	      }
	    else
	      if( item != null ) // one of outer classes
		var.mustBeField();
	  }
	if( var != null )
	  return var;
  	level--;
	if( level < 0 )
	  return null;
	item = (LevelStackItem)levelStack.elementAt(level);
	vars = item.variables;
      }
  }

  private CallAttachment findFunction( Identifier name, AbstractClass[] argTypes, BasicNode bnode )
  {
    int level = 0;
    CallAttachment att = new CallAttachment();
    int levelStackSize = levelStack.size();
    AbstractClass scope = currentClass;
    Hashtable funs = functions;
    boolean mustBeStatic = staticInitFlag;
    for( int i = levelStackSize - 1 ; i >= 0 || scope != null ; i-- )
      {
	// System.out.println( "Scope: " + scope );
	Vector acc = (Vector)funs.get( name );
	Enumeration extra = ( acc == null ? null : acc.elements() );
	AbstractMethod method;

	boolean thisClassOnly = (scope.getModifiers() & METHOD_SCOPE_FLAG) != 0;

	method = findMethod( scope, name, argTypes, bnode, true, extra, thisClassOnly );

	if( method != null )
	  {
	    att.method = method;
	    if( method != null && mustBeStatic && (method.getModifiers() & STATIC_FLAG) == 0 )
	      {
		Object[] info = { scope };
		throw new CompilerException( ERR_A_NONSTATICACC, bnode.getFilePos(), info );
	      }
	    return att;
	  }

	AbstractField outerField = scope.getOuterField();
	if( outerField == null )
	  mustBeStatic = true;

	scope = scope.getLexicallyOuter();

	if( i >= 0 )
	  {
	    LevelStackItem item = (LevelStackItem)levelStack.elementAt(i);
	    funs = item.functions;
	  }
	else
	  funs = null;
      }
    return null;
  }

  void typecheck( AbstractClass value, ExpressionNode valExpr, AbstractClass var, BasicNode node )
  {
    if( valExpr instanceof ConstNode )
      {
	Object cvo = ((ConstNode)valExpr).val;
	if( cvo instanceof Integer || cvo instanceof Character )
	  {
	    int cv;
	    if( cvo instanceof Integer )
	      cv = ((Integer)cvo).intValue();
	    else
	      cv = ((Character)cvo).charValue();
	    if( (var == classTable.byteType || var == classTable.byteClass) &&
		Byte.MIN_VALUE <= cv && Byte.MAX_VALUE >= cv )
	      return;
	    else if( (var == classTable.shortType || var == classTable.shortClass) &&
	        Short.MIN_VALUE <= cv && Short.MAX_VALUE >= cv )
	      return;
	    else if( (var == classTable.charType || var == classTable.characterClass) &&
	        Character.MIN_VALUE <= cv && Character.MAX_VALUE >= cv )
	      return;
	  }
      }
    if( classTable.canAssign( value, var ) )
      return;
    storeError( ERR_A_TYPEMISMATCH, node, var, value );
  }

  private boolean isIntType( AbstractClass clazz )
  {
    return clazz == classTable.intType || clazz == classTable.shortType ||
      clazz == classTable.byteType || clazz == classTable.charType || 
      clazz == classTable.anyClass;
  }

  private boolean isNumberType( AbstractClass clazz )
  {
    return clazz == classTable.intType || clazz == classTable.shortType ||
      clazz == classTable.longType ||
      clazz == classTable.byteType || clazz == classTable.charType ||
      clazz == classTable.floatType || clazz == classTable.doubleType ||
      clazz == classTable.anyClass;
  }

  public void storeError( CompilerException e )
  {
    if( errorStorage != null )
      errorStorage.storeError( e );
    else
      {
	System.err.println( "Line " + e.getLine() + ", pos " + 
			   e.getColumn() + ": " + e.getMessage() );
      }
  }  

  public void storeError( int code, int fpos )
  {
    storeError( new CompilerException( code, fpos ) );
  }  

  public void storeError( int code, int fpos, Object add )
  {
    Object[] addInfo = { add };
    storeError( new CompilerException( code, fpos, addInfo ) );
  }  

  public void storeError( int code, BasicNode node )
  {
    storeError( new CompilerException( code, node.getFilePos() ) );
  }  

  public void storeError( int code, BasicNode node, Object add )
  {
    Object[] addInfo = { add };
    storeError( new CompilerException( code, node.getFilePos(), addInfo ) );
  }  

  public void storeError( int code, BasicNode node, Object add1, Object add2 )
  {
    Object[] addInfo = { add1, add2 };
    storeError( new CompilerException( code, node.getFilePos(), addInfo ) );
  }  

  private ExpressionNode evalExpr( ExpressionNode expr )
  {
    try
      {
	return (ExpressionNode)expr.eval( this );
      }
    catch( CompilerException e )
      {
	storeError( e );
      }
    catch( RuntimeException e )
      {
	storeError( new CompilerException( ERR_A_INTERNAL, expr.getFilePos(), e ) );
      }
    return expr;
  }

  private AbstractClass variableType( IdentifierNode identNode, AbstractClass clazz )
  {
    ExpressionNode[] indexes = identNode.indexes;
    if( indexes != null && indexes.length != 0 )
      {
	int addDim = indexes.length;
	for( int i = 0 ; i < addDim ; i++ )
	  if( indexes[i] != null )
	    storeError( ERR_A_ARRNOTEMPTY, indexes[i] );
	clazz = classTable.getClass( new QualifiedIdentifier( clazz.getFullName(), addDim ) );
      }
    identNode.clazz = clazz;
    return clazz;
  }

  private AbstractClass findClass( IdentifierNode identNode, boolean expectEmptyIndex )
  {
    if( identNode == null )
      return classTable.anyClass;

    Identifier[] idList = identNode.idList;

    AbstractClass aClass = classTable.getClass( currentClass, idList, imports );
	
    if( aClass == null )
      {
	storeError( ERR_A_TYPENOTFOUND, identNode, identNode.toString() );
      }
    else
      {
	ExpressionNode[] indexes = identNode.indexes;
	if( indexes == null )
	  return aClass;
	for( int i = 0 ; i < indexes.length ; i++ )
	  {
	    if( expectEmptyIndex )
	      {
		if( indexes[i] != null )
		  storeError( ERR_A_ARRNOTEMPTY, indexes[i] );
	      }
	    else
	      {
		expectEmptyIndex = indexes[i] == null;
		if( i == 0 && expectEmptyIndex )
		  storeError( ERR_A_ARREMPTY, identNode );
	      }
	  }
	aClass = classTable.getClass( new QualifiedIdentifier( 
			aClass.getFullName(), indexes.length ) );
	if( aClass == null )
	  storeError( ERR_A_CANTMAKEARR, identNode.getFilePos() );
      }
    return aClass;
  }

  AbstractClass numberTypecheck( AbstractClass type1, AbstractClass type2, BasicNode node )
  {
    if( type1 == classTable.intType || type1 == classTable.shortType || 
	type1 == classTable.charType || type1 == classTable.byteType )
      {
	if( type2 == classTable.intType || type2 == classTable.shortType || 
	    type2 == classTable.byteType || type2 == classTable.charType )
	  return classTable.intType;
	if( type2 == classTable.longType || type2 == classTable.floatType ||
	    type2 == classTable.doubleType || type2 == classTable.anyClass )
	  return type2;
      }
    else if( type1 == classTable.floatType || type1 == classTable.longType )
      {
	if( type2 == classTable.intType || type2 == classTable.shortType || 
	    type2 == classTable.byteType || type2 == classTable.charType ||
	    type2 == classTable.longType )
	  return type1;
	else
	  if( type2 == classTable.floatType || type2 == classTable.doubleType ||
	      type2 == classTable.anyClass )
	    return type2;
      }
    else if( type1 == classTable.doubleType || type1 == classTable.anyClass )
      {
	if( type2 == classTable.intType || type2 == classTable.shortType || 
	    type2 == classTable.byteType || type2 == classTable.charType ||
	    type2 == classTable.longType || type2 == classTable.floatType ||
	    type2 == classTable.doubleType )
	  return type1;
	if( type2 == classTable.anyClass )
	  return type2;
	if( type2 == classTable.stringClass && type1 == classTable.anyClass )
	  return type1;
      }
    Object[] addInfo = { type1, type2 };
    throw new CompilerException( ERR_A_TYPEMISMATCH, node.getFilePos(), addInfo );
  }

  AbstractClass intTypecheck( AbstractClass type1, AbstractClass type2, BasicNode node )
  {
    if( type1 == classTable.intType || type1 == classTable.shortType || 
	type1 == classTable.charType || type1 == classTable.byteType )
      {
	if( type2 == classTable.intType || type2 == classTable.shortType || 
	    type2 == classTable.byteType || type2 == classTable.charType )
	  return classTable.intType;
	if( type2 == classTable.longType || type2 == classTable.anyClass )
	  return type2;
      }
    else if( type1 == classTable.longType || type1 == classTable.anyClass )
      {
	if( type2 == classTable.intType || type2 == classTable.shortType || 
	    type2 == classTable.byteType || type2 == classTable.charType ||
	    type2 == classTable.longType )
	    return type1;
	if( type2 == classTable.anyClass )
	  return type2;
      }
      
    Object[] addInfo = { type1, type2 };
    throw new CompilerException( ERR_A_NOTINTARGS, node.getFilePos(), addInfo );
  }
 
  private AbstractClass accessorClass()
  {
    AbstractClass clazz = currentClass;
    while( clazz != null && (clazz.getModifiers() & METHOD_SCOPE_FLAG) != 0 )
      clazz = clazz.getLexicallyOuter();
    return clazz;
  }

  private AbstractField findField( AbstractClass type, Identifier name )
  {
    return JintUtilities.findField( classTable, type, name, accessorClass() );
  }

  private AbstractField findField( AbstractClass type, Identifier name, AbstractClass accessor )
  {
    return JintUtilities.findField( classTable, type, name, accessor );
  }

  /**
   *  Find method named <i>name</i> with argument types matching <i>argTypes</i>
   *  in class <i>type</i>. Normally only one method is expected to match.
   *  Use <i>node</i> if need to report error; if <i>node</i> is not null,
   *  CompilerException is thrown when multiple methods match. If no methods match or
   *  <i>node</i> is null, null is returned. However if any of <i>argTypes</i> is
   *  class Any, <i>name</i> is not <i>&lt;init&gt;</i> and <i>notAny</i> is false,
   *  when multiple or no methods are found, AnyMethod is returned. Additional
   *  methods to search can be given by <i>extra</i> parameter (only parameters of
   *  those matter, their names can differ from <i>name</i>). If flag
   *  <i>thisClassOnly</i> is true, superclasses and implemented interfaces of
   *  class <i>type</i> are not searched (as they otherwise are).
   */
  private AbstractMethod findMethod( AbstractClass type, Identifier name,
	       AbstractClass[] argTypes, BasicNode node, boolean notAny,
	       Enumeration extra, boolean thisClassOnly )
  {
    Enumeration list = JintUtilities.findMethod( classTable, type, name, argTypes,
			     accessorClass(), extra, thisClassOnly );

    AbstractMethod m = null;

    if( list.hasMoreElements() )
      {
	m = (AbstractMethod)list.nextElement();
	if( ! list.hasMoreElements() )
	  return m; // exactly one match
      }

    if( !notAny )
      {
	for( int i = 0 ; i < argTypes.length ; i++ )
	  if( argTypes[i] == classTable.anyClass )
	    {
	      list = classTable.anyClass.getMethods( name, argTypes.length );
	      return (AbstractMethod)list.nextElement();
	    }
      }

    if( m != null && node != null )
      {
	StringBuffer sb = new StringBuffer();
	sb.append( " " + m );
	while( list.hasMoreElements() )
	  sb.append( " " + list.nextElement() );
	Object[] addInfo = { sb.toString() };
	throw new CompilerException( ERR_A_NOBESTMATCH, node.getFilePos(), addInfo );
      }

    return null;
  }

  private boolean canBeLValue( ExpressionNode expr )
  {
    if( expr instanceof InfixNode )
      {
	int op = ((InfixNode)expr).op;
	return op == '[' || op == '.';
      }
    else
      return expr instanceof IdentifierNode;
  }

  private void addImport( IdentifierNode addimp )
  {
    if( imports == null )
      {
	imports = new IdentifierNode[1];
      }
    else
      {
	IdentifierNode[] newImports = new IdentifierNode[ imports.length + 1 ];
	System.arraycopy( imports, 0, newImports, 0, imports.length );
	imports = newImports;
      }
    imports[imports.length-1] = addimp;
  }

  boolean verifyEvaluability( IdentifierNode ident, AbstractClass base, boolean genErr )
  {
    AbstractClass clazz;
    if( base == null )
      {
	AbstractVariable[] att = ((AbstractVariable[])ident.attachment);
	if( att.length == 0 )
	  clazz = ident.clazz;
	else
	  return true;
      }
    else
      {
	clazz = base;
      }

    // we rely on implicit 'this' pointer, check
    // if we actually can get through the chain

    AbstractClass t = currentClass;
    while( t != clazz && t != null )
      t = t.getOuter();

    if( genErr && t == null )
      {
	Object[] info = { clazz };
	throw new CompilerException( ERR_A_NONSTATICACC, ident.getFilePos(), info );
      }
    
    return t != null;
  }

  //----------------- implementation of Context --------------------

  public Object evalArrayValue( ArrayValueNode arrayValueNode )
  {
    AbstractClass type;
    AbstractClass elementType;
    if( arrayValueNode.type != null )
      type = findClass( arrayValueNode.type, true );
    else if( arrayValueNode.clazz != null )
      type = arrayValueNode.clazz;
    else
      type = classTable.anyClass;
    elementType = type.getElementClass();
    ExpressionNode[] vals = arrayValueNode.vals;
    for( int i = 0 ; i < vals.length ; i++ )
      {
	ExpressionNode vi = vals[i];
	if( vi instanceof ArrayValueNode && type.getDimension() > 0 )
	  ((ArrayValueNode)vi).clazz = elementType;
	vi = evalExpr( vi );
	vals[i] = vi;
	typecheck( vi.clazz, vi, elementType, vi );
      }
    arrayValueNode.clazz = type;
    return arrayValueNode;
  }

  public Object evalBody( BodyNode bodyNode )
  {
    StatementNode[] statements = bodyNode.statements;
    if( statements == null )
      return null;
    int nvar = variableAcc.size();
    int varOffset = currVarOffset;
    try
      {
	for( int i = 0 ; i < statements.length ; i++ )
	  statements[i].eval( this );
      }
    finally
      {
	for( int nvar1 = variableAcc.size() ; nvar1 > nvar ; nvar1-- )
	  {
	    Identifier varname = (Identifier)variableAcc.pop();
	    variables.remove( varname );
	  }
	if( cellReuse )
	  currVarOffset = varOffset;
      }
    return null;
  }

  public Object evalBreak( BreakNode breakNode )
  {
    if( breakNode.label != null )
      {
	breakNode.attachment = labels.get( breakNode.label.idList[0] );
	if( breakNode.attachment == null )
	  storeError( ERR_A_CANTACCLABEL, breakNode, breakNode.label.toString() );
      }
    return null;
  }

  /**
   *  Process method or constructor call expression. Such expression can be
   *  of the following kinds:
   *  <ul>
   *  <li>simple call: f(..), where f is not super or this
   *  <li>qualified call: a.b.f(..) or a.A.f(...), where a is not super
   *  <li>constructor call: this(...) or super(...)
   *  <li>superclass method call: super.f(...)
   *  <li>chained call: like second call in a().b()
   *  </ul>
   */
  public Object evalCall( CallNode callNode )
  {
    ExpressionNode[] args = callNode.args;

    methodFlag = true;
    callNode.func = (ExpressionNode)callNode.func.eval( this );
    AbstractClass targetClass = callNode.func.clazz; // class where to look for a method
    methodFlag = false;

    AbstractClass[] argTypes;
    if( args == null )
      argTypes = new AbstractClass[0];
    else
      {
	argTypes = new AbstractClass[args.length];
	for( int i = 0 ; i < args.length ; i++ )
	  {
	    ExpressionNode expr = (ExpressionNode)args[i].eval( this );
	    args[i] = expr;
	    argTypes[i] = expr.clazz;
	  }
      }

    IdentifierNode ident = null;

    if( callNode.func instanceof IdentifierNode )
      {
	// This is something like a.b.c(), not (blah-blah).a.b();
	ident = (IdentifierNode)callNode.func;
	Identifier nm = ident.idList[0];
	boolean shortFlag = ident.idList.length == 1; // simple (not qualified) identifier
	boolean superFlag = nm == SUPER; // first identifier is "super"
	boolean thisFlag = nm == THIS; // first identifier is "this"

	if( shortFlag && ( superFlag || thisFlag ) )
	  {
	    // this is either this(...) or super(...) -> constructor call
	    if( (currentClass.getModifiers() & STATIC_FLAG) != 0 )
	      throw new CompilerException( ERR_A_STATICTHIS, callNode.getFilePos() );
	    Identifier[] tmp = { nm, INIT };
	    nm = INIT;
	    // it's convenient to add actual constructor name <init> to the identifier list
	    ident = new IdentifierNode( ident.getFilePos(), tmp );
	    // target is "this" class, which is outer class of the current scope
	    targetClass = currentClass.getOuter();
	  }
	else
	  nm = ident.idList[ident.idList.length-1];

	// at this point nm is a method name

	if( thisFlag || superFlag || shortFlag )
	  {
	    CallAttachment att;
	    if( shortFlag && !( thisFlag || superFlag ) )
	      att = findFunction( nm, argTypes, callNode ); // something like f(...)
	    else
	      {
		if( superFlag )
		  {
		    // this works for both super(...) and super.f(...)
		    targetClass = targetClass.getSuperclass();
		  }
		if( nm == INIT )
		  {
		    // some type of constructor
		    AbstractClass need = targetClass.getOuter();
		    if( need != null )
		      {
			// constructor of the inner class needs an extra hidden parameter
			AbstractClass have = currentClass.getOuter();
			AbstractClass[] newArgTypes = new AbstractClass[ argTypes.length + 1 ];
			System.arraycopy( argTypes, 0, newArgTypes, 1, argTypes.length );
			argTypes = newArgTypes;
			while( have != null && have != need && !have.isSubclassOf(need) )
			  have = have.getOuter();
			argTypes[0] = have;
			if( have == null )
			  storeError( ERR_A_NOOUTERCLASS, callNode );
		      }
		  }

		AbstractMethod m = findMethod( targetClass, nm, argTypes, callNode, nm == INIT,
					       null, false );

		if( m == null )
		  att = null;
		else
		  {
		    att = new CallAttachment();
		    att.method = m;
		    att.targetClass = targetClass;
		  }
	      }
	    if( att != null )
	      {
		callNode.attachment = att;
		AbstractMethod m = att.method;
		callNode.clazz = m.getType();
		addExceptions( m );
		return callNode;
	      }
	    String fullname = fullName( nm, argTypes );
	    storeError( ERR_A_NOMETHOD, callNode, fullname, targetClass.getFullName() );
	    return null;
	  }
      }
    else if( callNode.func instanceof InfixNode )
      {
	InfixNode infix = (InfixNode)callNode.func;
	if( infix.op != '.' )
	  throw new CompilerException( ERR_A_BADMETHODNAME, infix.getFilePos() );
	ident = (IdentifierNode)infix.arg2;
      }
    else
      throw new CompilerException( ERR_A_BADMETHODNAME, callNode.func.getFilePos() );

    Identifier name = ident.idList[ident.idList.length-1];

    AbstractMethod method = findMethod( targetClass, name, argTypes, callNode,
					false, null, false );

    if( method == null )
      {
	String fullname = fullName( name, argTypes );	
	storeError( ERR_A_NOMETHOD, callNode, fullname, targetClass ); 
      }
    else
      addExceptions( method );

    callNode.attachment = method;
    if( method != null )
      {
	if( callNode.func instanceof IdentifierNode )
	  {
	    if( method instanceof AnyMethod )
	      {
		if( !verifyEvaluability( (IdentifierNode)callNode.func, null, false ) )
		  ((AnyMethod)method).makeStatic();
	      }
	    else
	      {
		if( (method.getModifiers() & STATIC_FLAG) == 0 )
		  verifyEvaluability( (IdentifierNode)callNode.func, null, true );
	      }
	  }
	callNode.clazz = method.getType();
      }
    return callNode;
  }

  public Object evalCase( CaseNode caseNode )
  {
    storeError( ERR_A_NEVER, caseNode );
    return null;
  }

  public Object evalCast( CastNode castNode )
  {
    AbstractClass castClass = findClass( castNode.type, true );
    castNode.clazz = castClass;
    castNode.arg = (ExpressionNode)castNode.arg.eval( this );
    AbstractClass aClass = castNode.arg.clazz;
    if( castClass != classTable.anyClass && aClass != classTable.anyClass &&
	aClass != null && castClass != null )
      {
	if( castClass.isPrimitive() )
	  if( aClass.isPrimitive() )
	    {
	      if( aClass == classTable.booleanType )
		storeError( ERR_A_BOOLCAST, castNode );
	      else if( aClass == classTable.voidType )
		storeError( ERR_A_VOIDCAST, castNode );
	      if( castClass == classTable.booleanType )
		storeError( ERR_A_TOBOOLCAST, castNode );
	      if( castClass == classTable.voidType )
		storeError( ERR_A_TOVOIDCAST, castNode );
	    }
	  else
	    storeError( ERR_A_OBJTOPRIMCAST, castNode );
	else
	  if( aClass.isPrimitive() )
	    storeError( ERR_A_PRIMTOOBJCAST, castNode );
	  else
	    {
	      if( !aClass.isInterface() && !castClass.isInterface() &&
		  !aClass.isSubclassOf( castClass ) && aClass != castClass &&
		  !castClass.isSubclassOf( aClass ) )
		storeError( ERR_A_CANTCAST , castNode, aClass, castClass );
	    }
      }
    return castNode;
  }

  public Object evalCatch( CatchNode catchNode )
  {
    AbstractClass type = null;
    int nvar = 0;
    int varOffset = 0; 

    nvar = variableAcc.size();
    varOffset = currVarOffset;

    if( catchNode.type != null )
      {
	type = findClass( catchNode.type, false );
	if( type == null || ! (type.isSubclassOf( classTable.throwableClass ) ||
			       type == classTable.throwableClass) )
	  storeError( ERR_A_NOTTHROWABLE, catchNode, type );
	Identifier name = catchNode.name.idList[0];
	
	AbstractVariable var = addVariable( false, type, name, catchNode.name );

	catchNode.name.attachment = var;
	catchNode.name.clazz = type;
      }
    else if( tmpVariable )
      {
	Identifier name = classTable.getNameTable().lookupAndInsert( "t$" + tmpVariableCount );
	tmpVariableCount++;
	catchNode.attachment = addVariable( false, classTable.objectClass, name, catchNode );
      }

    catchNode.body.eval( this );

    for( int nvar1 = variableAcc.size() ; nvar1 > nvar ; nvar1-- )
      {
	Identifier varname = (Identifier)variableAcc.pop();
	variables.remove( varname );
      }
    if( cellReuse )
      currVarOffset = varOffset;

    return null;
  }

  public Object evalClass( ClassNode classNode )
  {
    if( classNode.regex != null )
      {
	classNode.regex.eval( this );
	return null;
      }
    JintClass scope = (JintClass)classNode.attachment;
    pushLevel( scope );
    try
      {
	addVariable( true, scope, THIS, classNode );
	addVariable( true, scope, SUPER, classNode );

	StatementNode[] defs = classNode.statements;
	for( int i = 0 ; i < defs.length ; i++ )
	  defs[i].eval( this );
	return null;
      }
    finally
      {
	popLevel();
	scope.squeezeOuterChain();
      }
  }

  public Object evalCond( CondNode condNode )
  {
    condNode.cond = evalExpr( condNode.cond );
    AbstractClass clazz = condNode.cond.clazz;
    if( clazz != classTable.booleanType && clazz != classTable.anyClass )
      storeError( ERR_A_MUSTBEBOOL, condNode.cond );
    clazz = null;

    condNode.arg1 = (ExpressionNode)condNode.arg1.eval( this );
    condNode.arg2 = (ExpressionNode)condNode.arg2.eval( this );
    AbstractClass clazz1 = condNode.arg1.clazz;
    AbstractClass clazz2 = condNode.arg2.clazz;
    if( clazz1 == clazz2 )
      clazz = clazz1;
    else if( clazz1 == classTable.anyClass || clazz2 == classTable.anyClass )
      clazz = classTable.anyClass;
    if( clazz1 != null && clazz1.isPrimitive() )
      {
	if( clazz2 != null && clazz2.isPrimitive() )
	  {
	    ConstNode constNode = null;
	    if( condNode.arg1 instanceof ConstNode )
	      constNode = (ConstNode)condNode.arg1;
	    else if( condNode.arg2 instanceof ConstNode )
	      {
		AbstractClass tClazz = clazz1;
		clazz1 = clazz2;
		clazz2 = tClazz;
		constNode = (ConstNode)condNode.arg2;
	      }
	    if( clazz1 == classTable.byteType && clazz2 == classTable.shortType )
	      clazz = clazz2;
	    else if( clazz1 == classTable.shortType && clazz2 == classTable.byteType )
	      clazz = clazz1;
	    else if( clazz1 == classTable.intType && constNode != null )
	      {
		int iv = ((Integer)constNode.val).intValue();
		if( clazz2 == classTable.byteType )
		  {
		    if( Byte.MIN_VALUE <= iv && iv <= Byte.MAX_VALUE )
		      clazz = clazz2;
		  }
		else if( clazz2 == classTable.shortType )
		  {
		    if( Short.MIN_VALUE <= iv && iv <= Short.MAX_VALUE )
		      clazz = clazz2;
		  }
		else if( clazz2 == classTable.charType )
		  {
		    if( Character.MIN_VALUE <= iv && iv <= Character.MAX_VALUE )
		      clazz = clazz2;
		  }
	      }
	    if( clazz == null )
	      clazz = numberTypecheck( clazz1, clazz2, condNode );
	  }
	else
	  storeError( ERR_A_TYPEMISMATCH, condNode, clazz1, clazz2 );
      }
    else
      {
	if( clazz2 == null )
	  clazz = clazz1;
	else if( clazz1 == null )
	  clazz = clazz2;
	else if( clazz2.isPrimitive() )
	  storeError( ERR_A_TYPEMISMATCH, condNode, clazz1, clazz2 );
	else if( clazz1.isSubclassOf( clazz2 ) || clazz1 == clazz2 )
	  clazz = clazz2;
	else if( clazz2.isSubclassOf( clazz1 ) )
	  clazz = clazz1;
	else
	  storeError( ERR_A_TYPEMISMATCH, condNode, clazz1, clazz2 );
      }
    condNode.clazz = clazz;
    return condNode;
  }

  public Object evalConst( ConstNode constNode )
  {
    Object val = constNode.val;
    if( val instanceof Integer )
      constNode.clazz = classTable.intType;
    else if( val == null )
      constNode.clazz = null;
    else if( val instanceof Long )
      constNode.clazz = classTable.longType;
    else if( val instanceof Character )
      constNode.clazz = classTable.charType;
    else if( val instanceof Double )
      constNode.clazz = classTable.doubleType;
    else if( val instanceof Float )
      constNode.clazz = classTable.floatType;
    else if( val instanceof Short )
      constNode.clazz = classTable.shortType;
    else if( val instanceof Byte )
      constNode.clazz = classTable.byteType;
    else if( val instanceof Boolean )
      constNode.clazz = classTable.booleanType;
    else if( val instanceof String )
      constNode.clazz = classTable.stringClass;
    else
      constNode.clazz = classTable.getClass( constNode.val.getClass() );
    return constNode;
  }

  public Object evalContinue( ContinueNode continueNode )
  {
    if( continueNode.label != null )
      {
	continueNode.attachment = labels.get( continueNode.label.idList[0] );
	if( continueNode.attachment == null )
	  storeError( ERR_A_CANTACCLABEL, continueNode, continueNode.label.toString() );
      }
    return null;
  }

  public Object evalDo( DoNode doNode )
  {
    doNode.body.eval( this );
    doNode.cond = evalExpr( doNode.cond );
    AbstractClass clazz = doNode.cond.clazz;
    if( clazz != classTable.booleanType && clazz != classTable.anyClass )
      storeError( ERR_A_MUSTBEBOOL, doNode.cond );
    return null;
  }

  public Object evalExpressionStatement( ExpressionStatementNode esNode )
  {
    esNode.expr = evalExpr( esNode.expr );
    return null;
  }

  public Object evalFor( ForNode forNode )
  {
    int nvar = variableAcc.size();
    int varOffset = currVarOffset;
    if( forNode.init != null ) 
      forNode.init.eval( this );

    if( forNode.cond != null )
      {
	forNode.cond = evalExpr( forNode.cond );
	AbstractClass clazz = forNode.cond.clazz;
	if( clazz != classTable.booleanType && clazz != classTable.anyClass )
	  storeError( ERR_A_MUSTBEBOOL, forNode.cond );
      }

    if( forNode.next != null )
      forNode.next = evalExpr( forNode.next );
    if( forNode.body != null )
      forNode.body.eval( this );

    for( int nvar1 = variableAcc.size() ; nvar1 > nvar ; nvar1-- )
      {
	Identifier varname = (Identifier)variableAcc.pop();
	variables.remove( varname );
      }
    if( cellReuse )
      currVarOffset = varOffset;

    return null;
  }

  public Object evalIdentifier( IdentifierNode identifierNode )
  {
    Identifier name = identifierNode.idList[0];
    AbstractVariable var = null;
    Identifier[] idList = identifierNode.idList;

    Vector acc = new Vector();

    if( methodFlag && idList.length == 1 )
      {
	identifierNode.clazz = ( fieldOwner == null ? currentClass : fieldOwner );
	return identifierNode;
      }

    if( fieldOwner == null )
      {
	var = findVariable( name, false );
	if( var == null && lvalue && freshFlag && idList.length == 1 )
	  if( topOuterClass == classTable.anyClass )
	    {
	      // if outermost scope is Any, we do not know for sure if this
	      // name is a valid variable in the outermost scope, so we just
	      // assume it is
	      AbstractField field = findField( topOuterClass, name );
	      var = helper.newVariable( topOuterClass, field );
	    }
	  else
	    {
	      var = addVariable( false, classTable.anyClass, name, identifierNode );
	      storeError( ERR_A_IMPLICITDCL, identifierNode, name );
	    }
	if( var != null )
	  acc.addElement( var );
      }

    AbstractClass aClass;

    int firstAccess = 0; // first identifier in idList which is a field

    boolean mustBeStatic = staticInitFlag;

    int size = idList.length;
    
    if( methodFlag )
      size--;

    if( var != null )
      {
	aClass = var.getType();
	firstAccess = 1;
      }
    else if( fieldOwner != null )
      aClass = fieldOwner;
    else if( idList[0] == SELF )
      {
	aClass = currentClass;
	firstAccess = 1;
	currentClass.reportSelfUsage();
      }
    else
      {
	// this must be something like ClassName.member or package.ClassName.member	
	aClass = null;
	while( aClass == null )
	  {
	    firstAccess++;
	    if( firstAccess >= idList.length )
	      {
		// we reached the outermost level
		if( topOuterClass == classTable.anyClass )
		  {
		    AbstractField field = findField( topOuterClass, name );
		    var = helper.newVariable( topOuterClass, field );
		    acc.addElement( var );
		    break;
		  }
		else
		  {
		    Object[] addInfo = { identifierNode };
		    throw new CompilerException( ERR_A_CANTRESOLVE,
						 identifierNode.getFilePos(), addInfo );
		  }
	      }
	    aClass = classTable.getClass( currentClass, idList, 0, firstAccess, imports );
	  }

	if( var != null )
	  {
	    aClass = var.getType();
	    firstAccess = 1;
	  }
	else
	  {
	    // check for inner classes, like a.b.A.B
	    int fa = firstAccess;
	    for( fa++ ; fa < idList.length ; fa++ )
	      {
		AbstractClass ac = classTable.getClass( currentClass, idList, 0, fa, imports );
		if( ac != null )
		  aClass = ac;
		else
		  break;
	      }
	    firstAccess = fa - 1;
	    
	    if( idList[firstAccess] != CLASS )
	      {
		AbstractClass cc = currentClass;
		mustBeStatic = true;
		boolean firstStep = true;
		while( cc instanceof JintClass )
		  {
		    if( cc == aClass || ( firstStep && cc.isSubclassOf( aClass ) ) )
		      {
			mustBeStatic = false;
			break;
		      }
		    firstStep = false; // This is in Java 1.2
		    AbstractField field = ((JintClass)cc).getOuterField();
		    if( field == null )
		      {
			acc.removeAllElements();
			break;
		      }
		    cc = field.getType();
		  }
		
		if( idList[firstAccess] == THIS )
		  {
		    verifyEvaluability( identifierNode, aClass, true );
		    firstAccess++; // got an instance of that class
		    mustBeStatic = false;
		  }
		
	      }
	  }
      }
    
    int i = firstAccess;
    
    for( ; i < size && aClass != null ; i++ )
      {
	Identifier fname = idList[i];

	if( aClass.getDimension() > 0 && fname == LENGTH )
	  {
	    acc.addElement( helper.arrayLength( classTable, aClass ) );
	    aClass = classTable.intType;
	    continue;
	  }

	if( fname == CLASS )
	  {
	    acc.removeAllElements();
	    acc.addElement( helper.classAccess( classTable, aClass ) );
	    aClass = classTable.classClass;
	    continue;
	  }

	AbstractField field = findField( aClass, fname );
	if( field != null )
	  {
	    AbstractClass objClass = aClass;
	    aClass = field.getType();
	    var = helper.newVariable( objClass, field );
	    if( (field.getModifiers() & currentMethod.STATIC_FLAG ) != 0 )
	      acc.removeAllElements();
	    else if( mustBeStatic )
	      storeError( ERR_A_FIELDNOTSTAT, identifierNode, fname );
	    acc.addElement( var );
	  }
	else
	  {
	    storeError( ERR_A_NOFIELD, identifierNode, fname, aClass.getFullName() );
	    aClass = classTable.anyClass; // so no extra errors are generated
	  }
	mustBeStatic = false;
      }
  
    AbstractVariable[] att = new AbstractVariable[ acc.size() ];
    acc.copyInto( att );

    if( att.length == 1 )
      {
	// System.out.println( identifierNode + ": " + att[0] );
	Object v = constValue( att[0] );
	if( v != AbstractVariable.NOT_CONST )
	  {
	    ConstNode subst = new ConstNode( identifierNode.getFilePos(), v );
	    return subst.eval( this );
	  }
      }

    identifierNode.attachment = att;

    identifierNode.clazz = aClass;
    return identifierNode;
  }

  public Object evalFormat( FormatNode formatNode )
  {
    if( formatNode.bound )
      {
	formatNode.clazz = classTable.stringClass;
	String[] vars = formatNode.format.getVariables();
	if( vars != null )
	  {
	    NameTable nameTable = classTable.getNameTable();
	    AbstractVariable[][] avArr = new AbstractVariable[vars.length][];
	    for( int i = 0 ; i < vars.length ; i++ )
	      {
		Identifier nm = nameTable.lookupAndInsert( vars[i] );
		AbstractVariable var = findVariable( nm, true );
		if( var == null )
		  {
		    storeError( ERR_A_NOVAR, formatNode, nm );
		    continue;
		  }
		AbstractVariable[] av = { var };
		avArr[i] = av;
	      }
	    formatNode.attachment = avArr;
	  }
      }
    else
      formatNode.clazz = classTable.formatterClass;
    return formatNode;
  }

  public Object evalIf( IfNode ifNode )
  {
    ifNode.cond = evalExpr( ifNode.cond );
    AbstractClass clazz = ifNode.cond.clazz;
    if( clazz != classTable.booleanType && clazz != classTable.anyClass )
      storeError( ERR_A_MUSTBEBOOL, ifNode );
    ifNode.thenPart.eval( this );
    if( ifNode.elsePart != null )
      ifNode.elsePart.eval( this );
    return null;
  }

  public Object evalImportStatement( ImportStatementNode importStatement )
  {
    addImport( importStatement.path );
    return null;
  }

  public Object evalInfix( InfixNode infixNode )
  {
    try
      {
	AbstractClass t;
	int op = infixNode.op;
	switch( op )
	  {
	    // special rules
	    case '.' :
	      {
		boolean saveMethodFlag = methodFlag;
		methodFlag = false;
		infixNode.arg1 = (ExpressionNode)infixNode.arg1.eval( this );
		fieldOwner = infixNode.arg1.clazz;
		methodFlag = saveMethodFlag;
		if( ! (infixNode.arg2 instanceof IdentifierNode ) )
		  throw new CompilerException( ERR_A_FIELDNOTIDENT, infixNode.getFilePos() );
		infixNode.arg2.eval( this );
		t = infixNode.arg2.clazz;
		fieldOwner = null;
		infixNode.clazz = t;
	      }
	      return infixNode;
	    case TOKEN_INSTANCEOF :
	      {
		infixNode.arg1 = (ExpressionNode)infixNode.arg1.eval( this );
		AbstractClass type1 = infixNode.arg1.clazz;
		AbstractClass type2 = findClass( (IdentifierNode)infixNode.arg2, true );
		infixNode.arg2.clazz = type2;
		if( type1 != null && type1.isPrimitive() )
		  storeError( ERR_A_INSTOFPRIM1, infixNode.arg1 );
		if( type2 != null && type2.isPrimitive() )
		  storeError( ERR_A_INSTOFPRIM2, infixNode.arg2 );
		infixNode.clazz = classTable.booleanType;
	      }
	      return infixNode;
	    case '~' :
	    case TOKEN_TE :
	      {
		infixNode.arg1 = (ExpressionNode)infixNode.arg1.eval( this );
		infixNode.arg2 = (ExpressionNode)infixNode.arg2.eval( this );
		AbstractClass type1 = infixNode.arg1.clazz;
		AbstractClass type2 = infixNode.arg2.clazz;
		if( infixNode.arg2 instanceof RegexNode )
		  {
		    RegexNode tpn = (RegexNode)infixNode.arg2;
		    if( (tpn.flags & TOK_OPT_OFFLINE) == 0 && (tpn.flags & TOK_OPT_IGNORE) == 0 )
		      tpn.embed = true;
		  }

		if( type1 != classTable.stringClass && type1 != classTable.charTypeArray &&
		   type1 != classTable.charStringClass && type1 != classTable.anyClass )
		  storeError( ERR_A_MATCHARG1, infixNode );
		if( type2 != classTable.anyClass &&
		    type2 != classTable.regexClass && type2 != classTable.replacerClass &&
		    !type2.isSubclassOf(classTable.regexClass) && !type2.isSubclassOf(classTable.replacerClass) )
		  storeError( ERR_A_MATCHARG2, infixNode );
		if( type2 == classTable.regexClass || type2.isSubclassOf(classTable.regexClass) )
		  {
		    if( op == TOKEN_TE )
		      storeError( ERR_A_REGEXTE, infixNode );
		    infixNode.clazz = classTable.booleanType;
		  }
		else if( type2 == classTable.anyClass )
		  infixNode.clazz = type2;
		else
		  infixNode.clazz = type1;
	      }
	      return infixNode;

	    // normal case 
	    case '=' :
	    case AbstractTokenizer.TOKEN_GGE :
	    case AbstractTokenizer.TOKEN_GGGE :
	    case AbstractTokenizer.TOKEN_LLE :
	    case AbstractTokenizer.TOKEN_PE :
	    case AbstractTokenizer.TOKEN_ME :
	    case AbstractTokenizer.TOKEN_AE :
	    case AbstractTokenizer.TOKEN_SE :
	    case AbstractTokenizer.TOKEN_PCE :
	    case AbstractTokenizer.TOKEN_ORE :
	    case AbstractTokenizer.TOKEN_ANDE :
	    case AbstractTokenizer.TOKEN_XE :
	      freshFlag = op == '=';
	      lvalue = canBeLValue( infixNode.arg1 ); // lvalue is NOT local
	      if( !lvalue )
		storeError( ERR_A_NOTLVALUE, infixNode );
	      break;
	    default:
	      lvalue = false;
	      break;
	  }

	infixNode.arg1 = (ExpressionNode)infixNode.arg1.eval( this );
	lvalue = false;
	infixNode.arg2 = (ExpressionNode)infixNode.arg2.eval( this );
	AbstractClass type1 = infixNode.arg1.clazz;
	AbstractClass type2 = infixNode.arg2.clazz;

	ExpressionNode lhs = infixNode.arg1;

	if( (lhs instanceof ConstNode || infixNode.arg2 instanceof ConstNode) )
	  {
	    Object v = constValue( infixNode );
	    if( v != infixNode.NOT_CONST )
	      return evalConst(new ConstNode( infixNode.getFilePos(), v ));
	  }

	switch( op )
	  {
	    case '=' :
	      {
		typecheck( type2, infixNode.arg2, type1, infixNode );
		infixNode.clazz = type1;
	      }
	      break;
	    case '[' :
	      {
		int dim = type1.getDimension();
		boolean indexMustBeInteger;
		if( dim != 0 )
		  {
		    infixNode.clazz = type1.getElementClass();
		    indexMustBeInteger = true;
		  }
		else
		  {
		    if( (classTable.abstractListClass != null ?
			  type1 == classTable.abstractListClass || 
			          type1.isSubclassOf( classTable.abstractListClass ) :
			  type1 == classTable.vectorClass || type1.isSubclassOf( classTable.vectorClass ) )
			)
		      indexMustBeInteger = true;
		    else
		      {
			indexMustBeInteger = false;
			if( !type1.isSubclassOf( classTable.dictionaryClass )
			    && type1 != classTable.dictionaryClass
			    && type1 != classTable.anyClass )
			  storeError( ERR_A_NOTINDEXABLE, infixNode, type1 );
		      }
		    infixNode.clazz = classTable.anyClass;
		  }
		if( indexMustBeInteger )
		  {
		    if( ! isIntType( type2 ) || type2 == classTable.longType )
		      storeError( ERR_A_NOTINTEGER, infixNode, type2 );
		  }
	      }
	      break;
	    case AbstractTokenizer.TOKEN_GE :
	    case AbstractTokenizer.TOKEN_LE :
	    case '>' :
	    case '<' :
	      {
		infixNode.attachment = numberTypecheck( type1, type2, infixNode );
		infixNode.clazz = classTable.booleanType;
	      }
	      break;
	    case AbstractTokenizer.TOKEN_NE :
	    case AbstractTokenizer.TOKEN_EE :
	      {
		infixNode.clazz = classTable.booleanType;
		if( classTable.anyClass != null && 
		    (type1 == classTable.anyClass || type2 == classTable.anyClass) )
		  infixNode.attachment = classTable.anyClass;
		else if( type1 == type2 )
		  infixNode.attachment = type1;
		else if( type1 == null || type2 == null || 
			(!type1.isPrimitive() && !type2.isPrimitive() && 
			 type1.isSubclassOf(type2) || type2.isSubclassOf(type1) ) )
		  infixNode.attachment = classTable.objectClass;
		else
		  infixNode.attachment = numberTypecheck( type1, type2, infixNode );
	      }
	      break;
	    case AbstractTokenizer.TOKEN_AND :
	    case AbstractTokenizer.TOKEN_OR :
	      {
		if( type1 != classTable.booleanType && type1 != classTable.anyClass )
		  storeError( ERR_A_MUSTBEBOOL, infixNode.arg1 );
		if( type2 != classTable.booleanType && type2 != classTable.anyClass )
		  storeError( ERR_A_MUSTBEBOOL, infixNode.arg2 );
		infixNode.clazz = classTable.booleanType;
	      }
	      break;
	    case ',' :
	      infixNode.clazz = type2;
	      break;
	    case AbstractTokenizer.TOKEN_GGE :
	    case AbstractTokenizer.TOKEN_GGGE :
	    case AbstractTokenizer.TOKEN_LLE :
	    case AbstractTokenizer.TOKEN_GG :
	    case AbstractTokenizer.TOKEN_GGG :
	    case AbstractTokenizer.TOKEN_LL :
	      {
		if( !isIntType( type2 ) && type2 != classTable.longType )
		  storeError( ERR_A_NOTINTEGER, infixNode.arg2, type2 );
		if( isIntType( type1 ) )
		  infixNode.clazz = classTable.intType;
		else if( type1 == classTable.longType )
		  infixNode.clazz = type1;
		else if( type1 != null )
		  storeError( ERR_A_NOTINTEGER, infixNode.arg1, type1 );
		infixNode.attachment = infixNode.clazz;
	      }
	      break;
	    case AbstractTokenizer.TOKEN_ORE :
	    case AbstractTokenizer.TOKEN_ANDE :
	    case AbstractTokenizer.TOKEN_XE :
	    case '|' :
	    case '&' :
	    case '^' :
	      {
		infixNode.clazz = intTypecheck( type1, type2, infixNode );
		infixNode.attachment = infixNode.clazz;
	      }
	      break;
	    case AbstractTokenizer.TOKEN_PE :
	    case AbstractTokenizer.TOKEN_ME :
	    case AbstractTokenizer.TOKEN_AE :
	    case AbstractTokenizer.TOKEN_SE :
	    case AbstractTokenizer.TOKEN_PCE :
	    case '+' :
	    case '-' :
	    case '*' :
	    case '/' :
	    case '%' :
	      {
		if( (type1 == classTable.stringClass || type2 == classTable.stringClass) && 
		   ( infixNode.op == '+' || infixNode.op == AbstractTokenizer.TOKEN_PE ) )
		  infixNode.clazz = classTable.stringClass;
		else
		  infixNode.clazz = numberTypecheck( type1, type2, infixNode );
		infixNode.attachment = infixNode.clazz;
	      }
	      break;
	    default:
	      {
		Object[] info = { "'" + op + "'" };
		throw new CompilerException( ERR_A_BADOP, infixNode.getFilePos(), info );
	      }
	  }
      }
    catch( CompilerException e )
      {
	storeError( e );
      }
    catch( RuntimeException e )
      {
	storeError( new CompilerException( ERR_A_INTERNAL, infixNode.getFilePos(), e ) );
      }
    return infixNode;
  }

  public Object evalLabeled( LabeledNode labeledNode )
  {
    Identifier label = labeledNode.label.idList[0];
    if( labels.get( label ) != null )
      {
	storeError( ERR_A_LABELREDEF, labeledNode, label.toString() );
	labeledNode.statement.eval( this );
      }
    else
      {
	labels.put( label, labeledNode );
	labeledNode.statement.eval( this );
	labels.remove( label );
      }
    return null;
  }

  public Object evalMember( MemberNode memberNode )
  {
    staticInitFlag = (memberNode.getFlags() & memberNode.STATIC_FLAG) != 0;
    try
      {
	AbstractClass type = findClass( memberNode.type, true );
	boolean finalFlag = (memberNode.getFlags() & memberNode.FINAL_FLAG) != 0;
	ExpressionNode[] defs = memberNode.defs;
	for( int i = 0 ; i < defs.length ; i++ )
	  {
	    AbstractClass t = type;
	    IdentifierNode varName = null;
	    if( defs[i] instanceof InfixNode )
	      {
		InfixNode expr = (InfixNode)defs[i];
		if( expr.op == '=' && expr.arg1 instanceof IdentifierNode )
		  {
		    varName = (IdentifierNode)expr.arg1; 
		    t = variableType( varName, t );
		    if( expr.arg2 instanceof ArrayValueNode )
		      expr.arg2.clazz = t;
		  }
		else
		  storeError( ERR_A_NOTSIMPLEID, expr.arg1 );
	      }
	    else if( defs[i] instanceof IdentifierNode )
	      {
		varName = (IdentifierNode)defs[i];
		t = variableType( varName, t );
	      }
	    else
	      storeError( ERR_A_BADDCL, defs[i].getFilePos() );
	    if( varName != null )
	      {
		if( varName.idList.length > 1 )
		  storeError( ERR_A_NOTSIMPLEID, varName );
		else
		  addVariable( finalFlag, t, varName.idList[0], defs[i] );
		defs[i] = evalExpr( defs[i] );
	      }
	  }
      }
    finally
      {
	staticInitFlag = false;
      }
    return null;
  }

  public Object evalMethod( MethodNode methodNode )
  {
    // System.out.println( "evalMethod: " + methodNode.name + " " + methodNode.name.attachment );
    Identifier key = methodNode.name.idList[0];
    ParameterNode[] params = methodNode.params;
    if( params != null && params.length > 0 )
      {
	for( int i = 0 ; i < params.length ; i++ )
	  {
	    if( params[i].name.clazz != null )
	      continue;
	    AbstractClass pclazz = ( params[i].type == null ? classTable.anyClass : 
				     findClass( params[i].type, true ) );
	    if( pclazz == null )
	      storeError( ERR_A_TYPENOTFOUND, params[i].type, params[i].type );
	    params[i].name.clazz = pclazz;
	  }
      }

    AbstractMethod method = (AbstractMethod)methodNode.name.attachment;
    
    Vector list = (Vector)functions.get( key );
    if( list == null )
      {
	list = new Vector();
	functions.put( key, list );
      }
    list.addElement( method );

    if( currentClass != null )
      currentClass.markAsClass();

    JintClass scope = (JintClass)methodNode.attachment;
    pushLevel( scope );

    MethodNode oldMethod = currentMethod;

    try
      {
	currentMethod = methodNode;

	if( ( methodNode.getFlags() & methodNode.STATIC_FLAG ) == 0 )
	  {
	    AbstractClass thisClass = currentClass.getOuter();
	    addVariable( true, thisClass, THIS, methodNode );
	    addVariable( true, thisClass, SUPER, methodNode );
	    if( thisClass.getOuter() != null && method != null &&
	        method.getName() == INIT.toString() )
	      addVariable( true, thisClass.getOuter(), null, methodNode ); // hidden arg to inc maxvar
	  }

	if( methodNode.type != null )
	  methodNode.type.clazz = findClass( methodNode.type, true );

	if( params != null )
	  for( int i = 0 ; i < params.length ; i++ )
	    params[i].eval( this );

	StatementNode[] stmts = methodNode.body;
	if( stmts != null )
	  {
	    for( int i = 0 ; i < stmts.length ; i++ )
	      stmts[i].eval( this );
	  }

	if( exceptionList != null && exceptionList.size() > 0 )
	  {
	    AbstractClass[] thrownList = method.getExceptionTypes();
	    if( thrownList != null )
	      {
		for( int i = 0 ; i < thrownList.length ; i++ )
		  catchException( exceptionList, thrownList[i] );
	      }
	    catchException( exceptionList, classTable.runtimeExceptionClass );
	    catchException( exceptionList, classTable.errorClass );
	    if( exceptionList.size() > 0 )
	      storeError( ERR_A_NOTCAUGHT, methodNode.getFilePos(), exceptionList );
	  }

      }
    finally
      {
	popLevel();
	currentMethod = oldMethod;
      }
    return null;
  }

  public Object evalNew( NewNode newNode )
  {
    IdentifierNode type = newNode.type;
    AbstractClass aClass;
    if( newNode.inlineClass == null )
      aClass = findClass( type, false );
    else
      aClass = (AbstractClass)newNode.inlineClass.attachment;

    if( aClass == null )
      return newNode;

    ExpressionNode[] args = type.indexes;
    
    if( args != null )
      {
	for( int i = 0 ; i < args.length ; i++ )
	  if( args[i] != null )
	    {
	      args[i] = (ExpressionNode)args[i].eval( this );
	      AbstractClass iClass = args[i].clazz;
	      if( ! isIntType( iClass ) )
		storeError( ERR_A_NOTINTEGER , args[i], iClass );
	    }
      }

    args = newNode.args;

    if( args != null )
      {
	if( aClass.getDimension() > 0 )
	  throw new CompilerException( ERR_A_NEWARRARGS, newNode.getFilePos() );

	AbstractField outerField = aClass.getOuterField();
	int addParam = outerField != null ? 1 : 0;
	AbstractClass[] argTypes = new AbstractClass[args.length + addParam];
	if( outerField != null )
	  argTypes[0] = outerField.getType();
	for( int i = 0 ; i < args.length ; i++ )
	  {
	    args[i] = (ExpressionNode)args[i].eval( this );
	    argTypes[i+addParam] = args[i].clazz;
	  }

	Identifier name = classTable.getNameTable().INIT;

	AbstractMethod method;

	if( aClass == classTable.anyClass )
	  method = null; // Compiler will generate correct sequence
	else
	  {
	    method = findMethod( aClass, name, argTypes, newNode, outerField!=null, null, true );
	    if( method == null )
	      {
		storeError( ERR_A_NOCONSTRUCTOR, newNode, fullName( null, argTypes ), aClass );
		/*
		  if( aClass instanceof BasicClass )
		  ((BasicClass)aClass).dumpMethods( name );
		*/
	      }
	    else
	      addExceptions( method );
	  }

	NewAttachment att = new NewAttachment();

	att.method = method;

	newNode.attachment = att;

	if( newNode.inlineClass != null )
	  {
	    evalClass( newNode.inlineClass );
	    if( currentClass != null &&
		currentClass.isSubclassOf( classTable.abstractContainerClass ) )
	      {
		AbstractClass[] addArg = { aClass };
		att.addMethod = findMethod( currentClass, ADD, addArg, null, false, null, false );
		//System.out.println( "Trying to add " + aClass + " : " + att.addMethod );
		if( att.addMethod != null )
		  addExceptions( att.addMethod );
	      }
	  }

      }
    
    newNode.clazz = aClass;
    
    return newNode;
  }


  public Object evalPackageStatement( PackageStatementNode packageStatementNode )
  {
    Identifier[] idList = packageStatementNode.path.idList;
    Identifier[] idList1 = new Identifier[ idList.length + 1 ];
    System.arraycopy( idList, 0, idList1, 0, idList.length );
    addImport( new IdentifierNode( packageStatementNode.path.getFilePos(), idList1 ) );
    return null;
  }

  public Object evalParameter( ParameterNode parameterNode )
  {
    AbstractClass clazz = parameterNode.name.clazz;
    parameterNode.name.attachment = addVariable( parameterNode.finalFlag, clazz,
	       parameterNode.name.idList[0], parameterNode );
    return null;
  }

  public Object evalPostfix( PostfixNode postfixNode )
  {
    freshFlag = false;
    lvalue = canBeLValue( postfixNode.arg ); // lvalue is TyperContext field
    if( !lvalue )
      storeError( ERR_A_NOTLVALUE, postfixNode );
    postfixNode.arg = (ExpressionNode)postfixNode.arg.eval( this );
    AbstractClass type = postfixNode.arg.clazz;
    lvalue = false;
    switch( postfixNode.op )
      {
        case AbstractTokenizer.TOKEN_PP :
        case AbstractTokenizer.TOKEN_MM :
	  {	
	    if( ! isNumberType( type ) )
	      storeError( ERR_A_NOTNUMBER, postfixNode.arg, type );
	    postfixNode.clazz = type;
	    return postfixNode;
	  }
      }
    storeError( ERR_A_BADOP, postfixNode, "'"+(char)postfixNode.op+"'" );
    return postfixNode;
  }

  public Object evalPrefix( PrefixNode prefixNode )
  {
    if( prefixNode.op == AbstractTokenizer.TOKEN_PP ||
	prefixNode.op == AbstractTokenizer.TOKEN_MM )
      {
	freshFlag = false;
	lvalue = canBeLValue( prefixNode.arg );
	if( !lvalue )
	  storeError( ERR_A_NOTLVALUE, prefixNode );
      }

    prefixNode.arg = (ExpressionNode)prefixNode.arg.eval( this );
    AbstractClass type = prefixNode.arg.clazz;
      
    lvalue = false;

    if( (prefixNode.arg instanceof ConstNode) )
      {
	Object v = constValue( prefixNode );
	if( v != prefixNode.NOT_CONST )
	  return evalConst(new ConstNode( prefixNode.getFilePos(), v ));
      }

    switch( prefixNode.op )
      {
        case '!' :
	  if( type != classTable.booleanType && type != classTable.anyClass )
	    storeError( ERR_A_MUSTBEBOOL, prefixNode.arg );
	  prefixNode.clazz = type;
	  break;
	case '~' :
	  if( type == classTable.shortType || type == classTable.byteType || 
	       type == classTable.charType )
	    type = classTable.intType;
	  else if( type != classTable.longType )
	    if(	! isIntType( type ) )
	      storeError( ERR_A_NOTINTEGER, prefixNode, type );
	  prefixNode.clazz = type;
	  break;	  
        case AbstractTokenizer.TOKEN_PP :
        case AbstractTokenizer.TOKEN_MM :
        case '-' :
        case '+' :
	  {
	    if( ! isNumberType( type ) )
	      storeError( ERR_A_NOTNUMBER, prefixNode, type );
	    if( type == classTable.shortType || type == classTable.byteType || 
	       type == classTable.charType )
	      type = classTable.intType;
	    prefixNode.clazz = type;
	  }
	  break;
        default :
	  storeError( ERR_A_BADOP, prefixNode, "'"+(char)prefixNode.op+"'" );
      }
    return prefixNode;
  }

  public Object evalReturn( ReturnNode returnNode )
  {
    if( returnNode.result != null )
      {
	returnNode.result = evalExpr( returnNode.result );
	AbstractClass type = returnNode.result.clazz;
	AbstractClass methodType = ((AbstractMethod)currentMethod.name.attachment).getType();
	typecheck( type, returnNode.result, methodType, returnNode );
	if( tmpVariable )
	  {
	    Identifier name = classTable.getNameTable().lookupAndInsert( "t$" + tmpVariableCount );
	    tmpVariableCount++;
	    returnNode.attachment = addVariable( false, methodType, name, returnNode );
	  }
	else
	  returnNode.attachment = methodType;	  
      }
    return null;
  }

  public Object evalSwitch( SwitchNode switchNode )
  {
    switchNode.expr = evalExpr( switchNode.expr );
    AbstractClass type = switchNode.expr.clazz;
    if( type != null && !type.isPrimitive() && type != classTable.stringClass &&
	type != classTable.charStringClass && type != classTable.anyClass )
      {
	storeError( ERR_A_BADSWITCHTYPE, switchNode.expr, type );
	type = null;
      }

    if( type != classTable.anyClass )
      {
	if( isIntType( type ) )
	  type = classTable.intType;
	else if( isNumberType( type ) )
	  type = classTable.doubleType;
      }

    switchNode.attachment = type;

    CaseNode cases[] = switchNode.cases;
    
    int nvar = variableAcc.size();
    int varOffset = currVarOffset;

    for( int i = 0 ; i < cases.length ; i++ )
      {
	if( cases[i].expr != null )
	  {
	    cases[i].expr = evalExpr( cases[i].expr );
	    AbstractClass ctype = cases[i].expr.clazz;
	    if( type != ctype )
	      {
		try
		  {
		    if( type == classTable.charStringClass || type == classTable.stringClass )
		      {
			if( ctype != classTable.charStringClass &&
			    ctype != classTable.stringClass &&
			    ctype != classTable.regexClass )
			  storeError( ERR_A_TYPEMISMATCH, cases[i].expr, type, ctype );
		      }
		    else
		      {
			numberTypecheck( type, ctype, cases[i].expr );
		      }
		  }
		catch( CompilerException e )
		  {
		    storeError( e );
		  }
	      }
	  }
	StatementNode[] statements = cases[i].statements;
	for( int j = 0 ; j < statements.length ; j++ )
	  statements[j].eval( this );
      }
    
    for( int nvar1 = variableAcc.size() ; nvar1 > nvar ; nvar1-- )
      {
	Identifier varname = (Identifier)variableAcc.pop();
	variables.remove( varname );
      }
    if( cellReuse )
      currVarOffset = varOffset;
    
    return null;

  }

  public Object evalSynchronized( SynchronizedNode synchronizedNode )
  {
    synchronizedNode.lock = evalExpr( synchronizedNode.lock );
    AbstractClass type = synchronizedNode.lock.clazz;
    if( type != null && type.isPrimitive() )
      storeError( ERR_A_PRIMSYNC, synchronizedNode, type );
    synchronizedNode.body.eval( this );
    return null;
  }

  public Object evalRegex( RegexNode regexNode )
  {
    regexNode.clazz = (AbstractClass)regexNode.attachment;
    String[] pickList = regexNode.pickList;
    String[] refList = regexNode.refList;

    // where to look for variables
    AbstractClass myClass = ( (regexNode.flags & TOK_OPT_STANDALONE) != 0 ? regexNode.clazz : null );

    if( pickList != null || refList != null )
      {
	Hashtable ht = new Hashtable();
	NameTable nameTable = classTable.getNameTable();
	String[][] tbls = { pickList, refList };
	for( int k = 0 ; k < 2 ; k++ )
	  {
	    String[] tb = tbls[k];
	    if( tb == null )
	      continue;
	    for( int i = 0 ; i < tb.length ; i++ )
	      {
		if( ! Character.isLetter(tb[i].charAt(0)) )
		  continue;
		Identifier nm = nameTable.lookupAndInsert( tb[i] );
		AbstractVariable var = null;
		if( myClass != null )
		  {
		    AbstractField f = findField( myClass, nm, myClass );
		    if( f != null )
		      var = helper.newVariable( myClass, f );
		  }

		if( var == null )
		  var = findVariable( nm, true );

		if( var == null )
		  {
		    var = addVariable( false, classTable.charStringClass, nm, regexNode );
		    if( (regexNode.flags & TOK_OPT_DECLARE) == 0 || k == 1 )
		      storeError( ERR_A_IMPLICITDCL, regexNode, nm );
		  }
		AbstractVariable[] av = { var };
		ht.put( tb[i], av );
		if( ! regexNode.embed && tb == pickList )
		  var.mustBeField();
	      }
	  }
	regexNode.vars = ht;
      }
    return regexNode;
  }

  public Object evalReplacer( ReplacerNode replacerNode )
  {
    replacerNode.clazz = classTable.replacerClass;
    replacerNode.regex = (RegexNode)replacerNode.regex.eval( this );
    return replacerNode;
  }

  public Object evalThrow( ThrowNode throwNode )
  {
    throwNode.exception = (ExpressionNode)throwNode.exception.eval( this );
    AbstractClass type = throwNode.exception.clazz;
    if( type != null )
      if( type.isSubclassOf( classTable.throwableClass ) )
	addException( type );
      else
	storeError( ERR_A_NOTTHROWABLE, throwNode.getFilePos(), type );
    return null;
  }

  public Object evalTry( TryNode tryNode )
  {
    // System.out.println( "   %%[[[ " + tryNode.getLine() );
    Vector saveExceptionList = exceptionList;
    Vector thrownList;
    exceptionList = null;
    try
      {
	tryNode.body.eval( this );
	thrownList = exceptionList;
	// System.out.println( "   %%]]] " + thrownList );
      }
    finally
      {
	exceptionList = saveExceptionList;
      }
    CatchNode[] catches = tryNode.catches;
    for( int i = 0 ; i < catches.length ; i++ )
      {
	CatchNode c = catches[i];
	c.eval( this );
	if( c.name != null )
	  {
	    AbstractClass clz = c.name.clazz;
	    if( clz != null )
	      {
		if( ! catchException( thrownList, clz ) )
		  storeError( ERR_A_NOTTHROWN, c.getFilePos(), clz ); 
	      }
	  }
      }

    if( thrownList != null && thrownList.size() > 0 )
      {
	Enumeration tlist = thrownList.elements();
	while( tlist.hasMoreElements() )
	  addException( (AbstractClass)tlist.nextElement() );
      }

    return null;
  }

  public Object evalWhile( WhileNode whileNode )
  {
    whileNode.cond = evalExpr( whileNode.cond );
    AbstractClass clazz = whileNode.cond.clazz;
    if( clazz != classTable.booleanType && clazz != classTable.anyClass )
      storeError( ERR_A_MUSTBEBOOL, whileNode.cond );
    whileNode.body.eval( this );
    return null;
  }

}
