/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.util;

import java.lang.reflect.*;
import java.util.*;

import kmy.regex.util.Replacer;
import kmy.regex.util.Regex;
import kmy.jint.lang.ArgList;
import kmy.jint.lang.CharString;
import kmy.jint.constants.TokenConst;

/**
 *  AnyUtilities is a class with many static methods that are used
 *  to support kmy.jint.lang.Any pseudo-class at runtime.
 *  <p>
 *  Values that are explicitly or implicitly declared as Any are
 *  really just Objects, but all operations on them are deferred
 *  till runtime. For example, accessing a field of such object
 *  requires dynamic lookup through java reflection.
 *  <p>
 *  There is also a special case of Any - an object that implements
 *  kmy.jint.util.Dispatch interface. These objects are given
 *  opportunity to handle field access and method invocation themselves.
 */
public class AnyUtilities implements TokenConst
{

  private static final Method[] EMPTY = {};
  private static final Class[] OBJCLASS = { Object.class };
  private static final Object NULL = new Object();
  private static final Class[] EMPTYT = {};

  public static final int MATCH_OK      = 0;
  public static final int MATCH_PERFECT = 1;
  public static final int MATCH_FAIL    = -1;

  public static final boolean cacheOff = false;

  static private ClassUtilities utilities = new ClassUtilities( new JVMClassMethods() );
  static private ClassUtilities cutilities = 
     new ClassUtilities( new JVMConstructorClassMethods() );

  static class CacheEntry
  {
    int       cacheId;
    String    name;
    Class     target;
    Class[]   types;

    Method       method;
    Constructor  constr;
  }

  static class FCacheEntry
  {
    int       cacheId;
    String    name;
    Class     target;
    Field     field;
  }

  private static Object lock = new Object();
  private static int CACHE_BITS = 7;
  private static int CACHE_SIZE = (1<<CACHE_BITS);
  private static int CACHE_MASK = ~((~0)<<CACHE_BITS);
  private static CacheEntry[]  cache;
  private static FCacheEntry[] fcache;

  static
  {
    if( ! cacheOff )
      {
	cache = new  CacheEntry[CACHE_SIZE];
	for( int i = 0 ; i < cache.length ; i++ )
	  cache[i] = new CacheEntry();
	fcache = new FCacheEntry[CACHE_SIZE];
	for( int i = 0 ; i < fcache.length ; i++ )
	  fcache[i] = new FCacheEntry();
      }
  }

  /**
   *  Converts a number object <i>in</i> into an object specified by class <i>clazz</i>.
   *  Character is considered to be a number too.
   */
  public static Object convertNumber( Object in, Class clazz )
  {
    if( in.getClass() == clazz )
      return in;
    Number num;
    if( in instanceof Character )
      num = new Integer( ((Character)in).charValue() );
    else if( in instanceof Number )
      num = (Number)in;
    else
      throw new RuntimeException( "Cannot convert " + in + " to " + clazz.getName() );
    if( clazz == Integer.TYPE )
      return new Integer( num.intValue() );
    if( clazz == Long.TYPE )
      return new Long( num.longValue() );
    if( clazz == Short.TYPE )
      return new Short( num.shortValue() );
    if( clazz == Byte.TYPE )
      return new Byte( num.byteValue() );
    if( clazz == Character.TYPE )
      return new Character( (char)num.intValue() );
    if( clazz == Float.TYPE )
      return new Float( num.floatValue() );
    if( clazz == Double.TYPE )
      return new Double( num.doubleValue() );
    throw new ClassCastException( "Cannot convert " + num + " to " + clazz.getName() );     
  }

  /**
   *  Returns true if <i>obj</i> is a wrapper object, such as Boolean or Character.
   */
  public static boolean isWrapped( Object obj )
  {
    return obj instanceof Number || obj instanceof Boolean || obj instanceof Character;
  }

  /**
   *  Given two classes that represent numbers, returns the class that should be used to store
   *  the result of operation (like '+' or '*") applied to the objects of these classes.
   */
  public static Class commonClass( Class c1, Class c2 )
  {
    if( c1 == c2 )
      return c1;
    if( c1 == Integer.class || c1 == Character.class || c1 == Short.class || c1 == Byte.class )
      {
	if( c2 == Integer.class || c2 == Character.class || c2 == Short.class || c2 == Byte.class )
	  return Integer.class;
	else
	  return c2;
      }
    if( c1 == Long.class || c1 == Float.class )
      {
	if( c2 == Double.class || c2 == Float.class )
	  return c2;
	else
	  return c1;
      }
    if( c1 == Double.class )
      return c1;
    throw new RuntimeException( "Cannot handle Any here" );
  }

  /**
   *  Matches actual parameter or value type (value) with formal
   *  parameter or variable type (var).
   *  Returns MATCH_PERFECT if they are the same (type and its wrapper class are
   *  considered to be a perfect match), MATCH_OK, if they differ but assignement
   *  still can be made, and MATCH_FAIL if assignment cannot be made.
   */
  public static int match( Class value, Class var )
    {
      if( value == var )
	return MATCH_PERFECT;

      if( var.isPrimitive() )
	{
	  if( value == null )
	    return MATCH_FAIL;
	  if( var == Integer.TYPE )
	    var = Integer.class;
	  else if( var == Double.TYPE )
	    var = Double.class;
	  else if( var == Boolean.TYPE )
	    var = Boolean.class;
	  else if( var == Long.TYPE )
	    var = Long.class;
	  else if( var == Float.TYPE )
	    var = Float.class;
	  else if( var == Character.TYPE )
	    var = Character.class;
	  else if( var == Byte.TYPE )
	    var = Byte.class;
	  else if( var == Short.TYPE )
	    var = Short.class;
	  else
	    throw new RuntimeException( "Unknown primitive type: " + var );
	  if( var == value )
	    return MATCH_PERFECT;
	}
      else
	{
	  if( value == null )
	    return MATCH_PERFECT;
	}

      if( var.isAssignableFrom( value ) )
	return MATCH_OK;
      
      if( var == Object.class )
	return MATCH_OK;
      
      if( Number.class.isAssignableFrom( var ) )
	{
	  if( Number.class.isAssignableFrom( value ) )
	    {
	      if( var == Integer.class )
		{
		  if( value == Byte.class || value == Short.class )
		    return MATCH_OK;
		}
	      else if( var == Float.class )
		{
		  if( value != Double.class )
		    return MATCH_OK;
		}
	      else if( var == Double.class )
		{
		  return MATCH_OK;
		}
	      else if( var == Long.class )
		{
		  if( value != Double.class && value != Float.class )
		    return MATCH_OK;		  
		}
	      else if( var == Short.class )
		{
		  if( value == Byte.class )
		    return MATCH_OK;		  
		}
	    }
	  else if( value == Character.class )
	    {
	      if( var != Byte.class )
		return MATCH_OK;
	    }
	}
	
      return MATCH_FAIL;
    }

  /**
   *  Find a field in the class of <i>target</i> by its name. <i>cacheId</i> allows
   *  this class to cache found field information, so subsequent call with the same
   *  parameters can be faster.
   */
  public static Field findField( Object target, String name, int cacheId )
    throws DispatchException
  {
    Class targetClass = target.getClass();

    FCacheEntry entry;
    if( cacheOff )
      entry = null;
    else
      {
	// lookup in cache
	int ci = cacheId & CACHE_MASK;
	entry = fcache[ci];
	if( entry.cacheId == cacheId )
	  synchronized( entry )
	    {
	      /* entry may have changed meanwhile! */ 
	      if( entry.cacheId == cacheId && entry.name == name && entry.target == targetClass )
		return entry.field;
	    }
      }

    Field field;

    field = (Field)utilities.findField( targetClass, name, null );

    if( field == null )
      throw new DispatchException( "Field not found: " + name );

    if( ! cacheOff )
      {
	synchronized( entry )
	  {
	    entry.cacheId = cacheId;
	    entry.field = field;
	    entry.target = targetClass;
	    entry.name = name;
	  }
      }

    return field;
  }

  /**
   *  Returns value of the field <i>name</i> in object <i>target</i>. Integer parameter
   *  <i>cacheID</i> is used for caching.
   */
  public static Object getField0( Object target, String name, int cacheID )
    throws DispatchException
  {
    try
      {
	return findField( target, name, cacheID ).get( target );
      }
    catch( IllegalAccessException e )
      {
	throw new IllegalAccessError( e.getMessage() );
      }
  }

  /**
   *  Returns value of the field <i>name</i> in object <i>target</i>. If <i>target</i>
   *  is an instance of Dispatch interface, its getField method is called instead.
   *  If <i>target</i> is an array and name == "length", an Integer object that
   *  contains length of the array is returned. Integer parameter <i>cacheID</i> is used for caching.
   */
  public static Object getField( Object target, String name, int cacheID )
  {
    try
      {
	if( name == "length" && target.getClass().isArray() )
	  return new Integer( Array.getLength( target ) );
	if( target instanceof Dispatch )
	  return ((Dispatch)target).getField( name, cacheID );
	return getField0( target, name, cacheID );
      }
    catch( DispatchException e )
      {
	throw new NoSuchFieldError( e.getMessage() );
      }
  }

  /**
   *  Sets value of the field <i>name</i> in object <i>target</i>. Integer parameter
   *  <i>cacheID</i> is used for caching.
   */
  public static void setField0( Object target, Object value, String name, int cacheID )
    throws DispatchException
  {
    try
      {
	findField( target, name, cacheID ).set( target, value );
      }
    catch( IllegalAccessException e )
      {
	throw new IllegalAccessError( e.getMessage() );
      }
  }

  /**
   *  Sets value of the field <i>name</i> in object <i>target</i> to <i>value</i>.
   *  If <i>target</i> is an instance of Dispatch interface, its setField method
   *  is called instead. Integer parameter <i>cacheID</i> is used for caching.
   */
  public static void setField( Object target, Object value, String name, int cacheID )
  {
    try
      {
	if( target instanceof Dispatch )
	  ((Dispatch)target).setField( value, name, cacheID );
	else
	  setField0( target, value, name, cacheID );
      }
    catch( DispatchException e )
      {
	throw new NoSuchFieldError( e.getMessage() );
      }
  }

  /**
   *  Calls <i>target</i>'s method that is named <i>name</i> and has signature
   *  that matches parameter types given in argument <i>type</i>. Argument
   *  <i>args</i> specifies actual parameter values. Argument <i>cacheID</i> is
   *  used for caching, so the next time method can be found faster. Jint
   *  variable-length parameter list (ArgList) is supported. Actual call is
   *  done through Java reflection.
   */
  public static Object callMethod0( Object target, Class targetClass, String name,
      Class[] types, Object[] args, int cacheId ) throws Exception
  {
    if( targetClass.isArray() && name == "size" && types.length == 0 )
      {
	// we have object methods + .size() pseoudo method
	return new Integer( Array.getLength( target ) );
      }

    Method method = null;
    Constructor constr = null;
    Class[] formals = null;
    CacheEntry entry;

    if( cacheOff )
      entry = null;
    else
      {
	// lookup in cache
	int ci = cacheId & CACHE_MASK;
	entry = cache[ci];
	if( entry.cacheId == cacheId )
	  synchronized( entry )
	    {
	      // entry may have changed meanwhile!
	      if( entry.cacheId == cacheId && entry.name == name && entry.target == targetClass )
		{
		  if( types != null )
		    {
		      Class[] eTypes = entry.types;
		      if( eTypes != null && eTypes.length == types.length )
			{
			  method = entry.method;
			  constr = entry.constr;
			  for( int i = 0 ; i < types.length ; i++ )
			    if( types[i] != eTypes[i] )
			      {
				method = null;
				constr = null;
				break;
			      }
			}
		    }
		  else
		    if( entry.types == null )
		      {
			method = entry.method;
			constr = entry.constr;
		      }
		}
	    }
      }

    if( method == null && constr == null )
      {
	Enumeration list;
	if( name == "<init>" )
	  {
	    list = cutilities.findMethod( targetClass, name, types, null );
	    if( list.hasMoreElements() )
	      constr = (Constructor)list.nextElement();
	  }
	else
	  {
	    list = utilities.findMethod( targetClass, name, types, null );
	    if( list.hasMoreElements() )
	      method = (Method)list.nextElement();
	  }
	if( method != null || constr != null )
	  {
	    if( ! cacheOff )
	      {
		// insert in cache
		synchronized( entry )
		  {
		    entry.method = method;
		    entry.constr = constr;
		    entry.cacheId = cacheId;
		    entry.name = name;
		    entry.types = types;
		    entry.target = targetClass;
		  }
	      }
	  }
      }

    try
      {
	if( method != null )
	  {
	    formals = method.getParameterTypes();
	    return method.invoke( target, convertArgs( args, formals ) );
	  }
	if( constr != null )
	  {
	    formals = constr.getParameterTypes();
	    return constr.newInstance( convertArgs( args, formals ) );
	  }
	throw new DispatchException( "Method not found: " + targetClass.getName() + "." + name );
      }
    catch( InvocationTargetException e )
      {
	Throwable t = e.getTargetException();
	if( t instanceof Exception )
	  throw (Exception)t;
	if( t instanceof Error ) // doubt if ever 
	  throw (Error)t;
	else
	  throw new RuntimeException( "Unexpected throwable: " + t );
      }
  }

  /**
   *  Calls <i>targetClass</i>'s method that is named <i>name</i> and has signature
   *  that matches parameter types given in argument <i>type</i> using instance
   *  specified by <i>target</i>. <i>Target</i> can be null for constructors
   *  and static methods. Constructors can be accessed by specifying "<init>" as
   *  a method name. Argument
   *  <i>args</i> specifies actual parameter values. Argument <i>cacheID</i> is
   *  used for caching, so the next time method can be found faster. Any element
   *  in array <i>type</i> that has value null is filled by the
   *  corresponding actual parameter value's class. Jint
   *  variable-length parameter list (ArgList) is supported. Actual call is
   *  done through Java reflection if target is not an instance of
   *  kmy.jint.util.Dispatch; if it is, Dispatch.callMethod method is called.
   */
  public static Object callMethod( Object target, Class targetClass, String name,
	Class[] types, Object[] args, int cacheId ) throws Exception
  {
    try
      {
	if( types != null )
	  {
	    int nargs = types.length;
	    for( int i = 0 ; i < nargs ; i++ )
	      if( types[i] == null && args[i] != null )
		types[i] = args[i].getClass();
	  }
	else
	  types = EMPTYT;
	if( Dispatch.class.isAssignableFrom(targetClass) )
	  return ((Dispatch)target).callMethod( name, types, args, cacheId );
	return callMethod0( target, targetClass, name, types, args, cacheId );
      }
    catch( DispatchException e )
      {
	throw new NoSuchMethodException( e.getMessage() );
      }
  }

  /**
   *   Same as the other callMethod method, with targetClass = target.getClass().
   */
  public static Object callMethod( Object target, String name,
	Class[] types, Object[] args, int cacheId ) throws Exception
  {
    return callMethod( target, target.getClass(), name, types, args, cacheId );
  }

  /**
   *  Converts actaul parameter values in array <i>args</i> to
   *  types specified in array <i>types</i>. It involves converting
   *  objects representing numbers to corrrect type and collapsing
   *  argument list, if the last parameter type is ArgList and
   *  either actual parameter list length differs from parameter type
   *  list length or last actual parameter is not an instance of ArgList.
   *  Returns array of converted parameter values.
   */
  public static Object[] convertArgs( Object[] args, Class[] types )
  {
    if( args == null )
    args = EMPTY;
    Object[] res;
    int length = types.length;
    if( length > 0 && types[length-1] == ArgList.class && 
	(args.length != length || args[length-1] == null || 
	 args[length-1].getClass() != ArgList.class) )
      {
	res = new Object[length];
	length--;
	res[length] = new ArgList( args, length );
      }
    else
      res = args;
    for( int i = 0 ; i < length ; i++ )
      {
	Object val = args[i];
	if( val == null )
	  continue;
	Class type = types[i];
	if( type.isInstance( val ) )
	  {
	    if( args != res )
	      res[i] = val;
	  }
	else
	  res[i] = convertNumber( val, type );
      }
    return res;
  }

  /**
   *  Performs a prefix operation on an object. Prefix operations are: '+', '-', '~', '!'.
   */
  public static Object op( int op, Object arg1 )
  {
    Number n1;
    Number n2;

    if( op == '!' )
      return ((Boolean)arg1).booleanValue() ? Boolean.FALSE : Boolean.TRUE;

    if( arg1 instanceof Character )
      n1 = new Integer( ((Character)arg1).charValue() );
    else
      n1 = (Number)arg1;

    if( n1 instanceof Integer || n1 instanceof Short || n1 instanceof Byte )
      {
	int i1 = n1.intValue();
	switch( op )
	  {
	    case '+' :
	      break;
	    case '-' :
	      i1 = -i1;
	      break;
	    case '~' :
	      i1 = ~i1;
	      break;
	    default:
	      throw new RuntimeException( "Operation not supported: " + (char)op + " [" + op + "]" );
	  }
	return new Integer( i1 );
      }
    if( n1 instanceof Long )
      {
	long l1 = n1.longValue();
	switch( op )
	  {
	    case '+' :
	      break;
	    case '-' :
	      l1 = -l1;
	      break;
	    case '~' :
	      l1 = ~l1;
	      break;
	    default:
	      throw new RuntimeException( "Operation not supported: " + (char)op + " [" + op + "]" );
	  }
	return new Long( l1 );
      }
    if( n1 instanceof Double )
      {
	double d1 = n1.doubleValue();
	switch( op )
	  {
	    case '+' :
	      break;
	    case '-' :
	      d1 = -d1;
	      break;
	    case '~' :
	      throw new RuntimeException( "Operation '~' is not applicable to real numbers" );
	    default:
	      throw new RuntimeException( "Operation not supported: " + (char)op + " [" + op + "]" );
	  }
	return new Double( d1 );
      }
    if( n1 instanceof Float )
      {
	float f1 = n1.floatValue();
	switch( op )
	  {
	    case '+' :
	      break;
	    case '-' :
	      f1 = -f1;
	      break;
	    case '~' :
	      throw new RuntimeException( "Operation '~' is not applicable to real numbers" );
	    default:
	      throw new RuntimeException( "Operation not supported: " + (char)op + " [" + op + "]" );
	  }
	return new Float( f1 );
      }
    throw new RuntimeException( "Unexpected operand type + " + arg1.getClass().getName() );
  }

  /**
   *  Performs an infix operation on a pair of objects.
   */
  public static Object op( int op, Object arg1, Object arg2 )
  {
    Number n1;
    Number n2;
    if( arg1 instanceof Character )
      n1 = new Integer( ((Character)arg1).charValue() );
    else if( (op != '+' && op != TOKEN_PE) || arg1 instanceof Number )
      n1 = (Number)arg1;
    else
      return arg1.toString() + arg2;
    if( arg2 instanceof Character )
      n2 = new Integer( ((Character)arg2).charValue() );
    else if( op != '+' || arg2 instanceof Number )
      n2 = (Number)arg2;
    else
      return arg1.toString() + arg2;
    Class cc = commonClass( n1.getClass(), n2.getClass() );
    if( cc == Integer.class )
      {
	int i1 = n1.intValue();
	int i2 = n2.intValue();
	switch( op )
	  {
	    case '+' :
	    case TOKEN_PP :
	    case TOKEN_PE :
	      i1 += i2;
	      break;
	    case '-' :
	    case TOKEN_MM :
	    case TOKEN_ME :
	      i1 -= i2;
	      break;
	    case '*' :
	      i1 *= i2;
	      break;
	    case '/' :
	      i1 /= i2;
	      break;
	    case '%' :
	      i1 %= i2;
	      break;
	    case '^' :
	      i1 ^= i2;
	      break;
	    case '&' :
	      i1 &= i2;
	      break;
	    case '|' :
	      i1 |= i2;
	      break;
	    default:
	      throw new RuntimeException( "Operation not supported: " + (char)op + " [" + op + "]" );
	  }
	return new Integer( i1 );
      }
    if( cc == Long.class )
      {
	long l1 = n1.longValue();
	long l2 = n2.longValue();
	switch( op )
	  {
	    case '+' :
	    case TOKEN_PP :
	    case TOKEN_PE :
	      l1 += l2;
	      break;
	    case '-' :
	    case TOKEN_MM :
	    case TOKEN_ME :
	      l1 -= l2;
	      break;
	    case '*' :
	    case TOKEN_AE :
	      l1 *= l2;
	      break;
	    case '/' :
	    case TOKEN_SE :
	      l1 /= l2;
	      break;
	    case '%' :
	    case TOKEN_PCE :
	      l1 %= l2;
	      break;
	    case '^' :
	    case TOKEN_XE :
	      l1 ^= l2;
	      break;
	    case '&' :
	    case TOKEN_ANDE :
	      l1 &= l2;
	      break;
	    case '|' :
	    case TOKEN_ORE :
	      l1 |= l2;
	      break;
	    default:
	      throw new RuntimeException( "Operation not supported: " + (char)op + " [" + op + "]" );
	  }
	return new Long( l1 );
      }
    if( cc == Double.class )
      {
	double d1 = n1.doubleValue();
	double d2 = n2.doubleValue();
	switch( op )
	  {
	    case '+' :
	    case TOKEN_PP :
	    case TOKEN_PE :
	      d1 += d2;
	      break;
	    case '-' :
	    case TOKEN_MM :
	    case TOKEN_ME :
	      d1 -= d2;
	      break;
	    case '*' :
	    case TOKEN_AE :
	      d1 *= d2;
	      break;
	    case '/' :
	    case TOKEN_SE :
	      d1 /= d2;
	      break;
	    case '%' :
	    case TOKEN_PCE :
	      d1 %= d2;
	      break;
	    default:
	      throw new RuntimeException( "Operation not supported: " + (char)op + " [" + op + "]" );
	  }
	return new Double( d1 );
      }
    if( cc == Float.class )
      {
	float f1 = n1.floatValue();
	float f2 = n2.floatValue();
	switch( op )
	  {
	    case '+' :
	    case TOKEN_PP :
	    case TOKEN_PE :
	      f1 += f2;
	      break;
	    case '-' :
	    case TOKEN_MM :
	    case TOKEN_ME :
	      f1 -= f2;
	      break;
	    case '*' :
	    case TOKEN_AE :
	      f1 *= f2;
	      break;
	    case '/' :
	    case TOKEN_SE :
	      f1 /= f2;
	      break;
	    case '%' :
	    case TOKEN_PCE :
	      f1 %= f2;
	      break;
	    default:
	      throw new RuntimeException( "Operation not supported: " + (char)op + " [" + op + "]" );
	  }
	return new Float( f1 );
      }
    throw new RuntimeException( "Unexpected operand type + " + cc.getName() );
  }

  /**
   *  Evaluates condition (like '<' or '>') on a pair of objects.
   */
  public static boolean cond( int op, Object arg1, Object arg2 )
  {
    if( arg1 instanceof Character )
      arg1 = new Integer( ((Character)arg1).charValue() );
    if( arg2 instanceof Character )
      arg2 = new Integer( ((Character)arg2).charValue() );
    if( ! (arg1 instanceof Number && arg2 instanceof Number) )
      {
	if( op == TOKEN_EE || op == TOKEN_NE )
	  {
	    if( (arg1 instanceof String || arg1 instanceof CharString) &&
		(arg2 instanceof String || arg2 instanceof CharString) )
	      {
		return ( op == TOKEN_EE ) == arg1.toString().equals(arg2.toString());
	      }
	    else
	      {
		return ( op == TOKEN_EE ) == ( arg1 == arg2 );
	      }
	  }
      }
    Number n1 = (Number)arg1;
    Number n2 = (Number)arg2;

    Class cc = commonClass( n1.getClass(), n2.getClass() );
    if( cc == Integer.class )
      {
	int i1 = n1.intValue();
	int i2 = n2.intValue();
	switch( op )
	  {
	    case '<' :
	      return i1 < i2;
	    case '>' :
	      return i1 > i2;
	    case TOKEN_GE :
	      return i1 >= i2;
	    case TOKEN_LE :
	      return i1 <= i2;
	    case TOKEN_EE :
	      return i1 == i2;
	    case TOKEN_NE :
	      return i1 != i2;
	    default:
	      throw new RuntimeException( "Operation not supported: " + (char)op + " [" + op + "]" );
	  }
      }
    if( cc == Double.class )
      {
	double d1 = n1.doubleValue();
	double d2 = n2.doubleValue();
	switch( op )
	  {
	    case '<' :
	      return d1 < d2;
	    case '>' :
	      return d1 > d2;
	    case TOKEN_GE :
	      return d1 >= d2;
	    case TOKEN_LE :
	      return d1 <= d2;
	    case TOKEN_EE :
	      return d1 == d2;
	    case TOKEN_NE :
	      return d1 != d2;
	    default:
	      throw new RuntimeException( "Operation not supported: " + (char)op + " [" + op + "]" );
	  }
      }
    if( cc == Long.class )
      {
	long l1 = n1.longValue();
	long l2 = n2.longValue();
	switch( op )
	  {
	    case '<' :
	      return l1 < l2;
	    case '>' :
	      return l1 > l2;
	    case TOKEN_GE :
	      return l1 >= l2;
	    case TOKEN_LE :
	      return l1 <= l2;
	    case TOKEN_EE :
	      return l1 == l2;
	    case TOKEN_NE :
	      return l1 != l2;
	    default:
	      throw new RuntimeException( "Operation not supported: " + (char)op + " [" + op + "]" );
	  }
      }
    if( cc == Float.class )
      {
	float f1 = n1.floatValue();
	float f2 = n2.floatValue();
	switch( op )
	  {
	    case '<' :
	      return f1 < f2;
	    case '>' :
	      return f1 > f2;
	    case TOKEN_GE :
	      return f1 >= f2;
	    case TOKEN_LE :
	      return f1 <= f2;
	    case TOKEN_EE :
	      return f1 == f2;
	    case TOKEN_NE :
	      return f1 != f2;
	    default:
	      throw new RuntimeException( "Operation not supported: " + (char)op + " [" + op + "]" );
	  }
      }
    throw new RuntimeException( "Unexpected operand type + " + cc.getName() );
  }

  /**
   *  Performs shift ( '<<', '>>', or '>>>' ) on an object.
   */
  public static Object shift( int op, Object arg, int sh )
  {
    Number n;
    if( arg instanceof Character )
      n = new Integer( ((Character)arg).charValue() );
    else
      n = (Number)arg;
    if( n instanceof Double || n instanceof Float )
      throw new RuntimeException( "Unexpected operand type + " + arg.getClass().getName() );
    if( n instanceof Long )
      {
	long l = n.longValue();
	switch( op )
	  {
	    case TOKEN_GG :
	    case TOKEN_GGE :
	      l >>= sh;
	      break;
	    case TOKEN_GGG :
	    case TOKEN_GGGE :
	      l >>>= sh;
	      break;
	    case TOKEN_LL :
	    case TOKEN_LLE :
	      l <<= sh;
	      break;
	    default:
	      throw new RuntimeException( "Operation not supported: " + (char)op + " [" + op + "]" );
	  }
	return new Long( l );
      }
    else
      {
	int i = n.intValue();
	switch( op )
	  {
	    case TOKEN_GG :
	    case TOKEN_GGE :
	      i >>= sh;
	      break;
	    case TOKEN_GGG :
	    case TOKEN_GGGE :
	      i >>>= sh;
	      break;
	    case TOKEN_LL :
	    case TOKEN_LLE :
	      i <<= sh;
	      break;
	    default:
	      throw new RuntimeException( "Operation not supported: " + (char)op + " [" + op + "]" );
	  }
	return new Integer( i );
      }
  }

  /**
   *  Returns integer value of Character or Number object.
   */
  public static int intValue( Object v )
  {
    if( v instanceof Character )
      return ((Character)v).charValue();
    return ((Number)v).intValue();
  }

  /**
   *  Returns an element of a vector or dictionary <i>table</i> by its index or key <i>key</i>.
   *  If <i>table</i> is an instance of kmy.jint.util.Dispatch, then its getElement is called.  
   */
  public static Object getElement( Object table, Object key )
  {
    // TODO: handle Java 1.2 collections
    if( table instanceof Dispatch )
      return ((Dispatch)table).getElement( key );
    if( table instanceof Vector )
      return ((Vector)table).elementAt( intValue( key ) );
    if( table instanceof Dictionary )
      return ((Dictionary)table).get( key );
    if( table.getClass().isArray() )
      return Array.get( table, intValue( key ) );
    throw new IllegalArgumentException( "Not a dictionary or vector or list!" );
  }

  /**
   *  Sets element of a vector or dictionary <i>table</i> by its index or key <i>key</i> to
   *  <i>value</i>. If <i>table</i> is an instance of kmy.jint.util.Dispatch, then its
   *  setElement is called.  
   */
  public static void putElement( Object table, Object key, Object value )
  {
    // TODO: handle Java 1.2 collections
    if( table instanceof Dispatch )
      ((Dispatch)table).setElement( key, value );
    else if( table instanceof Vector )
      {
	Vector v = (Vector)table;
	int index = intValue( key );
	if( v.size() <= index )
	  v.setSize( index+1 );
        v.setElementAt( value, index );
      }
    else if( table instanceof Dictionary )
      ((Dictionary)table).put( key, value );
    else if( table.getClass().isArray() )
      Array.set( table, intValue( key ), value );
    else
      throw new IllegalArgumentException( "Neither a dictionary nor vector!" );
  }

  /**
   *  This method is called to implement Jint 'new Any()' expression. It creates
   *  an instance of DispatchObject.
   */
  public static Object newAny()
  {
    return new DispatchObject();
  }

  public static Object textSearch( Object text, Object textProc )
  {
    if( textProc instanceof Regex )
      return ((Regex)textProc).searchOnce( text ) ? Boolean.TRUE : Boolean.FALSE;
    if( textProc instanceof Replacer )
      return ((Replacer)textProc).process( text );
    throw new IllegalArgumentException( "Neither a Regex nor a Replacer!" );    
  }
}

