/*
    Copyright (c) Peter Sorotokin, 1998-2000
    See file "License.html" for terms of usage and
    redistribution.   
*/
package kmy.jint.util;

import java.util.Enumeration;
import java.util.Vector;

import kmy.jint.constants.DefinitionConst;

public class ClassUtilities implements DefinitionConst
{

  static final int MATCH_PERFECT = 0;
  static final int MATCH_OK = 1;
  static final int MATCH_FAIL = 2;

  static final int MORE_SPECIFIC = 0;
  static final int LESS_SPECIFIC = 1;
  static final int UNDEFINED = 2;
  static final int SAME = 3;

  ClassMethods methods;
  Object argListClass;
  Object anyClass;

  Object INIT;
  static Object[] EMPTY = {};

  public ClassUtilities( ClassMethods methods )
  {
    this.methods = methods;
    argListClass = methods.getArgListClass();
    anyClass = methods.getAnyClass();
    INIT = methods.getInitName();
  }

  private boolean samePackage( Object clazz1, Object clazz2 )
  {
    return methods.getPackage( clazz1 ).equals( methods.getPackage( clazz2 ) );
  }

  private boolean isMemberAccessible( Object clazz, int modifiers, Object accessorClass )
  {
    boolean publicFlag = (methods.getClassModifiers(clazz) & PUBLIC_FLAG) != 0;
    if( !publicFlag && ( accessorClass == null || !samePackage( clazz, accessorClass ) ) )
      return false;
    if( (modifiers & PUBLIC_FLAG) != 0 )
      return true;
    if( accessorClass == null )
      return false;
    if( (modifiers & PRIVATE_FLAG) != 0 )
      return methods.isPrivateAccessible( clazz, accessorClass );
    if( (modifiers & PROTECTED_FLAG) != 0 )
      {
	if( methods.isAssignableFrom( clazz, accessorClass ) )
	  return true;
      }
    // if not public, no need to check package, as it has already been checked!
    return !publicFlag || samePackage( clazz, accessorClass );
  }

  private boolean isMethodAccessible( Object method, Object accessorClass )
  {
    return isMemberAccessible( methods.getMethodDeclaringClass( method ),
			       methods.getMethodModifiers( method ), accessorClass );
  }

  private boolean isFieldAccessible( Object declaringClass, Object field, Object accessorClass )
  {
    return isMemberAccessible( declaringClass,
			       methods.getFieldModifiers( field ), accessorClass );
  }

  private int argListMatch( Object[] actual, Object[] formal )
  {
    if( formal == null )
      formal = EMPTY;
    int toTest = formal.length;
    int res = MATCH_PERFECT;
    if( toTest >= 1 && formal[formal.length-1] == argListClass )
      {
	toTest--;
	if( actual.length < toTest )
	  return MATCH_FAIL;
	res = MATCH_OK;
      }
    else
      {
	if( actual.length != toTest )
	  return MATCH_FAIL;
      }
    for( int i = 0 ; i < toTest ; i++ )
      {
	if( actual[i] == formal[i] )
	  continue;
	if( ! methods.isAssignableFrom( formal[i], actual[i] ) )
	  return MATCH_FAIL;
	res = MATCH_OK;
      }
    return res;
  }

  private int compareClasses( Object class1, Object class2 )
  {
    if( methods.areSame( class1, class2 ) )
      return SAME;
    if( methods.isAssignableFrom( class1, class2 ) )
      return LESS_SPECIFIC;
    if( methods.isAssignableFrom( class2, class1 ) )
      return MORE_SPECIFIC;
    return UNDEFINED;
  }

  private int compareClasses( int prev, Object c1, Object c2 )
  {
    int res = compareClasses( c1, c2 );
    if( res == SAME )
      return prev;
    if( prev == SAME || res == prev )
      return res;
    return UNDEFINED;
  }

  private int compareMethods( Object method1, Object method2 )
  {
    int result = SAME;

    Object class1 = methods.getMethodDeclaringClass( method1 );
    Object class2 = methods.getMethodDeclaringClass( method2 );

    Object[] param1 = methods.getParameterTypes( method1 );
    Object[] param2 = methods.getParameterTypes( method2 );

    int len1 = param1.length;
    int len2 = param2.length;
    int len = len1;

    if( len > 0 && param1[len-1] == argListClass )
      {
	len--;
	result = LESS_SPECIFIC;
      }

    if( len2 <= len+1 && len2 > 0 && param2[len2-1] == argListClass )
      {
	if( result != LESS_SPECIFIC )
	  result = MORE_SPECIFIC;
	else
	  if( len2 <= len )
	    result = MORE_SPECIFIC;
	  else if( len2 == len+1 )
	    result = SAME;
	len = len2-1;
      }

    result = compareClasses( result, class1, class2 );
    if( result == UNDEFINED )
      return UNDEFINED;

    for( int i = 0 ; i < len ; i++ )
      {
	result = compareClasses( result, param1[i], param2[i] );
	if( result == UNDEFINED )
	  return UNDEFINED;
      }

    return result;
  }

  private boolean processMethods( Object[] argTypes,
				     Enumeration list, Vector acc, Object accessor )
  {
    if( list == null )
      return false;
    while( list.hasMoreElements() )
      {
	Object m = list.nextElement();
	if( ! isMethodAccessible( m, accessor ) )
	  {
	    // System.out.println( "Method " + m + " is not accessible from " + accessor );
	    continue;
	  }
	int match = argListMatch( argTypes, methods.getParameterTypes( m ) );
	if( match == MATCH_PERFECT )
	  {
	    Object declaringClass = methods.getMethodDeclaringClass( m );
	    if( declaringClass != anyClass )
	      {
		//System.out.pgetMethodDeclaringClass(rintln( "Found perfect: " + m  );
		acc.removeAllElements();
		acc.addElement( m );
		return true;
	      }
	    match = MATCH_OK;
	  }
	if( match == MATCH_OK )
	  {
	    //System.out.println( "Found: " + m  );
	    int len = acc.size();
	    int res = MORE_SPECIFIC;
	    for( int i = 0 ; i < len ; i++ )
	      {
		Object mi = acc.elementAt(i);
		if( m == mi ) // can happen when multiply-inheriting from interfaces
		  {
		    res = LESS_SPECIFIC;
		    break;
		  }
		res = compareMethods( m, mi );
		if( res == SAME ) // can have m and mi representing the same method
		  {
		    res = LESS_SPECIFIC;
		    break;
		  }
		// System.out.println( "Comparing " + m + " " + mi + " -> " + res );
		if( res == LESS_SPECIFIC )
		  break;
		if( res == MORE_SPECIFIC )
		  {
		    acc.removeElementAt(i);
		    i--;
		    len--;
		  }
	      }
	    if( res == LESS_SPECIFIC )
	      continue;
	    acc.addElement( m );
	  }
      }
    return false;
  }

  private Enumeration getMethods( Object clazz, Object name, int nargs )
  {
    Enumeration list = methods.getMethods( clazz, name, nargs );
    Enumeration list1 = methods.getMethods( clazz, name, ClassMethods.VARIABLE );
    if( list1 != null && list1.hasMoreElements() )
      if( list == null )
	list = list1;
      else
	list = new SequenceEnumeration( list1, list );
    return list;
  }

  public Enumeration findMethod( Object type, Object name,
		     Object[] argTypes, Object accessor )
  {
    return findMethod( type, name, argTypes, accessor, null, false );
  }

  /**
   *  Search for methods that match given name and parameter type list and
   *  are accessible from a given class. Also looks in superclasses and
   *  implemented interfaces unless thisClassOnly parameter is true.
   *  @param type class being searched
   *  @param name method name
   *  @param argTypes list of parameter types
   *  @param accessor accessor class
   *  @param extraMethods list of extra methods that should be considered
   *  @param thisClassOnly if true only this class, but not its superclasses are searched 
   *  @returns the list of all matching methods
   */
  public Enumeration findMethod( Object type, Object name,
		     Object[] argTypes, Object accessor,
		     Enumeration extraMethods, boolean thisClassOnly )
  {
    //System.out.println( "-------- Looking for " + name  );
    Vector acceptable = new Vector();
    Enumeration interfaces = null;

    if( argTypes == null )
      argTypes = EMPTY;

    thisClassOnly = thisClassOnly || name == INIT;

    // walk inheritance tree breadth first
   OUTER:
    do
      {
	Vector interfaceAcc = null;
	Object type0 = type;
	while( true )
	  {
	    if( type != null || extraMethods != null )
	      {
		// System.out.println( "-------- Looking for " + name + " in " + type );
		Enumeration list;
		if( extraMethods == null )
		  list = getMethods( type, name, argTypes.length );
		else
		  {
		    if( type != null )
		      {
			list = getMethods( type, name, argTypes.length );
			if( list != null )
			  list = new SequenceEnumeration( list, extraMethods );
			else
			  list = extraMethods;
		      }
		    else
		      list = extraMethods;
		    extraMethods = null;
		  }
		if( processMethods( argTypes, list, acceptable, accessor ) )
		  break OUTER;
		if( type != null )
		  {
		    Object[] implemented = methods.getImplementedInterfaces( type );
		    if( implemented != null && implemented.length > 0 )
		      {
			if( interfaceAcc == null )
			  interfaceAcc = new Vector();
			for( int i = 0 ; i < implemented.length ; i++ )
			  interfaceAcc.addElement( implemented[i] );
		      }
		  }
		if( interfaces == null || !interfaces.hasMoreElements() )
		  break;
	      }
	    if( interfaces != null )
	      type = interfaces.nextElement();
	    else
	      break;
	  }
	if( interfaceAcc == null )
	  interfaces = null;
	else
	  interfaces = interfaceAcc.elements();
	if( type0 == null )
	  type = null;
	else
	  type = methods.getSuperclass( type0 );
      }
    while( !thisClassOnly && (type != null || interfaces != null) );

    return acceptable.elements();    
  }

  public Object findField( Object type0, Object name, Object accessor )
  {
    //System.out.println( "-------- Looking for " + name  );
    Object found = null;
    Enumeration interfaces = null;

    // walk inheritance tree breadth first
    do
      {
	Vector interfaceAcc = null;
	Object type = type0;
	while( true )
	  {
	    if( type != null )
	      {
		//System.out.println( "-------- Looking for " + name + " in " + type );
		Object f = methods.getField( type, name );
		if( f != null && isFieldAccessible( type, f, accessor) )
		  {
		    if( found != null && found != f )
		      {
			System.err.println( "Cannot resolve field " + name +
					    ", ambiguity: " + f + " " + found );
			return null;
		      }
		    found = f;
		  }
		Object[] implemented = methods.getImplementedInterfaces( type );
		if( implemented != null && implemented.length > 0 )
		  {
		    if( interfaceAcc == null )
		      interfaceAcc = new Vector();
		    for( int i = 0 ; i < implemented.length ; i++ )
		      interfaceAcc.addElement( implemented[i] );
		  }
		if( interfaces == null || !interfaces.hasMoreElements() )
		  break;
	      }
	    type = interfaces.nextElement();
	  }
	if( interfaceAcc == null )
	  interfaces = null;
	else
	  interfaces = interfaceAcc.elements();
	if( type0 != null )
	  type0 = methods.getSuperclass( type0 );
      }
    while( found == null && (type0 != null || interfaces != null) );
    return found;
  }


}
